/*  bivariat.cpp

    Implementation of bivariate data sets

    This file is part of Xtremes for Windows,
    Copyright (C) 1995, 2002 Xtremes Group, Siegen.
*/

#include <assert.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>

#include "mathfunc.hpp"
#include "samplist.hpp"
#include "sample.hpp"
#include "allest.hpp"

#ifndef TERMINAL
#include "boxplot.hpp"
#include "dbldlg.hpp"
#include "display.hpp"
#include "empiric.hpp"
#include "stddlg.hpp"
#include "scatter.hpp"
#include "stddef.hpp"
#include "ufodlg.hpp"
#include "wintypes.hpp"
#include "xplfunc.hpp"

#include "resource.h"
#include "privmsg.hpp"
#endif

/********************************************************************
                        Bivariate data sets
********************************************************************/

#ifndef TERMINAL

class TBivariateListWindow : public TSampleListWindow {
    typedef TSampleListWindow inherited;
    public:
        TBivariateListWindow (TMDIFrame *, TBivariateBaseSample *p);
    protected:
        virtual int GetSubMenuPos ();
        virtual String GetLine (int i);
};

TBivariateListWindow::TBivariateListWindow (TMDIFrame *MDIFrame, TBivariateBaseSample *p)
 : inherited (p, MDIFrame, p->GetFileName ()) {
    SetHeight (p->GetDataSize ());
}

int TBivariateListWindow::GetSubMenuPos () {
    return 10;
}

String TBivariateListWindow::GetLine (int i) {
    if (i >= s->GetDataSize ()) return "";
    return form (60, "%-15g %g", (*(TBivariateBaseSample *) s) [i].x, (*(TBivariateBaseSample *) s) [i].y);
}

#endif

TBivariateBaseSample::TBivariateBaseSample () {
    DataSize = 0;
    DataPtr = NULL;
    threshold = 0.0;
}

TBivariateBaseSample::~TBivariateBaseSample () {
    if (DataPtr) free (DataPtr);
    DataPtr = NULL;
}

void TBivariateBaseSample::SetValues (double t0) {
    threshold = t0;
}

int TBivariateBaseSample::GetDimension () {
    return 2;
}

int TBivariateBaseSample::GetDataSize () {
    return DataSize;
}

double TBivariateBaseSample::GetData (int i, int j) {
    assert (0 <= i && i < DataSize);
    return j == 0 ? DataPtr [i].x : DataPtr [i].y;
}

String TBivariateBaseSample::GetColumnName (int j) {
    if (j == 0) return "x";
    return "y";
}

int TBivariateBaseSample::GetRegisteredSize () {
    return RegisteredSize;
}

TDblPoint TBivariateBaseSample::operator [] (int i) {
    assert (i >= 0);
    assert (i < DataSize);
    return DataPtr [i];
}

BOOL TBivariateBaseSample::Append (double v, double w) {
    if (DataSize >= AllocatedSize) {
        AllocatedSize += 1000;
        if (AllocatedSize > MAXSAMPLESIZE) return FALSE;
        DataPtr = (TDblPoint *) realloc (DataPtr, AllocatedSize * sizeof (TDblPoint));
        if (!DataPtr) return FALSE;
    }
    DataPtr [DataSize++] = TDblPoint (v, w);
    return TRUE;
}

BOOL TBivariateBaseSample::ReadData (TTextFileReader &f) {
    BOOL b = TRUE;
    if (!ReadSampleSize (f, RegisteredSize)) return FALSE;
    DataSize = AllocatedSize = 0;
    do {
        double v = f.GetDouble (), w = 0.0;
        if (!f.IsError ()) w = f.GetDouble ();
        if (!f.IsError ()) b = Append (v, w);
    } while (!f.IsError () && b);
    if (f.GetError () == TTextFileReader::END_OF_FILE)
        return TRUE;
    else if (f.GetError ()) ReadMessage = "Error in number";
    else ReadMessage = "Insufficient memory to load data set";
    return FALSE;
}

BOOL TBivariateBaseSample::WriteData (FILE *f) {
    fprintf (f, "Samplesize: %d\n", DataSize);
    for (int i = 0; i < DataSize; i++)
        fprintf (f, "%g %g\n", DataPtr [i].x, DataPtr [i].y);
    return TRUE;
}

BOOL TBivariateBaseSample::GetMemory () {
    DataPtr = new TDblPoint [DataSize];
    return DataPtr != NULL;
}

void TBivariateBaseSample::SetData (int n, double *point, double *data, String fn) {
    DataSize = n;
    int Missing = 0;
    DataPtr = (TDblPoint *) malloc (DataSize * sizeof (TDblPoint));
    for (int i = 0; i < DataSize; i++) {
        if (point [i] == NOTAVAILABLE || data [i] == NOTAVAILABLE)
            Missing++;
        else
            DataPtr [i - Missing] = TDblPoint (point [i], data [i]);
    }
    RegisteredSize = (int) point [n - 1];
    filename = fn;
    if (filename != String ("")) {
	SaveData (0);           // Hier noch Handle auf Frame-Window
	SampleList.InsertSample (this);
    }
}

void TBivariateBaseSample::TruncateSample (String filename, double threshold, int Column) {
    int NewDataSize = 0;
    int i;
    for (i = 0; i < GetDataSize (); i++)
	if (DataPtr [i].y >= threshold) NewDataSize++;
    double *x = new double [NewDataSize],
	   *y = new double [NewDataSize];
    TBivariateBaseSample *p = DuplicateType ();
    p->SetName ("Exceedances of " & GetName ());
    TList s;
    s.Append (new TStringEl ("Data set generated by Xtremes applying the Store Exceedances option to"));
    s.Append (new TStringEl (GetName ()));
    p->SetDescription (&s);
    s.DeleteAll (FALSE);
    int j = 0;
    for (i = 0; i < GetDataSize (); i++)
	if (DataPtr [i].y >= threshold) {
	    x [j] = DataPtr [i].x;
	    y [j++] = DataPtr [i].y;
	}
    p->SetData (NewDataSize, x, y, filename);
    delete [] x; delete [] y;
}

#ifndef TERMINAL

void TBivariateBaseSample::ScatterPlot2D (TMDIFrame *MDIFrame) {
    new TScatterPlotWindow (MDIFrame, this);
}

void TBivariateBaseSample::ListData (TMDIFrame *MDIFrame) {
    new TBivariateListWindow (MDIFrame, this);
}

void TBivariateBaseSample::TransformSample (TWindow *parent) {
    TWriteFileField e (IDD_Name, ModifyName ('a'));
    TCalcDoubleField e1 (IDD_Mu, 0.0),
                 e2 (IDD_Sigma, 1.0);
    TDialog d (IDD_AffineTrafoSample);
    d << e << e1 << e2;
    unsigned ch = d.Handle (parent);
    if (ch == IDOK) {
        double *x = new double [GetDataSize ()],
               *y = new double [GetDataSize ()],
               a = e1.GetValue (), b = e2.GetValue ();
        for (int i = 0; i < GetDataSize (); i++) {
            x [i] = DataPtr [i].x;
            y [i] = a + b * DataPtr [i].y;
        }
        TList s;
        s.Append (new TStringEl (form (150, "Data set generated by Xtremes applying the transformation x -> %g x + %g to", b, a)));
        s.Append (new TStringEl (GetName ()));
        TBivariateBaseSample *p = DuplicateType ();
        p->SetName ("Transformed " & GetName ());
        p->SetDescription (&s);
	((TTimeSeriesSample *) p)->SetData (GetDataSize (), x, y, e.ReadBuf ());
        s.DeleteAll (FALSE);
        delete [] x; delete [] y;
    }
}

void TBivariateBaseSample::LogScaleSample (TWindow *parent) {
    TWriteFileField e (IDD_Name, ModifyName ('l'));
    TDialog d (IDD_ConverseSample);
    d.SetTitel ("Price, Log-Scaled");
    d << e;
    unsigned ch = d.Handle (parent);
    if (ch == IDOK) {
	int n = GetDataSize (), i;
        double *t = new double [n],
               *r = new double [n];
	for (i = 0; i < n; i++) {
	    t [i] = DataPtr [i].x;
	    r [i] = DataPtr [i].y > 0.0 ? log (DataPtr [i].y) : -1000.0;
	}
	double *c = PolynomialRegression (1, n, t, r);
	for (i = 0; i < n; i++)
	    r [i] -= (c [0] + c [1] * t [i]);
        TList s;
        s.Append (new TStringEl ("Data set generated by Xtremes applying a log scale transformation to"));
        s.Append (new TStringEl (GetName ()));
        TBivariateBaseSample *p = DuplicateType ();
        p->SetName ("Log-Scaled " & GetName ());
        p->SetDescription (&s);
//      if (p->GetSampleType () == TIMESERIES_SAMPLE)
            ((TTimeSeriesSample *) p)->SetData (GetDataSize (), t, r, e.ReadBuf ());
//      else
//          ((TBivariateSample *) p)->SetData (GetDataSize (), x, y, e.ReadBuf ());
        s.DeleteAll (FALSE);

	delete [] c;
	delete [] t;
	delete [] r;
    }
}

void TBivariateBaseSample::ConverseSample (TWindow *parent) {
//    TEditField e (IDD_Name, ModifyName ('c') , MAXFILENAME);
    TWriteFileField e (IDD_Name, ModifyName ('c'));
    TDialog d (IDD_ConverseSample);
    d << e;
    unsigned ch = d.Handle (parent);
    if (ch == IDOK) {
        double *y = new double [GetDataSize ()],
               *x = new double [GetDataSize ()];
        for (int i = 0; i < GetDataSize (); i++) {
            x [i] = DataPtr [i].x;
            y [i] = -DataPtr [i].y;
        }
        TList s;
        s.Append (new TStringEl ("Data set generated by Xtremes changing the sign of"));
        s.Append (new TStringEl (GetName ()));
        TBivariateBaseSample *p = DuplicateType ();
        p->SetName ("Conversed " & GetName ());
        p->SetDescription (&s);
//      if (p->GetSampleType () == TIMESERIES_SAMPLE)
            ((TTimeSeriesSample *) p)->SetData (GetDataSize (), x, y, e.ReadBuf ());
//      else
//          ((TBivariateSample *) p)->SetData (GetDataSize (), x, y, e.ReadBuf ());
        s.DeleteAll (FALSE);
        delete [] x; delete [] y;
    }
}

void TBivariateBaseSample::UserTransformation (TWindow *parent) {
    static TList PrevList1, PrevList2;
    Expression *c1, *c2;
    TUfoEditField e0 (IDD_Formula, &PrevList1),
                  e2 (IDD_Formula2, &PrevList2);
    TWriteFileField e1 (IDD_Name, ModifyName ('u'));
    TDialog d (IDD_BivariateUserTrafo);
    d.SetHelpNumber (IDD_UserTrafo);
    d << e0 << e1 << e2;
    unsigned ch = d.Handle (parent);
    if (ch == IDOK) {
        c1 = e0.GetExpression ();
        c2 = e2.GetExpression ();
        TList s;
        s.Append (new TStringEl ("Data set created from " & GetName ()));
        s.Append (new TStringEl ("by applying the transformations"));
        s.Append (new TStringEl ("(t,x)->(" & e0.ReadBuf () & "," & e2.ReadBuf () & ")"));
        TBivariateBaseSample *p = DuplicateType ();
        p->SetName ("Transformed " & GetName ());
        p->SetDescription (&s);
        double *x = new double [GetDataSize ()],
               *y = new double [GetDataSize ()];
        VarSpace LocalVarcontext (&GlobalVarcontext);
        UfoString Dummy;
        LocalVarcontext.insert ("t", 0.0, Dummy);
        LocalVarcontext.insert ("x", 0.0, Dummy);
        double *tptr = LocalVarcontext.getref ("t", Dummy),
               *xptr = LocalVarcontext.getref ("x", Dummy);
        for (int i = 0; i < GetDataSize (); i++) {
            *tptr = DataPtr [i].x;
            *xptr = DataPtr [i].y;
            x [i] = c1->evaluate (GlobalFunccontext, LocalVarcontext, 1000);
            y [i] = c2->evaluate (GlobalFunccontext, LocalVarcontext, 1000);
        }
        ((TTimeSeriesSample *) p)->SetData (GetDataSize (), x, y, e1.ReadBuf ());
        delete [] x; delete [] y;
        s.DeleteAll (FALSE);
        if (c1) delete c1;
        if (c2) delete c2;
    }
}

void TBivariateBaseSample::TruncateSample (TWindow *parent) {
    TCalcDoubleField f (IDD_Threshold, threshold);
    TWriteFileField e (IDD_Name, ModifyName ('e'));
    TDialog d (IDD_ExtremesDialog);
    d.SetTitel ("Save exceedances");
    d << e << f;
    unsigned ch = d.Handle (parent);
    if (ch == IDOK) {
        threshold = f.GetValue ();
        TruncateSample (e.ReadBuf (), threshold, 0);
    }
}

TSample *TBivariateBaseSample::ConvertToGrouped (TWindow *parent) {
    TWriteFileField e (IDD_Name, ModifyName ('p'));
    TDialog d (IDD_ConverseSample);
    d.SetTitel ("Convert to Grouped Data");
    d << e;
    unsigned ch = d.Handle (parent);
    TGroupedSample *p = NULL;
    if (ch == IDOK) {
        double *x = new double [DataSize],
               *y = new double [DataSize];
        p = new TGroupedSample;
        p->SetName ("Grouped " & GetName ());
        TList s;
        s.Append (new TStringEl ("Data set generated by Xtremes applying the Convert to Grouped Data set option to"));
        s.Append (new TStringEl (GetName ()));
        p->SetDescription (&s);
        s.DeleteAll (FALSE);
        for (int i = 0; i < DataSize; i++) {
            x [i] = DataPtr [i].x;
            y [i] = DataPtr [i].y;
        }
        p->SetData (DataSize, x, y, e.ReadBuf ());
        filename = e.ReadBuf ();
        delete [] x; delete [] y;
    }
    return p;
}

TSample *TBivariateBaseSample::ConvertToUnivariate (TWindow *parent) {
//    TEditField e (IDD_Name, ModifyName ('p'), MAXFILENAME);
    TWriteFileField e (IDD_Name, ModifyName ('p'));
    TDialog d (IDD_ConverseSample);
    d.SetTitel ("Convert to univariate");
    d << e;
    unsigned ch = d.Handle (parent);
    TUnivariateSample *p;
    if (ch == IDOK) {
        double *x = new double [DataSize];
        p = new TUnivariateSample;
        p->SetName ("Univariate " & GetName ());
        TList s;
        s.Append (new TStringEl ("Data set generated by Xtremes applying the Convert to Univariate Data set option to"));
        s.Append (new TStringEl (GetName ()));
        p->SetDescription (&s);
        s.DeleteAll (FALSE);
        for (int i = 0; i < DataSize; i++)
            x [i] = DataPtr [i].y;
        p->SetData (DataSize, x, e.ReadBuf ());
        delete [] x;
    }
    return NULL;
}

TSample *TBivariateBaseSample::ConvertToCopulaScale (TWindow *parent)
{
 		  int n = GetDataSize();
        double *y = new double [n],
               *x = new double [n];
        for (int i = 0; i < n; i++) {
	        	double s1=0,s2=0;
        		for (int j = 0; j < n; j++)
            		{
            		if (DataPtr[i].x>DataPtr[j].x)
                  	s1++;
						if (DataPtr[i].y>DataPtr[j].y)
                  	s2++;
                  }
            x [i] = s1/n;
            y [i] = s2/n;
        }
        TList s;
        s.Append (new TStringEl ("Data set generated by Xtremes convert to copula scale"));
        s.Append (new TStringEl (GetName ()));
        TBivariateBaseSample *p = DuplicateType ();
        p->SetName ("Copula Scale " & GetName ());
	     p->SetData (GetDataSize (), x, y, "copulascale.dat");
        s.DeleteAll (FALSE);
        delete [] x; delete [] y;
return p;
}

TSample *TBivariateBaseSample::ConvertTo (TSampleType st, TWindow *parent) {
    if (st == UNIVARIATE_SAMPLE)
        return ConvertToUnivariate (parent);
    if (st == GROUPED_SAMPLE)
        return ConvertToGrouped (parent);
    if (st == COPULA_SAMPLE)
        return ConvertToCopulaScale(parent);
    return TSample::ConvertTo (st, parent);
}

BOOL TBivariateBaseSample::CreateData (TWindow *parent) {
    return FALSE;
}

unsigned TBivariateBaseSample::EditParameter (TWindow *parent) {
    return IDCANCEL;
}

void TBivariateBaseSample::ShowFunctionalParameters (TWindow *parent) {
    TStaticField fSize (IDD_SampleSize),
                 fMin1 (IDD_SampleMin1),
                 fMax1 (IDD_SampleMax1),
                 fMean1 (IDD_SampleMean1),
                 fVar1 (IDD_SampleVariance1),
                 fSkew1 (IDD_SampleSkewness1),
                 fMin2 (IDD_SampleMin2),
                 fMax2 (IDD_SampleMax2),
                 fMean2 (IDD_SampleMean2),
                 fVar2 (IDD_SampleVariance2),
                 fSkew2 (IDD_SampleSkewness2),
                 fCorr (IDD_SampleCorrelation),
                 fCov (IDD_SampleCovariance);

    double sum1 = 0.0, sum2 = 0.0, sum1sq = 0.0, sum2sq = 0.0, sumxy = 0.0,
           xmin = DBL_MAX, xmax = -DBL_MAX,
           ymin = DBL_MAX, ymax = -DBL_MAX;
    for (int i = 0; i < DataSize; i++) {
        sum1 += DataPtr [i].x;
        sum2 += DataPtr [i].y;
        sumxy += DataPtr [i].x * DataPtr [i].y;
        sum1sq += square (DataPtr [i].x);
        sum2sq += square (DataPtr [i].y);
        if (DataPtr [i].x > xmax) xmax = DataPtr [i].x;
        if (DataPtr [i].x < xmin) xmin = DataPtr [i].x;
        if (DataPtr [i].y > ymax) ymax = DataPtr [i].y;
        if (DataPtr [i].y < ymin) ymin = DataPtr [i].y;
    }
    double xm1 = sum1 / DataSize,
           xm2 = sum2 / DataSize,
           s1 = sum1sq / DataSize - square (xm1),
           s2 = sum2sq / DataSize - square (xm2),
           cov = sumxy / DataSize - xm1 * xm2;

    double corr = 0.0;
    if (sum1sq * sum2sq)
        corr = cov / sqrt (s1 * s2);

    fSize.SetText (form (20, "%d", DataSize));
    fMin1.SetText (form (20, "%g", xmin));
    fMin2.SetText (form (20, "%g", ymin));
    fMax1.SetText (form (20, "%g", xmax));
    fMax2.SetText (form (20, "%g", ymax));
    fMean1.SetText (form (20, "%g", xm1));
    fMean2.SetText (form (20, "%g", xm2));
    fVar1.SetText (form (20, "%g", sqrt (s1)));
    fVar2.SetText (form (20, "%g", sqrt (s2)));
    fCov.SetText (form (20, "%g", cov));
    fCorr.SetText (form (20, "%g", corr));

    TDialog d (IDD_BivariateFunctionalParameters);
    d << fSize << fMin1 << fMax1 << fMean1 << fVar1 << fSkew1
	       << fMin2 << fMax2 << fMean2 << fVar2 << fSkew2
               << fCorr << fCov;
    d.Handle (parent);
}

void TBivariateBaseSample::BoxPlot (TMDIFrame *MDIFrame) {
    double *x = new double [DataSize], ymin, ymax;
    for (int i = 0; i < DataSize; i++)
    	x [i] = DataPtr [i].y;
    TPlotWindow *w = PlotWindowManager.GetOpenWindow ("Boxplot of " & Name ());
    TFunction *f = new TBoxPlot (DataSize, x, 1, Name ());
    f->GetYRange (ymin, ymax);
    w->InsertFunction (f);
    w->SetNewCoordinates (TDblRectangle (0, ymin - (ymax - ymin) / 5.0, 2, ymax + (ymax - ymin) / 5.0), FALSE);
    delete [] x;
}

#endif

/********************************************************************
                        Time Series
********************************************************************/

TSampleType TTimeSeriesSample::GetSampleType () {
    return TIMESERIES_SAMPLE;
}

String TTimeSeriesSample::GetHeader () {
    return SampleHeaders [TIMESERIES_SAMPLE];
}

TBivariateBaseSample *TTimeSeriesSample::DuplicateType () {
    TTimeSeriesSample *p = new TTimeSeriesSample;
    p->SetRegisteredSize (GetRegisteredSize ());
    return p;
}


double TTimeSeriesSample::MeanClusterSize (int mingap, double threshold) {
    BOOL InCluster = FALSE;
    int ClusterCount = 0, ClusterSize = 0,
        Len = 0, Gap = 0, PointsInGap = 0;
    for (int i = 0; i < GetDataSize (); i++)
        if (InCluster)
	    if (DataPtr [i].y >= threshold) {
                Len++; Gap = 0;
            } else {
                PointsInGap++; Gap++; Len++;
                if (Gap >= mingap) {
                    ClusterSize += (Len - PointsInGap);
                    ClusterCount++;
                    PointsInGap = 0;
                    InCluster = FALSE;
                }
            }
        else
            if (DataPtr [i].y >= threshold) {
                InCluster = TRUE; Len = 1; Gap = 0;
            }
    if (InCluster) {
	ClusterSize += (Len - PointsInGap);
        ClusterCount++;
    }
    if (ClusterCount) return (double) ClusterSize / (double) ClusterCount;
    return 0.0;
}

#ifndef TERMINAL

TSample *TTimeSeriesSample::ConvertToMultivariate (TWindow *parent) {
    TWriteFileField e (IDD_Name, ModifyName ('b'));
    TDialog d (IDD_ConverseSample);
    d.SetTitel ("Convert to Multivariate");
    d << e;
    unsigned ch = d.Handle (parent);
    TMultivariateSample *p = NULL;
    if (ch == IDOK) {
        double *x = new double [DataSize],
               *y = new double [DataSize];
        p = new TMultivariateSample;
        p->SetName ("Multivariate" & GetName ());
        p->SetData (2, e.ReadBuf (), DataSize);
        TList s;
        s.Append (new TStringEl ("Multivariate data set generated by Xtremes from"));
        s.Append (new TStringEl (GetName ()));
        p->SetDescription (&s);
        s.DeleteAll (FALSE);
        for (int i = 0; i < DataSize; i++) {
            x [i] = DataPtr [i].x;
            y [i] = DataPtr [i].y;
        }
	p->SetColumn (0, x, "Time");
        p->SetColumn (1, y, "Observation");
        p->SaveData (0);
	delete [] x; delete [] y;
    }
    return p;
}

TSample *TTimeSeriesSample::ConvertTo (TSampleType st, TWindow *parent) {
    return TBivariateBaseSample::ConvertTo (st, parent);
}

void TTimeSeriesSample::PlotSampleACF (TWindow *parent) {
    PlotFunction (parent, new TEmpiricalAutocovarianceFunction (this), "Sample Autocovariance Function");
}

void TTimeSeriesSample::PlotSampleAutoCorrelation (TWindow *parent) {
    PlotFunction (parent, new TEmpiricalAutocorellationFunction (this), "Sample Autocorellation Function");
}

void TTimeSeriesSample::PlotSampleVar (TWindow *parent, int type, int distr) {
    switch (type) {
        case TMovingVARFunction::VAR_VAR:
            PlotFunction (parent, new TMovingVARFunction (this, type, distr), "Temporary Estimates (VAR)");
            break;
        case TMovingVARFunction::VAR_SHAPE:
            PlotFunction (parent, new TMovingVARFunction (this, type, distr), "Temporary Estimates (Shape)");
            break;
        case TMovingVARFunction::VAR_MEAN:
            PlotFunction (parent, new TMovingVARFunction (this, type, distr), "Temporary Estimates (Mean)");
            break;
        case TMovingVARFunction::VAR_DEV:
            PlotFunction (parent, new TMovingVARFunction (this, type, distr), "Temporary Estimates (Standard Deviation)");
            break;
        case TMovingVARFunction::VAR_VARIANCE:
            PlotFunction (parent, new TMovingVARFunction (this, type, distr), "Temporary Estimates (Variance)");
            break;
        default:
            break; // keep gcc happy
    }
}

void TTimeSeriesSample::DeclusterSample (TWindow *parent, int mingap, double threshold) {
    TWriteFileField e (IDD_Name, ModifyName ('e'));
    TDialog d (IDD_ClusterMaximaDialog);
    TIntField f (IDD_Gap, mingap, 1, INT_MAX);
    TCalcDoubleField g (IDD_Threshold, threshold);
    d << e << f << g;
    unsigned ch = d.Handle (parent);
    if (ch == IDOK) {
        double *Maxima = new double [DataSize],
               *Points = new double [DataSize];
        threshold = g.GetValue ();
        mingap = f.GetValue ();
	BOOL InCluster = FALSE;
        int Gap = 0, ClusterNum = 0;
        for (int i = 0; i < GetDataSize (); i++)
            if (InCluster)
                if (DataPtr [i].y >= threshold) {
                    Gap = 0;
                    if (DataPtr [i].y > Maxima [ClusterNum]) {
                        Maxima [ClusterNum] = DataPtr [i].y;
                        Points [ClusterNum] = DataPtr [i].x;
		    }
                } else {
                    Gap++;
                    if (Gap >= mingap) {
                        InCluster = FALSE;
                        ClusterNum++;
		    }
		}
            else
                if (DataPtr [i].y >= threshold) {
                    InCluster = TRUE; Gap = 0;
                    Maxima [ClusterNum] = DataPtr [i].y;
                    Points [ClusterNum] = DataPtr [i].x;
                }
        if (InCluster) ClusterNum++;
        if (!ClusterNum) {
	    ErrorBox (parent, "No exceedances above selected threshold.");
	    return;
	}

	TTimeSeriesSample *s = new TTimeSeriesSample;
	s->SetRegisteredSize (GetRegisteredSize ());
	s->SetName ("Cluster maxima from " & GetFileName ());
	TList l;
	l.Append (new TStringEl ("Data set generated by Xtremes extracting cluster maxima from"));
	l.Append (new TStringEl (Name ()));
	l.Append (new TStringEl (form (100, "Threshold: %g, run length: %d", threshold, mingap)));
	s->SetDescription (&l);
	s->SetData (ClusterNum, Points, Maxima, e.ReadBuf ());
	l.DeleteAll (FALSE);
	delete [] Maxima; delete [] Points;
    }
}

#endif

/********************************************************************/
/*			Hydrograph				    */
/********************************************************************/

int TTimeSeriesSample::SearchMaxIndex (int t, double val, int sign0, int maxgap) {
    int low, up, i, maxindex, gap;
    double maxval;
    BOOL Done;
    low = t; Done = FALSE; gap = 0;
    while (!Done)
	if ((sign (DataPtr [low].y - val) == sign0) || (gap < maxgap)) {
	    if (sign (val - DataPtr [low].y) == sign0) gap++;
	    else gap = 0;
	    if (low == 0) Done = TRUE;
	    else low--;
	} else Done = TRUE;
    up = t; Done = FALSE; gap = 0;
    while (!Done)
	if ((sign (DataPtr [up].y - val) == sign0) || (gap < maxgap)) {
	    if (sign (val - DataPtr [up].y) == sign0) gap++;
	    else gap = 0;
	    if (up == DataSize - 1) Done = TRUE;
	    else up++;
	} else Done = TRUE;

    maxindex = low; maxval = DataPtr [low].y;
    for (i = low + 1; i <= up; i++)
	if (sign (DataPtr [i].y - maxval) == sign0) {
	    maxindex = i;
	    maxval = DataPtr [i].y;
	}
    return maxindex;
}

void TTimeSeriesSample::Entry (double x, double y, double * &xp, double * &yp, int &count, int &allocated) {
    if (count == allocated) {
	allocated += 20;
	xp = (double *) realloc (xp, allocated * sizeof (double));
	yp = (double *) realloc (yp, allocated * sizeof (double));
    }
    xp [count] = x;
    yp [count] = y;
    count++;
}

void TTimeSeriesSample::TestEntry (double x, double y, double * &xp, double * &yp, int postgap, int &gapcount, int &gapallocated) {
    int i;
    for (i = 0; i < gapcount; i++)
	if (fabs (x - xp [i]) < postgap)
	    return;
    Entry (x, y, xp, yp, gapcount, gapallocated);
}

void TTimeSeriesSample::DetectCluster (int sign0, int maxgap, double c, int postgap,
				       double * &xgap, double * &ygap, int &gapcount, int &gapallocated) {
    int i, j, count = 0, allocated = 0;
    double *xhydro = NULL, *yhydro = NULL;
    for (i = 0; i < DataSize; i++) {
	j = SearchMaxIndex (i, (1.0 - sign0 * sign (DataPtr [i].y) * c) * DataPtr [i].y, sign0, maxgap);
	if (i == j)
	    Entry (DataPtr [i].x, DataPtr [i].y, xhydro, yhydro, count, allocated);
    }
    sort (0, count - 1, yhydro, xhydro);
    if (sign0 == 1)
	for (i = count - 1; i >= 0; i--)
	    TestEntry (xhydro [i], yhydro [i], xgap, ygap, postgap, gapcount, gapallocated);
    else
	for (i = 0; i < count; i++)
	    TestEntry (xhydro [i], yhydro [i], xgap, ygap, postgap, gapcount, gapallocated);
    if (xhydro) free (xhydro);
    if (yhydro) free (yhydro);
}

void TTimeSeriesSample::CalculateHydrograph (int r0, double c, int r1, double * &x, double * &y, int &count) {
    double *xlow = NULL, *ylow = NULL, *xup = NULL, *yup = NULL;
    int lowcount = 0, lowalloc = 0, upcount = 0, upalloc = 0;
    DetectCluster (1, r0, c, r1, xup, yup, upcount, upalloc);
    DetectCluster (-1, r0, c, r1, xlow, ylow, lowcount, lowalloc);
    count = lowcount + upcount + 2;
    x = new double [count]; y = new double [count];
    memcpy (x + 1, xlow, lowcount * sizeof (double));
    memcpy (y + 1, ylow, lowcount * sizeof (double));
    memcpy (x + 1 + lowcount, xup, upcount * sizeof (double));
    memcpy (y + 1 + lowcount, yup, upcount * sizeof (double));
    x [0] = DataPtr [0].x; y [0] = DataPtr [0].y;
    x [count - 1] = DataPtr [DataSize - 1].x; y [count - 1] = DataPtr [DataSize - 1].y;
    sort (0, count - 1, x, y);
    free (xlow); free (ylow);
    free (xup); free (yup);
}


#ifndef TERMINAL

// graphical interface for hydrograph

void TTimeSeriesSample::CalculateHydrograph (TWindow *parent) {
    TDialog d (IDD_Hydrograph);
    TIntField fr0 (IDD_r0, GlobalData.HydroR0, 1, INT_MAX),
	      fr1 (IDD_r1, GlobalData.HydroR1, 1, INT_MAX);
    TCalcDoubleField fc (IDD_c, GlobalData.HydroC, 0.0, 1.0);
    d << fr0 << fr1 << fc;
    if (d.Handle (parent) == IDOK) {
	int r0 = GlobalData.HydroR0 = fr0.GetValue ();
	int r1 = GlobalData.HydroR1 = fr1.GetValue ();
	double c = GlobalData.HydroC = fc.GetValue ();
	double *x = NULL, *y = NULL;
	int count = 0;
	CalculateHydrograph (r0, c, r1, x, y, count);
	new TScatterPlotWindow (PlotWindowManager.GetFramePtr (), "Hydrograph of " & GetName (), count, x, y,
        FALSE);
    }
}

#endif

/********************************************************************/
/*		   Partial Duration Series			    */
/********************************************************************/

#define L 365

void TTimeSeriesSample::CumulateData () {
    int i;
    nc = new int [L];
    gamma = new double [L];
    sigma = new double [L];
    nPeriods = (int) (floor (DataPtr [DataSize - 1].x / L) + 1);
    x = new double * [L];
    for (i = 0; i < L; i++) {
	x [i] = new double [nPeriods];
	nc [i] = 0;
    }
    for (i = 0; i < DataSize; i++)
	if (DataPtr [i].x <= nPeriods * L) {
	    int index = (int) DataPtr [i].x % L;
	    x [index][nc [index]] = DataPtr [i].y;
	    nc [index]++;
	}
}

void TTimeSeriesSample::FreeMemory () {
    for (int i = 0; i < L; i++) delete [] x [i];
    delete [] x;
    delete [] sigma;
    delete [] gamma;
    delete [] nc;
}

void TTimeSeriesSample::Entry (double *z, int &Nr, int c) {
    while (c < 0) c += L;
    c = c % L;
    for (int i = 0; i < nc [c]; i++)
	z [Nr + i] = x [c][i];
    Nr += nc [c];
}

void TTimeSeriesSample::Estimate (int i, int k, double u) {
    int Nr = 1, Offset = 0;
    z [0] = u;
    do {
	Entry (z, Nr, i % L + Offset);
	if (Offset <= 0) Offset = -Offset + 1;
	else Offset = -Offset;
    } while (Nr < k);
    TUnivariateSample *q = new TUnivariateSample;
    q->SetData (Nr, z, "");
    TEstResult r = MLEGP->Estimate (Nr, q);
    if (r != EST_OK) r = MomentGP->Estimate (Nr, q);
    delete q;
    if (r == EST_OK) {
	gamma [i] = MLEGP->GetEstimatedParameterValue (0);
	sigma [i] = MLEGP->GetEstimatedParameterValue (2);
    } else {
	gamma [i] = NOTAVAILABLE;
	sigma [i] = 0.0;
	TUnivariateSample *q = new TUnivariateSample;
	q->SetData (Nr, z, "");
	SampleList.Append (q);
    }
}

void TTimeSeriesSample::CalcEstimates (int k, double u) {
    z = new double [k + 2 * nPeriods];
    for (int i = 0; i < L; i++)
	if (nc [i]) Estimate (i, k, u);
	else gamma [i] = sigma [i] = 0.0;
    delete [] z;
}

void TTimeSeriesSample::StoreEstimates (String fn) {
    TMultivariateSample *s = new TMultivariateSample;
    s->SetData (3, L);
    double *day = new double [L];
    for (int i = 0; i < L; i++) day [i] = i + 1;
    s->SetColumn (0, day, "Day");
    s->SetColumn (1, gamma, "gamma");
    s->SetColumn (2, sigma, "sigma");
    s->SetName ("Daily estimates");
    s->SetFileName (fn);
    s->SaveData (0);
    SampleList.InsertSample (s);
    delete [] day;
}

double TTimeSeriesSample::f (double theta, double u) {
    double sum = 0.0;
    for (int i = 0; i < L; i++)
	if (sigma [i])
	    sum += nc [i] * (1.0 - XGPDF ((theta - u) / sigma [i], gamma [i]));
    return sum / nPeriods;
}

void TTimeSeriesSample::MakePlots (double u, double tmin, double tmax, int tpoints) {
#ifndef TERMINAL
    int i, j;
    TFunction *g;

    double xw [52], yw [52];
    for (i = 0; i < 52; i++) {
	yw [i] = 0.0;
	xw [i] = i + 1;
    }
    for (i = 0; i < L - 1; i++)
	yw [i / 7] += nc [i];
    g = new TXPLFunction (52, xw, yw, "Weekly frequency " & GetName (), TRUE);
    g->SetPlotMode (PLOT_BARS);
    PlotWindowManager.GetOpenWindow ("Weekly frequencies")->InsertFunction (g);

    int nges = 0;
    for (i = 0; i < L; i++)
	nges += nc [i];
    double *xt = new double [nges], *yt = new double [nges];
    int k = 0;
    for (i = 0; i < L; i++)
	for (j = 0; j < nc [i]; j++) {
	    xt [k] = i + 1;
	    yt [k] = x [i][j];
	    k++;
	}
    new TScatterPlotWindow (PlotWindowManager.GetFramePtr (), "Observations", nges, xt, yt, FALSE, FALSE);

    double *xf = new double [tpoints],
	   *yf = new double [tpoints];
    for (i = 0; i < tpoints; i++) {
	yf [i] = tmin + (double) i / (tpoints - 1) * (tmax - tmin);
	xf [i] = 1.0 / f (yf [i], u);
    }
    g = new TXPLFunction (tpoints, xf, yf, "T-Year-Value" & GetName (), FALSE);
    PlotWindowManager.GetOpenWindow ("T-Year value")->InsertFunction (g);

#endif
}

void TTimeSeriesSample::AnalysePartialDurationSeries (double u, int k, String fn, double tmin, double tmax, int tpoints) {
    CumulateData ();
    CalcEstimates (k, u);
    StoreEstimates (fn);
    MakePlots (u, tmin, tmax, tpoints);
    FreeMemory ();
}

#ifndef TERMINAL

void TTimeSeriesSample::AnalysePartialDurationSeries (TWindow *parent) {
    TDialog d (IDD_ParDurSeries);
    TCalcDoubleField e1 (IDD_Threshold, GlobalData.ParDurThreshold),
		     e4 (IDD_TMin, GlobalData.ParDurTMin),
		     e5 (IDD_TMax, GlobalData.ParDurTMax);
    TIntField e2 (IDD_Exceedances, GlobalData.ParDurExceedances, 5, DataSize),
    	      e6 (IDD_TPoints, GlobalData.ParDurTPoints, 2, INT_MAX);
    TWriteFileField e3 (IDD_FileName, GlobalData.ParDurFilename);
    d << e1 << e2 << e3 << e4 << e5 << e6;
    if (d.Handle (parent) == IDOK) {
	GlobalData.ParDurThreshold = e1.GetValue ();
	GlobalData.ParDurExceedances = e2.GetValue ();
	GlobalData.ParDurFilename = e3.ReadBuf ();
	GlobalData.ParDurTMin = e4.GetValue ();
	GlobalData.ParDurTMax = e5.GetValue ();
	GlobalData.ParDurTPoints = e6.GetValue ();
	AnalysePartialDurationSeries (GlobalData.ParDurThreshold, GlobalData.ParDurExceedances, GlobalData.ParDurFilename,
				      GlobalData.ParDurTMin, GlobalData.ParDurTMax, GlobalData.ParDurTPoints);
    }
}

#endif

