/*  compiler.cpp
 *
 *  Implementation of base StatPascal compiler.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998, 2000 Xtremes Group, Siegen. 
 */

#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
// Test
#include <limits.h>

#include "compiler.hpp"
#include "errors.hpp"
#include "mathfunc.hpp"
#include "scanner.hpp"
#include "stddef.hpp"
#include "path.hpp"

/*********************************************************/
/*		Statements				 */
/*********************************************************/

class TStatementBase : public TObject {
    public:
	TStatementBase ();
	virtual ~TStatementBase ();
	virtual void OutputCode () = 0;
	static TStatementBase *Parse ();
    protected:
	int line;
};

class TEmptyStatement : public TStatementBase {
    public:
	virtual void OutputCode ();
};

class TSimpleStatement : public TStatementBase {
    public:
	TSimpleStatement ();
	virtual ~TSimpleStatement ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
	TExpressionBase *first, *second;
};

class TIfStatement : public TStatementBase {
    public:
	TIfStatement ();
	virtual ~TIfStatement ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
	TExpressionBase *expr;
	TStatementBase *st1, *st2;
};

class TStatementSequence : public TStatementBase {
    public:
	TStatementSequence ();
	virtual ~TStatementSequence ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
	TList StatementList;
};

class TRepeatStatement : public TStatementBase {
    public:
	TRepeatStatement ();
	virtual ~TRepeatStatement ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
        TList StatementList;
	TExpressionBase *cond;
};

class TWhileStatement : public TStatementBase {
    public:
	TWhileStatement ();
	virtual ~TWhileStatement ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
 	TStatementBase *statement;
	TExpressionBase *cond;
};

class TForStatement : public TStatementBase {
    public:
        TForStatement ();
	virtual ~TForStatement ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
	TSymbol *control;
	TExpressionBase *start, *stop;
	BOOL Increment;
	TStatementBase *statement;
};

class TReturnStatement : public TStatementBase {
    public:
	TReturnStatement ();
	virtual ~TReturnStatement ();
	TStatementBase *Parse ();
        virtual void OutputCode ();
    private:
	TExpressionBase *expr;
};

class TGotoStatement : public TStatementBase {
    public:
	TGotoStatement ();
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
	TSymbol *Label;
	int CallSSL;
};

class TLabelDefinition : public TStatementBase {
    public:
	TLabelDefinition (TSymbol *Label);
	TStatementBase *Parse ();
	virtual void OutputCode ();
    private:
	TSymbol *Label;
	TStatementBase *ControlledStm;
};

class TCaseLabel : public TObject {
    public:
   	TCaseLabel (int val);
	TCaseLabel (int val0, int val1);
	void OutputCode (int dest);
	int Size ();
    private:
	int a, b;
	BOOL both;
	int line;
};

class TCaseLine : public TObject {
    public:
	TCaseLine ();
	virtual ~TCaseLine ();
	void OutputCode (int startpos);
	void SetStatement (TStatementBase *);
	void AddLabel (TCaseLabel *);
        void FixupFinal (int dest);
    private:
	TList LabelList;
	int Final;
	TStatementBase *statement;
	int line;
};

class TCaseStatement : public TStatementBase {
    public:
	TCaseStatement ();
	virtual ~TCaseStatement ();
	TStatementBase *Parse ();
        virtual void OutputCode ();
    private:
	TExpressionBase *expr;
	TList CaseLineList;
	TStatementBase *otherwise;
};

/*********************************************************/
/*		Global Variables			 */
/*********************************************************/

TCompiler *Compiler;
TScanner *Scanner;
TSymbolTable *SymbolTable;
TPredefinedFunctionList *PredefinedFunctionList;

/*********************************************************/
/*	Implementation of Expressions			 */
/*********************************************************/

TExpressionBase::TExpressionBase () {
    line = Scanner->GetLine ();
}

TExpressionBase::~TExpressionBase () {
}

TType *TExpressionBase::GetType () {
    return type;
}

TType *TExpressionBase::CheckOperatorType (TExpressionBase *&first, TExpressionBase *&second, BOOL IsMultiplication) {
    TType *left = ConvertToBaseType (first),
          *right = ConvertToBaseType (second),
	  *origleft = left,
	  *origright = right;

    // an error has occured, do nothing:
    if (left == SymbolTable->TERROR || right == SymbolTable->TERROR)
	return SymbolTable->TERROR;

    // convert arrays used in expressions to vectors:
    if (left->IsArray ()) {
        if (left->GetBaseType ()->IsArray ()) 
	    first = new TMakeMatrix (first);
	else
	    first = new TMakeVector (first);
	left = first->GetType ();
    }
    if (right->IsArray ()) {
	if (right->GetBaseType ()->IsArray ()) 
	    second = new TMakeMatrix (second);
	else
	    second = new TMakeVector (second);
	right = second->GetType ();
    }
    if (left == right) return left;
    if (left == SymbolTable->TINT && right == SymbolTable->TREAL) {
	first = new TConvertIntReal (first);
	return SymbolTable->TREAL;
    }
    if (left == SymbolTable->TREAL && right == SymbolTable->TINT) {
	second = new TConvertIntReal (second);
	return SymbolTable->TREAL;
    }

    if (left->IsSet () && right->IsSet ()) {
        TType *t1base = left->GetBaseType (),
              *t2base = right->GetBaseType ();
        if (t1base->IsSubrange ()) t1base = t1base->GetBaseType ();
        if (t2base->IsSubrange ()) t2base = t2base->GetBaseType ();
        if (t1base == t2base) return left;
    }
    if (left->IsPointer () && right == SymbolTable->TINT)
        return left;
    // vector operations

    if (left->IsVector () || left->IsMatrix () || right->IsVector () || right->IsMatrix ()) {
	TType *baseleft = left;
	if (left->IsVector () || left->IsMatrix ()) {
	    baseleft = left->GetBaseType ();
	    if (baseleft->IsSubrange ()) baseleft = baseleft->GetBaseType ();
	}
	TType *baseright = right;
	if (right->IsVector () || right->IsMatrix ()) {
	    baseright = right->GetBaseType ();
	    if (baseright->IsSubrange ()) baseright = baseright->GetBaseType ();
	}
	if (baseleft == SymbolTable->TINT && baseright == SymbolTable->TREAL) {
	    if (left->IsVector ()) first = new TConvertIntVectRealVect (first);
	    else if (left->IsMatrix ()) first = new TConvertIntMatrixRealMatrix (first);
	    else first = new TConvertIntReal (first);
	}
	if (baseleft == SymbolTable->TREAL && baseright == SymbolTable->TINT) {
	    if (right->IsVector ()) second = new TConvertIntVectRealVect (second);
	    else if (right->IsMatrix ()) second = new TConvertIntMatrixRealMatrix (second);
	    else second = new TConvertIntReal (second);
	}
	// A scalar is converted to a vector and handled using the componentwise vector/matrix
        // operations, except if a multiplication with a matrix is performed.
	if (!(right->IsVector () || right->IsMatrix ())) {
	    if (left->IsVector ()) second = new TMakeVector (second);
	    if (left->IsMatrix () && !IsMultiplication) second = new TMakeVector (second);
	    right = second->GetType ();
	}
	if (!(left->IsVector () || left->IsMatrix ())) {
	    if (right->IsVector ()) first = new TMakeVector (first);
	    if (right->IsMatrix () && !IsMultiplication) first = new TMakeVector (first);
	    left = first->GetType ();
	}

	if (IsMultiplication && left->IsMatrix () && right->IsVector ()) return right;
	if (left->IsMatrix ()) return first->GetType ();
	if (right->IsMatrix ()) return second->GetType ();
	return left->IsVector () ? second->GetType () : first->GetType ();
    }	
/*
    if ((left->IsVector () || left->IsMatrix ()) && !(right->IsVector () || right->IsMatrix ())) {
	second = new TMakeVector (second);
	right = second->GetType ();
    }
    if (!(left->IsVector () || left->IsMatrix ()) && (right->IsVector () || right->IsMatrix ())) {
 	first = new TMakeVector (first);
	left = first->GetType ();
    }

    if ((left->IsVector () || left->IsMatrix ()) && (right->IsVector () || right->IsMatrix ())) {
	if (left->GetBaseType () == right->GetBaseType ()) return left->IsMatrix () ? left : right;
	if (left->GetBaseType () == SymbolTable->TINT && right->GetBaseType () == SymbolTable->TREAL) {
	    first = new TConvertIntVectRealVect (first);
	    return right;
	}
	if (right->GetBaseType () == SymbolTable->TINT && left->GetBaseType () == SymbolTable->TREAL) {
	    second = new TConvertIntVectRealVect (second);
	    return left;
	}
    }
*/
/*
    if (left->IsVector () && CheckTypeConversion (left->GetBaseType (), second)) {
	second = new TMakeVector (second);
	return left;
    }
    if (right->IsVector () && CheckTypeConversion (right->GetBaseType (), first)) {
  	first = new TMakeVector (first);
	return right;
    }
*/
    if (left->IsPointer () && right->IsPointer () && (left == SymbolTable->TGENERICPOINTER || right == SymbolTable->TGENERICPOINTER))
	return SymbolTable->TGENERICPOINTER;

    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot combine \'" & origleft->Name () & "\' and \'" & origright->Name () & "\'");
    return SymbolTable->TERROR;	
}

TType *TExpressionBase::ConvertToBaseType (TExpressionBase *&first) {
    TType *p = first->GetType ();
    if (p->IsRoutine () && !((TRoutine *) p)->GetNumberOfParameters ()) {
	first = new TFunctionCall (first, NULL);
 	p = first->GetType ();
    }
    if (p->IsSubrange ()) return p->GetBaseType ();
    return p;
}

BOOL TExpressionBase::CheckTypeConversion (TType *Required, TExpressionBase *&p) {
    if (Required == SymbolTable->TERROR || p->GetType () == SymbolTable->TERROR) return TRUE;
    if (Required == p->GetType ()) return TRUE;
    if (Required->IsSubrange () || p->GetType ()->IsSubrange ()) {
        if (p->GetType ()->IsSubrange ()) p = new TChangeType (p->GetType ()->GetBaseType (), p);
        BOOL f = CheckTypeConversion (Required->IsSubrange () ? Required->GetBaseType () : Required, p);
        if (Compiler->CodeRangeCheck && Required->IsSubrange ())
	    p = new TConvertSubrange (Required, p);
        return f;
    }
    if (Required->IsRoutine () && p->GetType ()->IsRoutine ()) {
        if (((TRoutine *) Required)->GetNumberOfParameters () ==
            ((TRoutine *) (p->GetType ()))->GetNumberOfParameters ()) {
            for (TType *a = ((TRoutine *) Required)->GetFirstParaType (),
                       *b = ((TRoutine *) (p->GetType ()))->GetFirstParaType ();
                 a;    a = ((TRoutine *) Required)->GetNextParaType (),
                       b = ((TRoutine *) (p->GetType ()))->GetNextParaType ()) {
                if (a == b) return TRUE;
	        if (a->IsReference () && b->IsReference () && a->GetBaseType () == b->GetBaseType ()) return TRUE;
		return FALSE;
	    }
            return TRUE;
        }
        return FALSE;
    }
    if (p->GetType ()->IsRoutine () && ((TRoutine *) (p->GetType ()))->GetNumberOfParameters () == 0) {
	p = (new TFunctionCall (p, NULL));
	return CheckTypeConversion (Required, p);
    }
    if (Required->IsPointer () && p->GetType ()->IsPointer () && Required->GetBaseType () == p->GetType ()->GetBaseType ())
        return TRUE;
    if (Required->IsPointer () && p->GetType () == SymbolTable->TGENERICPOINTER)
	return TRUE;
    if (Required->IsSet () && p->GetType ()->IsSet ()) {
        TType *t1 = Required->GetBaseType (),
              *t2 = p->GetType ()->GetBaseType ();
        if (t1->IsSubrange ()) t1 = t1->GetBaseType ();
        if (t2->IsSubrange ()) t2 = t2->GetBaseType ();
        return t1 == t2;
    }
    if (Required == SymbolTable->TREAL && p->GetType () == SymbolTable->TINT) {
	p = new TConvertIntReal (p);
        return TRUE;
    }
    if (Required->IsVector ()) {
	TType *q = p->GetType ();
	if (q->IsArray ()) {
            if (q->GetBaseType () == Required->GetBaseType () || q == Required->GetBaseType ()) {
	        p = new TMakeVector (p);
		return TRUE;
	    } else if (q->GetBaseType () == SymbolTable->TINT && Required->GetBaseType () == SymbolTable->TREAL) {
		p = new TConvertIntVectRealVect (new TMakeVector (p));
		return TRUE;
	    }
        } else if (q->IsVector ()) {
            if (Required->GetBaseType () == q->GetBaseType ())
                return TRUE;
            if (Required->GetBaseType () == SymbolTable->TREAL && q->GetBaseType () == SymbolTable->TINT) {
		p = new TConvertIntVectRealVect (p);
                return TRUE;
            }
        } else if (CheckTypeConversion (Required->GetBaseType (), p)) {
	    p = new TMakeVector (p);
	    return TRUE;
	}
    }
    if (Required->IsMatrix ()) {
	TType *q = p->GetType ();
	if (q->IsArray ()) {
	    if (q->GetBaseType ()->IsArray () && q->GetBaseType ()->GetBaseType () == Required->GetBaseType ()) {
		p = new TMakeMatrix (p);
		return TRUE;
	    }
 	} else if (q->IsMatrix () || q->IsVector ()) {
	    if (q->GetBaseType () == Required->GetBaseType ())
		return TRUE;
	}
    }
    return FALSE;
}

void TExpressionBase::DoTypeConversion (TType *Required, TExpressionBase *&p) {
    if (!CheckTypeConversion (Required, p))
        ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot convert \'" & p->GetType ()->Name () & "\' to \'" & Required->Name () & "\'");
}

BOOL TExpressionBase::IsLValue () {
    return FALSE;
}

BOOL TExpressionBase::IsLValueDeref () {
    return FALSE;
}

BOOL TExpressionBase::IsConstant () {
    return FALSE;
}

BOOL TExpressionBase::IsSymbol () {
    return FALSE;
}

TCompExpression::TCompExpression (TExpressionBase *first0, TExpressionBase *second0, TToken op0) {
    first = first0;
    second = second0;
    op = op0;
    if (op == _IN && second->GetType ()->IsSet ()) {
 	TType *left = first->GetType ();
 	TType *right = second->GetType ()->GetBaseType ();
 	if (left->IsSubrange ()) left = left->GetBaseType ();
 	if (right->IsSubrange ()) right = right->GetBaseType ();
 	if (left != right) {
	    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot combine '" & first->GetType ()->Name () & "' and '" & second->GetType ()->Name () & "'");
	    type = SymbolTable->TERROR;
	} else 
 	    type = SymbolTable->TBOOL;
     } else {
 	type = TExpressionBase::CheckOperatorType (first, second, false);
 	// type checking (for equal types)
 	if (type->IsVector () && type != SymbolTable->TSTRING) type = SymbolTable->TBOOLVECT;
 	else if (type != SymbolTable->TERROR) type = SymbolTable->TBOOL;
     }
}

TCompExpression::~TCompExpression () {
    delete first;
    delete second;
}

TExpressionBase *TCompExpression::Parse () {
    TExpressionBase *p = TExpression::Parse ();
    TToken op = Scanner->GetToken ();
    while (op == EQUAL || op == NOTEQUAL || op == GREATERTHAN || op == LESSTHAN ||
           op == GREATEREQUAL || op == LESSEQUAL || op == _IN) {
	Scanner->GetNextToken ();
	TExpressionBase *q = TExpression::Parse ();
	p = new TCompExpression (p, q, op);
	op = Scanner->GetToken ();
    }
    return p;
}

void TCompExpression::OutputCode () {
    if (op == _IN) {
	Compiler->OutCode (line, SETCREATE);
	first->OutputCode ();
	if (Compiler->CodeRangeCheck) Compiler->OutCode (line, CHECK, 0, 8 * N_SET_CELLS * sizeof (double));
	Compiler->OutCode (line, SETINC);
        second->OutputCode ();
	Compiler->OutCode (line, SETAND);
	Compiler->OutCode (line, SETEMPTY);
    } else {
        first->OutputCode ();
	second->OutputCode ();
        int index = -1;
        TType *t = ConvertToBaseType (first);
	if (t == SymbolTable->TINT || t == SymbolTable->TCHAR || t == SymbolTable->TBOOL || t->IsPointer ()) index = 0;
	else if (t == SymbolTable->TREAL) index = 1;
	else if (t == SymbolTable->TSTRING) index = 2;
	else if (t->IsVector ()) {
	    TType *s = t->GetBaseType ();
	    if (s->IsSubrange ()) s = s->GetBaseType ();
	    if (s->IsEnumerated ()) index = 3;
	    else if (s == SymbolTable->TREAL) index = 4;
	}
	if (index == -1) {
	    ErrorMessage (INTERNAL_ERROR, Scanner->GetLine (), "TCompExpression::OutputCode failed for type " & t->Name ());
	    index = 0;
	}
        static TCode TCodeTab [5][6] =
             {{ IEQ, IGT, ILT, IGE, ILE, INE },
              { FEQ, FGT, FLT, FGE, FLE, FNE },
              { SEQ, SGT, SLT, SGE, SLE, SNE },
              { VIEQ, VIGT, VILT, VIGE, VILE, VINE },
              { VFEQ, VFGT, VFLT, VFGE, VFLE, VFNE }
             };
	Compiler->OutCode (line, TCodeTab [index][op - EQUAL]);
    }
}


TExpression::TExpression (TExpressionBase *first0, TExpressionBase *second0, TToken op0) {

    // op: OR XOR ADD SUB

    first = first0;
    second = second0;
    op = op0;
    code = -1;

    type = TExpressionBase::CheckOperatorType (first, second, false);

    if (type == SymbolTable->TBOOL) 
	switch (op) {
	    case OR:
		code = IOR;
		break;
	    case XOR:		
		code = IXOR;
	    default:
		break;
	}
    else if (type == SymbolTable->TINT) 
	switch (op) {
            case ADD:  
		code = IADD; 
		break;
            case SUB:  
		code = ISUB; 
		break;
            case OR:   
		code = IOR;  
		break;
            case XOR:  
		code = IXOR; 
	    default:
		break;
	}
    else if (type == SymbolTable->TREAL) 
	switch (op) {
            case ADD:  
		code = FADD; 
		break;
            case SUB: 
		code = FSUB; 
	    default:
		break;
	}
    else if (type == SymbolTable->TCOMPLEX) {
        if (op == ADD) code = CADD;
        if (op == SUB) code = CSUB;

    } else if (type == SymbolTable->TSTRING && op == ADD)
	code = VCVEC;
    else if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TREAL) 
	switch (op) {
            case ADD: 
		code = VFADD; 
		break;
            case SUB: 
		code = VFSUB;
	    default:
		break;
	}
    else if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TINT) 
	switch (op) {
            case ADD: 
		code = VIADD; 
		break;
            case SUB: 
		code = VISUB; 
		break;
	    case OR:
		code = VIOR;
		break;
	    case XOR:
		code = VIXOR;
	    default:
		break;
	}
    else if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TBOOL) 
	switch (op) {
	    case OR:
		code = VIOR;
		break;
	    case XOR:
		code = VIXOR;
	    default:
		break;
	}
    else if (type->IsPointer ()) {
	TType *s = type->GetBaseType ();
	if (s->GetSize () > 1) 
	    second = new TTerm (second, new TIntConstant (s->GetSize (), SymbolTable->TINT), MUL);
	switch (code) {
	    case ADD:
		code = IADD;
		break;
	    case SUB:
		code = ISUB;
	    default:
		break;
	}
    } else if (type->IsSet ()) 
	switch (op) {
	    case ADD:
		code = SETOR;
		break;
	    case SUB:
		code = 0;	// special treatment in OutputCode
	    default:
		break;
	}
    if (code == -1)
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot combine \'" & first0->GetType ()->Name () & "\' and \'" & second0->GetType ()->Name () & "\'");
}

TExpression::~TExpression () {
    delete first;
    delete second;
}

void TExpression::OutputCode () {
    if (type == SymbolTable->TBOOL && op == OR && Compiler->CodeBoolShortcuts) {
        first->OutputCode ();
	Compiler->OutCode (line, DUP, 0, 0);
	int FixupPos = Compiler->PCodeInterpreter->GetPos ();
	Compiler->OutCode (line, JNZ, 0, 0); 	// will be replaced later
	second->OutputCode ();
	Compiler->OutCode (line, IOR);
        Compiler->PCodeInterpreter->Fixup (FixupPos, JNZ, Compiler->PCodeInterpreter->GetPos ());
    } else {
	first->OutputCode ();
	second->OutputCode ();
	if (type->IsSet () && op == SUB) {
	    Compiler->OutCode (line, SETNEG);
	    Compiler->OutCode (line, SETAND);
	} else
	    Compiler->OutCode (line, code);
    }
}

TExpressionBase *TExpression::Parse () {
    TExpressionBase *p = Scanner->Try (SUB) ? new TNegativeTerm (TTerm::Parse ()) : TTerm::Parse ();
    while (Scanner->GetToken () == ADD || Scanner->GetToken () == SUB || Scanner->GetToken () == OR || Scanner->GetToken () == XOR) {
	TToken op = Scanner->GetToken ();
	Scanner->GetNextToken ();
	TExpressionBase *q = TTerm::Parse ();
	p = new TExpression (p, q, op);
    }
    return p;
}

// TTerm

TTerm::TTerm (TExpressionBase *first0, TExpressionBase *second0, TToken op0) {

    // op: MUL DIV AND DIVIDE MODULO

    first = first0;
    second = second0;
    op = op0;
    code = -1;

/*
    if (op == MUL && CheckTypeConversion (SymbolTable->TREALMATR, first)) {
        if (CheckTypeConversion (SymbolTable->TREAL, second)) {
	    type = SymbolTable->TREALMATR;
	    code = MSFMUL;
	} else if (CheckTypeConversion (SymbolTable->TREALVECT, second)) {
	    type = SymbolTable->TREALVECT;
	    code = MVFMUL;
	} else if (CheckTypeConversion (SymbolTable->TREALMATR, second)) {
	    type = SymbolTable->TREALMATR;
	    code = MFMUL;
	}
    } else if (op == MUL && second->GetType ()->IsMatrix () && second->GetType ()->GetBaseType () == SymbolTable->TREAL && CheckTypeConversion (SymbolTable->TREAL, first)) {
	// need scalar as second argument
	Swap (first, second);
	type = SymbolTable->TREALMATR;
	code = MSFMUL;
    } else {
*/

    type = TExpressionBase::CheckOperatorType (first, second, op == MUL);

    if (op == MUL && ((first->GetType ()->IsMatrix () && (second->GetType ()->IsMatrix () || second->GetType ()->IsVector ()))
                      || (first->GetType ()->IsVector () && second->GetType ()->IsMatrix ())))
	if (type->GetBaseType () == SymbolTable->TREAL)
	    code = MFMUL;
	else
	    code = MIMUL;


    // convert scalar into vector and apply vector operation
    if (code == -1 && type->IsMatrix ()) {
	TType *t;
	if (first->GetType ()->IsMatrix ()) {
	    t = ConvertToBaseType (second);
	    if (t == SymbolTable->TINT || t == SymbolTable->TREAL)
		second = new TMakeVector (second);
	} else if (second->GetType ()->IsMatrix ()) {
	    t = ConvertToBaseType (first);
     	    if (t == SymbolTable->TINT || t == SymbolTable->TREAL)
		first = new TMakeVector (first);
	}
    }

    if (code == -1) 
	if (type == SymbolTable->TBOOL) {
	    if (op == AND)
		code = IAND;	    
        } else if (type == SymbolTable->TINT)
            switch (op) {
                case AND:    
        	    code = IAND; 
        	    break;
                case MUL:    
        	    code = IMUL; 
        	    break;
            	case DIVIDE: 
        	    code = Compiler->CodeDivisionCheck ? IDIVC : IDIV; 
        	    break;
            	case MODULO: 
        	    code = Compiler->CodeDivisionCheck ? IMODC : IMOD; 
        	    break;
            	case DIV:
        	    first = new TConvertIntReal (first);
        	    second = new TConvertIntReal (second);
        	    type = SymbolTable->TREAL;
        	    code = Compiler->CodeDivisionCheck ? FDIVC : FDIV;
	   	default:
        	    break;
            }
        else if (type == SymbolTable->TREAL)
            switch (op) {
            	case MUL:
        	    code = FMUL; 
        	    break;
                case DIV: 
        	    code = Compiler->CodeDivisionCheck ? FDIVC : FDIV; 
		default:
        	    break;
            }
        else if (type == SymbolTable->TCOMPLEX)
            switch (op) {
            	case MUL:
        	    code = CMUL; 
        	    break;
                case DIV: 
        	    code = Compiler->CodeDivisionCheck ? CDIVC : CDIV; 
		default:
        	    break;
            }
        else if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TINT)
            switch (op) {
                case AND:
        	    code = VIAND;
        	    break;
                case MUL:
        	    code = VIMUL;
        	    break;
                case DIVIDE:
        	    code = Compiler->CodeDivisionCheck ? VIDIVC : VIDIV;
        	    break;
                case MODULO:
        	    code = Compiler->CodeDivisionCheck ? VIMODC : VIMOD;
        	    break;
                case DIV:
        	    code = Compiler->CodeDivisionCheck ? VIDIVFC : VIDIVF;
        	    type = SymbolTable->TREALVECT;
		default:
        	    break;
            }
        else if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TBOOL) {
            if (op == AND)
                code = VIAND;
        } else if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TREAL)
            switch (op) {
                case MUL:
        	    code = VFMUL;
        	    break;
                case DIV:
        	    code = Compiler->CodeDivisionCheck ? VFDIVC : VFDIV;
		default:
         	    break;
            }
        else if (type->IsSet () && op == MUL)
            code = SETAND;

    if (code == -1) 
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot combine " & first0->GetType ()->Name () & " and " & second0->GetType ()->Name ());
}

TTerm::~TTerm () {
    delete first;
    delete second;
}

void TTerm::OutputCode () {
    if (type == SymbolTable->TBOOL && op == AND && Compiler->CodeBoolShortcuts) {
        first->OutputCode ();
	Compiler->OutCode (line, DUP, 0, 0);
        int FixupPos =  Compiler->PCodeInterpreter->GetPos ();
        Compiler->OutCode (line, JZ, 0, 0);     // will be replaced later
        second->OutputCode ();
    	Compiler->OutCode (line, IAND);
        if (FixupPos) Compiler->PCodeInterpreter->Fixup (FixupPos, JZ, Compiler->PCodeInterpreter->GetPos ());
    } else {
	first->OutputCode ();
	second->OutputCode ();
        Compiler->OutCode (line, code);
    }
}

TExpressionBase *TTerm::Parse () {
    TExpressionBase *p = TFactor::Parse ();
    while (Scanner->GetToken () == MUL || Scanner->GetToken () == DIV || Scanner->GetToken () == DIVIDE ||
           Scanner->GetToken () == MODULO || Scanner->GetToken () == AND) {
	TToken op = Scanner->GetToken ();
	Scanner->GetNextToken ();
	TExpressionBase *q = TFactor::Parse ();
	p = new TTerm (p, q, op);
    }
    return p;
}

// TNegativeTerm

TNegativeTerm::TNegativeTerm (TExpressionBase *a0) {
    a = a0;
    if (CheckTypeConversion (SymbolTable->TINT, a))
	code = INEG;
    else if (CheckTypeConversion (SymbolTable->TREAL, a))
	code = FNEG;
    else if (CheckTypeConversion (SymbolTable->TINTVECT, a))
	code = VINEG;
    else if (CheckTypeConversion (SymbolTable->TREALVECT, a))
	code = VFNEG;
    else
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot apply negative sign to value of type \'" & a0->GetType ()->Name () & "\'");
    type = a->GetType ();
}

TNegativeTerm::~TNegativeTerm () {
    delete a;
}

void TNegativeTerm::OutputCode () {
    a->OutputCode ();
    Compiler->OutCode (line, code);
}   

// TPower

TPower::TPower (TExpressionBase *first0, TExpressionBase *second0) {
    first = first0;
    second = second0;
    TType *t = CheckOperatorType (first, second, false);
    if (t != SymbolTable->TINT)
 	    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), 
 		      "Cannot create vector from \'" & first->GetType ()->Name () & "\' and \'" & second->GetType ()->Name () & "\'");
    type = SymbolTable->TINTVECT;
}

TPower::~TPower () {
    delete first;
    delete second;
}

TExpressionBase *TPower::Parse () {
    TExpressionBase *a = TFunctionCall::Parse ();
    return Scanner->Try (TWOPOINT) ? 
	new TPower (a, TFunctionCall::Parse ()) : a;
}

void TPower::OutputCode () {
    first->OutputCode ();
    second->OutputCode ();
    Compiler->OutCode (line, VINT);
}	

// TNotFactor

TNotFactor::TNotFactor (TExpressionBase *p0) {
    p = p0;
    type = p->GetType ();
    if (type == SymbolTable->TBOOL) return;
    if ((type->IsVector () || type->IsMatrix ()) && type->GetBaseType () == SymbolTable->TBOOL) return;
    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Boolean or boolean vector expected after 'not'");
}

TNotFactor::~TNotFactor () {
    delete p;
}

void TNotFactor::OutputCode () {
    p->OutputCode ();
    Compiler->OutCode (line, type->IsVector () || type->IsMatrix () ? VINOT : INOT);
}

// TFactor

TFactor::TFactor (TExpressionBase *first0, TExpressionBase *second0) {
    first = first0;
    second = second0;
    code = -1;
    type = TExpressionBase::CheckOperatorType (first, second, false);
    if (type == SymbolTable->TREAL)
	code = FPOW;
    else if (type == SymbolTable->TINT)
	code = IPOW;
    else if (type->IsVector () && type->GetBaseType () == SymbolTable->TREAL)
	code = VFPOW;
    else if (type->IsVector () && type->GetBaseType () == SymbolTable->TINT)
	code = VIPOW;
    if (code == -1) 
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot combine " & first0->GetType ()->Name () & " and " & second0->GetType ()->Name () & " in power operation");
}

TFactor::~TFactor () {
    delete first;
    delete second;
}

void TFactor::OutputCode () {
    first->OutputCode ();
    second->OutputCode ();
    Compiler->OutCode (line, code); 
}

TExpressionBase *TFactor::Parse () {
    TExpressionBase *p = Scanner->Try (NOT) ? new TNotFactor (TPower::Parse ()) : TPower::Parse ();
    while (Scanner->Try (POW))
	p = new TFactor (p, TFunctionCall::Parse ());
    return p;
}

// TFunctionCall

TFunctionCall::TFunctionCall (TExpressionBase *f0, TList *ArgList0) {
    f = f0;
    type = ((TRoutine *) f->GetType ())->GetReturnType ();
    ImplicitLooping = FALSE;
    TRoutine *r = (TRoutine *) f->GetType ();
    int Parameters = r->GetNumberOfParameters (),
        ParsedParaCount = ArgList0 ? ArgList0->Entries () : 0;

    if (ParsedParaCount != Parameters) {
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Number of arguments does not match declaration");
	ArgList0->DeleteAll (TRUE);
	type = SymbolTable->TERROR;
 	return;
    }
    if (Parameters) {
	TType *t = r->GetFirstParaType ();
        TExpressionBase *p = (TExpressionBase *) ArgList0->GetFirst ();

	if (Parameters == 1 && (p->GetType ()->IsVector () || p->GetType ()->IsArray ()) &&
	    p->GetType ()->GetBaseType () == t) {
	    ImplicitLooping = TRUE;
	    if (p->GetType ()->IsArray ())
		p = new TMakeVector (p);
	    type = p->GetType ();
	    ArgList.Append (p);
	} else 
            while (Parameters--) {
                if (t->IsReference ()) {
            	    if (p->IsLValueDeref ()) p = ((TLValueDeref *) p)->UnDeref ();
            	    else ErrorMessage (IDENTIFIER_EXPECTED,  Scanner->GetLine (), "L-value required for reference parameter");
            	    if (t->GetBaseType () != p->GetType ()) 
			ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (),
            	   		"Expected reference of type \'" & t->GetBaseType ()->Name () & "\', but got \'" & p->GetType ()->Name () & "\'");
                } else
            	    DoTypeConversion (t, p);
                ArgList.Append (p);
                t = r->GetNextParaType ();
                p = (TExpressionBase *) ArgList0->GetNext ();
            }
    }
}

TFunctionCall::~TFunctionCall () {
    ArgList.DeleteAll (TRUE);
    delete f;
}

void TFunctionCall::OutputCode () {
    if (ImplicitLooping) {
	((TExpressionBase *) ArgList.GetFirst ())->OutputCode ();
	f->OutputCode ();
	Compiler->OutCode (line, ITERINIT);
	Compiler->OutCode (line, ITERDO);
	Compiler->OutCode (line, ITERLOOP);
    } else {
        for (TExpressionBase *p = (TExpressionBase *) ArgList.GetFirst (); p; p = (TExpressionBase *) ArgList.GetNext ()) {
	    p->OutputCode ();
 	    if (p->IsLValue ()) 
		Compiler->OutCode (line, PUSH, 1);
            else if (p->GetType ()->ContainsVector ())
		p->GetType ()->CodePush (line, Compiler);
	    else 
		Compiler->OutCode (line, PUSH, p->GetType ()->GetSize (), 0);
        }
        f->OutputCode ();
        Compiler->OutCode (line, CALL);	
    }
}

TExpressionBase *TFunctionCall::Parse () {
    TExpressionBase *p = TFunctionExpr::Parse ();
    while (Scanner->Try (BRACKETOPEN)) {
        TList ParsedArgList;
	do 
	    ParsedArgList.Append (TCompExpression::Parse ());
	while (Scanner->Try (COMMA));
	if (!Scanner->Try (BRACKETCLOSE)) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Function call missing closing ')'");
	    ParsedArgList.DeleteAll (TRUE);
	    return new TErrorExpr;
	}
	p = new TFunctionCall (p, &ParsedArgList);
    }	
    return p;
}
    
// TFunctionExpr

TExpressionBase *TFunctionExpr::Parse () {
    TExpressionBase *p;
    if (Scanner->GetToken () == IDENTIFIER) {
	return TVarAccess::Parse ();
    } else if (Scanner->GetToken () == DBLCONST) {
 	p = new TDblConstant (Scanner->GetDouble ());
	Scanner->GetNextToken ();
    } else if (Scanner->GetToken () == INTCONST) {
        p = new TIntConstant (Scanner->GetInteger (), SymbolTable->TINT);
        Scanner->GetNextToken ();
    } else if (Scanner->Try ( NIL)) 
	return new TIntConstant (0, SymbolTable->TGENERICPOINTER);
    else if (Scanner->GetToken () == CHARCONST) {
        p = new TIntConstant (Scanner->GetInteger (), SymbolTable->TCHAR);
        Scanner->GetNextToken ();
    } else if (Scanner->GetToken () == STRCONST) {
        p = new TStringConstant (Scanner->GetString ());
        Scanner->GetNextToken ();
    } else if (Scanner->Try (BRACKETOPEN)) {
	p = TCompExpression::Parse ();
	if (!Scanner->Try (BRACKETCLOSE)) 
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Bracketed expression missing ')'");
	while (Scanner->Try (SQUAREBRACKETOPEN)) {
	    p = new TVectorIndex (p, TCompExpression::Parse ());
	    if (!Scanner->Try (SQUAREBRACKETCLOSE))
		ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Vector index missing closing ']'");
	}
    } else if (Scanner->Try (SQUAREBRACKETOPEN)) {
 	p = new TSetExpression;
	((TSetExpression *) p)->Parse ();	
    } else {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Missing part in expression");
	p = new TErrorExpr;
    }
    return p;
}

// TVarAccess

TExpressionBase *TVarAccess::Parse () {
    TExpressionBase *p;
    String s = Scanner->GetString ();
    Scanner->GetNextToken ();
    TSymbol *Symbol = SymbolTable->Search (s, Scanner->GetLine (), FALSE);
    if (Symbol) 
        if (Symbol->IsConstant ())
    	    p = new TDeclaredConstant (Symbol);
        else if (Symbol->GetType ()->IsReference ())
    	    p = new TReferenceVariable (Symbol);
    	else
    	    p = new TLValue (Symbol);
    else {
        if (s == String ("write") || s == String ("writeln")) {
    	    p = new TWriteFunction (s == String ("writeln"));
    	    ((TWriteFunction *) p)->Parse ();
        } else if (s == String ("read") || s == String ("readln")) {
    	    p = new TReadFunction (s == String ("readln"));
    	    ((TReadFunction *) p)->Parse ();
        } else if (s == String ("combine")) {
    	    p = new TCombineFunction;
    	    ((TCombineFunction *) p)->Parse ();
	} else if (s == String ("cbind")) {
	    p = new TCBindFunction;
	    ((TCBindFunction *) p)->Parse ();
	} else if (s == String ("rbind")) {
	    p = new TRBindFunction;
	    ((TRBindFunction *) p)->Parse ();
        } else if (s == String ("scan")) {
    	    p = new TScanFunction;
    	    ((TScanFunction *) p)->Parse();
        } else p = TPredefinedFunction::Parse (s);
        if (!p) {
	    ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine (), s & " is undeclared");
	    return new TErrorExpr;
	}
	return p;
    }

    BOOL Goon = TRUE;
    while (Goon) 
        if (Scanner->Try (SQUAREBRACKETOPEN)) {
	    do {
	        TExpressionBase *q = TCompExpression::Parse ();
                p = p->GetType ()->IsVector () ? (TExpressionBase *) new TVectorIndex (p, q) : (TExpressionBase *) new TArrayIndex (p, q);
	    } while (Scanner->Try (COMMA));
            if (!Scanner->Try (SQUAREBRACKETCLOSE)) 
		ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Array index missing closing ']'");
	} else if (Scanner->Try (ENDPOINT)) 
	    p = (new TRecordIndex (p))->Parse ();
	else if (Scanner->Try (DEREF)) 
	    p = (new TPointerDereference (p))->Parse ();
        else Goon = FALSE;
    if (p->IsLValue ()) p = new TLValueDeref (p);
    return p;
}

// TConstant

BOOL TConstantBase::IsConstant () {
    return TRUE;
}

TDeclaredConstant::TDeclaredConstant (TSymbol *s0) {
    s = s0;
    type = s->GetType ();
}

void TDeclaredConstant::OutputCode () {
    if (type == SymbolTable->TREAL)
        Compiler->OutCode (line, FLOAD, s->GetDoubleVal ());
    else if (type == SymbolTable->TSTRING)
        Compiler->OutCode (line, LOADSTR, s->GetStringVal ());
    else
        Compiler->OutCode (line, LOAD, s->GetIntVal (), 0);
}

// TDblConstant

TDblConstant::TDblConstant (double val0) {
    val = val0;
    type = SymbolTable->TREAL;
}

void TDblConstant::OutputCode () {
    Compiler->OutCode (line, FLOAD, val);
}

// TStringConstant

TStringConstant::TStringConstant (String s0) {
    s = s0;
    type = SymbolTable->TSTRING;
}

void TStringConstant::OutputCode () {
    Compiler->OutCode (line, LOADSTR, s);
}

// TIntConstant

TIntConstant::TIntConstant (int val0, TType *type0) {
    val = val0;
    type = type0;
}

void TIntConstant::OutputCode () {
    Compiler->OutCode (line, LOAD, val);
}

// TConvertIntReal

TConvertIntReal::TConvertIntReal (TExpressionBase *p0) {
    p = p0;
    type = SymbolTable->TREAL;
}

TConvertIntReal::~TConvertIntReal () {
    delete p;
}

void TConvertIntReal::OutputCode () {
    p->OutputCode ();
    Compiler->OutCode (line, CIF);
}

// TConvertIntVectRealVect 

TConvertIntVectRealVect::TConvertIntVectRealVect (TExpressionBase *p0) {
    p = p0;
    type = SymbolTable->TREALVECT;
}

TConvertIntVectRealVect::~TConvertIntVectRealVect () {
    delete p;
}

void TConvertIntVectRealVect::OutputCode () {
    p->OutputCode ();
    Compiler->OutCode (line, VCIF);
}

// TConvertIntMatrixRealMatrix

TConvertIntMatrixRealMatrix::TConvertIntMatrixRealMatrix (TExpressionBase *p0) 
 : inherited (p0) {
    type = SymbolTable->TREALMATR;
}

//

TChangeType::TChangeType (TType *Req, TExpressionBase *p0) {
    type = Req;
    p = p0;
}

TChangeType::~TChangeType () {
    delete p;
}

void TChangeType::OutputCode () {
    p->OutputCode ();
}    

// TConvertSubRange

TConvertSubrange::TConvertSubrange (TType *Req, TExpressionBase *p0) {
    p = p0;
    type = Req;
}

TConvertSubrange::~TConvertSubrange () {
    delete p;
}

void TConvertSubrange::OutputCode () {
    p->OutputCode ();
    if (Compiler->CodeRangeCheck) {
	int a, b;
	type->GetIndexRange (a, b);
	Compiler->OutCode (line, CHECK, a, b);
    }
}

// TMakeVector

TMakeVector::TMakeVector (TExpressionBase *p0) {
    p = p0;
    if (p->GetType ()->IsArray ())
	type = new TVector (p->GetType ()->GetBaseType ());
    else
	type = new TVector (p->GetType ());
}

TMakeVector::~TMakeVector () {
    delete p;
}

void TMakeVector::OutputCode () {
    p->OutputCode ();
    TType *base = type->GetBaseType ();	// base type of vector
    Compiler->OutCode (line, LOADT, p->GetType ()->GetSize () / base->GetSize (), 1);
    Compiler->OutCode (line, LOADT, p->GetType ()->GetSize () + 1, base->GetSize ());
}

   
// TMakeMatrix

TMakeMatrix::TMakeMatrix (TExpressionBase *p0) {
    p = p0;
    type = new TSPMatrix (p->GetType ()->GetBaseType ()->GetBaseType ());
}

TMakeMatrix::~TMakeMatrix () {
    delete p;
}

void TMakeMatrix::OutputCode () {
    p->OutputCode ();
    int a, b, c, d;
    p->GetType ()->GetBaseType ()->GetIndexType ()->GetIndexRange (a, b);
    p->GetType ()->GetIndexType ()->GetIndexRange (c, d);
    Compiler->OutCode (line, LOADT, d - c + 1, b - a + 1);
    Compiler->OutCode (line, LOADT, p->GetType ()->GetSize () + 1, p->GetType ()->GetBaseType ()->GetBaseType ()->GetSize ());
}
   
// L-Values

TLValueDeref::TLValueDeref (TExpressionBase *p0) {
    p = p0;
    LoadVector = TRUE;
    type = p->GetType ();
}

void TLValueDeref::OutputCode () {
    if (p->IsSymbol () && p->GetType ()->GetSize () == 1) {
	((TLValue *) p)->OutputLoadVal (); 
        if (LoadVector && (p->GetType ()->IsVector () || p->GetType ()->IsMatrix ())) 
            Compiler->OutCode (line, VLOAD);
    } else {
        p->OutputCode ();
	if (LoadVector && p->GetType ()->ContainsVector ()) {
	    Compiler->OutCode (line, POPIND);
	    p->GetType ()->CodeLoad (line, Compiler);
	} else 
            Compiler->OutCode (line, LOADMEM, type->GetSize (), 0);
    }
}

BOOL TLValueDeref::IsLValueDeref () {
    return TRUE;
}

void TLValueDeref::SetLoadVector (BOOL f) {
    LoadVector = f;
}

TExpressionBase *TLValueDeref::UnDeref () {
    TExpressionBase *q = p;
    delete this;
    return q;
}

TLValue::TLValue (TSymbol *s0) {
    s = s0;
    type = s->GetType ();
}

void TLValue::OutputCode () {
    Compiler->OutCode (line, LOADPTR, s->GetAddress (), s->GetLevel ());
}

BOOL TLValue::IsLValue () {
    return TRUE;
}

BOOL TLValue::IsSymbol () {
    return TRUE;
}

void TLValue::OutputLoadVal () {
    Compiler->OutCode (line, LOADVAL, s->GetAddress (), s->GetLevel ());
}

void TLValue::OutputStoreVal () {
    Compiler->OutCode (line, STOREVAL, s->GetAddress (), s->GetLevel ());
}




TReferenceVariable::TReferenceVariable (TSymbol *s)
 : TLValue (s) {
    type = s->GetType ()->GetBaseType ();	// Base type of reference
}

void TReferenceVariable::OutputCode () {
    Compiler->OutCode (line, LOADVAL, s->GetAddress (), s->GetLevel ());
}

BOOL TReferenceVariable::IsSymbol () {
    return FALSE;
}


// TVectorIndex

TVectorIndex::TVectorIndex (TExpressionBase *vector0, TExpressionBase *index0) {
    BOOL Error = FALSE;
    LValue = FALSE;
    if (!vector0->GetType ()->IsVector ()) {
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot apply index operation to value of type " & vector0->GetType ()->Name ());
	Error = TRUE;
    }
    vector = vector0;
    if (vector->IsLValue ()) vector = new TLValueDeref (vector);
    index = index0;
    type = vector0->GetType ();
    if (Error) return;
    TType *IndexType = ConvertToBaseType (index);
    VectorForced = FALSE;
    if (IndexType == SymbolTable->TINT) {
	type = type->GetBaseType ();
	if (vector0->IsLValue ()) {	// keep L-value property of indexed vector for assignments
	    LValue = TRUE;
	    return;
	} else {
	    index = new TMakeVector (index);
	    IndexType = index->GetType ();
	    VectorForced = TRUE;
        }
    }
    if (IndexType->IsVector () && (IndexType->GetBaseType () == SymbolTable->TINT || IndexType->GetBaseType () == SymbolTable->TBOOL)) ;
    else ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Type \'" & IndexType->Name () & "\' not allowed as index type of vector");
}

TVectorIndex::~TVectorIndex () {
    delete vector;
    delete index;
}

BOOL TVectorIndex::IsLValue () {
    return LValue;
}

void TVectorIndex::OutputCode () {
    TType *indextype = index->GetType ();
    if (vector->IsLValueDeref ()) {
	((TLValueDeref *) vector)->SetLoadVector (FALSE);	// load only address of vector data
 	index->OutputCode ();
 	vector->OutputCode ();
	Compiler->OutCode (line, LOAD, vector->GetType ()->GetBaseType ()->GetSize ());
        if (LValue)						// L-value indexed by integer; load address of vector element
	    Compiler->OutCode (line, VLVINDINT);  
	else if (indextype->GetBaseType () == SymbolTable->TINT) 
	    Compiler->OutCode (line,  Compiler->CodeRangeCheck ? VINDINTC : VINDINT);
	else 
	    Compiler->OutCode (line, VINDBOOL);
    } else 
	if (indextype->GetBaseType () == SymbolTable->TINT) {
	    index->OutputCode ();
	    vector->OutputCode ();
	    Compiler->OutCode (line, LOAD, vector->GetType ()->GetBaseType ()->GetSize ());
	    Compiler->OutCode (line, Compiler->CodeRangeCheck ? VSINDINTC : VSINDINT);
	} else {
	    vector->OutputCode ();
	    index->OutputCode ();
	    Compiler->OutCode (line, LOAD, vector->GetType ()->GetBaseType ()->GetSize ());
	    Compiler->OutCode (line, VSINDBOOL);
    	}
    if (VectorForced) Compiler->OutCode (line, ADDCSP, -2);		// scalar expected, remove vector data
}


TArrayIndex::TArrayIndex (TExpressionBase *p0, TExpressionBase *index0) {
    p = p0;
    index = index0;
    TType *q = p->GetType ();
    if (q->IsArray ())
        IndexType = q->GetIndexType ();
    else if (q->IsPointer ())
        IndexType = SymbolTable->TCARDINAL;
    else {
        ErrorMessage (ARRAY_EXPECTED, Scanner->GetLine ());
	type = IndexType = SymbolTable->TERROR;
	return;
    }
    DoTypeConversion (IndexType, index);
    type = q->GetBaseType ();
}

TArrayIndex::~TArrayIndex () {
    delete p;
    if (index) delete index;
}

void TArrayIndex::OutputCode () {
    p->OutputCode ();
    if (p->GetType ()->IsPointer ()) {
        Compiler->OutCode (line, LOADMEM, 1);		// load pointer table index
	Compiler->OutCode (line, LOADMEM, 1);		// pointer table lookup
    }
    index->OutputCode ();
    int a, b;
    IndexType->GetIndexRange (a, b);
//    if (Compiler->CodeRangeCheck) Compiler->OutCode (line, CHECK, a, b);
    if (a || type->GetSize () > 1) 
	Compiler->OutCode (line, ISMA, a, type->GetSize ());
    else 
	Compiler->OutCode (line, IADD);
}

BOOL TArrayIndex::IsLValue () {
    return TRUE;
}

TRecordIndex::TRecordIndex (TExpressionBase *p0) {
    p = p0;
}

TRecordIndex::~TRecordIndex () {
    delete p;
}

TExpressionBase *TRecordIndex::Parse () {
    if (!p->GetType ()->IsRecord ()) {
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Applying selection to non-record type");
	type = SymbolTable->TERROR;
	return this;
    }
    if (Scanner->GetToken () != IDENTIFIER) {
	ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine ());
	type = SymbolTable->TERROR;
	return this;
    }
    if (!((TRecord *) p->GetType ())->GetComponent (Scanner->GetString (), offset, type, Scanner->GetLine ())) {
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), Scanner->GetString () & " is not a component of the record");
	type = SymbolTable->TERROR;
    }
    Scanner->GetNextToken ();
    return this;
}

void TRecordIndex::OutputCode () {
    p->OutputCode ();
    Compiler->OutCode (line, LOAD, offset);
    Compiler->OutCode (line, IADD);
}

BOOL TRecordIndex::IsLValue () {
    return TRUE;
}

TPointerDereference::TPointerDereference (TExpressionBase *p0) {
    p = p0;
    type = p->GetType ()->GetBaseType ();
}

TPointerDereference::~TPointerDereference () {
    delete p;
}

TExpressionBase *TPointerDereference::Parse () {
    if (!p->GetType ()->IsPointer ()) {
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Pointer dereference operator applied to non-pointer type \'" & p->GetType ()->Name () & "\'");
	type = SymbolTable->TERROR;
    } else
    	type = p->GetType ()->GetBaseType ();
    return this;
}

void TPointerDereference::OutputCode () {
    p->OutputCode ();
    Compiler->OutCode (line, LOADMEM, 1);
    Compiler->OutCode (line, LOADMEM, 1);
}

BOOL TPointerDereference::IsLValue () {
    return TRUE;
}

// TSetExpression

TSetExpression::~TSetExpression () {
    ArgList.DeleteAll (TRUE);
}

void TSetExpression::Parse () {
    type = SymbolTable->TEMPTYSET;
    if (Scanner->Try (SQUAREBRACKETCLOSE)) return;
    do {
	TExpressionBase *p = TCompExpression::Parse ();
	TType *s = TExpressionBase::ConvertToBaseType (p);
	if (type == SymbolTable->TEMPTYSET) type = new TSet (s);
	if (!s->IsEnumerated ()) ErrorMessage (INVALID_TYPE, Scanner->GetLine (), "Ordinal type required in set constructor.");
	ArgList.Append (p);
    } while (Scanner->Try (COMMA));
    if (!Scanner->Try (SQUAREBRACKETCLOSE)) {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Set expression not terminated with ']'");
	type = SymbolTable->TERROR;
    }
}

void TSetExpression::OutputCode () {
    Compiler->OutCode (line, SETCREATE);
    for (TExpressionBase *p = (TExpressionBase *) ArgList.GetFirst (); p; p = (TExpressionBase *) ArgList.GetNext ()) {
  	p->OutputCode ();
	if (Compiler->CodeRangeCheck) Compiler->OutCode (line, CHECK, 0, N_MAX_SETORD);
	Compiler->OutCode (line, SETINC);
    }
}

// predefined functions

// combine

TCombineFunction::TCombineFunction () {
}

TCombineFunction::~TCombineFunction () {
    ArgList.DeleteAll (TRUE);
}

void TCombineFunction::Parse () {
    if (!Scanner->Try (BRACKETOPEN)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'(' required in call to 'combine'");
    TExpressionBase *p = TCompExpression::Parse ();
    TType *t = TExpressionBase::ConvertToBaseType (p), *t1, *t2;
    ArgList.Append (p);
    if (t->IsVector ()) {
	t1 = t; t2 = t->GetBaseType ();
    } else {
	t1 = new TVector (t); t2 = t;
    }
    while (Scanner->Try (COMMA)) {
	p = TCompExpression::Parse ();
	TType *s = TExpressionBase::ConvertToBaseType (p);
	if (s->IsVector ()) TExpressionBase::DoTypeConversion (t1, p);
	else TExpressionBase::DoTypeConversion (t2, p);
	ArgList.Append (p);
    }
    if (!Scanner->Try (BRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "')' required in call to 'combine'");
    type = t1;
}

void TCombineFunction::OutputCode () {
    BOOL First = TRUE;
    for (TExpressionBase *p = (TExpressionBase *) ArgList.GetFirst (); p; p = (TExpressionBase *) ArgList.GetNext ()) {
  	p->OutputCode ();
	if (First) {
	    First = FALSE;
	    if (!p->GetType ()->IsVector ()) {
		Compiler->OutCode (line, LOADT, 1, 1);
		Compiler->OutCode (line, LOADT, p->GetType ()->GetSize () + 1, p->GetType ()->GetSize ());
	    }
 	} else
	    if (p->GetType ()->IsVector ()) 
		Compiler->OutCode (line, VCVEC);
	    else 
		Compiler->OutCode (line, VCVAL, p->GetType ()->GetSize ());
    }
}

// cbind

TBindFunction::~TBindFunction () {
    ArgList.DeleteAll (TRUE);
}

void TBindFunction::Parse (String msg) {
    if (!Scanner->Try (BRACKETOPEN)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'(' required in call to '" & msg & "'");
    TExpressionBase *p = TCompExpression::Parse ();
    TType *t = TExpressionBase::ConvertToBaseType (p), *t1 = NULL, *t2 = NULL, *t3;
    ArgList.Append (p);
    if (t->IsMatrix ()) {
	t1 = t; t2 = t->GetBaseType ();
    } else if (t->IsVector ()) {
	t2 = t->GetBaseType ();
	t1 = new TSPMatrix (t2);
    } else ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Matrix or vector required in call to '" & msg & "'");
    t3 = new TVector (t2);
    while (Scanner->Try (COMMA)) {
	p = TCompExpression::Parse ();
	if (!CheckTypeConversion (t3, p))
	    TExpressionBase::DoTypeConversion (t1, p);
	ArgList.Append (p);
    }
    if (!Scanner->Try (BRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "')' required in call to '" & msg & "'");
    type = t1;
}

void TBindFunction::ProduceCode (TCode code) {
    BOOL First = TRUE;
    for (TExpressionBase *p = (TExpressionBase *) ArgList.GetFirst (); p; p = (TExpressionBase *) ArgList.GetNext ()) {
  	p->OutputCode ();
	if (First) 
	    First = FALSE;
 	else
	    Compiler->OutCode (line, code);
    }
}

void TCBindFunction::Parse () {
    inherited::Parse ("cbind");
}

void TRBindFunction::OutputCode () {
    inherited::ProduceCode (MRBIND);
}

void TRBindFunction::Parse () {
    inherited::Parse ("rbind");
}

void TCBindFunction::OutputCode () {
    inherited::ProduceCode (MCBIND);
}

// scan

TScanFunction::TScanFunction () {
    Count = 0;
    formatstr = NULL;
    scanstr = NULL;
}

TScanFunction::~TScanFunction () {
    if (formatstr) delete formatstr;
    if (scanstr) delete scanstr;
    for (int i = 0; i < Count; i++)
	delete arg [i];
}

void TScanFunction::Parse () {
    if (!Scanner->Try (BRACKETOPEN)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'(' expected in call to 'scan'");
    scanstr = TCompExpression::Parse ();
    if (scanstr->GetType () != SymbolTable->TSTRING) 
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "First argument of 'scan' must be a string");
    if (!Scanner->Try (COMMA)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "',' expected in call to 'scan'");
    formatstr = TCompExpression::Parse ();
    if (formatstr->GetType () != SymbolTable->TSTRING) 
	ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "First argument of 'scan' must be a string");
    if (!Scanner->Try (COMMA)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "',' expected in call to 'scan'");
    Count = 0;
    do {
	if (Count == 30) ErrorMessage (PARAMETERTYPE_NOT_ALLOWED, Scanner->GetLine (), "Too many arguments in 'scan'");
	else {
	    arg [Count] = TCompExpression::Parse ();
	    if (!arg [Count]->IsLValueDeref () ||
	        (arg [Count]->GetType () != SymbolTable->TINT &&
                 arg [Count]->GetType () != SymbolTable->TREAL &&
                 arg [Count]->GetType () != SymbolTable->TCHAR))
	        ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Argument type " & arg [Count]->GetType ()->Name () & " not allowed in 'scan'");
            arg [Count] = ((TLValueDeref *) arg [Count])->UnDeref ();
	    Count++;
	}
    } while (Scanner->Try (COMMA));
    if (!Scanner->Try (BRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "')' expected in call to 'scan'");
    type = SymbolTable->TVOID;
}

void TScanFunction::OutputCode () {
    scanstr->OutputCode ();
    formatstr->OutputCode ();
    for (int i = Count - 1; i >= 0; i--) 
	arg [i]->OutputCode ();
    Compiler->OutCode (line, LOAD, Count);
    Compiler->OutCode (line, SCAN);
}

// write

TWriteItem::TWriteItem (TExpressionBase *exp0, TExpressionBase *length0, TExpressionBase *digits0, TExpressionBase *columns0) {
    exp = exp0; length = length0; digits = digits0; columns = columns0;
}

TWriteItem::~TWriteItem () {
    delete exp;
    if (length) delete length;
    if (digits) delete digits;
    if (columns) delete columns;
}

void TWriteItem::OutputCode () {
    exp->OutputCode ();
    if (length) length->OutputCode ();
    else Compiler->OutCode (line, LOAD, -1);
    if (digits) digits->OutputCode ();
    else Compiler->OutCode (line, LOAD, -1);
    if (columns) columns->OutputCode ();
    else Compiler->OutCode (line, LOAD, -1);
    if (exp->GetType () == SymbolTable->TINT) Compiler->OutCode (line, PRINTINT);
    else if (exp->GetType () == SymbolTable->TREAL) Compiler->OutCode (line, PRINTDBL);
    else if (exp->GetType () == SymbolTable->TCHAR) Compiler->OutCode (line, PRINTCHAR);
    else if (exp->GetType () == SymbolTable->TBOOL) Compiler->OutCode (line, PRINTBOOL);
    else if (exp->GetType () == SymbolTable->TCOMPLEX) Compiler->OutCode (line, PRINTC);
    else if (exp->GetType ()->IsVector ()) {
	if (exp->GetType ()->GetBaseType () == SymbolTable->TINT) Compiler->OutCode (line, VPRINTINT);
        else if (exp->GetType ()->GetBaseType () == SymbolTable->TBOOL) Compiler->OutCode (line, VPRINTBOOL);
        else if (exp->GetType ()->GetBaseType () == SymbolTable->TREAL) Compiler->OutCode (line, VPRINTDBL);
	else if (exp->GetType ()->GetBaseType () == SymbolTable->TCHAR) Compiler->OutCode (line, PRINTSTR);
    } else if (exp->GetType ()->IsMatrix ()) {
        if (exp->GetType ()->GetBaseType () == SymbolTable->TREAL) Compiler->OutCode (line, MPRINTDBL);
        if (exp->GetType ()->GetBaseType () == SymbolTable->TINT) Compiler->OutCode (line, MPRINTINT);
    }
}

TWriteFunction::TWriteFunction (BOOL WriteLine0) {
    WriteLine = WriteLine0;
    type = SymbolTable->TVOID;
}

TWriteFunction::~TWriteFunction () {
    ArgList.DeleteAll (TRUE);
}

void TWriteFunction::Parse () {
    BOOL First = TRUE;
    if (Scanner->Try (BRACKETOPEN)) {
        do {
    	    TExpressionBase *p = TCompExpression::Parse (), 
		            *length = NULL, *digits = NULL, *columns = NULL;
            if (First && p->GetType () == SymbolTable->TTEXT) 
		ArgList.Append (p);
	    else if (CheckTypeConversion (SymbolTable->TBOOL, p) ||
                CheckTypeConversion (SymbolTable->TCHAR, p) ||
    	    	CheckTypeConversion (SymbolTable->TINT, p) ||
    	    	CheckTypeConversion (SymbolTable->TREAL, p) ||
                CheckTypeConversion (SymbolTable->TCOMPLEX, p) ||
    	    	CheckTypeConversion (SymbolTable->TSTRING, p) ||
    	    	CheckTypeConversion (SymbolTable->TBOOLVECT, p) ||
    	    	CheckTypeConversion (SymbolTable->TINTVECT, p) ||
    	    	CheckTypeConversion (SymbolTable->TREALVECT, p) ||
		CheckTypeConversion (SymbolTable->TBOOLMATR, p) ||
		CheckTypeConversion (SymbolTable->TINTMATR, p) || 
		CheckTypeConversion (SymbolTable->TREALMATR, p)) {
		    if (Scanner->Try (COLON)) {
			length = TCompExpression::Parse ();			
			DoTypeConversion (SymbolTable->TINT, length);
		    }
		    if (Scanner->Try (COLON)) {
			digits = TCompExpression::Parse ();
			DoTypeConversion (SymbolTable->TINT, digits);
		    }
		    if (Scanner->Try (COLON)) {
			columns = TCompExpression::Parse ();
			DoTypeConversion (SymbolTable->TINT, columns);
		    }
		ArgList.Append (new TWriteItem (p, length, digits, columns));
	    }
	    else ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot print value of type " & p->GetType ()->Name ());
	    First = FALSE;
	} while (Scanner->Try (COMMA));
	if (!Scanner->Try (BRACKETCLOSE))
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Missing ')' in call to 'write(ln)'");
    }
}

void TWriteFunction::OutputCode () {
    if (ArgList.Entries () || WriteLine) {
	int Done = 0;
	TExpressionBase *p = (TExpressionBase *) (ArgList.Entries () ? ArgList.GetFirst () : NULL);
	if (p && p->GetType () == SymbolTable->TTEXT) {
	    p->OutputCode ();
	    p = (TExpressionBase *) ArgList.GetNext ();
	    Done++;
	} else {
	    TSymbol *s = SymbolTable->Search ("output", 0, TRUE, 0);
	    Compiler->OutCode (line, LOADVAL, s->GetAddress (), s->GetLevel ());
        }
	for ( ; Done < ArgList.Entries (); Done++) {
	    if (Done < ArgList.Entries () - 1 || WriteLine) Compiler->OutCode (line, DUP, 0, 0);
	    p->OutputCode ();
	    p = (TExpressionBase *) ArgList.GetNext ();
	}
    }
    if (WriteLine) Compiler->OutCode (line, PRINTLF); 
}

TReadFunction::TReadFunction (BOOL ReadLine0) {
    FileGiven = NULL;
    ReadLine = ReadLine0;
    type = SymbolTable->TVOID;
}

TReadFunction::~TReadFunction () {
    if (FileGiven) delete FileGiven;
    ArgList.DeleteAll (TRUE);
}

void TReadFunction::Parse () {
    if (Scanner->Try (BRACKETOPEN)) {
	BOOL Goon = TRUE;
	TExpressionBase *p = TCompExpression::Parse ();
	if (p->GetType () == SymbolTable->TTEXT) {
	    FileGiven = p;
	    Goon = Scanner->Try (COMMA);
	    if (Goon) p = TCompExpression::Parse ();	
	}
	while (Goon) {
     	    if (p->IsLValueDeref ()) p = ((TLValueDeref *) p)->UnDeref ();
	    else ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Variable parameter required as argument of 'readln'");
	    if (p->GetType () != SymbolTable->TINT && p->GetType () != SymbolTable->TREAL &&
		p->GetType () != SymbolTable->TCHAR && p->GetType () != SymbolTable->TSTRING)
		ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Cannot read value of type " & p->GetType ()->Name ());
	    ArgList.Append (p);
	    Goon = Scanner->Try (COMMA);
	    if (Goon) p = TCompExpression::Parse ();
	}
	if (!Scanner->Try (BRACKETCLOSE)) 
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Missing ')' in call to 'read'");
    }
}

void TReadFunction::OutputCode () {
    if (FileGiven) 
	FileGiven->OutputCode ();
    else {
	TSymbol *s = SymbolTable->Search ("input", 0, TRUE, 0);
	Compiler->OutCode (line, LOADVAL, s->GetAddress (), s->GetLevel ());
    }
    for (TExpressionBase *p = (TExpressionBase *) ArgList.GetFirst (); p; p = (TExpressionBase *) ArgList.GetNext ()) {
	p->OutputCode ();
	Compiler->OutCode (line, DUP, 1);
	TType *t = p->GetType ();
        if (t == SymbolTable->TINT) Compiler->OutCode (line, GETINT);
        else if (t == SymbolTable->TCHAR) Compiler->OutCode (line, GETCHAR);
        else if (t == SymbolTable->TREAL) Compiler->OutCode (line, GETDBL);
        else if (t == SymbolTable->TSTRING) Compiler->OutCode (line, GETSTR);
        if (t == SymbolTable->TSTRING) Compiler->OutCode (line, VSTORE);
        else Compiler->OutCode (line, STOREMEM, 1, 0);
    }
    if (ReadLine) Compiler->OutCode (line, GETLF);
    else Compiler->OutCode (line, ADDCSP, -1);
}

TPredefinedFunction::TPredefinedFunction (int Code0, TExpressionBase **a0, TType *ReturnType, BOOL ImplInput, BOOL ImplOutput) {
    Code = Code0;
    type = ReturnType;
    for (int i = 0; i < ARGMAX; i++)
	a [i] = a0 [i];
    ImplicitInput = ImplInput;
    ImplicitOutput = ImplOutput;
}

TPredefinedFunction::~TPredefinedFunction () {
    for (int i = 0; i < ARGMAX; i++)
	if (a [i]) delete a [i];
}

TExpressionBase *TPredefinedFunction::Parse (String Name) {
    TExpressionBase *ArgVector [ARGMAX] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
    if (Scanner->GetToken () == BRACKETOPEN) {
	Scanner->GetNextToken ();
	int i = 0;
	BOOL Goon;
	do {
	    ArgVector [i++] = TCompExpression::Parse ();
	    Goon = (Scanner->GetToken () == COMMA);
	    if (Goon && i < ARGMAX) Scanner->GetNextToken ();
	} while (Goon && i < ARGMAX);
	if (!Scanner->Try (BRACKETCLOSE)) 
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Missing ')' in call to '" & Name & "'");
    }
    int Code, ArgCheckCode;
    TType *ReturnType;
    BOOL ImplInput, ImplOutput;
    PredefinedFunctionList->Get (Name, ArgVector, Code, ArgCheckCode, ReturnType, ImplInput, ImplOutput);
    if (Code > 0)
	return new TPredefinedFunction (Compiler->CodeArgCheck ? ArgCheckCode : Code, ArgVector, ReturnType, ImplInput, ImplOutput);
    else if (Code == -1) { 
	ErrorMessage (PARAMETERTYPE_NOT_ALLOWED, Scanner->GetLine (), "\"" & Name & "\" is called with wrong arguments.");
	return new TErrorExpr;
    } else if (Name == String ("size") || Name == String ("rev") || Name == String ("ord") || Name == String ("chr") ||
             Name == String ("new") || Name == String ("dispose") || Name == String ("succ") || Name == String ("pred") ||
	     Name == String ("inc") || Name == String ("dec"))
	    return (new TSpecialFunction (Name, ArgVector))->Parse ();
    return NULL;
}

void TPredefinedFunction::OutputCode () {
    if (ImplicitInput) {
        TSymbol *s = SymbolTable->Search ("input", 0, TRUE, 0);
        Compiler->OutCode (line, LOADVAL, s->GetAddress (), s->GetLevel ());
    } 
    if (ImplicitOutput) {
        TSymbol *s = SymbolTable->Search ("output", 0, TRUE, 0);
        Compiler->OutCode (line, LOADVAL, s->GetAddress (), s->GetLevel ());
    }
    for (int i = 0; i < ARGMAX; i++)
	if (a [i]) a [i]->OutputCode ();
    Compiler->OutCode (line, Code);
}

TSpecialFunction::TSpecialFunction (String Name0, TExpressionBase **a0) {
    ArgCount = 0;
    for (int i = 0; i < ARGMAX; i++) {
        a [i] = a0 [i];
	if (a [i]) ArgCount++;
    }
    Name = Name0;
}

TSpecialFunction::~TSpecialFunction () {
    for (int i = 0; i < ARGMAX; i++)
        if (a [i]) delete a [i];
}

TExpressionBase *TSpecialFunction::Parse () {
    if (Name == String ("size") && ArgCount == 1 &&
	(a [0]->GetType ()->IsArray () || a [0]->GetType ()->IsVector ())) {
	type = SymbolTable->TINT;
	return this;
    }
    if (Name == String ("rev") && ArgCount == 1) {
	if (a [0]->GetType ()->IsArray ())
	    a [0] = new TMakeVector (a [0]);
	if (a [0]->GetType ()->IsVector ()) {
	    type = a [0]->GetType ();
	    return this;
	}
    }
    if (Name == String ("new") && (ArgCount == 1 || ArgCount == 2)) {
	if (a [0]->IsLValueDeref () && a [0]->GetType ()->IsPointer ()) a [0] = ((TLValueDeref *) a [0])->UnDeref ();
	else ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Pointer variable required as first argument of 'new'");
	if (ArgCount == 2) TExpressionBase::DoTypeConversion (SymbolTable->TINT, a [1]);
	type = SymbolTable->TVOID;
        return this;
    }
    if (Name == String ("dispose") && ArgCount == 1 && a [0]->GetType ()->IsPointer ()) {
	type = SymbolTable->TVOID;
	return this;
    }
    if (Name == String ("ord") && ArgCount == 1)  {
	if (a [0]->GetType ()->IsEnumerated ()) {
	    type = SymbolTable->TINT;
            return this;
	} else if (a [0]->GetType ()->IsVector () && a [0]->GetType ()->GetBaseType ()->IsEnumerated ()) {
	    type = SymbolTable->TINTVECT;
            return this;
        }
    }
    if ((Name == String ("succ") || Name == String ("pred")) && ArgCount == 1) {
        if (a [0]->GetType ()->IsEnumerated () ||
            (a [0]->GetType ()->IsVector () && a [0]->GetType ()->GetBaseType ()->IsEnumerated ())) {
	    type = a [0]->GetType ();
            return this;
        }
    }
    if ((Name == String ("inc") || Name == String ("dec")) && ArgCount == 1) {
	if (a [0]->IsLValueDeref ()) a [0] = ((TLValueDeref *) a [0])->UnDeref ();
	if (a [0]->GetType ()->IsEnumerated ()) {
	    type = SymbolTable->TVOID;
            return this;
    	}
    }
    if (Name == String ("chr") && ArgCount == 1 && CheckTypeConversion (SymbolTable->TINT, a [0])) {
	type = SymbolTable->TCHAR;
	return this;
    }
    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (),
	      "Invalid types in call of \'" & Name & "\'");
    return NULL;
}

void TSpecialFunction::OutputCode () {
    TType *t = a [0]->GetType ();
    if (Name == String ("size")) {
	if (t->IsArray ()) {
	    int a, b;
	    t->GetIndexType ()->GetIndexRange (a, b);
	    Compiler->OutCode (line, LOAD, b - a + 1);
	} else if (t->IsVector ()) {
	    if (a [0]->IsLValueDeref ())
		((TLValueDeref *) a [0])->SetLoadVector (FALSE);       // load only address of vector data
	    a [0]->OutputCode ();
	    Compiler->OutCode (line, a [0]->IsLValueDeref () ? VSIZE : VSSIZE);
	}
    } else if (Name == String ("rev")) {
	a [0]->OutputCode ();
	Compiler->OutCode (line, VREV);
    } else if (Name == String ("addr"))
        a [0]->OutputCode ();
    else if (Name == String ("new")) {
	a [0]->OutputCode ();
        Compiler->OutCode (line, LOAD, a [0]->GetType ()->GetBaseType ()->GetSize ());
	if (ArgCount == 2) {
	    a [1]->OutputCode ();
	    if (a [0]->GetType ()->GetBaseType ()->ContainsVector ()) {
	        Compiler->OutCode (line, DUP);
		Compiler->OutCode (line, LOADLS);
	    }
	    Compiler->OutCode (line, IMUL);
	}
	Compiler->OutCode (line, ALLOC);
	
	if (a [0]->GetType ()->GetBaseType ()->ContainsVector ()) {
	    Compiler->OutCode (line, DUP);
	    Compiler->OutCode (line, LOADMEM, 1);
	    Compiler->OutCode (line, POPIND);
	    int pos = 0;
	    if (ArgCount == 2) pos = Compiler->GetPos ();
	    a [0]->GetType ()->GetBaseType ()->CodeConstructor (line, Compiler);
	    if (ArgCount == 2) Compiler->OutCode (line, LOOP, pos);
	}

	Compiler->OutCode (line, STOREMEM, 1);
    } else if (Name == String ("dispose")) {
	a [0]->OutputCode ();
	// Achtung: new (x, n) nicht behandelt !!!!
	if (a [0]->GetType ()->GetBaseType ()->ContainsVector ()) {
	    Compiler->OutCode (line, DUP);
	    Compiler->OutCode (line, LOADMEM, 1);
	    Compiler->OutCode (line, POPIND);
	    a [0]->GetType ()->GetBaseType ()->CodeDestructor (line, Compiler);
	}
	Compiler->OutCode (line, DISPOSE);
    } else if (Name == String ("ord") || Name == String ("chr"))
	a [0]->OutputCode ();
    else if (Name == String ("succ")) {
        a [0]->OutputCode ();
        if (a [0]->GetType ()->IsVector ()) {
            Compiler->OutCode (line, LOAD, 1);
            Compiler->OutCode (line, LOAD, 1);
            Compiler->OutCode (line, VIADD);
        } else {
            Compiler->OutCode (line, LOAD, 1);
            Compiler->OutCode (line, IADD);
        }
    } else if (Name == String ("pred")) {
        a [0]->OutputCode ();
        if (a [0]->GetType ()->IsVector ()) {
            Compiler->OutCode (line, LOAD, 1);
            Compiler->OutCode (line, LOAD, 1);
            Compiler->OutCode (line, VISUB);
        } else {
            Compiler->OutCode (line, LOAD, 1);
            Compiler->OutCode (line, ISUB);
        }
    } else if (Name == String ("inc")) {
	a [0]->OutputCode ();
	Compiler->OutCode (line, INCPTR, 1);
    } else if (Name == String ("dec")) {
        a [0]->OutputCode ();
        Compiler->OutCode (line, INCPTR, -1);
    }
}


TPredefinedFunctionList::TPredefinedFunctionList () {
    Count = 0;
}

TPredefinedFunctionList::~TPredefinedFunctionList () {
}

void TPredefinedFunctionList::Append (String Name, int Code, int ArgCheckCode, TType *ReturnType0, 
	 TType *Arg1, TType *Arg2, TType *Arg3, TType *Arg4, TType *Arg5, TType *Arg6, TType *Arg7, TType *Arg8) {
    assert (Count < PREMAX);
    ListName [Count] = Name;
    ReturnType [Count] = ReturnType0;
    ListCode [Count] = Code;
    ListArgCheckCode [Count] =  ArgCheckCode;
    ListType [Count][0] = Arg1;
    ListType [Count][1] = Arg2;
    ListType [Count][2] = Arg3;
    ListType [Count][3] = Arg4;
    ListType [Count][4] = Arg5;
    ListType [Count][5] = Arg6;
    ListType [Count][6] = Arg7;
    ListType [Count][7] = Arg8;
    for (int i = 0; i < ARGMAX; i++)
        IsReference [Count][i] = FALSE;
    ImplicitInput [Count] = FALSE;
    ImplicitOutput [Count] = FALSE;
    Count++;
}

void TPredefinedFunctionList::SetReference (BOOL IsRef1, BOOL IsRef2, BOOL IsRef3, BOOL IsRef4, BOOL IsRef5, BOOL IsRef6, BOOL IsRef7, BOOL IsRef8) {
    Count--;
    IsReference [Count][0] = IsRef1;
    IsReference [Count][1] = IsRef2;
    IsReference [Count][2] = IsRef3;
    IsReference [Count][3] = IsRef4;
    IsReference [Count][4] = IsRef5;
    IsReference [Count][5] = IsRef6;
    IsReference [Count][6] = IsRef7;
    IsReference [Count][7] = IsRef8;
    Count++;
}

void TPredefinedFunctionList::CodeImplicitInput () {
    ImplicitInput [Count - 1] = TRUE;
}

void TPredefinedFunctionList::CodeImplicitOutput () {
    ImplicitOutput [Count - 1] = TRUE;
}

void TPredefinedFunctionList::Get (String Name, TExpressionBase **Arg, int &Code, int &ArgCheckCode, TType *&ReturnType0, BOOL &ImplInput, BOOL &ImplOutput) {
    BOOL NameFound = FALSE;
    Code = 0;
    TType *type [ARGMAX];
    int i;
    for (i = 0; i < ARGMAX; i++) 
	type [i] = Arg [i] ? TExpressionBase::ConvertToBaseType (Arg [i]) : (TType *) NULL;
    for (i = 0; i < Count; i++) 
	if (Name == ListName [i]) {
	    NameFound = TRUE;
	    if (CheckTypes (Arg, ListType [i], IsReference [i])) {
	        Code = ListCode [i];
		ArgCheckCode = ListArgCheckCode [i];
	        ReturnType0 = ReturnType [i];
		ImplInput = ImplicitInput [i];
		ImplOutput = ImplicitOutput [i];
	        return;
	    }
	}
    if (NameFound) Code = -1;
}

BOOL TPredefinedFunctionList::CheckTypes (TExpressionBase **a, TType **wanted, BOOL *IsRef) {
    for (int i = 0; i < ARGMAX; i++)  {
        if (wanted [i] == NULL || a [i] == NULL)
            return (void *) wanted [i] == (void *) a [i];
        if (IsRef [i]) {
	    if (!a [i]->IsLValueDeref ()) return FALSE;
	    if (a [i]->GetType () != wanted [i]) return FALSE;
	    a [i] = ((TLValueDeref *) a [i])->UnDeref ();
        } else 
	    if (!TExpressionBase::CheckTypeConversion (wanted [i], a [i])) return FALSE;
    }
    return TRUE;
}

void TErrorExpr::OutputCode () {
}

TType *TErrorExpr::GetType () {
    return SymbolTable->TERROR;
}

/*********************************************************/
/*	Implementation of Statements			 */
/*********************************************************/

TStatementBase::TStatementBase () {
    line = Scanner->GetLine ();
}

TStatementBase::~TStatementBase () {
}

TStatementBase *TStatementBase::Parse () {
    if (Scanner->Try (BEGIN)) 	return (new TStatementSequence)->Parse ();
    if (Scanner->Try (IF))	return (new TIfStatement)->Parse ();
    if (Scanner->Try (REPEAT))	return (new TRepeatStatement)->Parse ();
    if (Scanner->Try (WHILE))	return (new TWhileStatement)->Parse ();
    if (Scanner->Try (FOR))	return (new TForStatement)->Parse ();
    if (Scanner->Try (RETURN))  return (new TReturnStatement)->Parse ();
    if (Scanner->Try (CASE))	return (new TCaseStatement)->Parse ();
    if (Scanner->Try (GOTO))	return (new TGotoStatement)->Parse ();
    if (Scanner->GetToken () == IDENTIFIER || Scanner->GetToken () == INTCONST) {
	TSymbol *s;
	if (Scanner->GetToken () == IDENTIFIER)
	    s = SymbolTable->Search (Scanner->GetString (), Scanner->GetLine (), FALSE);
	else 
	    s = SymbolTable->Search (form (50, "%d", Scanner->GetInteger ()), Scanner->GetLine (), FALSE);
	if (s && s->IsLabel ())
	    return (new TLabelDefinition (s))->Parse ();
	else
	    return (new TSimpleStatement)->Parse ();
    }
    return new TEmptyStatement;

}

// empty statement

void TEmptyStatement::OutputCode () {
}

// simple statement

TSimpleStatement::TSimpleStatement () {
    first = second = NULL;
}

TSimpleStatement::~TSimpleStatement () {
    if (first) delete first;
    if (second) delete second;
}

TStatementBase *TSimpleStatement::Parse () {
    String id = Scanner->GetString ();
    first = TCompExpression::Parse ();
    if (Scanner->Try (DEFINE))
	second = TCompExpression::Parse ();
    if (!second)
	if (TExpressionBase::ConvertToBaseType (first) == SymbolTable->TVOID) return this;
	else ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Procedure call expected");
    else if (first->IsLValueDeref ()) {
	first = ((TLValueDeref *) first)->UnDeref ();
	TExpressionBase::DoTypeConversion (first->GetType (), second);
	return this;
    } else { // assignment to function name
	TSymbol *s = SymbolTable->Search (id, 0, FALSE);
	if (first->IsConstant () && s) {
	    TType *t = s->GetType ();
	    if (s->IsConstant () && t->IsRoutine ()) {
	        TType *u = ((TRoutine *) t)->GetReturnType ();
	        if (u != SymbolTable->TVOID) {
		    delete first;
		    first = new TLValue (SymbolTable->Search ("$" & id, Scanner->GetLine ()));
		    return this;
	        }
	    }
	}
	ErrorMessage (INVALID_USE_OF_SYMBOL, Scanner->GetLine (), "Assignment or function called expected");
    }
    return this;
}

void TSimpleStatement::OutputCode () {
// store optimization !!!!
    if (first->IsSymbol () && first->GetType ()->GetSize () == 1 && second && !second->GetType ()->IsVector () && !second->GetType ()->IsMatrix ()) {
	second->OutputCode ();
	((TLValue *) first)->OutputStoreVal ();
	return;
    }
//	
    TType *t = NULL;
    if (second) {
        t = second->GetType ();
        if (t->ContainsVector () && !t->IsVector ()) {
	    second->OutputCode ();
	    first->OutputCode ();
	    Compiler->OutCode (line, POPIND);
	    Compiler->OutCode (line, ADDIND, t->GetSize () - 1);
	    second->GetType ()->CodeSave (line, Compiler);
	    return;
	}
    }

    first->OutputCode ();
    if (second) {
	second->OutputCode ();
	if (t->IsVector ())
	    Compiler->OutCode (line, VSTORE);
	else
	    Compiler->OutCode (line, STOREMEM, t->GetSize ());
    }
}

// if

TIfStatement::TIfStatement () {
    expr = NULL;
    st1 = st2 = NULL;
}

TIfStatement::~TIfStatement () {
    if (expr) delete expr;
    if (st1) delete st1;
    if (st2) delete st2;
}

TStatementBase *TIfStatement::Parse () {
    expr = TCompExpression::Parse ();
    if (!TExpressionBase::CheckTypeConversion (SymbolTable->TBOOL, expr)) ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Boolean required in condition of 'if'-statement");
    if (!Scanner->Try (THEN)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'then' expected");
    st1 = TStatementBase::Parse ();
    if (Scanner->Try (ELSE))
	st2 = TStatementBase::Parse ();
    return this;
}
    
void TIfStatement::OutputCode () {
    expr->OutputCode ();
    int FixupPos1 = Compiler->GetPos (), FixupPos2 = 0;
    Compiler->OutCode (line, JZ, 0);
    st1->OutputCode ();
    if (st2) {
	FixupPos2 = Compiler->GetPos ();
	Compiler->OutCode (line, JUMP, 0);
    }
    Compiler->Fixup (FixupPos1, JZ, Compiler->GetPos ());
    if (st2) {
	st2->OutputCode ();
    	Compiler->Fixup (FixupPos2, JUMP, Compiler->GetPos ());
    }
}

// sequence

TStatementSequence::TStatementSequence () {
}

TStatementSequence::~TStatementSequence () {
    StatementList.DeleteAll (TRUE);
}

TStatementBase *TStatementSequence::Parse () {
    BOOL Goon;
    do {
	StatementList.Append (TStatementBase::Parse ());
	unsigned ch = Scanner->GetToken ();
	if (ch != END && ch != SEMICOLON) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "A semicolon or 'end' is missing");
	    Compiler->PanicRecovery (END, SEMICOLON, UNTIL, ELSE, BEGIN, PROCEDURE, FUNCTION, TERMINATOR, END_OF_LIST);
	}
        Goon = Scanner->Try (SEMICOLON);
    } while (Goon);
    Scanner->Try (END);
    return this;
}

void TStatementSequence::OutputCode () {
    for (TStatementBase *p = (TStatementBase *) StatementList.GetFirst (); p; p = (TStatementBase *) StatementList.GetNext ())
    	p->OutputCode ();
}

// repeat

TRepeatStatement::TRepeatStatement () {
    cond = NULL;
}

TRepeatStatement::~TRepeatStatement () {
    if (cond) delete cond;
    StatementList.DeleteAll (TRUE);
}

TStatementBase *TRepeatStatement::Parse () {
    BOOL Goon;
    do {
	StatementList.Append (TStatementBase::Parse ());
	unsigned ch = Scanner->GetToken ();
	if (ch != UNTIL && ch != SEMICOLON) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "A semicolon or 'until' is missing");
	    Compiler->PanicRecovery (END, SEMICOLON, UNTIL, ELSE, BEGIN, PROCEDURE, FUNCTION, TERMINATOR, END_OF_LIST);
	}
	Goon = Scanner->Try (SEMICOLON);
    } while (Goon);
    if (Scanner->Try (UNTIL)) {
        cond = TCompExpression::Parse ();
        TExpressionBase::DoTypeConversion (SymbolTable->TBOOL, cond);
    }
    return this;
}

void TRepeatStatement::OutputCode () {
    int Pos = Compiler->GetPos ();
    for (TStatementBase *p = (TStatementBase *) StatementList.GetFirst (); p; p = (TStatementBase *) StatementList.GetNext ())
    	p->OutputCode ();
    cond->OutputCode ();
    Compiler->OutCode (line, JZ, Pos);
}

// while

TWhileStatement::TWhileStatement () {
    statement = NULL;
    cond = NULL;
}

TWhileStatement::~TWhileStatement () {
    if (statement) delete statement;
    if (cond) delete cond;
}

TStatementBase *TWhileStatement::Parse () {
    cond = TCompExpression::Parse ();
    TExpressionBase::DoTypeConversion (SymbolTable->TBOOL, cond);
    if (!Scanner->Try (DO)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'do' expected in 'while'-statement");
    statement = TStatementBase::Parse ();
    return this;
}

void TWhileStatement::OutputCode () {
    int Pos1 = Compiler->GetPos ();
    cond->OutputCode ();
    int Pos2 = Compiler->GetPos ();
    Compiler->OutCode (line, JZ, 0);
    statement->OutputCode ();
    Compiler->OutCode (line, JUMP, Pos1);
    Compiler->Fixup (Pos2, JZ, Compiler->GetPos ());
}

// for

TForStatement::TForStatement () {
    start = stop = NULL;
    statement = NULL;
}

TForStatement::~TForStatement () {
    if (start) delete start;
    if (stop) delete stop;
    if (statement) delete statement;
}

TStatementBase *TForStatement::Parse () {
    if (Scanner->GetToken () != IDENTIFIER) {
	ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine (), "An identifier is required after FOR");
	control = NULL;
    } else {
	control = SymbolTable->Search (Scanner->GetString (), Scanner->GetLine (), FALSE);
	if (!control) ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine (), Scanner->GetString () & " is undeclared");
    }
    Scanner->GetNextToken ();
    if (!Scanner->Try (DEFINE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':=' required in 'for'-statement");
    start = TCompExpression::Parse ();
    if (Scanner->Try (TO)) Increment = TRUE;
    else if (Scanner->Try (DOWNTO)) Increment = FALSE;
    else ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'to' or 'downto' expected in 'for'-loop");
    stop = TCompExpression::Parse ();
    if (!Scanner->Try (DO)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'do' expected in 'for'-loop");
    if (control) {
        if (control->IsConstant () || !control->GetType ()->IsEnumerated ())
	    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Control variable of \'for\'-loop must be of an ordinal type");
        if (start) TExpression::DoTypeConversion (control->GetType (), start);
        if (stop) TExpression::DoTypeConversion (control->GetType (), stop);
    }
    Compiler->IncStatementStackLevel ();
    statement = TStatementBase::Parse ();
    Compiler->DecStatementStackLevel ();
    return this;
}

void TForStatement::OutputCode () {
    start->OutputCode ();
    Compiler->OutCode (line, STOREVAL, control->GetAddress (), control->GetLevel ());
    stop->OutputCode ();
    int Pos1 = Compiler->GetPos ();
    Compiler->OutCode (line, DUP, 0, 0);
    Compiler->OutCode (line, LOADVAL, control->GetAddress (), control->GetLevel ());
    Compiler->OutCode (line, Increment ? IGE : ILE);
    int Pos2 = Compiler->GetPos ();
    Compiler->OutCode (line, JZ, 0);
    statement->OutputCode ();
    Compiler->OutCode (line, Increment ? INC : DEC, control->GetAddress (), control->GetLevel ());
    Compiler->OutCode (line, JUMP, Pos1);
    Compiler->Fixup (Pos2, JZ, Compiler->GetPos ());
    Compiler->OutCode (line, ADDCSP, -1);
}

// return

TReturnStatement::TReturnStatement () {
    expr = NULL;
}

TReturnStatement::~TReturnStatement () {
    if (expr) delete expr; 
}

TStatementBase *TReturnStatement::Parse () {
    TRoutine *r = Compiler->GetCurrentRoutine ();
    if (r->GetReturnType () != SymbolTable->TVOID) {
	expr = TCompExpression::Parse ();
	TExpressionBase::DoTypeConversion (r->GetReturnType (), expr);
    }
    return this;
}

void TReturnStatement::OutputCode () {
    int level = Compiler->GetCurrentLevel ();
    Compiler->OutCode (line, RSP, level);
    if (expr) expr->OutputCode ();
    SymbolTable->CodeDestructors (line, Compiler);
    Compiler->OutCode (line, LEAVE, SymbolTable->GetNumParaSize (level), level);
}

// goto

TGotoStatement::TGotoStatement () {
    Label = NULL;
    CallSSL = Compiler->GetStatementStackLevel ();
}

TStatementBase *TGotoStatement::Parse () {
    String t;
    if (Scanner->GetToken () == IDENTIFIER)
	t = Scanner->GetString ();
    else if (Scanner->GetToken () == INTCONST)
	t = form (50, "%d", Scanner->GetInteger ());
    else {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Label expected");
	return this;
    }
    Label = SymbolTable->Search (t, Scanner->GetLine (), FALSE);
    if (!Label || !Label->IsLabel ()) 
	ErrorMessage (INVALID_USE_OF_SYMBOL, Scanner->GetLine (), "Label '" & t & "' undeclared'");
    Scanner->GetNextToken ();
    return this;
}

void TGotoStatement::OutputCode () {
    int a = Label->GetIntVal ();
    // We know at least the static nested level at which
    // the label is defined, because it must be defined within
    // the same block
    int dstSSL = Label->GetStatementStackLevel ();
    if (dstSSL > CallSSL)
	ErrorMessage (INVALID_USE_OF_SYMBOL, Scanner->GetLine (), "Cannot perform jump to label '" & Label->Name () & "' because it is deeper nested");
    if (CallSSL > dstSSL) Compiler->OutCode (line, ADDCSP, dstSSL - CallSSL);
    if (!a) // not yet defined
        Label->SetUseAddress (Compiler->PCodeInterpreter->GetPos ());
    Compiler->OutCode (line, JUMP, a);
}

// label definition

TLabelDefinition::TLabelDefinition (TSymbol *Label0) {
    Label = Label0;
    ControlledStm = NULL;
    Label->SetStatementStackLevel (Compiler->GetStatementStackLevel ());
}

TStatementBase *TLabelDefinition::Parse () {
    Scanner->GetNextToken ();
    if (!Scanner->Try (COLON))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':' expected after label definition");
    ControlledStm = TStatementBase::Parse ();
    return this;
}

void TLabelDefinition::OutputCode () {
    if (Label->GetIntVal ())
	ErrorMessage (INVALID_USE_OF_SYMBOL, Scanner->GetLine (), "Label '" & Label->Name () & "' already defined");
    int a = Compiler->PCodeInterpreter->GetPos ();
    Label->SetIntVal (a);
    int b = Label->GetFirstUseAddress ();
    while (b) {
	Compiler->Fixup (b, JUMP, a);
	b = Label->GetNextUseAddress ();
    }
    if (ControlledStm) ControlledStm->OutputCode ();
}

// case

TCaseLabel::TCaseLabel (int val) {
    both = FALSE;
    a = val;
    line = Scanner->GetLine ();
}

TCaseLabel::TCaseLabel (int val0, int val1) {
    both = TRUE;
    a = val0; b = val1;
}

void TCaseLabel::OutputCode (int dest) {
    if (both) {
	Compiler->OutCode (line, CMPRANGE, a, b);
	Compiler->OutCode (line, JNZ, dest);
    } else
	Compiler->OutCode (line, CMPJEQ, a, dest);
}

int TCaseLabel::Size () {
    return 1 + (int) both;
}

TCaseLine::TCaseLine () {
    line = Scanner->GetLine ();
    statement = NULL;
}

TCaseLine::~TCaseLine () {
    if (statement) delete statement;
    LabelList.DeleteAll (TRUE);
}

void TCaseLine::OutputCode (int startpos) {
    int dest = startpos + 1;	// calculate address of statement
    TCaseLabel *clab;
    for (clab = (TCaseLabel *) LabelList.GetFirst (); clab; clab = (TCaseLabel *) LabelList.GetNext ())
	dest += clab->Size ();
    for (clab = (TCaseLabel *) LabelList.GetFirst (); clab; clab = (TCaseLabel *) LabelList.GetNext ())
	clab->OutputCode (dest);
    int Pos = Compiler->GetPos ();		// jump around statement
    Compiler->OutCode (line, JUMP, 0);
    statement->OutputCode ();
    Final = Compiler->GetPos ();		// jump to end of case
    Compiler->OutCode (line, JUMP, 0);
    Compiler->Fixup (Pos, JUMP, Compiler->GetPos ());    
}

void TCaseLine::SetStatement (TStatementBase *s) {
    statement = s;
}

void TCaseLine::AddLabel (TCaseLabel *clab) {
    LabelList.Append (clab);
}

void TCaseLine::FixupFinal (int dest) {
    Compiler->Fixup (Final, JUMP, dest);
}

TCaseStatement::TCaseStatement () {
    expr = NULL;
    otherwise = NULL;
}

TCaseStatement::~TCaseStatement () {
    if (expr) delete expr;
    if (otherwise) delete otherwise;
    CaseLineList.DeleteAll (TRUE);
}

TStatementBase* TCaseStatement::Parse () {
    expr = TCompExpression::Parse ();
    TType *exprtype = TExpressionBase::ConvertToBaseType (expr);
    if (!Scanner->Try (OF))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'of' expected in 'case'-statement");
    BOOL Goon;
    Compiler->IncStatementStackLevel ();
    do {
	TCaseLine *cl = new TCaseLine;
	do {
	    TConstant c, d;
	    c.v.intval = d.v.intval = 0;
	    Compiler->ConstExpression (c);
	    if (c.t != exprtype) ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Type of case label does not match expression");
	    if (Scanner->Try (TWOPOINT)) {
		Compiler->ConstExpression (d);
		if (d.t != exprtype) ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine (), "Type of case label does not match expression");
		cl->AddLabel (new TCaseLabel (c.v.intval, d.v.intval));
	    } else cl->AddLabel (new TCaseLabel (c.v.intval));
	    Goon = Scanner->Try (COMMA);
	} while (Goon);
	if (!Scanner->Try (COLON))
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':' required in 'case'-statement");
	cl->SetStatement (TStatementBase::Parse ());
	CaseLineList.Append (cl);
	Goon = Scanner->Try (SEMICOLON);
        if (Scanner->GetToken () == END) Goon = FALSE;	// Hmm... not LL(1)?
    } while (Goon);
    if (Scanner->Try (ELSE)) otherwise = TStatementBase::Parse ();
    if (!Scanner->Try (END))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'case'-statement must be terminated with 'end'");
    Compiler->DecStatementStackLevel ();
    return this;
}

void TCaseStatement::OutputCode () {
    expr->OutputCode ();
    TCaseLine *cl;
    for (cl = (TCaseLine *) CaseLineList.GetFirst (); cl; cl = (TCaseLine *) CaseLineList.GetNext ()) 
	cl->OutputCode (Compiler->GetPos ());
    if (otherwise) otherwise->OutputCode ();
    for (cl = (TCaseLine *) CaseLineList.GetFirst (); cl; cl = (TCaseLine *) CaseLineList.GetNext ()) 
	cl->FixupFinal (Compiler->GetPos ());
    Compiler->OutCode (line, ADDCSP, -1);
}

/*********************************************************/
/*		TCompiler				 */
/*********************************************************/

TCompiler::TCompiler (TPCodeInterpreter *p, int ProgramSize0, String PathName0) {
    PCodeInterpreter = p;
    FindRunTimeError = FALSE;
    CodeArgCheck = CodeRangeCheck = CodeDivisionCheck = TRUE;
    CodeBoolShortcuts = FALSE;
    ProgramSize = ProgramSize0;
    ErrorCount = 0;
    MemFullReported = FALSE;
    PathName = PathName0;
    StatementStackLevel = 0;
}

TCompiler::~TCompiler () {
    LoadedUnits.DeleteAll (TRUE);
}

void TCompiler::PanicRecovery (TToken t, ...) {
    va_list ap;
    BOOL Found, Done;
    do {
        TToken ch = Scanner->GetToken ();
	Found = (ch == t);
	Done = FALSE;
	va_start (ap, t);
	while (!Found && !Done) {
#if (__GNUC__ == 2)  && (__GNUC_MINOR__ == 96) || (__GNUC__ >= 3)
          TToken t1 = TToken( va_arg(ap, int));
#else
	    TToken t1 = va_arg (ap, TToken);
#endif
    	    Found = (ch == t1);
	    Done = (t1 == END_OF_LIST);
	}
    	va_end (ap);
	if (!Found) Scanner->GetNextToken ();
    } while (!Found);
}
 

String TCompiler::GetName () {
    return name;
}

void TCompiler::BuildUserSymbols () {
}

void TCompiler::BuildPredefinedFunctions () {
// Fehlen: succ, pred
    TType *R = SymbolTable->TREAL,
          *C = SymbolTable->TCOMPLEX, 
          *N = SymbolTable->TINT,
          *RV = SymbolTable->TREALVECT,
          *NV = SymbolTable->TINTVECT,
	  *BV = SymbolTable->TBOOLVECT,
	  *B = SymbolTable->TBOOL,
	  *S = SymbolTable->TSTRING,
	  *V = SymbolTable->TVOID,
	  *RM = SymbolTable->TREALMATR;
    PredefinedFunctionList->Append ("sin", FSIN, FSIN, R, R);
    PredefinedFunctionList->Append ("sin", VSIN, VSIN, RV, RV);
    PredefinedFunctionList->Append ("arcsin", FASIN, FASIN, R, R);
    PredefinedFunctionList->Append ("arcsin", VARCSIN, VARCSIN, RV, RV);
    PredefinedFunctionList->Append ("cos", FCOS, FCOS, R, R);
    PredefinedFunctionList->Append ("cos", VCOS, VCOS, RV, RV);
    PredefinedFunctionList->Append ("arccos", FACOS, FACOS, R, R);
    PredefinedFunctionList->Append ("arccos", VARCCOS, VARCCOS, RV, RV);
    PredefinedFunctionList->Append ("tan", FTAN, FTAN, R, R);
    PredefinedFunctionList->Append ("tan", VTAN, VTAN, RV, RV);
    PredefinedFunctionList->Append ("arctan", FATAN, FATAN, R, R);
    PredefinedFunctionList->Append ("arctan", VARCTAN, VARCTAN, RV, RV);
    PredefinedFunctionList->Append ("cosh", FCOSH, FCOSH, R, R);
    PredefinedFunctionList->Append ("cosh", VCOSH, VCOSH, RV, RV);
    PredefinedFunctionList->Append ("sinh", FSINH, FSINH, R, R);
    PredefinedFunctionList->Append ("sinh", VSINH, VSINH, RV, RV);
    PredefinedFunctionList->Append ("tanh", FTANH, FTANH, R, R);
    PredefinedFunctionList->Append ("tanh", VTANH, VTANH, RV, RV);
    PredefinedFunctionList->Append ("log", FLOG, FLOG, R, R);
    PredefinedFunctionList->Append ("log", VLOG, VLOGC, RV, RV);
    PredefinedFunctionList->Append ("ln", FLOG, FLOG, R, R);
    PredefinedFunctionList->Append ("ln", VLOG, VLOGC, RV, RV);
    PredefinedFunctionList->Append ("exp", FEXP, FEXP, R, R);
    PredefinedFunctionList->Append ("exp", VEXP, VEXP, RV, RV);
    PredefinedFunctionList->Append ("sqrt", FSQRT, FSQRT, R, R);
    PredefinedFunctionList->Append ("sqrt", VSQRT, VSQRTC, RV, RV);
    PredefinedFunctionList->Append ("sign", FSIGN, FSIGN, R, R);
    PredefinedFunctionList->Append ("sign", VSIGN, VSIGN, RV, RV);
    PredefinedFunctionList->Append ("frac", FFRAC, FFRAC, R, R);
    PredefinedFunctionList->Append ("frac", VFRAC, VFRAC, RV, RV);
    PredefinedFunctionList->Append ("odd", IODD, IODD, B, N);
    PredefinedFunctionList->Append ("odd", VIODD, VIODD, BV, NV);
    PredefinedFunctionList->Append ("random", FRANDOM, FRANDOM, R);
    PredefinedFunctionList->Append ("memavail", AVAIL, AVAIL, N);

    PredefinedFunctionList->Append ("cumsum", VCUMSUM, VCUMSUM, RV, RV);
    PredefinedFunctionList->Append ("sum", VSUM, VSUM, R, RV);
    PredefinedFunctionList->Append ("sort", VFSRT, VFSRT, RV, RV);
    PredefinedFunctionList->Append ("rank", VFRANK, VFRANK, NV, RV);
    PredefinedFunctionList->Append ("exists", VEXIST, VEXIST, B, BV);
    PredefinedFunctionList->Append ("all", VFORALL, VFORALL, B, BV);
    PredefinedFunctionList->Append ("realvector", VREAL, VREAL, RV, R, R, N);
    PredefinedFunctionList->Append ("realvect", VREAL, VREAL, RV, R, R, N);
    PredefinedFunctionList->Append ("unitvector", VUNIT, VUNIT, RV, N, N);

    PredefinedFunctionList->Append ("complex", CMAKE, CMAKE, C, R, R);

    PredefinedFunctionList->Append ("transpose", MTRANS, MTRANS, RM, RM);
    PredefinedFunctionList->Append ("invert", MINV, MINV, RM, RM);
    PredefinedFunctionList->Append ("unitmatrix", MUNIT, MUNIT, RM, N);
    PredefinedFunctionList->Append ("chol", MSQRT, MSQRT, RM, RM);
    PredefinedFunctionList->Append ("makematrix", MREAL, MREAL, RM, RV, N, N);

    PredefinedFunctionList->Append ("round", FINT, FINT, N, R);
    PredefinedFunctionList->Append ("trunc", FTRUNC, FTRUNC, N, R);
    PredefinedFunctionList->Append ("max", IMAX, IMAX, N, N, N);
    PredefinedFunctionList->Append ("max", VIMAX2, VIMAX2, NV, NV, NV);
    PredefinedFunctionList->Append ("max", VIMAX, VIMAX, N, NV);
    PredefinedFunctionList->Append ("max", FMAX, FMAX, R, R, R);
    PredefinedFunctionList->Append ("max", VMAX2, VMAX2, RV, RV, RV);
    PredefinedFunctionList->Append ("max", VMAXV, VMAXV, R, RV);
    PredefinedFunctionList->Append ("min", IMIN, IMIN, N, N, N);
    PredefinedFunctionList->Append ("min", VIMIN2, VIMIN2, NV, NV, NV);
    PredefinedFunctionList->Append ("min", VIMIN, VIMIN, N, NV);
    PredefinedFunctionList->Append ("min", FMIN, FMIN, R, R, R);
    PredefinedFunctionList->Append ("min", VMIN2, VMIN2, RV, RV, RV);
    PredefinedFunctionList->Append ("min", VMINV, VMINV, R, RV);
    PredefinedFunctionList->Append ("indicator", FCHR, FCHR, R, R, R, R);
    PredefinedFunctionList->Append ("str", CFS, CFS, S, R);

    PredefinedFunctionList->Append ("abs", IABS, IABS, N, N);
    PredefinedFunctionList->Append ("abs", FABS, FABS, R, R);
    PredefinedFunctionList->Append ("abs", VIABS, VIABS, NV, NV);
    PredefinedFunctionList->Append ("abs", VFABS, VFABS, RV, RV);
    PredefinedFunctionList->Append ("sqr", ISQR, ISQR, N, N);
    PredefinedFunctionList->Append ("sqr", FSQR, FSQR, R, R);
    PredefinedFunctionList->Append ("sqr", VISQR, VISQR, NV, NV);
    PredefinedFunctionList->Append ("sqr", VFSQR, VFSQR, RV, RV);
    PredefinedFunctionList->Append ("eof", FEOF, FEOF, B, SymbolTable->TTEXT);
    PredefinedFunctionList->Append ("eof", FEOF, FEOF, B, SymbolTable->TFILE);
    PredefinedFunctionList->Append ("eof", FEOF, FEOF, B);
    PredefinedFunctionList->CodeImplicitOutput ();

    // predefined procedures

    PredefinedFunctionList->Append ("assign", FASSIGN, FASSIGN, V, SymbolTable->TFILE, SymbolTable->TSTRING);
    PredefinedFunctionList->SetReference (TRUE);
    PredefinedFunctionList->Append ("assign", FTASSIGN, FTASSIGN, V, SymbolTable->TTEXT, SymbolTable->TSTRING);
    PredefinedFunctionList->SetReference (TRUE);
    PredefinedFunctionList->Append ("reset", FRESET, FRESET, V, SymbolTable->TTEXT);
    PredefinedFunctionList->Append ("reset", FRESET, FRESET, V, SymbolTable->TFILE);
    PredefinedFunctionList->Append ("rewrite", FREWRITE, FREWRITE, V, SymbolTable->TTEXT);
    PredefinedFunctionList->Append ("rewrite", FREWRITE, FREWRITE, V, SymbolTable->TFILE);
    PredefinedFunctionList->Append ("flush", FFLUSH, FFLUSH, V, SymbolTable->TTEXT);
    PredefinedFunctionList->Append ("flush", FFLUSH, FFLUSH, V, SymbolTable->TFILE);
    PredefinedFunctionList->Append ("flush", FFLUSH, FFLUSH, V);
    PredefinedFunctionList->CodeImplicitOutput ();
    PredefinedFunctionList->Append ("close", FCLOSE, FCLOSE, V, SymbolTable->TTEXT);
    PredefinedFunctionList->Append ("close", FCLOSE, FCLOSE, V, SymbolTable->TFILE);
    PredefinedFunctionList->Append ("page", PAGE, PAGE, V, SymbolTable->TTEXT);
    PredefinedFunctionList->Append ("page", PAGE, PAGE, V);
    PredefinedFunctionList->CodeImplicitOutput ();

    PredefinedFunctionList->Append ("textcolor", TXTCOL, TXTCOL, V, SymbolTable->TTEXT, N);
    PredefinedFunctionList->Append ("textcolor", TXTCOL, TXTCOL, V, N);
    PredefinedFunctionList->CodeImplicitOutput ();
    PredefinedFunctionList->Append ("textbackground", TXTBCK, TXTBCK, V, SymbolTable->TTEXT, N);
    PredefinedFunctionList->Append ("textbackground", TXTBCK, TXTBCK, V, N);
    PredefinedFunctionList->CodeImplicitOutput ();
    PredefinedFunctionList->Append ("gotoxy", GOTOXY, GOTOXY, V, SymbolTable->TTEXT, N, N);
    PredefinedFunctionList->Append ("gotoxy", GOTOXY, GOTOXY, V, N, N);
    PredefinedFunctionList->CodeImplicitOutput ();
    PredefinedFunctionList->Append ("getkey", KEYP, KEYP, N);
    PredefinedFunctionList->CodeImplicitOutput ();
    PredefinedFunctionList->Append ("system", SYSTEM, SYSTEM, N, S);
    PredefinedFunctionList->Append ("date", SYSDATE, SYSDATE, V, N, N, N);
    PredefinedFunctionList->SetReference (TRUE, TRUE, TRUE);
    PredefinedFunctionList->Append ("time", SYSTIME, SYSTIME, V, N, N, N);
    PredefinedFunctionList->SetReference (TRUE, TRUE, TRUE);
    PredefinedFunctionList->Append ("paramcount", ARGC, ARGC, N);
    PredefinedFunctionList->Append ("paramstr", ARGV, ARGV, S, N);

}

void TCompiler::Compile () {
    SymbolTable = new TSymbolTable;
    PredefinedFunctionList = new TPredefinedFunctionList;
    Compiler = this;
    Program ();
    delete SymbolTable;
    delete PredefinedFunctionList;
}

void TCompiler::CheckAddress (int line) {
    if (PCodeInterpreter->GetPos () > RunTimeErrorPos) {
	ErrorMessage (POSITION_FOUND, line);
	FindRunTimeError = FALSE;
    }
}

void TCompiler::SearchRunTimeError (int n) {
    RunTimeErrorPos = n;
    FindRunTimeError = TRUE;
}

BOOL TCompiler::CheckMemFull () {
    if (PCodeInterpreter->GetPos () >= ProgramSize) {
        if (!MemFullReported) {
	    ErrorMessage (MEMORY_FULL, Scanner->GetLine ());
	    MemFullReported = TRUE;
	}
	return FALSE;
    }
    return TRUE;
}

void TCompiler::OutCode (int line, TCode c) {
    if (CheckMemFull ()) {
        PCodeInterpreter->OutCode (c);
        if (FindRunTimeError) CheckAddress (line);
    }
}

void TCompiler::OutCode (int line, TCode c, int a, int b) {
    if (CheckMemFull ()) {
        PCodeInterpreter->OutCode (c, a, b);
        if (FindRunTimeError) CheckAddress (line);
    }
}

void TCompiler::OutCode (int line, TCode c, double a) {
    if (CheckMemFull ()) {
        PCodeInterpreter->OutCode (c, a);
        if (FindRunTimeError) CheckAddress (line);
    }
}

void TCompiler::OutCode (int line, TCode c, String a) {
    if (CheckMemFull ()) {
        PCodeInterpreter->OutCode (c, a);
        if (FindRunTimeError) CheckAddress (line);
    }
}

int TCompiler::GetPos () {
    return PCodeInterpreter->GetPos ();
}

void TCompiler::Fixup (int Pos, TCode c, int a) {
    PCodeInterpreter->Fixup (Pos, c, a);
}

void TCompiler::Run () {
    PCodeInterpreter->Run ();
}

//void TCompiler::Check (TToken c) {
//    if (Scanner->GetToken () == c) Scanner->GetNextToken ();
//    else SetError (SYNTAX_ERROR, Scanner->GetLine ());
//}

/********************************************************************
			Type handling
********************************************************************/


TType *TCompiler::ParseType () {
    if (Scanner->GetToken () == BRACKETOPEN) {
	Scanner->GetNextToken ();
	return ParseEnumeration ();
    } else if (Scanner->GetToken () == ARRAY) {
	Scanner->GetNextToken ();
	if (!Scanner->Try (SQUAREBRACKETOPEN)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'[' required in array declaration");
	return ParseArray ();
    } else if (Scanner->GetToken () == RECORD) {
	Scanner->GetNextToken ();
	return ParseRecord ();
    } else if (Scanner->GetToken () == SET) {
	Scanner->GetNextToken ();
	return ParseSet ();
    } else if (Scanner->GetToken () == VECTOR) {
  	Scanner->GetNextToken ();
	return ParseVector ();
    } else if (Scanner->GetToken () == MATRIX) {
  	Scanner->GetNextToken ();
	return ParseMatrix ();
    } else if (Scanner->GetToken () == FUNCTION) {
	Scanner->GetNextToken ();
	return ParseFunction ();
    } else if (Scanner->GetToken () == DEREF) {
	Scanner->GetNextToken ();
	return ParsePointer ();
    } else if (Scanner->GetToken () == PROCEDURE) {
	Scanner->GetNextToken ();
	return ParseProcedure ();
    } else if (Scanner->GetToken () == IDENTIFIER) {
	String Name = Scanner->GetString ();
	if (SymbolTable->Search (Name, Scanner->GetLine (), FALSE)) return ParseSubrange ();
	Scanner->GetNextToken ();
	return SymbolTable->SearchType (Name, Scanner->GetLine ());
    } else if (Scanner->GetToken () == INTCONST || Scanner->GetToken () == SUB) 
	return ParseSubrange ();
    ErrorMessage (INVALID_TYPE, Scanner->GetLine ());
    return SymbolTable->TERROR;
}

TType *TCompiler::ParseEnumeration () {
    TEnumerated *p = new TEnumerated;
    int v = 0;
    BOOL Goon;
    do {
	if (Scanner->GetToken () != IDENTIFIER) ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine (), "Identifier required in declaration of enumerated type");
	else {
  	    String s = Scanner->GetString ();
	    Scanner->GetNextToken ();
	    TSymbol *q = SymbolTable->NewSymbol (s, p, Scanner->GetLine ());
	    q->SetConstant (TRUE);
	    q->SetIntVal (v++);
	}
	Goon = Scanner->Try (COMMA);
    } while (Goon);
    p->SetIndexRange (0, v - 1);
    if (!Scanner->Try (BRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Missing ')' at end of enumeration");
    return p;
}

TType *TCompiler::ParseSubrange () {
    TConstant c, d;
    ConstExpression (c);
    if (!Scanner->Try (TWOPOINT)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'..' required in subrange declaration");
    ConstExpression (d);
    if (c.t != d.t || !c.t->IsEnumerated ()) ErrorMessage (INVALID_TYPE, Scanner->GetLine ());
    return new TSubrange (c.t, c.v.intval, d.v.intval);
}

TType *TCompiler::ParseSet () {
    if (!Scanner->Try (OF)) {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'of' required in set declaration");
	return SymbolTable->TERROR;
    }
    TType *t = ParseType ();
    if (!t->IsEnumerated ()) ErrorMessage (INVALID_TYPE, Scanner->GetLine (), "Base type of set must be ordinal");
    int a, b;
    t->GetIndexRange (a, b);
    if (a < 0 || b > (int)(N_MAX_SETORD)) ErrorMessage (INVALID_TYPE, Scanner->GetLine (), "Ordinal values of base type of set must be less than 256");
    return new TSet (t);
}

TType *TCompiler::ParseArray () {
    TType *IndexType = ParseType ();
    if (Scanner->Try (COMMA)) 
	return new TArray (IndexType, ParseArray ());
    if (!Scanner->Try (SQUAREBRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "']' required in array declaration");
    if (!Scanner->Try (OF)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'of' required in array declaration");
    return new TArray (IndexType, ParseType ());
}

TType *TCompiler::ParseRecord () {
    TRecord *p = new TRecord;
    BOOL Goon;
    do {
	TList Components;
	do 
	    if (Scanner->GetToken () != IDENTIFIER) 
	        ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine ());
	    else {
	        Components.Append (new TStringEl (Scanner->GetString ()));
	        Scanner->GetNextToken ();
	    }
	while (Scanner->Try (COMMA));
	if (!Scanner->Try (COLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':' expected in record declaration");
	TType *t = ParseType ();
	for (TStringEl *s = (TStringEl *) Components.GetFirst (); s; s = (TStringEl *) Components.GetNext ())
	    p->AppendComponent (s->Name (), t);
	Components.DeleteAll (TRUE);
	Goon = Scanner->Try (SEMICOLON);
    } while (Goon);
    if (!Scanner->Try (END)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'end' required to terminate record declaration");
    return p;
}

TType *TCompiler::ParsePointer () {
    String s = "";
    TPointer *p;
    if (Scanner->GetToken () == IDENTIFIER && !SymbolTable->SearchType ((s = Scanner->GetString ()), 0)) {
        // pointer to a (not necessarily defined) type, will be resolved at end of type declaration.
	p = new TPointer (SymbolTable->TFORWARDPTR);
	Scanner->GetNextToken ();
    } else 
        p = new TPointer (ParseType ());
    p->SetBaseName (s);
    return p;
}

TType *TCompiler::ParseProcedure () {
    TRoutine *f = new TRoutine;
    ParseRoutineType (f);
    f->SetReturnType (SymbolTable->TVOID);
    return f;
}

TType *TCompiler::ParseFunction () {
    TRoutine *f = new TRoutine;
    ParseRoutineType (f);
    if (!Scanner->Try (COLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':' required in declaration of functional type");
    f->SetReturnType (ParseType ());
    return f;
}

TType *TCompiler::ParseVector () {
    if (!Scanner->Try (OF))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'of' required in vector declaration");
    TType *t = ParseType ();
    if (t->ContainsVector ()) ErrorMessage (INVALID_TYPE, Scanner->GetLine (), "A vector containing vectors or matrices is not supported by the current implementation");
    return new TVector (t);
}

TType *TCompiler::ParseMatrix () {
    if (!Scanner->Try (OF))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'of' required in matrix declaration");
    TType *t = ParseType ();
    if (t->ContainsVector ()) ErrorMessage (INVALID_TYPE, Scanner->GetLine (), "A matrix containing vectors or matrices is not supported by the current implementation");
    return new TSPMatrix (t);
}

void TCompiler::ParseRoutineType (TRoutine *f) {
    if (Scanner->Try (BRACKETOPEN)) {
	do {
	    BOOL VarType = Scanner->Try (VAR);
	    TType *t = ParseType ();
	    if (VarType) t = new TReference (t);
	    f->AppendParaType (t, String (""));
	} while (Scanner->Try (COMMA));
	if (!Scanner->Try (BRACKETCLOSE)) 
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "')' required in declaration of functional type");
    }
}

void TCompiler::TypeDeclaration (BOOL Temporary) {
    Scanner->GetNextToken ();
    do {
	if (Scanner->GetToken () != IDENTIFIER) {
	    ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine (), "Type declaration must start with identifier");
	    Scanner->GetNextToken ();
	} else {
	    String Name = Scanner->GetString ();
	    Scanner->GetNextToken ();
	    if (!Scanner->Try (EQUAL)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'=' required in type declaration");
	    TType *t = ParseType ();
	    SymbolTable->InsertType (new TTypeListEl (t, Name));
	}
	if (!Scanner->Try (SEMICOLON)) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Type declaration must be terminated with ';'");
	    Compiler->PanicRecovery (SEMICOLON, VAR, _CONST, TYPE, PROCEDURE, FUNCTION, BEGIN, TERMINATOR, END_OF_LIST);
	    Scanner->Try (SEMICOLON);
	}
    } while (Scanner->GetToken () == IDENTIFIER);
    SymbolTable->FillPointerTypes (Scanner->GetLine ());
}



void TCompiler::Program () {
    PCodeInterpreter->Reset ();
    Scanner->GetNextToken ();
    if (Scanner->GetToken () == PROGRAM) {
	Scanner->GetNextToken ();
	name = Scanner->GetString ();
	if (!Scanner->Try (IDENTIFIER)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Program name expected");
	if (!Scanner->Try (SEMICOLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected");
    }
    SymbolTable->EnterLevel (0);
    SymbolTable->BuildTypes ();

    MainBlock ();
}

void TCompiler::MainBlock () {
    BuildUserSymbols ();
    BuildPredefinedFunctions ();
    SymbolTable->AssignStackPositions (0);
    OutCode (Scanner->GetLine (), ENTER, SymbolTable->GetNumStackSize (), 0);
    SymbolTable->PrintLevel (0);
    SymbolTable->CodeConstructors (Scanner->GetLine (), this);

    TSymbol *s = SymbolTable->Search ("output", 0, 0);
    OutCode (Scanner->GetLine (), LOAD, 1, 0);
    OutCode (Scanner->GetLine (), STOREVAL, s->GetAddress (), s->GetLevel ());
    s = SymbolTable->Search ("input", 0, 0);
    OutCode (Scanner->GetLine (), LOAD, 0, 0);
    OutCode (Scanner->GetLine (), STOREVAL, s->GetAddress (), s->GetLevel ());
    int FixupPos = PCodeInterpreter->GetPos ();
    OutCode (Scanner->GetLine (), JUMP, 0, 0);

    SymbolTable->EnterLevel (1);
    if (Scanner->Try (USES)) 
	UsesDeclaration (1);
    Block (1, NULL, FixupPos);
    if (!Scanner->Try (ENDPOINT)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'.' expected at end of program");
    OutCode (Scanner->GetLine (), SPHALT);
    PCodeInterpreter->SetProgramSize ();
    SymbolTable->LeaveLevel (1);
//    SymbolTable->LeaveLevel (0);
}

void TCompiler::PerformReturn (int Level) {
    OutCode (Scanner->GetLine (), LEAVE, SymbolTable->GetNumParaSize (Level), Level);
}

void TCompiler::UsesDeclaration (int StaticLevel) {
//    puts ("Entering UsesDecl");
    TScanner *StoredScanner;
    do {
	if (Scanner->GetToken () == IDENTIFIER) {
	    String fn = Scanner->GetString ();
	    BOOL Found = FALSE;
	    for (TStringEl *s = (TStringEl *) LoadedUnits.GetFirst (); s && !Found; s = (TStringEl *) LoadedUnits.GetNext ())
		if (s->Name () == fn)
		    Found = TRUE;
	    if (!Found) {
		LoadedUnits.Append (new TStringEl (fn));
	        StoredScanner = Scanner;
	        Scanner = new TScanner (SearchFile (fn & ".sp", PathName));
	        if (Scanner->IsValid ()) {
		    Scanner->GetNextToken ();
	            Unit (StaticLevel);
	        } else
	            ErrorMessage (FILE_NOT_FOUND, StoredScanner->GetLine (), "Cannot open " & fn & ".sp", StoredScanner->GetFile ());
	        delete Scanner;
	        Scanner = StoredScanner;
	    } else printf ("Unit %s already loaded\n", (const char *) fn);
	} else ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Unit name expected.");
	Scanner->GetNextToken ();
    } while (Scanner->Try (COMMA));
    if (!Scanner->Try (SEMICOLON)) {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of 'uses' declaration.");
        PanicRecovery (SEMICOLON, VAR, _CONST, TYPE, PROCEDURE, FUNCTION, BEGIN, TERMINATOR, END_OF_LIST);
	Scanner->Try (SEMICOLON);
    }
}

void TCompiler::Unit (int StaticLevel) {
    String Name;
    if (!Scanner->Try (UNIT)) {
        ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'unit' expected");
    } else
 	Name = Scanner->GetString ();
    Scanner->GetNextToken ();
    if (!Scanner->Try (SEMICOLON)) {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of 'unit' declaration.");
        PanicRecovery (SEMICOLON, VAR, _CONST, TYPE, PROCEDURE, FUNCTION, BEGIN, TERMINATOR, END_OF_LIST);
	Scanner->Try (SEMICOLON);
    }
    if (!Scanner->Try (INTERFACE))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'interface' expected.");
    if (Scanner->Try (USES))
	UsesDeclaration (StaticLevel);
    DeclarationPart (StaticLevel, TRUE, FALSE);
    if (!Scanner->Try (IMPLEMENTATION))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'implementation' expected.");
    DeclarationPart (StaticLevel, FALSE, TRUE);
    SymbolTable->HideTemporaries ();
    if (!Scanner->Try (END) || !Scanner->Try (ENDPOINT))
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "unit must end with 'end.'");
}

void TCompiler::DeclarationPart (int StaticLevel, BOOL ForceForward, BOOL Temporary) {
   while (Scanner->GetToken () == _CONST || Scanner->GetToken () == TYPE || Scanner->GetToken () == VAR ||
	   Scanner->GetToken () == FUNCTION || Scanner->GetToken () == PROCEDURE || Scanner->GetToken () == LABEL) {
	if (Scanner->GetToken () == LABEL) LabelDeclaration (Temporary);
	else if (Scanner->GetToken () == _CONST) ConstDeclaration (Temporary);
	else if (Scanner->GetToken () == TYPE) TypeDeclaration (Temporary);
	else if (Scanner->GetToken () == VAR) {
	    VarDeclaration (Temporary);
	    SymbolTable->AssignStackPositions (Scanner->GetLine ());
	}
	else if (Scanner->GetToken () == FUNCTION) Function (StaticLevel, ForceForward, Temporary);
	else Procedure (StaticLevel, ForceForward, Temporary);
    }
 }

void TCompiler::Block (int StaticLevel, TSymbol *retsym, int FixupPos) {
    if (FixupPos == -1) {
	// for the outermost block, this has been done in
 	// MainBlock before including units
    	FixupPos = PCodeInterpreter->GetPos ();
    	OutCode (Scanner->GetLine (), JUMP, 0, 0);
    }

    int OldLevel = CurrentLevel;
    CurrentLevel = StaticLevel;

    SymbolTable->AssignStackPositions (Scanner->GetLine ());
    DeclarationPart (StaticLevel, FALSE, FALSE);
    SymbolTable->AssignStackPositions (Scanner->GetLine ());
    PCodeInterpreter->Fixup (FixupPos, JUMP, PCodeInterpreter->GetPos ());
    OutCode (Scanner->GetLine (), ENTER, SymbolTable->GetNumStackSize (), StaticLevel);
    SymbolTable->PrintLevel (StaticLevel);

    SymbolTable->CodeConstructors (Scanner->GetLine (), this);
    Statement ();
    if (retsym)	{	// load result of function onto stack
    	Compiler->OutCode (Scanner->GetLine (), LOADPTR, retsym->GetAddress (), retsym->GetLevel ());
    	Compiler->OutCode (Scanner->GetLine (), POPIND);
	retsym->GetType ()->CodeLoad (Scanner->GetLine (), this);
    }
    SymbolTable->CodeDestructors (Scanner->GetLine (), this);

    CurrentLevel = OldLevel;
}

void TCompiler::SingleDeclaration (BOOL CreateReference, BOOL IsParameter, BOOL Temporary) {
    BOOL Goon;
    TType type;
    TList IdentList;
    do {
	if (Scanner->GetToken () == IDENTIFIER) {
	    IdentList.Append (new TStringEl (Scanner->GetString ()));
	    Scanner->GetNextToken ();
	    Goon = Scanner->GetToken () == COMMA;
	    if (Goon) Scanner->GetNextToken ();
	} else {
	    ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine ());
	    Goon = FALSE;
	}
    } while (Goon);
    if (!Scanner->Try (COLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':' expected in variable declaration");
    TType *t = ParseType ();
    if (t == SymbolTable->TFORWARDPTR) ErrorMessage (IDENTIFIER_NOT_FOUND, Scanner->GetLine ());
    if (CreateReference) t = new TReference (t);
    for (TListEl *p = IdentList.GetFirst (); p; p = IdentList.GetNext ()) {
	TSymbol *q = SymbolTable->NewSymbol (p->Name (), t, Scanner->GetLine ());
	q->SetConstant (FALSE);
	q->SetParameter (IsParameter);
	if (Temporary) q->SetTemporary ();
    }
    IdentList.DeleteAll (TRUE);
}

void TCompiler::Declaration (BOOL IsParameter, BOOL Temporary) {
    do {
	SingleDeclaration (FALSE, IsParameter, Temporary);
	if (!Scanner->Try (SEMICOLON)) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Declaration must be terminated with ';'");
	    PanicRecovery (SEMICOLON, VAR, _CONST, TYPE, PROCEDURE, FUNCTION, BEGIN, TERMINATOR, END_OF_LIST);
	    Scanner->Try (SEMICOLON);
	}
    } while (Scanner->GetToken () == IDENTIFIER);
}

void TCompiler::LabelDeclaration (BOOL Temporary) {
    Scanner->GetNextToken ();
    do {
	String t;
	if (Scanner->GetToken () == IDENTIFIER) 
	    t = Scanner->GetString ();
	else if (Scanner->GetToken () == INTCONST) 
	    t = form (50, "%d", Scanner->GetInteger ());
	else {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Invalid label declaration");
	    t = "$DUMMY";
	}

	printf ("Got label: %s\n", (const char *) t);

	TSymbol *p = SymbolTable->NewSymbol (t, SymbolTable->TERROR, Scanner->GetLine ());
	p->SetConstant (TRUE);
	p->SetLabel (TRUE);
	p->SetIntVal (0);
	if (Temporary) p->SetTemporary ();
	Scanner->GetNextToken ();
    } while (Scanner->Try (COMMA));
    if (!Scanner->Try (SEMICOLON)) {
        ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of label declaration");
        PanicRecovery (SEMICOLON, TYPE, VAR, _CONST, PROCEDURE, FUNCTION, BEGIN, TERMINATOR, END_OF_LIST);
        Scanner->Try (SEMICOLON);
    }
}

void TCompiler::ConstDeclaration (BOOL Temporary) {
    TConstant c;
    Scanner->GetNextToken ();
    while (Scanner->GetToken () == IDENTIFIER) {
	String t = Scanner->GetString ();
	Scanner->GetNextToken ();
	if (Scanner->GetToken () == EQUAL) Scanner->GetNextToken ();
	else ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Constant declaration missing \'=\'");
	ConstExpression (c);
	if (!Scanner->Try (SEMICOLON)) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of constant declaration");
	    PanicRecovery (SEMICOLON, TYPE, VAR, _CONST, PROCEDURE, FUNCTION, BEGIN, TERMINATOR, END_OF_LIST);
	    Scanner->Try (SEMICOLON);
	}
	TSymbol *p = SymbolTable->NewSymbol (t, c.t, Scanner->GetLine ());
	p->SetConstant (TRUE);
// ????	p->SetType (c.t);
	if (Temporary) p->SetTemporary ();
	if (c.t == SymbolTable->TBOOL || c.t == SymbolTable->TCHAR || c.t == SymbolTable->TINT)
	    p->SetIntVal (c.v.intval);
	else if (c.t == SymbolTable->TREAL)
	    p->SetDoubleVal (c.v.doubleval);
	else p->SetStringVal (c.s);
    }
}

void TCompiler::VarDeclaration (BOOL Temporary) {
    Scanner->GetNextToken ();
    Declaration (FALSE, Temporary);
}

void TCompiler::ParameterList () {
    BOOL Reference;
    do {
	Reference = (Scanner->GetToken () == VAR);
	if (Reference) Scanner->GetNextToken ();
	SingleDeclaration (Reference, TRUE, FALSE);
    } while (Scanner->Try (SEMICOLON));
    if (!Scanner->Try (BRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Parameter list must be closed with ')'");
}

void TCompiler::Function (int StaticLevel, BOOL ForceForward, BOOL Temporary) {
    Scanner->GetNextToken ();
    String Name;
    if (Scanner->GetToken () != IDENTIFIER) {
	ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine ());
	Name = "$DUMMY";
    } else Name = Scanner->GetString ();

    TSymbol *f = SymbolTable->Search (Name, Scanner->GetLine (), FALSE);
    BOOL ForwardResolved = f && f->GetLevel () == StaticLevel && f->IsForward ();
    if (ForwardResolved) {	// resolve forward calls
	int addr = f->GetFixupAddress ();
//	printf ("Resolving forward to %s\n", (const char *) Name);
//      printf ("Patching jump at %d\n", addr);	    
 	Compiler->PCodeInterpreter->Fixup (addr, JUMP, PCodeInterpreter->GetPos ());
	f->SetForwardResolved ();
    } else {
        f = SymbolTable->NewSymbol (Name, NULL, Scanner->GetLine ());
        f->SetConstant (TRUE);
        f->SetIntVal (PCodeInterpreter->GetPos ());
	if (Temporary) f->SetTemporary ();
    }

    Scanner->GetNextToken ();
    SymbolTable->EnterLevel (StaticLevel + 1);
    TRoutine *t;

    if (ForwardResolved) {
	t = (TRoutine *) f->GetType ();
	String s; TType *r;
	for (s = t->GetFirstParaName (), r = t->GetFirstParaType (); r; s = t->GetNextParaName (), r = t->GetNextParaType ()) {
//   	    printf ("Restoring parameter %s, type %s\n", (const char *) s, (const char *) r->Name ());
	    TSymbol *u = SymbolTable->NewSymbol (s, r, Scanner->GetLine ());
	    u->SetConstant (FALSE);
	    u->SetParameter (TRUE);
	}
    } else {
        if (Scanner->GetToken () == BRACKETOPEN) {
	    Scanner->GetNextToken ();
	    ParameterList ();
        } 
        t = new TRoutine;
        SymbolTable->AssignParameters (t);
        if (!Scanner->Try (COLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "':' expected in function declaration");
        TType *rettype = ParseType ();
        if (rettype == SymbolTable->TFORWARDPTR) {
	    ErrorMessage (IDENTIFIER_NOT_FOUND, Scanner->GetLine (), "Return type of function '" & Name &"' is undeclared");
	    rettype = SymbolTable->TERROR;
        }
        t->SetReturnType (rettype);
        f->SetType (t);
    }
    if (!Scanner->Try (SEMICOLON)) {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of function header");
	PanicRecovery (SEMICOLON, _CONST, TYPE, VAR, FUNCTION, PROCEDURE, BEGIN, FORWARD, TERMINATOR, END_OF_LIST);
	Scanner->Try (SEMICOLON);
    }
    if (ForceForward || Scanner->Try (FORWARD)) {
	f->SetForward ();
	f->SetFixupAddress (PCodeInterpreter->GetPos ());
	OutCode (Scanner->GetLine (), JUMP, 0, 0);
	if (ForwardResolved) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "routine already declared forward");
	}
    } else {
        TRoutine *PrevRoutine = CurrentRoutine;
        CurrentRoutine = t;

        TSymbol *retsym = SymbolTable->NewSymbol (String ("$") & f->Name (), t->GetReturnType (), Scanner->GetLine ());
    	Block (StaticLevel + 1, retsym);
//    	Compiler->OutCode (Scanner->GetLine (), LOADPTR, retsym->GetAddress (), retsym->GetLevel ());
//    	Compiler->OutCode (Scanner->GetLine (), LOADMEM, t->GetReturnType ()->GetSize ());
//    	if (t->GetReturnType ()->IsVector ()) Compiler->OutCode (Scanner->GetLine (), VLOAD);
        PerformReturn (StaticLevel + 1);
    	CurrentRoutine = PrevRoutine;
    }
    SymbolTable->LeaveLevel (StaticLevel + 1);
    if (!ForceForward && !Scanner->Try (SEMICOLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of function");
}

void TCompiler::Procedure (int StaticLevel, BOOL ForceForward, BOOL Temporary) {
    Scanner->GetNextToken ();
    String Name;
    if (Scanner->GetToken () != IDENTIFIER) {
	ErrorMessage (IDENTIFIER_EXPECTED, Scanner->GetLine (), "A procedure name was expected");
	Name = "$DUMMY";
    } else Name = Scanner->GetString ();

    TSymbol *f = SymbolTable->Search (Name, Scanner->GetLine (), FALSE);
    BOOL ForwardResolved = f && f->GetLevel () == StaticLevel && f->IsForward ();
    if (ForwardResolved) {	// resolve forward calls
	int addr = f->GetFixupAddress ();
//	printf ("Resolving forward to %s\n", (const char *) Name);
//        printf ("Patching jump at %d\n", addr);	    
 	Compiler->PCodeInterpreter->Fixup (addr, JUMP, PCodeInterpreter->GetPos ());
	f->SetForwardResolved ();
    } else {
        f = SymbolTable->NewSymbol (Name, NULL, Scanner->GetLine ());
        f->SetConstant (TRUE);
        f->SetIntVal (PCodeInterpreter->GetPos ());
	if (Temporary) f->SetTemporary ();
    }
    Scanner->GetNextToken ();
    SymbolTable->EnterLevel (StaticLevel + 1);
    TRoutine *t;
    if (ForwardResolved) {
	t = (TRoutine *) f->GetType ();
	String s; TType *r;
	for (s = t->GetFirstParaName (), r = t->GetFirstParaType (); r; s = t->GetNextParaName (), r = t->GetNextParaType ()) {
//   	    printf ("Restoring parameter %s, type %s\n", (const char *) s, (const char *) r->Name ());
	    TSymbol *u = SymbolTable->NewSymbol (s, r, Scanner->GetLine ());
	    u->SetConstant (FALSE);
	    u->SetParameter (TRUE);
	}
    }
    else {
        if (Scanner->GetToken () == BRACKETOPEN) {
	    Scanner->GetNextToken ();
	    ParameterList ();
        }
        t = new TRoutine;
        SymbolTable->AssignParameters (t);
        t->SetReturnType (SymbolTable->TVOID);
        f->SetType (t);
    }
    if (!Scanner->Try (SEMICOLON)) {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of procedure header");
	PanicRecovery (SEMICOLON, _CONST, TYPE, VAR, FUNCTION, PROCEDURE, BEGIN, TERMINATOR, FORWARD, END_OF_LIST);
	Scanner->Try (SEMICOLON);
    }
    if (ForceForward || Scanner->Try (FORWARD)) {
	f->SetForward ();
	f->SetFixupAddress (PCodeInterpreter->GetPos ());
	OutCode (Scanner->GetLine (), JUMP, 0, 0);
	if (ForwardResolved) {
	    ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "routine already declared forward");
	}
    } else {
        TRoutine *PrevRoutine = CurrentRoutine;
        CurrentRoutine = t;
        Block (StaticLevel + 1);
        PerformReturn (StaticLevel + 1);
        CurrentRoutine = PrevRoutine;
    } 
    SymbolTable->LeaveLevel (StaticLevel + 1);
    if (!ForceForward && !Scanner->Try (SEMICOLON)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected at end of procedure");
}

void TCompiler::Statement () {
    if (!Scanner->Try (BEGIN)) 
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "'begin' expected");
    else {
        TStatementBase *s = (new TStatementSequence ())->Parse ();
        if (!ErrorCount) s->OutputCode ();
        delete s;
    }
}

/*********************************************************/
/*		Constant expressions			 */
/*********************************************************/

BOOL TCompiler::MakeConstDouble (TConstant &a, TConstant &b) {
    if (a.t == SymbolTable->TINT) {
	a.v.doubleval = (double) a.v.intval;
	a.t = SymbolTable->TREAL;
    }
    if (b.t == SymbolTable->TINT) {
	b.v.doubleval = (double) b.v.intval;
	b.t = SymbolTable->TREAL;
    }
    if (a.t != SymbolTable->TREAL || b.t != SymbolTable->TREAL) return FALSE;
    return TRUE;
}

void TCompiler::ConstExpression (TConstant &c) {
    BOOL Minus = Scanner->Try (SUB);
    ConstTerm (c);
    if (Minus)
	if (c.t == SymbolTable->TINT) c.v.intval = -c.v.intval;
	else if (c.t == SymbolTable->TREAL) c.v.doubleval = -c.v.doubleval;
	else if (c.t != SymbolTable->TERROR) {
	    ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine ());
	    c.t = SymbolTable->TERROR;
	}
    while (Scanner->GetToken () == ADD || Scanner->GetToken () == SUB) {
	TToken t = Scanner->GetToken ();
	Scanner->GetNextToken ();
	TConstant d;
	ConstTerm (d);
	if (c.t == SymbolTable->TINT && d.t == SymbolTable->TINT)
	    if (t == ADD) c.v.intval += d.v.intval;
	    else c.v.intval -= d.v.intval;
	else if (c.t == SymbolTable->TSTRING && d.t == SymbolTable->TSTRING)
	    if (t == ADD) c.s = c.s & d.s;
	    else {
		ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine ());
		c.t = SymbolTable->TERROR;
	    }
	else if (c.t != SymbolTable->TERROR && d.t != SymbolTable->TERROR) {
	    if (MakeConstDouble (c, d)) {
	    	if (t == ADD) c.v.doubleval += d.v.doubleval;
	    	else c.v.doubleval -= d.v.doubleval;
	    	c.t = SymbolTable->TREAL;
	    } else {
		ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine ());
		c.t = SymbolTable->TERROR;
	    }		
	}
    }
}

void TCompiler::ConstTerm (TConstant &c) {
    ConstFactor (c);
    while (Scanner->GetToken () == MUL || Scanner->GetToken () == DIV) {
	TToken t = Scanner->GetToken ();
	Scanner->GetNextToken ();
	TConstant d;
	ConstFactor (d);
	if (c.t == SymbolTable->TINT && d.t == SymbolTable->TINT) {
	    if (t == MUL) c.v.intval *= d.v.intval;
	    else {
		if (!d.v.intval) ErrorMessage (DIVISION_BY_ZERO, Scanner->GetLine ());
		else c.v.intval /= d.v.intval;
	    }
	} else {
	    if (MakeConstDouble (c, d)) {
	    	if (t == MUL) c.v.doubleval *= d.v.doubleval;
	    	else {
		    if (!d.v.doubleval) ErrorMessage (DIVISION_BY_ZERO, Scanner->GetLine ());
		    else c.v.doubleval /= d.v.doubleval;
	        }
	    } else {
		ErrorMessage (INCOMPATIBLE_TYPES, Scanner->GetLine ());
		c.t = SymbolTable->TERROR;
	    }		
	}
    }
}

void TCompiler::ConstFactor (TConstant &c) {
    if (Scanner->GetToken () == IDENTIFIER) {
	TSymbol *Symbol = SymbolTable->Search (Scanner->GetString (), Scanner->GetLine (), FALSE);
	Scanner->GetNextToken ();
	if (!Symbol) {
	    ErrorMessage (IDENTIFIER_NOT_FOUND, Scanner->GetLine ());
	    c.t = SymbolTable->TERROR;
	    return;
	}
	if (Symbol->IsConstant ()) {
	    c.t = Symbol->GetType ();
	    if (c.t == SymbolTable->TSTRING) c.s = Symbol->GetStringVal ();
	    else if (c.t == SymbolTable->TREAL) c.v.doubleval = Symbol->GetDoubleVal ();
	    else c.v.intval = Symbol->GetIntVal ();
	} else {
	    ErrorMessage (CONST_EXPECTED, Scanner->GetLine ());
	    c.t = SymbolTable->TERROR;
	}
    } else if (Scanner->GetToken () == INTCONST) {
	c.t = SymbolTable->TINT;
	c.v.intval = Scanner->GetInteger ();
	Scanner->GetNextToken ();
    } else if (Scanner->GetToken () == DBLCONST) {
	c.t = SymbolTable->TREAL;
	c.v.doubleval = Scanner->GetDouble ();
	Scanner->GetNextToken ();
    } else if (Scanner->GetToken () == CHARCONST) {
	c.t = SymbolTable->TCHAR;
	c.v.intval = Scanner->GetInteger ();
	Scanner->GetNextToken ();
    } else if (Scanner->GetToken () == BRACKETOPEN) {
	Scanner->GetNextToken ();
	ConstExpression (c);
	if (!Scanner->Try (BRACKETCLOSE)) ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "')' expected in constant expression");
    } else if (Scanner->GetToken () == STRCONST) {
	c.t = SymbolTable->TSTRING;
	c.s = Scanner->GetString ();
	Scanner->GetNextToken ();
    } else {
	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine ());
	c.t = SymbolTable->TERROR;
    }
}




#ifdef TEST
/*
int main (int argc, char **argv) {
    const int BufSize = 10000;

    char *buf, s [80];
    buf = new char [BufSize];
    memset (buf, 0, BufSize);
    if (argc <	2) {
	puts ("This is the terminal version of XPL compiler. Parameters are:\n\nCOMPILER filename	[...]\n\n"
	      "Arbitrary second parameter produces compiler listing (debugging purposes)\n");
	exit (1);
    }
    FILE *f = fopen (argv [1],	"rt");
    if (!f) { perror ("XPL"); exit (1); }
    fread (buf, 1, BufSize, f);
    fclose (f);

    TPCodeInterpreter p;
    if (!p.CreateMemory (1100000, 10000)) abort ();
    TFileList *fl = new TFileList;
    fl->SetStandardOutput (new TConsoleFileDescriptor);
    fl->SetStandardInput (new TConsoleFileDescriptor);
    p.SetFileList (fl);
    TCompiler c (&p, 10000);
    Scanner = new TScanner;
    Scanner->SetProgram (buf);
    if (argc == 4 && *argv [2] == 'f') c.SearchRunTimeError (atoi (argv [3]));
    try {
	c.Compile ();
	if (argc == 3) p.Listing ();
	puts ("Compilation successful - now running program");
	TRunTimeErrorType t = p.Run ();
	switch (t) {
	    case RT_DIVZERO: puts ("Division by zero"); break;
	    case RT_OPCODE: puts ("Invalid operation code"); break;
	    case RT_RANGECHECK: puts ("Range check error"); break;
	    case RT_STACKOVERFLOW: puts ("Stack overflow"); break;
	    case RT_ARGUMENT: puts ("Invalid argument"); break;
	}
	if (t != RT_OK) printf ("at address %d\n", p.GetPC ());
    } 
    catch (TErrorClass e) {
	puts (e.GetErrorMessage ());
    }
//    delete [] buf;
    return 0;
}
*/
#endif
