/*  compiler.hpp
 *
 *  Base classes for StatPascal compiler.
 *  All Xtremes or Windows specific routines are added in xpl.cpp.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998, 2000 Xtremes Group, Siegen.
 */

#ifndef COMPILER_HPP
#define COMPILER_HPP

#include "boolean.hpp"
#include "list.hpp"
#include "pcode.hpp"
#include "scanner.hpp"
#include "symbols.hpp"

/*********************************************************/
/*		Expressions				 */
/*********************************************************/

// Sollte auf TExpressionBase reduziert werden !!!!

class TExpressionBase : public TObject {
    public:
        TExpressionBase ();
	virtual ~TExpressionBase ();
	virtual void OutputCode () = 0;
	virtual TType *GetType ();
	static TType *CheckOperatorType (TExpressionBase *&first, TExpressionBase *&second, BOOL IsMultiplication);
        static TType *ConvertToBaseType (TExpressionBase *&first);
	static BOOL CheckTypeConversion (TType *Required, TExpressionBase *&p);
	static void DoTypeConversion (TType *Required, TExpressionBase *&p);
	
	virtual BOOL IsLValue ();
	virtual BOOL IsLValueDeref ();	// needed for vectors
	virtual BOOL IsConstant ();
	virtual BOOL IsSymbol ();	// for loading/storing optimization
    protected:
	TType *type;
	int line;			// line number of statement for locating errors
};

class TCompExpression : public TExpressionBase {
    public:
	TCompExpression (TExpressionBase *first, TExpressionBase *second, TToken op);
	virtual ~TCompExpression ();
	virtual void OutputCode ();
	static TExpressionBase *Parse ();
    protected:
	TExpressionBase *first, *second;
	TToken op;
};

class TExpression : public TExpressionBase {
    public:
	TExpression (TExpressionBase *first, TExpressionBase *second, TToken op);
	virtual ~TExpression ();
	virtual void OutputCode ();
	static TExpressionBase *Parse ();
    protected:
	TExpressionBase *first, *second;
	TToken op;
	int code;
};

class TTerm : public TExpressionBase {
    public:
        TTerm (TExpressionBase *first, TExpressionBase *second, TToken op);
        virtual ~TTerm ();
        virtual void OutputCode ();
	static TExpressionBase *Parse ();
    protected:
        TExpressionBase *first, *second;
        TToken op;
	int code;
};

class TNegativeTerm : public TExpressionBase {
    public:
	TNegativeTerm (TExpressionBase *a);
	virtual ~TNegativeTerm ();
	virtual void OutputCode ();
    protected:
	TExpressionBase *a;
	int code;
};

class TPower : public TExpressionBase {
   public:
	TPower (TExpressionBase *first, TExpressionBase *second);
	virtual ~TPower ();
	virtual void OutputCode ();
	static TExpressionBase *Parse ();
    protected:
	TExpressionBase *first, *second;
};

class TVectorIndex : public TExpressionBase {
    public:
	TVectorIndex (TExpressionBase *vector, TExpressionBase *index);
	virtual ~TVectorIndex ();
	virtual BOOL IsLValue ();
	virtual void OutputCode ();
    protected:
	TExpressionBase *vector, *index;
	BOOL VectorForced, LValue;
};

class TNotFactor : public TExpressionBase {
    public:
	TNotFactor (TExpressionBase *p);
	virtual ~TNotFactor ();
	virtual void OutputCode ();
    protected:
	TExpressionBase *p;
};

class TFactor : public TExpressionBase {
    public:
        TFactor (TExpressionBase *first, TExpressionBase *second);
        virtual ~TFactor ();
        virtual void OutputCode ();
	static TExpressionBase *Parse ();
    protected:
        TExpressionBase *first, *second;
	int code;
};

class TFunctionExpr : public TExpressionBase {
    public:
	static TExpressionBase *Parse ();
};

class TVarAccess : public TExpressionBase {
    public:
	static TExpressionBase *Parse ();
};

class TConstantBase : public TExpressionBase {
    public:
	virtual BOOL IsConstant ();
};

class TDeclaredConstant : public TConstantBase {
    public:
	TDeclaredConstant (TSymbol *s);
        virtual void OutputCode ();
    protected:
	TSymbol *s;
};

class TIntConstant : public TConstantBase {
    // stores also nil, boolean and char
    public:
        TIntConstant (int val, TType *type);
        virtual void OutputCode ();
    protected:
        int val;
};

class TDblConstant : public TConstantBase {
    public:
	TDblConstant (double val);
        virtual void OutputCode ();
    protected:
	double val;
};

class TStringConstant : public TConstantBase {
    public:
	TStringConstant (String s);
	virtual void OutputCode ();
    protected:
 	String s;
};

class TFunctionCall : public TExpressionBase {
    public:
	TFunctionCall (TExpressionBase *f, TList *ArgList0);
	virtual ~TFunctionCall ();
        virtual void OutputCode ();
	static TExpressionBase *Parse ();
    protected:
	TExpressionBase *f;
	TList ArgList;
	BOOL ImplicitLooping;
};

class TConvertIntReal : public TExpressionBase {
    public:
	TConvertIntReal (TExpressionBase *p);
	virtual ~TConvertIntReal ();
        virtual void OutputCode ();
    protected:
	TExpressionBase *p;
};

class TConvertIntVectRealVect : public TExpressionBase {
    public:
	TConvertIntVectRealVect (TExpressionBase *p);
	virtual ~TConvertIntVectRealVect ();
	virtual void OutputCode ();
    protected:
        TExpressionBase *p;
};

class TConvertIntMatrixRealMatrix : public TConvertIntVectRealVect {
    typedef TConvertIntVectRealVect inherited;
    public:
	TConvertIntMatrixRealMatrix (TExpressionBase *p);
};

class TChangeType : public TExpressionBase {
    public:
	TChangeType (TType *Req, TExpressionBase *p);
	virtual ~TChangeType ();
	virtual void OutputCode ();
    protected:
	TExpressionBase *p;
};

class TConvertSubrange : public TExpressionBase {
    public:
        TConvertSubrange (TType *Req, TExpressionBase *p);
        virtual ~TConvertSubrange ();
        virtual void OutputCode ();
    protected:
        TExpressionBase *p;
};

class TMakeVector : public TExpressionBase {
    public:
	TMakeVector (TExpressionBase *p);
	virtual ~TMakeVector ();
	virtual void OutputCode ();
    protected:
	TExpressionBase *p;
};

class TMakeMatrix : public TExpressionBase {
    public:
	TMakeMatrix (TExpressionBase *p);
	virtual ~TMakeMatrix ();
	virtual void OutputCode ();
    protected:
	TExpressionBase *p;
};

class TLValueDeref : public TExpressionBase {
    public:
    	TLValueDeref (TExpressionBase *p);
    	virtual void OutputCode ();
	virtual BOOL IsLValueDeref ();
	void SetLoadVector (BOOL f);
        TExpressionBase *UnDeref ();
    protected:
	BOOL LoadVector;
	TExpressionBase *p;
};

class TLValue : public TExpressionBase {
    public:
        TLValue (TSymbol *s);
        virtual void OutputCode ();
	virtual BOOL IsLValue ();

	virtual BOOL IsSymbol ();	// optimization of load/store
	void OutputLoadVal ();		
	void OutputStoreVal ();
    protected:
        TSymbol *s;
};

class TReferenceVariable : public TLValue {
    public:
	TReferenceVariable (TSymbol *s);
	virtual void OutputCode ();
	virtual BOOL IsSymbol ();
};

class TArrayIndex : public TExpressionBase {
    public:
	TArrayIndex (TExpressionBase *p, TExpressionBase *index);
	virtual ~TArrayIndex ();
	virtual void OutputCode ();
	virtual BOOL IsLValue ();
    protected:
	TExpressionBase *p, *index;
	TType *IndexType;
};

class TRecordIndex : public TExpressionBase {
    public:
 	TRecordIndex (TExpressionBase *p);
	virtual ~TRecordIndex ();
	virtual void OutputCode ();
	TExpressionBase *Parse ();
	virtual BOOL IsLValue ();
    protected:
	TExpressionBase *p;
	int offset;
};

class TPointerDereference : public TExpressionBase {
    public:
	TPointerDereference (TExpressionBase *p);
	virtual ~TPointerDereference ();
	TExpressionBase *Parse ();
	virtual void OutputCode ();
	virtual BOOL IsLValue ();
    protected:
	TExpressionBase *p;
};

class TSetExpression : public TExpressionBase {
    public:
	virtual ~TSetExpression ();
	void Parse ();
	virtual void OutputCode ();
    protected:
	TList ArgList;
};

// predefined functions

class TCombineFunction : public TExpressionBase {
    public:
	TCombineFunction ();
	virtual ~TCombineFunction ();
	void Parse ();
	virtual void OutputCode ();
    private:
	TList ArgList;
};

class TBindFunction : public TExpressionBase {
    public:
	virtual ~TBindFunction ();
	void Parse (String msg);
	void ProduceCode (TCode code);
    private:
	TList ArgList;
};

class TCBindFunction : public TBindFunction {
    typedef TBindFunction inherited;
    public:
	void Parse ();
	virtual void OutputCode ();
};

class TRBindFunction : public TBindFunction {
    typedef TBindFunction inherited;
    public:
	void Parse ();
	virtual void OutputCode ();
};

class TScanFunction : public TExpressionBase {
    public:
	TScanFunction ();
	virtual ~TScanFunction ();
	void Parse ();
	virtual void OutputCode ();
    private:
	int Count;
	TExpressionBase *arg [31];
	TExpressionBase *formatstr, *scanstr;
};

class TWriteItem : public TExpressionBase {
    public:
	TWriteItem (TExpressionBase *exp, TExpressionBase *length, TExpressionBase *digits, TExpressionBase *columns);
	virtual ~TWriteItem ();
	virtual void OutputCode ();
    private:
	TExpressionBase *exp, *length, *digits, *columns;
};

class TWriteFunction : public TExpressionBase {
    public:
	TWriteFunction (BOOL WriteLine);
	virtual ~TWriteFunction ();
	void Parse ();
	virtual void OutputCode ();
    protected:
   	TList ArgList;
	BOOL WriteLine;
};

class TReadFunction : public TExpressionBase {
    public:
	TReadFunction (BOOL ReadLine);
	virtual ~TReadFunction ();
	void Parse ();
	virtual void OutputCode ();
    protected:
	TList ArgList;
	BOOL ReadLine;
	TExpressionBase *FileGiven;
};

#define PREMAX 300
#define ARGMAX 8

class TPredefinedFunction : public TExpressionBase {
    public:
	TPredefinedFunction (int Code, TExpressionBase **a, TType *ReturnType, BOOL ImplInput, BOOL ImplOutput);
	virtual ~TPredefinedFunction ();
	static TExpressionBase *Parse (String Name);
	virtual void OutputCode ();
    protected:
	int Code;
	TExpressionBase *a [ARGMAX];
	BOOL ImplicitInput, ImplicitOutput;
};

class TSpecialFunction : public TExpressionBase {
    public:
	TSpecialFunction (String Name, TExpressionBase **a);
	virtual ~TSpecialFunction ();
	TExpressionBase *Parse ();
	virtual void OutputCode ();
    protected:
	String Name;
	TExpressionBase *a [ARGMAX];
	int ArgCount;
};

class TPredefinedFunctionList : public TObject {
    public:
	TPredefinedFunctionList ();
	~TPredefinedFunctionList ();
	void Append (String Name, int Code, int ArgCheckCode, TType *ReturnType, 
		TType *Arg1 = NULL, TType *Arg2 = NULL, TType *Arg3 = NULL, TType *Arg4 = NULL,
		TType *Arg5 = NULL, TType *Arg6 = NULL, TType *Arg7 = NULL, TType *Arg8 = NULL);
	void SetReference (
		BOOL IsRef1 = FALSE, BOOL IsRef2 = FALSE, BOOL IsRef3 = FALSE, BOOL IsRef4 = FALSE,
		BOOL IsRef5 = FALSE, BOOL IsRef6 = FALSE, BOOL IsRef7 = FALSE, BOOL IsRef8 = FALSE);
	void CodeImplicitInput ();
	void CodeImplicitOutput ();
	void Get (String Name, TExpressionBase **Arg, int &Code, int &ArgCheckCode, TType *&ReturnType, BOOL &ImplInput, BOOL &ImplOutput);
    private:
	BOOL CheckTypes (TExpressionBase **Arg, TType **wanted, BOOL *IsRef);
	String ListName [PREMAX];
	TType *ReturnType [PREMAX],
	      *ListType [PREMAX][ARGMAX];
	BOOL IsReference [PREMAX][ARGMAX];
	BOOL ImplicitInput [PREMAX], ImplicitOutput [PREMAX];
	int ListCode [PREMAX], ListArgCheckCode [PREMAX];
	int Count;
};

class TErrorExpr : public TExpressionBase {
    public:
	virtual void OutputCode ();
	virtual TType *GetType ();
};

/*********************************************************/
/*              Global Variables                         */
/*********************************************************/

extern TCompiler *Compiler;
extern TScanner *Scanner;
extern TSymbolTable *SymbolTable;
extern TPredefinedFunctionList *PredefinedFunctionList;


//////////////
/////////////



struct TConstant {
    TType *t;
    TStorage v;
    String s;
};

class TCompiler { 
    public:
        TCompiler (TPCodeInterpreter *, int ProgramSize, String PathName);
	virtual ~TCompiler ();

	virtual void Compile ();
        void Run ();
	void SearchRunTimeError (int Pos);
        String GetName ();

        void OutCode (int line, TCode);
        void OutCode (int line, TCode, int, int = 0);
        void OutCode (int line, TCode, double);
        void OutCode (int line, TCode, String);
        int GetPos ();
	void Fixup (int Pos, TCode, int);
        TRoutine *GetCurrentRoutine () 	{ return CurrentRoutine; }
	int GetCurrentLevel () 		{ return CurrentLevel; }
        void ConstExpression (TConstant &c);

	void IncErrorCount ()		{ ErrorCount++; }
 	int GetErrorCount ()		{ return ErrorCount; }

	void IncStatementStackLevel ()  { StatementStackLevel++; }
	void DecStatementStackLevel ()  { StatementStackLevel--; }
 	int GetStatementStackLevel ()   { return StatementStackLevel; }

	BOOL CodeRangeCheck, CodeDivisionCheck, CodeBoolShortcuts, CodeArgCheck;
        TPCodeInterpreter *PCodeInterpreter;

	void PanicRecovery (TToken ch, ...);
    protected:
	virtual void BuildUserSymbols ();		// overload to include constants
	virtual void BuildPredefinedFunctions ();	// overload to include functions
	       

	void CheckAddress (int line);
        BOOL CheckMemFull ();

        void Program ();
        void MainBlock ();		// initializes pred. functions, creates input/output variables
	void UsesDeclaration (int StaticLevel);
	void Unit (int StaticLevel);
	void DeclarationPart (int StaticLevel, BOOL ForceForward, BOOL Temporary);
        void Block (int StaticLevel, TSymbol *retsym = NULL, int FixupPos = -1);
        void SingleDeclaration (BOOL CreateReference, BOOL IsParameter, BOOL Temporary);
        void Declaration (BOOL IsParameter, BOOL Temporary);
        void VarDeclaration (BOOL Temporary);
	void LabelDeclaration (BOOL Temporary);
        void ConstDeclaration (BOOL Temporary);
        void ParameterList ();

        TType *ParseType ();
	TType *ParseEnumeration ();
	TType *ParseSubrange ();
        TType *ParseArray ();
	TType *ParseRecord ();
	TType *ParsePointer ();
	TType *ParseProcedure ();
	TType *ParseFunction ();
	TType *ParseSet ();
	TType *ParseVector ();
	TType *ParseMatrix ();
	void ParseRoutineType (TRoutine *f);
        void TypeDeclaration (BOOL Temporary);

        void PerformReturn (int Level);
        void Function (int StaticLevel, BOOL ForceForward, BOOL Temporary);
        void Procedure (int StaticLevel, BOOL ForceForward, BOOL Temporary);
        void Statement ();

        BOOL MakeConstDouble (TConstant &a, TConstant &b);
        void ConstTerm (TConstant &c);
        void ConstFactor (TConstant &c);

        TRoutine *CurrentRoutine;
        int CurrentLevel;
	int StatementStackLevel;	// to unload temps from stack in jumps

        BOOL FindRunTimeError,
             MemFullReported;
        int RunTimeErrorPos;
	String name;
        int ProgramSize;
	int ErrorCount;

	TList LoadedUnits;
        String PathName;	// for units
};

#endif
