/*  estimate.cpp

    Class definitions for estimators.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 2000 Xtremes Group, Siegen.
*/

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>

#include "estimate.hpp"
#include "mathfunc.hpp"

#ifndef TERMINAL
    #include "diagram.hpp"
    #include "distr.hpp"
    #include "iidsamp.hpp"
    #include "mrlf.hpp"
    #include "qqplot.hpp"
    #include "wintypes.hpp"
    #include "protocol.hpp"
    #include "ufodlg.hpp"
    #include "resource.h"
    #include "privmsg.hpp"
#endif


static char *ErrorMessages [] = {
    "",
    "Negative data in GP 1 or EV 1 model",
    "Newton-Raphson iteration found no point of zero",
    "Number of extremes is too small",
    "Number of extremes is larger than sample size",
    "Order statistics have different sign",
    "Data set is too small for this estimator",
    "Fatal error in XPL program",
    "Estimator failed - Please report to info@xtremes.stat.math.uni-siegen.de"
};

#ifdef TERMINAL
TEstimator::TEstimator (String EstimatorName) {
    name = EstimatorName;
}
#else
TEstimator::TEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddMenu)
 : TModelessDialog (IDD_DialogBox),
   SampleField (IDD_SampleData, ""),
   ResultField (IDD_Result, "") {
    name = EstimatorName;
    menuId = MenuId;
    if (!EstimatorListPtr) EstimatorListPtr = new TEstimatorList;
    if (AddMenu) EstimatorListPtr->Append (this);
    SetIcon (IDI_Estimate);
    SetHelpNumber (MenuId);
    *this << SampleField << ResultField;
}
#endif

#ifndef TERMINAL

TSample *TEstimator::GetBootstrapSample () {
    return NULL;
}

#endif

String TEstimator::Name () {
    return name;
}

double TEstimator::GetEstimatedParameterValue (int Nr) {
    return Parameters [Nr];
}

BOOL TEstimator::IsSumEstimator () {
    return FALSE;
}

BOOL TEstimator::IsGPEstimator () {
    return FALSE;
}

BOOL TEstimator::IsEVEstimator () {
    return FALSE;
}

double TEstimator::GetEstimatedShapeGamma () {
    assert (0);		// these should never be called
    return 0.0;
}

double TEstimator::GetEstimatedMean () {
    assert (0);		// these should never be called
    return 0.0;
}

double TEstimator::GetEstimatedVariance (){
    assert (0);		// these should never be called
    return 0.0;
}

void TEstimator::PerformEstimation () {
#ifndef TERMINAL
    EvaluateAdditionalParameters ();
#endif
    LastResult = EST_OK;
    LastResult = Estimate (0, p);
#ifndef TERMINAL
    ShowCurrentEstimation ();
#endif
}

#ifndef TERMINAL

void TEstimator::NotifyInitDialog () {
    inherited::NotifyInitDialog ();
    SetTitel (Name ());
    p = SampleList.GetActiveSample (this);
    SampleField.SetText (p->Name ());
}

void TEstimator::EvaluateAdditionalParameters () {
}

void TEstimator::ShowCurrentEstimation () {
    if (LastResult != EST_OK)
	ResultField.SetText (ErrorMessages [(int) LastResult]);
    else {
        int n = GetNumberOfEstimatedParameters ();
        String t;
        for (int i = 0; i < n; i++)
            t = t & GetEstimatedParameterName (i) & form (25, ": %g  ", GetEstimatedParameterValue (i));
        ResultField.SetText (t);
    }
}

int TEstimator::GetSubMenuPos () {
    return 0;
}

void TEstimator::BootstrapConfidenceInterval (int NrSims, int Parameter, double alpha, int k,
        double &left, double &right, double &mse) {
    double *x = new double [NrSims];
    TSample *OldSample = p;	// p is destroyed by some estimators
    TSample *q = GetBootstrapSample ();
    double estval = GetEstimatedParameterValue (Parameter);
    mse = 0.0;
    for (int i = 0; i < NrSims; i++) {
        Estimate (k, q);
        x [i] = GetEstimatedParameterValue (Parameter);
        mse += square (x [i]- estval);
        q->CreateMemory (0);
    }
    delete q;
    p = OldSample;
    Estimate (k, p);
    qsort (x, NrSims, sizeof (double), realcomp);
    alpha = 1.0 - alpha;
    left = x [(int) (alpha / 2.0 * NrSims)];
    right = x [(int) ((1.0 - alpha / 2.0) * NrSims)];
    mse /= NrSims;
    delete [] x;
}

double TEstimator::TYearThreshold (double t) {
    return 0.0;
}

class TTYearDialog : public TDialog {
    typedef TDialog inherited;
    public:
        TTYearDialog (TEstimator *e);
        virtual void NotifyActionEvent (TDialogMember *m);
    private:
        TEstimator *e;
        TCalcDoubleField e1;
        TStaticField s;
};

TTYearDialog::TTYearDialog (TEstimator *e0)
 : inherited (IDD_TYearValue),
   s (IDD_Result),
   e1 (IDD_TValue, GlobalData.TYearValue, DBL_MIN, DBL_MAX) {
    e = e0;
    *this << s << e1;
    s.SetText (form (50, "%g", e->TYearThreshold (GlobalData.TYearValue)));
}

void TTYearDialog::NotifyActionEvent (TDialogMember *m) {
    if (m->GetID () == IDD_Evaluate) {
        EvaluateDialog ();
        if (IsCorrect ()) {
            GlobalData.TYearValue = e1.GetValue ();
            s.SetText (form (50, "%g", e->TYearThreshold (GlobalData.TYearValue)));
        }
    } else inherited::NotifyActionEvent (m);
}


class TConfIntervalDialog : public TDialog {
    typedef TDialog inherited;
    public:
        TConfIntervalDialog (TEstimator *e);
        ~TConfIntervalDialog ();
        virtual void NotifyActionEvent (TDialogMember *m);
    private:
        TEstimator *e;
        TListField l;
        TIntField e1;
        TCalcDoubleField e2;
        TStaticField s, t;
        TList ParList;
};

TConfIntervalDialog::TConfIntervalDialog (TEstimator *e0)
 : TDialog (IDD_BootstrapConf),
   l (IDD_ParBox, NULL),
   e1 (IDD_NrSims, GlobalData.BootstrapNrSims, 1, INT_MAX),
   e2 (IDD_ConfLevel, GlobalData.BootstrapConfLevel, 0.0, 1.0),
   s (IDD_Result), t (IDD_MSE) {
    e = e0;
    for (int i = 0; i < e->GetNumberOfEstimatedParameters (); i++)
        ParList.Append (new TStringEl (e->GetEstimatedParameterName (i)));
    l.SetList (&ParList);
    *this << l << e1 << e2 << s << t;
}

TConfIntervalDialog::~TConfIntervalDialog () {
    ParList.DeleteAll (TRUE);
}

void TConfIntervalDialog::NotifyActionEvent (TDialogMember *m) {
    if (m->GetID () == IDD_StartSim) {
        EvaluateDialog ();
        if (IsCorrect ()) {
//            THourglas d (DialogWindow);
            GlobalData.BootstrapNrSims = e1.GetValue ();
            GlobalData.BootstrapConfLevel = e2.GetValue ();
            double left, right, mse;
            e->BootstrapConfidenceInterval (GlobalData.BootstrapNrSims, l.Selection (), GlobalData.BootstrapConfLevel,
                                   e->GetUsedK (), left, right, mse);
            s.SetText (form (50, "[%g, %g]", left, right));
            t.SetText (form (50, "MSE: %g", mse));
        }
    } else inherited::NotifyActionEvent (m);
}

void TEstimator::NotifyActionEvent (TDialogMember *m) {
    switch (m->GetID ()) {
        case IDM_BootstrapConfidence: {
            TConfIntervalDialog d (this);
            d.Handle (this); }
            break;
        case IDM_TYearValue: {
            TTYearDialog d (this);
            d.Handle (this); }
            break;
        default:
            inherited::NotifyActionEvent (m);
            break;
    }
}

int TEstimator::GetUsedK () {
    return 0;
}
#endif


#ifndef TERMINAL
void TEstimator::NotifyEvent (int ev) {
    switch (ev) {
        case M_VALIDATE_FUNCTIONS:
            if (p->IsDeleted ()) p = NULL;
            break;
        case M_VALIDATE_WINDOW :
            if (!p) CloseDialog ();
            break;
        default:
    	    inherited::NotifyEvent (ev);
    }
}
#endif

BOOL TEstimator::SearchMax (int Dimension) {
    const double eps = 1e-5;
    const int maxstep = 20;
    double **H = new double * [Dimension],
           **A = new double * [Dimension],
           *fv = new double [Dimension],
           *fs = new double [Dimension],
           *xold = new double [Dimension],
           *xnew = new double [Dimension],
           *xtemp = new double [Dimension];
    int i;
    for (i = 0; i < Dimension; i++) {
        H [i] = new double [Dimension];
        A [i] = new double [Dimension];
    }

    memcpy (xnew, Parameters, Dimension * sizeof (double));
    double fnew = SearchL (xnew), fold;
    int steps = 0;
    do {
        memcpy (xold, xnew, Dimension * sizeof (double));
	fold = fnew;
        for (i = 0; i < Dimension; i++) {
            memcpy (xtemp, xold, Dimension * sizeof (double));
            xtemp [i] += eps;
            fv [i] = SearchL (xtemp);
        }
        for (i = 0; i < Dimension; i++) {
            fs [i] = (fv [i] - fold) / eps;
            for (int j = 0; j < Dimension; j++)
                if (j >= i) {
                    memcpy (xtemp, xold, Dimension * sizeof (double));
                    xtemp [i] += eps;
                    xtemp [j] += eps;
                    H [i][j] = (SearchL (xtemp) - fv [i] - fv [j] + fold) / (eps * eps);
                } else
                    H [i][j] = H [j][i];
        }
        if (MatInvert (H, A, Dimension)) {
            for (i = 0; i < Dimension; i++) {
                xtemp [i] = 0.0;
                for (int j = 0; j < Dimension; j++)
                    xtemp [i] += A [i][j] * fs [j];
            }
            double s = 1.0;
            BOOL valid;
            do {
                for (i = 0; i < Dimension; i++)
                    xnew [i] = xold [i] - s * xtemp [i];
                s *= 0.5;
                valid = SearchBounds (xnew);
                if (valid) fnew = SearchL (xnew);
            } while (!((valid && fnew > fold) || s < 1e-30));
        } else fnew = fold - 1.0;
        if (fnew < fold) { 	// Newton iteration failed
            double s = 1.0;
            BOOL valid;
            do {
                for (i = 0; i < Dimension; i++)
                    xnew [i] = xold [i] - s * fs [i];
                s *= 0.5;
                valid = SearchBounds (xnew);
                if (valid) fnew = SearchL (xnew);
            } while (!((valid && fnew > fold) || s < 1e-30));
	}
        steps++;
    } while (steps < maxstep && fabs (fnew - fold) > 1e-9);
    memcpy (Parameters, xnew, Dimension * sizeof (double));

    for (i = 0; i < Dimension; i++) {
        delete [] H [i];
        delete [] A [i];
    }
    delete [] H;
    delete [] A;
    delete [] xold;
    delete [] xnew;
    delete [] xtemp;
    delete [] fs;
    delete [] fv;
    return steps < maxstep;
}

double TEstimator::SearchL (double *) {
    return 0.0;
}

BOOL TEstimator::SearchBounds (double *) {
    return TRUE;
}


/********************************************************************
        Base class of estimators for normal distribution
********************************************************************/

#ifdef TERMINAL
TNormalEstimator::TNormalEstimator (String EstimatorName)
 : TEstimator (EstimatorName) {
}
#else
TNormalEstimator::TNormalEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : ProgressField (IDD_ProgressField, ""),
   QQPlotLine (IDD_QQPlotLine, GlobalData.QQPlotLine),
   CloseBox (IDD_Close, TRUE),
   TEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
   *this << ProgressField << QQPlotLine << CloseBox;
}
#endif

TSampleType TNormalEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

int TNormalEstimator::GetNumberOfEstimatedParameters () {
    return 2;
}

BOOL TNormalEstimator::IsSumEstimator () {
    return TRUE;
}

double TNormalEstimator::GetEstimatedMean () {
    return Parameters [0];
}

double TNormalEstimator::GetEstimatedVariance () {
    return Parameters [1] * Parameters [1];
}

String TNormalEstimator::GetEstimatedParameterName (int Nr) {
    if (Nr == 1) return "sigma";
    return "mu";
}

#ifndef TERMINAL

TSample *TNormalEstimator::GetBootstrapSample () {
    TIIDGaussianSample *q = new TIIDGaussianSample;
    q->SetParameter (Parameters [1], Parameters [0], p->GetDataSize ());
    q->CreateMemory (0);
    return q;
}

double TNormalEstimator::TYearThreshold (double t) {
    return Parameters [0] + Parameters [1] * XGaussianQF (1.0 - 1.0 / t);
}

void TNormalEstimator::NotifyActionEvent (TDialogMember *m) {
    switch (m->GetID ()) {
        case IDD_Density :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TNormalDensity (Parameters [1], Parameters [0], 1));
            break;
        case IDD_Quantile :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TNormalQuantile (Parameters [1], Parameters [0], 1));
            break;
        case IDD_Distribution:
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TNormalDistribution (Parameters [1], Parameters [0], 1));
            break;
        case IDD_QQPlot: {
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow ("Q-Q Plot")
                  ->InsertFunction (new TNormalQQPlot (Parameters [1], Parameters [0], (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ())); }
            break;
    }
    if (m->GetID () == IDD_Density || m->GetID () == IDD_Quantile ||
        m->GetID () == IDD_Distribution || m->GetID () == IDD_QQPlot)
        if (CloseBox.Selected ()) CloseDialog ();
    inherited::NotifyActionEvent (m);
}

#endif

#ifdef USE_STABLE

/********************************************************************
        Base class of estimators for stable distribution
********************************************************************/

#ifdef TERMINAL
TStableEstimator::TStableEstimator (String EstimatorName)
 : TEstimator (EstimatorName) {
}
#else
TStableEstimator::TStableEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : ProgressField (IDD_ProgressField, ""),
   QQPlotLine (IDD_QQPlotLine, GlobalData.QQPlotLine),
   CloseBox (IDD_Close, TRUE),
   TEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
   *this << ProgressField << QQPlotLine << CloseBox;
}
#endif

TSampleType TStableEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

int TStableEstimator::GetNumberOfEstimatedParameters () {
    return 4;
}

String TStableEstimator::GetEstimatedParameterName (int Nr) {
    switch (Nr) {
        case 0: return "alpha";
        case 1: return "beta";
        case 2: return "mu";
    }
    return "sigma";
}

#ifndef TERMINAL

TSample *TStableEstimator::GetBootstrapSample () {
    TSumStableSample *q = new TSumStableSample;
    q->SetParameter (Parameters [0], Parameters [1], Parameters [2], Parameters [3], p->GetDataSize ());
    q->CreateMemory (0);
    return q;
}

double TStableEstimator::TYearThreshold (double t) {
    return Parameters [2] + Parameters [3] * Xstableqf (1.0 - 1.0 / t, Parameters [0], Parameters [1]);
}

void TStableEstimator::NotifyActionEvent (TDialogMember *m) {
    switch (m->GetID ()) {
        case IDD_Density :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TSumStableDensity (Parameters [0], Parameters [1], Parameters [3], Parameters [2], 1));
            break;
        case IDD_Quantile :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TSumStableQuantile (Parameters [0], Parameters [1], Parameters [3], Parameters [2], 1));
            break;
        case IDD_Distribution:
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TSumStableDistribution (Parameters [0], Parameters [1], Parameters [3], Parameters [2], 1));
            break;
        case IDD_QQPlot: {
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow ("Q-Q Plot")
                  ->InsertFunction (new TStableQQPlot (Parameters [0], Parameters [1], Parameters [2], Parameters [3], (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ())); }
            break;
    }
    if (m->GetID () == IDD_Density || m->GetID () == IDD_Quantile ||
        m->GetID () == IDD_Distribution || m->GetID () == IDD_QQPlot)
        if (CloseBox.Selected ()) CloseDialog ();
    inherited::NotifyActionEvent (m);
}

#endif
#endif

/********************************************************************
        Base class of estimators for T distribution
********************************************************************/

#ifdef TERMINAL
TTEstimator::TTEstimator (String EstimatorName)
 : TEstimator (EstimatorName) {
}
#else
TTEstimator::TTEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : ProgressField (IDD_ProgressField, ""),
   QQPlotLine (IDD_QQPlotLine, GlobalData.QQPlotLine),
   CloseBox (IDD_Close, TRUE),
   TEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
   *this << ProgressField << QQPlotLine << CloseBox;
}
#endif

TSampleType TTEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

int TTEstimator::GetNumberOfEstimatedParameters () {
    return 2;
}

String TTEstimator::GetEstimatedParameterName (int Nr) {
    if (Nr == 0) return "alpha";
    return "sigma";
}

#ifndef TERMINAL

TSample *TTEstimator::GetBootstrapSample () {
    TStudentTSample *q = new TStudentTSample;
    q->SetParameter (Parameters [0], Parameters [1], p->GetDataSize ());
    q->CreateMemory (0);
    return q;
}

double TTEstimator::TYearThreshold (double t) {
    return Parameters [1] * XTQF (Parameters [0], 1.0 - 1.0 / t);
}

void TTEstimator::NotifyActionEvent (TDialogMember *m) {
    switch (m->GetID ()) {
        case IDD_Density :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TStudentTDensity (Parameters [0], Parameters [1], 1));
            break;
        case IDD_Quantile :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TStudentTQuantile (Parameters [0], Parameters [1], 1));
            break;
        case IDD_Distribution:
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TStudentTDistribution (Parameters [0], Parameters [1], 1));
            break;
        case IDD_QQPlot: {
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow ("Q-Q Plot")
                  ->InsertFunction (new TStudentTQQPlot (Parameters [0], Parameters [1], (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ())); }
            break;
    }
    if (m->GetID () == IDD_Density || m->GetID () == IDD_Quantile ||
        m->GetID () == IDD_Distribution || m->GetID () == IDD_QQPlot)
        if (CloseBox.Selected ()) CloseDialog ();
    inherited::NotifyActionEvent (m);
}

#endif

#ifdef TERMINAL
TNCTEstimator::TNCTEstimator (String EstimatorName)
 : TEstimator (EstimatorName) {
}
#else
TNCTEstimator::TNCTEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : ProgressField (IDD_ProgressField, ""),
   QQPlotLine (IDD_QQPlotLine, GlobalData.QQPlotLine),
   CloseBox (IDD_Close, TRUE),
   TEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
   *this << ProgressField << QQPlotLine << CloseBox;
}
#endif

TSampleType TNCTEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

int TNCTEstimator::GetNumberOfEstimatedParameters () {
    return 3;
}

String TNCTEstimator::GetEstimatedParameterName (int Nr) {
    if (Nr == 0) return "alpha";
    if (Nr == 1) return "beta";
    return "sigma";
}

#ifndef TERMINAL

TSample *TNCTEstimator::GetBootstrapSample () {
    TNCStudentTSample *q = new TNCStudentTSample;
    q->SetParameter (Parameters [0], Parameters [1], Parameters [2], p->GetDataSize ());
    q->CreateMemory (0);
    return q;
}

double TNCTEstimator::TYearThreshold (double t) {
    return Parameters [1] * XNCTQF (Parameters [0], Parameters [1], 1.0 - 1.0 / t);
}

void TNCTEstimator::NotifyActionEvent (TDialogMember *m) {
    switch (m->GetID ()) {
        case IDD_Density :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TNCStudentTDensity (Parameters [0], Parameters [1], Parameters [2], 1));
            break;
        case IDD_Quantile :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TNCStudentTQuantile (Parameters [0], Parameters [1], Parameters [2], 1));
            break;
        case IDD_Distribution:
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TNCStudentTDistribution (Parameters [0], Parameters [1], Parameters [2], 1));
            break;
        case IDD_QQPlot: {
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow ("Q-Q Plot")
                  ->InsertFunction (new TNCStudentTQQPlot (Parameters [0], Parameters [1], Parameters [2], (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ())); }
            break;
    }
    if (m->GetID () == IDD_Density || m->GetID () == IDD_Quantile ||
        m->GetID () == IDD_Distribution || m->GetID () == IDD_QQPlot)
        if (CloseBox.Selected ()) CloseDialog ();
    inherited::NotifyActionEvent (m);
}

#endif

/********************************************************************
        Base class of estimators for Gumbel distribution
********************************************************************/

#ifdef TERMINAL
TGumbelEstimator::TGumbelEstimator (String EstimatorName)
 : TEstimator (EstimatorName) {
}
#else
TGumbelEstimator::TGumbelEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : ProgressField (IDD_ProgressField, ""),
   QQPlotLine (IDD_QQPlotLine, GlobalData.QQPlotLine),
   CloseBox (IDD_Close, TRUE),
   TEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
   *this << ProgressField << QQPlotLine << CloseBox;
}
#endif

TSampleType TGumbelEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

int TGumbelEstimator::GetNumberOfEstimatedParameters () {
    return 2;
}

String TGumbelEstimator::GetEstimatedParameterName (int Nr) {
    if (Nr == 1) return "sigma";
    return "mu";
}

#ifndef TERMINAL

TSample *TGumbelEstimator::GetBootstrapSample () {
    TGumbelSample *q = new TGumbelSample;
    q->SetParameter (Parameters [1], Parameters [0], p->GetDataSize ());
    q->CreateMemory (0);
    return q;
}

double TGumbelEstimator::TYearThreshold (double t) {
    return Parameters [0] + Parameters [1] * XGumbelQF (1.0 - 1.0 / t);
}

void TGumbelEstimator::NotifyActionEvent (TDialogMember *m) {
    int id = m->GetID ();
    switch (id) {
        case IDD_Density :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TGumbelDensity (Parameters [1], Parameters [0], 1));
            break;
        case IDD_Quantile :
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TGumbelQuantile (Parameters [1], Parameters [0], 1));
            break;
        case IDD_Distribution:
            if (LastResult == EST_OK && Parameters [1])
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TGumbelDistribution (Parameters [1], Parameters [0], 1));
            break;
        case IDD_QQPlot:
            if (LastResult == EST_OK)
                PlotWindowManager.GetOpenWindow ("Q-Q Plot")->InsertFunction (new TEVDQQPlot (
                    0.0, GetEstimatedParameterValue (1),
                    GetEstimatedParameterValue (0), (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ()));
            break;
    }
    if (id == IDD_Density || id == IDD_Quantile ||
        id == IDD_Distribution || id == IDD_QQPlot)
        if (CloseBox.Selected ()) CloseDialog ();
    inherited::NotifyActionEvent (m);
}
#endif

/********************************************************************
      Base class of estimators for extreme value and
      generalized Pareto distributions
********************************************************************/

#ifdef TERMINAL
TEVDGPDBaseEstimator::TEVDGPDBaseEstimator (String EstimatorName)
 : TEstimator (EstimatorName) {
}
#else
TEVDGPDBaseEstimator::TEVDGPDBaseEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddMenu)
 : ModeField (GlobalData.GPDEstimatorMode, 2, IDD_AlphaMode, IDD_BetaLS),
   CloseBox (IDD_Close, TRUE),
   TEstimator (EstimatorName, MenuId, IDD_DialogBox, AddMenu) {
    *this << ModeField << CloseBox;
}
#endif

void TEVDGPDBaseEstimator::BuildParameterList () {
}

int TEVDGPDBaseEstimator::GetNumberOfEstimatedParameters () {
    return 3;
}

void TEVDGPDBaseEstimator::TransformModels (int k, TUnivariateSample *q) {
    TEstMode EstMode = GetEstimatorMode ();
    CompleteModels (EstMode, k, q);
    memcpy (Parameters, RawParameters, 3 * sizeof (double));
    if ((EstMode == ALPHA_MODE || EstMode == FULL_ALPHA_MODE) && GetParameter ()->GetMode () == BETA_MODE)
        if (Parameters [0] > 0.0) {
            Parameters [0] = 1.0 / Parameters [0];
            Parameters [1] += Parameters [2];
            Parameters [2] *= Parameters [0];
        } else {
            Parameters [0] = Parameters [0] ? 1.0 / Parameters [0] : MYMAXDOUBLE;
            Parameters [1] -= Parameters [2];
            Parameters [2] *= -Parameters [0];
        }
    else if ((EstMode == BETA_MODE || EstMode == FULL_BETA_MODE) && GetParameter ()->GetMode () == ALPHA_MODE)
        if (Parameters [0] > 0.0) {
            Parameters [0] = 1.0 / Parameters [0];
            Parameters [2] *= Parameters [0];
            Parameters [1] -= Parameters [2];
	} else if (Parameters [0] < 0.0) {
	    Parameters [0] = Parameters [0] ? 1.0 / Parameters [0] : MYMAXDOUBLE;
	    Parameters [2] *= -Parameters [0];
	    Parameters [1] += Parameters [2];
	} else
	    Parameters [0] = Parameters [1] = Parameters [2] = NOTAVAILABLE;
}

String TEVDGPDBaseEstimator::GetEstimatedParameterName (int i) {
    assert (0 <= i && i < 3);
    switch (i) {
        case 0 :
            if (GetParameter ()->GetMode () == ALPHA_MODE) return "alpha";
            else return "gamma";
        case 1 : return "mu";
	case 2 : return "sigma";
    }
    return "";
}

#ifndef TERMINAL

void TEVDGPDBaseEstimator::PlotDensity () {
}

void TEVDGPDBaseEstimator::PlotQuantile () {
}

void TEVDGPDBaseEstimator::PlotDistribution () {
}

void TEVDGPDBaseEstimator::PlotMeanEF () {
}

void TEVDGPDBaseEstimator::PlotMedianEF () {
}

void TEVDGPDBaseEstimator::PlotQQPlot () {
}

void TEVDGPDBaseEstimator::NotifyActionEvent (TDialogMember *m) {
    int id = m->GetID ();
    switch (id) {
        case IDD_AlphaMode :
        case IDD_BetaLS :
        case IDD_BetaRobust :
            EvaluateDialog ();
            GetParameter ()->SetMode ((TEstMode) ModeField.Selector ());
            TransformModels (k, (TUnivariateSample *) p);
            ShowCurrentEstimation ();
            BuildParameterList ();
            break;
	case IDD_Density :
	    if (GetEstimatedParameterValue (0) == NOTAVAILABLE) return;
	    PlotDensity ();
	    break;
	case IDD_Quantile :
	    if (GetEstimatedParameterValue (0) == NOTAVAILABLE) return;
	    PlotQuantile ();
	    break;
	case IDD_Distribution:
	    if (GetEstimatedParameterValue (0) == NOTAVAILABLE) return;
	    PlotDistribution ();
	    break;
	case IDD_MeanEF :
	    if (GetEstimatedParameterValue (0) == NOTAVAILABLE) return;
	    PlotMeanEF ();
	    break;
	case IDD_MedianEF :
	    if (GetEstimatedParameterValue (0) == NOTAVAILABLE) return;
	    PlotMedianEF ();
	    break;
	case IDD_QQPlot:
	    if (GetEstimatedParameterValue (0) == NOTAVAILABLE) return;
	    PlotQQPlot ();
	    break;
	case IDD_Estimate :
	    EvaluateDialog ();
	    PerformEstimation ();
	    break;
	default :
	    inherited::NotifyActionEvent (m);
    }
    if (id == IDD_Density || id == IDD_Quantile || id == IDD_Distribution ||
	id == IDD_MeanEF || id == IDD_MedianEF || id == IDD_Diagram || id == IDD_QQPlot)
	if (CloseBox.Selected ()) CloseDialog ();
}

#endif

/********************************************************************
      Base class of estimators for extreme value distributions
********************************************************************/

#ifdef TERMINAL
TEVDEstimator::TEVDEstimator (String EstimatorName)
 : TEVDGPDBaseEstimator (EstimatorName) {
}
#else
TEVDEstimator::TEVDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : ProgressField (IDD_ProgressField, ""),
   QQPlotLine (IDD_QQPlotLine, GlobalData.QQPlotLine),
   TEVDGPDBaseEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
    *this << ProgressField << QQPlotLine;
}
#endif

TSampleType TEVDEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

BOOL TEVDEstimator::IsEVEstimator () {
    return TRUE;
}

void TEVDEstimator::LeastSquaresEstimate (TUnivariateSample *p) {
    TUnivariateSample &Data = *(TUnivariateSample *) p;
    int k = Data.GetDataSize (), count = 0;
//      d = (Parameters [0] <= 0.5) ? k : MulDiv (4, k, 5);

    double xq = 0.0, yq = 0;
    int i;
    for (i = k / 10; i < k - k / 10; i++) {
        count++;
        yq += Data [i - 1];
        xq += vonMisesEXTQF (double (i) / double (k + 1), RawParameters [0], 1.0, 0.0);
    }
    xq /= count; yq /= count;
    double zaehler = 0.0, nenner = 0.0, x_xq, y_yq;
    for (i = k / 10; i < k - k / 10; i++) {
        x_xq = (vonMisesEXTQF (double (i) / double (k + 1), RawParameters [0], 1.0, 0.0) - xq);
        y_yq = Data[i - 1] - yq;
//      zaehler  += x_xq * y_yq;
//      nenner += x_xq * x_xq;
        zaehler += y_yq * y_yq;
        nenner += x_xq * y_yq;
    }
    if (fabs (nenner) < MYMINDOUBLE) RawParameters [2] = MYMAXDOUBLE;
    else RawParameters [2] = zaehler / nenner;
    RawParameters [1] = yq - RawParameters [2] * xq;
}

double TEVDEstimator::sigmaEst (TUnivariateSample *p) {
    double alpha = RawParameters [0];
    if (alpha == 0.0) return 1.0;
    int i = p->GetDataSize () / 2;
    double v = (*p) [i] * pow (-log ((double) i / (p->GetDataSize () + 1)), 1 / alpha);
    return alpha < 0.0 ? -v : v;
}

void TEVDEstimator::CompleteModels (TEstMode EstMode, int k, TUnivariateSample *q) {
    if (EstMode == ALPHA_MODE) {
        RawParameters [2] = sigmaEst (q);
        RawParameters [1] = 0.0;
    } else if (EstMode == BETA_MODE)
        LeastSquaresEstimate (q);
}


#ifndef TERMINAL

TSample *TEVDEstimator::GetBootstrapSample () {
    TSample *q;
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        if (Parameters [0] > 0.0) {
            q = new TFrechetSample;
            ((TFrechetSample *) q)->SetParameter (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1),
                p->GetDataSize ());
        } else {
            q = new TWeibullSample;
            ((TWeibullSample *) q)->SetParameter (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1),
                p->GetDataSize ());
        }
    else {
        q = new TvonMisesEXTSample;
        ((TvonMisesEXTSample *) q)->SetParameter (
            GetEstimatedParameterValue (0),
            GetEstimatedParameterValue (2),
            GetEstimatedParameterValue (1),
            p->GetDataSize ());
    }
    q->CreateMemory (0);
    return q;
}

double TEVDEstimator::TYearThreshold (double t) {
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        if (Parameters [0] > 0.0)
            return Parameters [1] + Parameters [2] * XFrechetQF (1.0 - 1.0 / t, Parameters [0]);
        else
            return Parameters [1] + Parameters [2] * XWeibullQF (1.0 - 1.0 / t, Parameters [0]);
    else
        return Parameters [1] + Parameters [2] * XEVQF (1.0 - 1.0 / t, Parameters [0]);
}

void TEVDEstimator::ShowCurrentEstimation () {
    if (LastResult != EST_OK)
	ResultField.SetText (ErrorMessages [(int) LastResult]);
    else {
        int n = GetNumberOfEstimatedParameters ();
        String t;
        if (GetEstimatedParameterValue (0) == NOTAVAILABLE) {
            ResultField.SetText ("Alpha mode representation does not exist for gamma = 0");
            return;
        }
        for (int i = 0; i < n; i++)
            t = t & GetEstimatedParameterName (i) & form (25, ": %g  ", GetEstimatedParameterValue (i));
        if (Parameters [0] < 0.0)
            t = t & form (50, "Right endpoint: %g", GetParameter ()->GetMode () == ALPHA_MODE ? Parameters [1] : Parameters [1] - Parameters [2] / Parameters [0]);
        ResultField.SetText (t);
    }
}

void TEVDEstimator::PlotDensity () {
    if (LastResult == EST_OK)
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (GetEstimatedParameterValue (0) > 0.0)
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TFrechetDensity (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
            else
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TWeibullDensity (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
        else
            PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TEVDDensity (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), 1));
}

void TEVDEstimator::PlotQuantile () {
    if (LastResult == EST_OK)
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (Parameters [0] > 0.0)
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TFrechetQuantile (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
            else
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TWeibullQuantile (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
        else
            PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TEVDQuantile (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), 1));
}

void TEVDEstimator::PlotDistribution () {
    if (LastResult == EST_OK)
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (Parameters [0] > 0.0)
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TFrechetDistribution (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
            else
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TWeibullDistribution (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
        else
            PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TEVDDistribution (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), 1));
}

void TEVDEstimator::PlotQQPlot () {
    if (LastResult == EST_OK) {
        TPlotWindow *w = PlotWindowManager.GetOpenWindow ("Q-Q Plot");
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (Parameters [0] > 0.0)
                w->InsertFunction (new TFrechetQQPlot (
                    GetEstimatedParameterValue (0), GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ()));
            else
                w->InsertFunction (new TWeibullQQPlot (
                    GetEstimatedParameterValue (0), GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ()));
        else
            w->InsertFunction (new TEVDQQPlot (
                GetEstimatedParameterValue (0), GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), (TUnivariateSample *) p, Name (), !QQPlotLine.Selected ()));
    }
}


void TEVDEstimator::NotifyInitDialog () {
    inherited::NotifyInitDialog ();
    ModeField.SetSelector (GetParameter ()->GetMode () != ALPHA_MODE);
}

#endif

/********************************************************************
    Base class of estimators for generalized Pareto distributions
********************************************************************/

#ifdef TERMINAL
TGPDEstimator::TGPDEstimator (String EstimatorName)
 : TEVDGPDBaseEstimator (EstimatorName) {
}
#else
TGPDEstimator::TGPDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddMenu)
 : EditExtremes (IDD_Extremes, 10, 0, INT_MAX),
   EditP (IDD_P, GlobalData.trimming_p, DBL_MIN, 1.0),
   Automatic (IDD_Automatic, FALSE),
   ParameterList (IDD_ParameterList, NULL),
   Scrollbar (IDD_ScrollBar, 10, 2, 100),
   TEVDGPDBaseEstimator (EstimatorName, MenuId, IDD_DialogBox, AddMenu) {
    *this << EditExtremes << EditP
          << ParameterList << Automatic << Scrollbar;
}
#endif

TGPDEstimator::~TGPDEstimator () {
    ParameterNames.DeleteAll (TRUE);
}

#ifndef TERMINAL

TSample *TGPDEstimator::GetBootstrapSample () {
    TSample *q;
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        if (Parameters [0] > 0.0) {
            q = new TParetoSample;
            ((TParetoSample *) q)->SetParameter (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1),
                p->GetDataSize ());
        } else {
            q = new TParetoIISample;
            ((TParetoIISample *) q)->SetParameter (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1),
                p->GetDataSize ());
        }
    else {
        q = new TvonMisesGPDSample;
        ((TvonMisesGPDSample *) q)->SetParameter (
            GetEstimatedParameterValue (0),
            GetEstimatedParameterValue (2),
            GetEstimatedParameterValue (1),
            p->GetDataSize ());
    }
    q->CreateMemory (0);
    return q;
}

double TGPDEstimator::TYearThreshold (double t) {
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        if (Parameters [0] > 0.0)
            return Parameters [1] + Parameters [2] * XParetoQF (1.0 - 1.0 / t, Parameters [0]);
        else
            return Parameters [1] + Parameters [2] * XBetaQF (1.0 - 1.0 / t, Parameters [0]);
    else
        return Parameters [1] + Parameters [2] * XGPQF (1.0 - 1.0 / t, Parameters [0]);
}

#endif

double TGPDEstimator::GetEstimatedShapeGamma () {
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        return 1.0 / Parameters [0];
    else
        return Parameters [0];
}

double TGPDEstimator::GetEstimatedMean () {
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        if (Parameters [0] > 0.0)
            if (Parameters [0] <= 1.0)
                return MYMAXDOUBLE;
            else
                return Parameters [1] + Parameters [2] * Parameters [0] / (Parameters [0] - 1.0);
        else
            return Parameters [1] - Parameters [2] * Parameters [0] / (Parameters [0] - 1.0);
    else
        if (Parameters [0] >= 1.0)
            return MYMAXDOUBLE;
        else
            return Parameters [1] + Parameters [2] / (1.0 - Parameters [0]);
}

double TGPDEstimator::GetEstimatedVariance () {
    if (GetParameter ()->GetMode () == ALPHA_MODE)
        if (1.0 / Parameters [0] < 0.5)
            return sqr (Parameters [2]) * Parameters [0] / (sqr (Parameters [0] - 1.0) * (Parameters [0] - 2.0));
	else
            return MYMAXDOUBLE;
    else
        if (Parameters [0] >= 0.5)
            return MYMAXDOUBLE;
        else
            return sqr (Parameters [2]) / (sqr (1.0 - Parameters [0]) * (1.0 - 2.0 * Parameters [0]));
}

TEstResult TGPDEstimator::aQuick (TUnivariateSample &Data, int n, int k, double q, double &alpha, TEstMode Mode) {
    if (k >= n) return EST_PAR;
    if (Data [n - k - 1] <= 0.0) return EST_NEG;
    int index = (int) floor (k * (1.0 - q));
    if (index <= 1) return EST_EXT;
    double nenner = log (Data [n - index] / Data [n - k - 1]);
    if (fabs (nenner) < 1e-300) alpha = MYMAXDOUBLE;
    else alpha = -log (1.0 - q) / nenner;
    if (Mode != ALPHA_MODE) {
        if (alpha != 0.0) alpha = 1.0 / alpha;
        else alpha = MYMAXDOUBLE;
    }
    return EST_OK;
}

double TGPDEstimator::sHill (TUnivariateSample *s, int size, int k, double a, BOOL EstAlphaRec) {
    return (*s)[size - k] * pow ((double) k / size,
                              EstAlphaRec ? fabs (a) : 1.0 / fabs (a));
}

double TGPDEstimator::sDQEst (TUnivariateSample *s, int size, int k, double a, BOOL EstAlphaRec) {
    if ((*s)[size - k] < 0.0) {
        double v = pow ((double) k / size, EstAlphaRec ? fabs (a) : 1.0 / fabs (a));
        return v != 0.0 ? -(*s)[size - k] / v : MYMAXDOUBLE;
    }
    else return sHill (s, size, k, a, EstAlphaRec);
}

// Transformation of models

// Beta mode estimator for mu and sigma

void TGPDEstimator::BetaModeEst (int n, int k, TUnivariateSample *s, double &beta, double &sigma, double &mu) {

    double t = (*s) [n - k];
    double sum1 = 0.0, sum2 = 0.0, help;
    for (int i = 1; i < 2 * k / 3; i++) {
	help = vonMisesGPDQF ((double) i / (double) k, beta, 1.0, 0.0);
	sum1 += help * ((*s) [n - k + i] - t);
	sum2 += help * help;
    }
    help = sum1 / sum2;
    if (fabs (beta) > 1e-10) {
	sigma = help * pow ((double) k / (double) n, beta);
	mu = t - (help - sigma) / beta;
    } else {
	sigma = help;
	mu = t + sigma * log (double (k) / double (n));
    }
}

void TGPDEstimator::CompleteModels (TEstMode EstMode, int k, TUnivariateSample *q) {
    if (EstMode == ALPHA_MODE) {
        RawParameters [2] = sDQEst (q, q->GetDataSize (), k, RawParameters [0], FALSE);
        RawParameters [1] = 0.0;
    } else if (EstMode == BETA_MODE)
        BetaModeEst (q->GetDataSize (), k, q, RawParameters [0], RawParameters [2], RawParameters [1]);
}

#ifndef TERMINAL

void TGPDEstimator::BuildParameterList () {
    ParameterNames.DeleteAll (TRUE);
    for (int i = 0; i < GetNumberOfEstimatedParameters (); i++)
        ParameterNames.Append (new TStringEl (GetEstimatedParameterName (i)));
    ParameterNames.GetFirst ();
    ParameterNames.ToggleActive ();
    ParameterList.SetList (&ParameterNames);
}

int TGPDEstimator::GetUsedK () {
    return EditExtremes.GetValue ();
}

void TGPDEstimator::PerformEstimation () {
    k = EditExtremes.GetValue ();
    Scrollbar.SetSlider (2 + ((TUnivariateSample *) p)->GetDataSize () - k);
    EvaluateAdditionalParameters ();
    LastResult = Estimate (k, p);
    if (LastResult == EST_OK) ShowCurrentEstimation ();
    else ResultField.SetText (ErrorMessages [(int) LastResult]);
}

void TGPDEstimator::ShowCurrentEstimation () {
    if (LastResult != EST_OK)
	ResultField.SetText (ErrorMessages [(int) LastResult]);
    else {
        int n = GetNumberOfEstimatedParameters ();
        String t;
        if (GetEstimatedParameterValue (0) == NOTAVAILABLE) {
            ResultField.SetText ("Alpha mode representation does not exist for gamma = 0");
            return;
        }
        for (int i = 0; i < n; i++)
            t = t & GetEstimatedParameterName (i) & form (25, ": %g  ", GetEstimatedParameterValue (i));
        if (GetParameter ()->GetMode () != ALPHA_MODE && GetEstimatedParameterValue (0) < 0.0)
            t = t & form (50, "Right endpoint: %g", GetEstimatedParameterValue (1) - GetEstimatedParameterValue (2) / GetEstimatedParameterValue (0));
        ResultField.SetText (t);
        Protocol << (Name () & ", k = " & form (10, "%d", GetUsedK ()) & ": " & t);
    }
}

int TGPDEstimator::GetMinK () {
    return -1;
}

void TGPDEstimator::CalcEstimator (Array<double> &, int &kmax, TSample *) {
    kmax = 5;
}

double TGPDEstimator::GetExponent () {
    return 0.0;
}

int TGPDEstimator::EstimateOptK (TSample *q) {
    const int m = 15,
              b = 15;
    Array<double> a (-m, q->GetDataSize ()),
                  as (-m, q->GetDataSize ());
    int kmax, kmin, i, j;
    CalcEstimator (a, kmax, q);
    kmin = GetMinK ();
    for (i = 1; i <= m; i++)
        a [kmin - i] = a [kmin + i];
    if (kmax < kmin) return 0;
    Array<double> median (kmin, kmax);
    for (i = kmin; i <= kmax; i++) {
        j = i;
        as [i] = a [i];
        while (j > kmin && as [j] < as [j - 1]) {
            Swap (as [j], as [j - 1]);
            j--;
        }
        median [i] = as [kmin + (i - kmin) / 2];
    }

    double exponent = GetExponent ();
    for (i = kmin - m; i <= kmax; i++) {
        as [i] = 0.0;
        for (j = kmin - m; j <= i; j++)
            as [i] += pow (j - 1 + m, exponent) * fabs (a [j] - median [max (i, kmin)]);
        as [i] /= (m + i);
    }

    for (i = kmin + b / 2 - m; i <= kmax - b / 2; i++) {
        a [i] = 0.0;
        for (j = -b / 2; j <= b / 2; j++)
            a [i] += as [i + j];
        a [i] /= b;
    }
    for (i = kmin - m; i < kmin + b / 2 - m; i++)
        a [i] = a [kmin + b / 2 - m];
    for (i = kmax - b / 2 + 1; i <= kmax; i++)
        a [i] = a [kmax - b / 2];

    int minindex = kmin;
    double minval = a [kmin];
    for (i = kmin + 1; i <= kmax; i++)
        if (a [i] < minval) {
            minval = a [i];
            minindex = i;
        }
    return minindex;
}

void TGPDEstimator::NotifyInitDialog () {
    inherited::NotifyInitDialog ();
    ModeField.SetSelector (GetParameter ()->GetMode () != ALPHA_MODE);
    BuildParameterList ();
    if (GetMinK () == -1) Automatic.Disable ();
}

void TUnivariateGPDEstimator::NotifyScrollEvent (TScrollBar *m) {
    int k0 = EditExtremes.GetValue ();
    k = 2 + ((TUnivariateSample *) p)->GetDataSize () - Scrollbar.GetSlider ();
    if (k != k0) {
        EditExtremes.SetValue (k);
        PerformEstimation ();
    }
}

void TGPDEstimator::PlotDensity () {
    PerformEstimation ();
    if (LastResult == EST_OK)
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (GetEstimatedParameterValue (0) > 0.0)
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TParetoDensity (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
            else
                PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TBetaDensity (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
        else
            PlotWindowManager.GetOpenWindow (DENSITY)->InsertFunction (new TGPDDensity (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), 1));
}

void TGPDEstimator::PlotQuantile () {
    PerformEstimation ();
    if (LastResult == EST_OK)
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (Parameters [0] > 0.0)
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TParetoQuantile (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
            else
                PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TBetaQuantile (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
        else
            PlotWindowManager.GetOpenWindow (QUANTILE)->InsertFunction (new TGPDQuantile (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), 1));
}

void TGPDEstimator::PlotDistribution () {
    PerformEstimation ();
    if (LastResult == EST_OK)
        if (GetParameter ()->GetMode () == ALPHA_MODE)
            if (Parameters [0] > 0.0)
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TParetoDistribution (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
            else
                PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TBetaDistribution (
                    GetEstimatedParameterValue (0),
                    GetEstimatedParameterValue (2),
                    GetEstimatedParameterValue (1), 1));
        else
            PlotWindowManager.GetOpenWindow (DISTRIBUTION)->InsertFunction (new TGPDDistribution (
                GetEstimatedParameterValue (0),
                GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), 1));
}

void TGPDEstimator::PlotMeanEF () {
    TFunction *f = NULL;
    double p;
    if (EditP.IsCorrect ()) {
        PerformEstimation ();
        if (LastResult == EST_OK) {
            GlobalData.trimming_p = p = EditP.GetValue ();
            if (GetParameter ()->GetMode () == ALPHA_MODE)
                if (GetEstimatedParameterValue (0) > 1.0)
                    f = new TAlphaModeMeanRLF ("Alpha Mode Mean Excess", GetEstimatedParameterValue (0), GetEstimatedParameterValue (1), GetEstimatedParameterValue (2), p);
                else
                    ErrorBox (this, "Estimated alpha must be larger than 1");
            else
                if (GetEstimatedParameterValue (0) < 1.0)
		    f = new TBetaModeMeanRLF ("Gamma Mode Mean Excess",
			GetEstimatedParameterValue (0),
			GetEstimatedParameterValue (2),
			GetEstimatedParameterValue (1), p);
		else
		    ErrorBox (this, "Estimated gamma must be less than 1");
	    if (f)
		if (f->EditParameter (this))
		    PlotWindowManager.GetOpenWindow (MEANEXCESS)->InsertFunction (f);
		else delete f;
	}
    }
}

void TGPDEstimator::PlotMedianEF () {
    TFunction *f;
    PerformEstimation ();
    if (LastResult == EST_OK) {
	if (GetParameter ()->GetMode () == ALPHA_MODE)
	    f = new TAlphaModeMedianRLF ("Alpha Mode Median Excess", GetEstimatedParameterValue (0), GetEstimatedParameterValue (1), GetEstimatedParameterValue (2));
	else
	    f = new TBetaModeMedianRLF ("Gammaa Mode Median Excess",
		GetEstimatedParameterValue (0),
		GetEstimatedParameterValue (2),
		GetEstimatedParameterValue (1));
	PlotWindowManager.GetOpenWindow (MEDIANEXCESS)->InsertFunction (f);
    }
}

#endif


#ifdef TERMINAL
TUnivariateGPDEstimator::TUnivariateGPDEstimator (String EstimatorName)
 : TGPDEstimator (EstimatorName) {
}
#else
TUnivariateGPDEstimator::TUnivariateGPDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : TGPDEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList) {
}
#endif

TSampleType TUnivariateGPDEstimator::GetSampleType () {
    return UNIVARIATE_REQUIRED;
}

BOOL TUnivariateGPDEstimator::IsGPEstimator () {
    return TRUE;
}

#ifndef TERMINAL
void TUnivariateGPDEstimator::PlotQQPlot () {
    PerformEstimation ();
    TPlotWindow *w = PlotWindowManager.GetOpenWindow ("Q-Q Plot");
    if (GetParameter ()->GetMode () == ALPHA_MODE)
	if (Parameters [0] > 0.0)
            w->InsertFunction (new TParetoQQPlot (
                GetEstimatedParameterValue (0), GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), (TUnivariateSample *) p, Name (), FALSE));
        else
            w->InsertFunction (new TBetaQQPlot (
                GetEstimatedParameterValue (0), GetEstimatedParameterValue (2),
                GetEstimatedParameterValue (1), (TUnivariateSample *) p, Name (), FALSE));
    else
        w->InsertFunction (new TGPDQQPlot (
            GetEstimatedParameterValue (0), GetEstimatedParameterValue (2),
            GetEstimatedParameterValue (1), (TUnivariateSample *) p, Name (), FALSE));
}

void TUnivariateGPDEstimator::NotifyActionEvent (TDialogMember *m) {
    int id = m->GetID ();
    switch (id) {
	case IDD_Diagram: {
	    PerformEstimation ();
	    for (int i = 0; i < GetNumberOfEstimatedParameters (); i++)
		if (ParameterList.IsEntrySelected (i)) {
		    BOOL NewWindow = !PlotWindowManager.IsOpen (GetEstimatedParameterName (i));
		    TPlotWindow *w = PlotWindowManager.GetOpenWindow (GetEstimatedParameterName (i));
		    w->InsertFunction (new TEstDiagram ((TGPDEstimator *) Duplicate (), (TUnivariateSample *) p, i));
		    if (NewWindow) {
			w->InsertFunction (new TTextField (TDblPoint (0, 0), "Number of exceedances", FALSE, TRUE), TRUE);
			w->InsertFunction (new TTextField (TDblPoint (0, 0), "Estimates", TRUE, FALSE), TRUE);
			w->Repaint ();
		    }
		}
	    }
	    inherited::NotifyActionEvent (m);
	    break;
	case IDD_Automatic: {
	    BOOL Sel = Automatic.Selected ();
            if (Sel) {
                EditExtremes.SetValue (EstimateOptK (p));
                EditExtremes.Disable ();
            } else EditExtremes.Enable ();
            if (!Sel) break; }
            // fall-through if Automatic Selection
        case IDD_Estimate:
            k = EditExtremes.GetValue ();
	    if (k > ((TUnivariateSample *) p)->GetDataSize () || k < 2)
                EditExtremes.SetValue (((TUnivariateSample *) p)->GetDataSize ());
	    PerformEstimation ();
            break;
        default:
            inherited::NotifyActionEvent (m);
    }
}

void TUnivariateGPDEstimator::NotifyInitDialog () {
    inherited::NotifyInitDialog ();
    k = ((TUnivariateSample *) p)->GetDataSize () / 10;
    Scrollbar.SetRange (2, ((TUnivariateSample *) p)->GetDataSize ());
    EditExtremes.SetValue (k);
}

#endif


#ifdef TERMINAL
TGroupedGPDEstimator::TGroupedGPDEstimator (String EstimatorName)
 : TGPDEstimator (EstimatorName) {
}
#else
TGroupedGPDEstimator::TGroupedGPDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList)
 : TGPDEstimator (EstimatorName, MenuId, IDD_DialogBox, AddList),
   ProgressField (IDD_ProgressField) {
    *this << ProgressField;
}
#endif

TSampleType TGroupedGPDEstimator::GetSampleType () {
    return GROUPED_SAMPLE;
}

#ifndef TERMINAL
void TGroupedGPDEstimator::NotifyInitDialog () {
    inherited::NotifyInitDialog ();
    k = 2;
    EditExtremes.SetValue (k);
    Scrollbar.SetRange (2, ((TGroupedSample *) p)->GetDataSize ());
}

/* !!!!
void TGroupedGPDEstimator::OnVScroll(HWND hwnd, HWND hwndCtl, UINT code, int pos) {
    int d = 0;
    if (code == SB_LINEUP) d = 1;
    else if (code == SB_LINEDOWN) d = -1;
    int k0 = EditExtremes.GetValue ();
    k = k0 + d;
    Adjust (k, 2, ((TGroupedSample *) p)->GetGroupCount ());
    if (k != k0) {
        EditExtremes.SetValue (k);
        PerformEstimation ();
    }
}
*/

void TGroupedGPDEstimator::PlotQQPlot () {
    PerformEstimation ();
    PlotWindowManager.GetOpenWindow ("Q-Q Plot")->InsertFunction (
        new TQQPlotGrouped (k, Parameters [0], Parameters [2], Parameters [1], (TGroupedSample *) p,
                               p->Name (), GetParameter ()->GetMode (), FALSE));
}

void TGroupedGPDEstimator::Diagram () {
     PerformEstimation ();
     for (int i = 0; i < GetNumberOfEstimatedParameters (); i++)
         if (ParameterList.IsEntrySelected (i)) {
             TPlotWindow *w = PlotWindowManager.GetOpenWindow (GetEstimatedParameterName (i));
             w->InsertFunction (new TGroupedEstDiagram ((TGroupedGPDEstimator *) Duplicate (), (TGroupedSample *) p, i));
         }

}

void TGroupedGPDEstimator::NotifyActionEvent (TDialogMember *m) {
    switch (m->GetID ()) {
        case IDD_Diagram:
            Diagram ();
            if (CloseBox.Selected ()) CloseDialog ();
	    break;
	case IDM_BootstrapConfidence:
	    ErrorBox (this, "A bootstrap confidence interval only available for univariate estimators");
	    break;
	default:
            inherited::NotifyActionEvent (m);
            break;
    }
}

void TGroupedGPDEstimator::NotifyScrollEvent (TScrollBar *m) {
    int k0 = EditExtremes.GetValue ();
    k = 2 + ((TGroupedSample *) p)->GetDataSize () - Scrollbar.GetSlider ();
    if (k != k0) {
        EditExtremes.SetValue (k);
        PerformEstimation ();
    }
}

#endif

TEstimatorList *EstimatorListPtr = NULL;

TEstimator *TEstimatorList::GetEntryEst (int id) {
    TEstimator *p = GetFirstEst ();
    while (p && p->GetMenuId () != id)
        p = GetNextEst ();
    return p;
}



