/*  estimate.hpp

    Class definitions for estimators.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 2000 Xtremes Group, Siegen.
*/

#ifndef ESTIMATE_HPP
#define ESTIMATE_HPP

#include "sample.hpp"
#include "samplist.hpp"
#include "stddef.hpp"

#ifndef TERMINAL
#include "dbldlg.hpp"
#include "stddlg.hpp"
#endif

enum TEstMode    { ALPHA_MODE, BETA_MODE, FULL_ALPHA_MODE, FULL_BETA_MODE };
enum TEstResult  { EST_OK, EST_NEG, EST_POZ, EST_EXT, EST_PAR, EST_SIGN, EST_SAMPSIZE, EST_XPL_FATAL,
		   EST_FAILURE };

class TEstimatorData : public TObject {
    public:
	void SetMode (TEstMode m)	{ Mode = m; }
	TEstMode GetMode ()		{ return Mode; }
    private:
	TEstMode Mode;
};

#ifdef TERMINAL
class TEstimator : public TObject {
#else
class TEstimator : public TModelessDialog {
    typedef TModelessDialog inherited;
#endif
    public:
#ifdef TERMINAL
	TEstimator (String EstimatorName);
#else
	TEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddMenu);
	virtual int GetUsedK ();
#endif
	virtual void PerformEstimation ();
	virtual String Name ();

	virtual TSampleType GetSampleType () = 0;
	virtual TEstimator *Duplicate () = 0;
	int GetMenuId ()	{ return menuId; }

	virtual int GetNumberOfEstimatedParameters () = 0;
	virtual double GetEstimatedParameterValue (int);
	virtual String GetEstimatedParameterName (int) = 0;
	virtual TEstResult Estimate (int k, TSample *) = 0;

        // Classification
        virtual BOOL IsSumEstimator ();
        virtual BOOL IsEVEstimator ();
        virtual BOOL IsGPEstimator ();

        // Parameters
        virtual double GetEstimatedShapeGamma ();
        virtual double GetEstimatedMean ();
        virtual double GetEstimatedVariance ();

#ifndef TERMINAL
	void BootstrapConfidenceInterval (int NrSims, int Parameter, double alpha, int k, double &left, double &right, double &mse);
	virtual double TYearThreshold (double t);
#endif
    protected:
	TSample *p;
#ifndef TERMINAL
	virtual void NotifyEvent (int ev);
	virtual void NotifyActionEvent (TDialogMember *m);
	virtual int GetSubMenuPos ();
	virtual TSample *GetBootstrapSample ();
	virtual void EvaluateAdditionalParameters ();

	virtual void NotifyInitDialog ();
	virtual void ShowCurrentEstimation ();	// assumes IDD_SampleData and IDD_Result
	TStaticField SampleField, ResultField;
#endif
	// SearchMax searches a maximum of the function SearchL within SearchBounds.
        // Start values and the result are stored in Parameters.
        BOOL SearchMax (int Dimension);
        virtual double SearchL (double *);
        virtual BOOL SearchBounds (double *);

	double Parameters [10];
	double pValue;
	String name;
	int menuId;
	TEstResult LastResult;		// stores estimator result from dialog operations
};

class TEstimatorList : public TList {
    public:
	TEstimator *GetFirstEst () 		{ return (TEstimator *) GetFirst (); }
	TEstimator *GetNextEst ()	 	{ return (TEstimator *) GetNext (); }
	TEstimator *GetEntryEst (int MenuId);
};

/********************************************************************
			Gaussian distribution
********************************************************************/

class TNormalEstimator : public TEstimator {
    typedef TEstimator inherited;
    public:
#ifdef TERMINAL
	TNormalEstimator (String EstimatorName);
#else
	TNormalEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
	virtual int GetNumberOfEstimatedParameters ();
	virtual String GetEstimatedParameterName (int);
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
        virtual BOOL IsSumEstimator ();
        virtual double GetEstimatedMean ();
        virtual double GetEstimatedVariance ();
    protected:
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void NotifyActionEvent (TDialogMember *m);
	TStaticField ProgressField;
	TSelectField QQPlotLine, CloseBox;
#endif
};

/********************************************************************
			Stable distribution
********************************************************************/

#ifdef USE_STABLE

class TStableEstimator : public TEstimator {
    typedef TEstimator inherited;
    public:
#ifdef TERMINAL
	TStableEstimator (String EstimatorName);
#else
	TStableEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
	virtual int GetNumberOfEstimatedParameters ();
	virtual String GetEstimatedParameterName (int);
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
    protected:
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void NotifyActionEvent (TDialogMember *m);
	TStaticField ProgressField;
	TSelectField QQPlotLine, CloseBox;
#endif
};

#endif

/********************************************************************
			T distribution
********************************************************************/

class TTEstimator : public TEstimator {
    typedef TEstimator inherited;
    public:
#ifdef TERMINAL
	TTEstimator (String EstimatorName);
#else
	TTEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
	virtual int GetNumberOfEstimatedParameters ();
	virtual String GetEstimatedParameterName (int);
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
    protected:
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void NotifyActionEvent (TDialogMember *m);
	TStaticField ProgressField;
	TSelectField QQPlotLine, CloseBox;
#endif
};

class TNCTEstimator : public TEstimator {
    typedef TEstimator inherited;
    public:
#ifdef TERMINAL
	TNCTEstimator (String EstimatorName);
#else
	TNCTEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
	virtual int GetNumberOfEstimatedParameters ();
	virtual String GetEstimatedParameterName (int);
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
    protected:
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void NotifyActionEvent (TDialogMember *m);
	TStaticField ProgressField;
	TSelectField QQPlotLine, CloseBox;
#endif
};

/********************************************************************
			Gumbel distribution
********************************************************************/

class TGumbelEstimator : public TEstimator {
    typedef TEstimator inherited;
    public:
#ifdef TERMINAL
	TGumbelEstimator (String EstimatorName);
#else
	TGumbelEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
	virtual int GetNumberOfEstimatedParameters ();
	virtual String GetEstimatedParameterName (int);
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
    protected:
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void NotifyActionEvent (TDialogMember *m);
	TStaticField ProgressField;
	TSelectField QQPlotLine, CloseBox;
#endif
};

/********************************************************************
      Base class of estimators for extreme value and
      generalized Pareto distributions
********************************************************************/

class TEVDGPDBaseEstimator : public TEstimator {
    typedef TEstimator inherited;
    public:
#ifdef TERMINAL
	TEVDGPDBaseEstimator (String EstimatorName);
#else
	TEVDGPDBaseEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TEstimatorData *GetParameter () = 0;
	virtual int GetNumberOfEstimatedParameters ();
	virtual String GetEstimatedParameterName (int);
    protected:
	int k;					// GPD only

	double RawParameters [3];		// shape, scale, location parameter as returned by estimator
	virtual TEstMode GetEstimatorMode () = 0;			// mode used estimator
	virtual void TransformModels (int k, TUnivariateSample *q);	// Adjustment of models
	virtual void CompleteModels (TEstMode EstMode, int k, TUnivariateSample *q) = 0;	// Default scale/loc. param.
	virtual void BuildParameterList ();

#ifndef TERMINAL
	TAlterField ModeField;
	TSelectField CloseBox;
	virtual void NotifyActionEvent (TDialogMember *m);

	virtual void PlotDensity ();
	virtual void PlotQuantile ();
	virtual void PlotDistribution ();
	virtual void PlotMeanEF ();
	virtual void PlotMedianEF ();
	virtual void PlotQQPlot ();
#endif
};

class TEVDEstimator : public TEVDGPDBaseEstimator {
    typedef TEVDGPDBaseEstimator inherited;
    public:
#ifdef TERMINAL
	TEVDEstimator (String EstimatorName);
#else
	TEVDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
        virtual BOOL IsEVEstimator ();
    protected:
	void LeastSquaresEstimate (TUnivariateSample *p);	// location/scale parameters in beta-mode
	double sigmaEst (TUnivariateSample *p);			// scale parameter in alpha-mode
	virtual void CompleteModels (TEstMode EstMode, int k, TUnivariateSample *q);	// Default scale/loc. param.
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void ShowCurrentEstimation ();
	virtual void NotifyInitDialog ();
	virtual void PlotDensity ();
	virtual void PlotQuantile ();
	virtual void PlotDistribution ();
	virtual void PlotQQPlot ();

	TStaticField ProgressField;
	TSelectField QQPlotLine;
#endif
};

class TGPDEstimator : public TEVDGPDBaseEstimator {
    typedef TEVDGPDBaseEstimator inherited;
    public:
#ifdef TERMINAL
	TGPDEstimator (String EstimatorName);
#else
	TGPDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddMenu);
	virtual int GetUsedK ();
#endif
	virtual ~TGPDEstimator ();
#ifndef TERMINAL
	virtual double TYearThreshold (double t);
#endif
        virtual double GetEstimatedShapeGamma ();
        virtual double GetEstimatedMean ();
        virtual double GetEstimatedVariance ();
    protected:
	TEstResult aQuick (TUnivariateSample &Data, int n, int k, double q, double &alpha, TEstMode Mode);
	void BetaModeEst (int n, int k, TUnivariateSample *, double &beta, double &sigma, double &mu);
	double sHill (TUnivariateSample *s, int size, int k, double a, BOOL EstAlphaRec);
	double sDQEst (TUnivariateSample *s, int size, int k, double a, BOOL EstAlphaRec);


	virtual void CompleteModels (TEstMode EstMode, int k, TUnivariateSample *q);	// Default scale/loc. param.
	TList ParameterNames;
#ifndef TERMINAL
	virtual TSample *GetBootstrapSample ();
	virtual void BuildParameterList ();
	virtual void PerformEstimation ();
	virtual void ShowCurrentEstimation ();

	virtual void PlotDensity ();
	virtual void PlotQuantile ();
	virtual void PlotDistribution ();
	virtual void PlotMeanEF ();
	virtual void PlotMedianEF ();

	virtual int EstimateOptK (TSample *);	// automatic selection of number of extremes
	virtual double GetExponent ();
	virtual int GetMinK ();
	virtual void CalcEstimator (Array <double> &, int &kmax, TSample *);

	virtual void NotifyInitDialog ();

	TIntField EditExtremes;
	TDoubleField EditP;
	TSelectField Automatic;
	TListField ParameterList;
        TScrollBar Scrollbar;
	String ErrorMessage;
#endif
};

class TUnivariateGPDEstimator : public TGPDEstimator {
    typedef TGPDEstimator inherited;
    public:
#ifdef TERMINAL
	TUnivariateGPDEstimator (String EstimatorName);
#else
	TUnivariateGPDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
        virtual BOOL IsGPEstimator ();
    protected:
#ifndef TERMINAL
	virtual void NotifyInitDialog ();
	virtual void PlotQQPlot ();
	virtual void NotifyActionEvent (TDialogMember *m);
	virtual void NotifyScrollEvent (TScrollBar *m);
#endif
};

class TGroupedGPDEstimator : public TGPDEstimator {
    typedef TGPDEstimator inherited;
    public:
#ifdef TERMINAL
	TGroupedGPDEstimator (String EstimatorName);
#else
	TGroupedGPDEstimator (String EstimatorName, int MenuId, int IDD_DialogBox, BOOL AddList);
#endif
	virtual TSampleType GetSampleType ();
    protected:
#ifndef TERMINAL
	TStaticField ProgressField;
	virtual void NotifyInitDialog ();
	virtual void NotifyActionEvent (TDialogMember *m);
	virtual void NotifyScrollEvent (TScrollBar *m);
	virtual void PlotQQPlot ();
	void Diagram ();
#endif
};

extern TEstimatorList *EstimatorListPtr;

#endif
