/*  evdest.cpp

    Estimators for extreme value distributions.

    This file is part of Xtremes for Windows,
    Copyright (C) 1994 Xtremes Group, Siegen.
*/

// b, m, s

#include <math.h>
#include <stdlib.h>

#include "allest.hpp"
#include "mathfunc.hpp"
#include "stddef.hpp"

#ifndef TERMINAL
    #include "qqplot.hpp"
    #include "wintypes.hpp"

    #pragma hdrstop
    #include "resource.h"
#endif

/********************************************************************
                Ratio of Differences (RD) estimtor
********************************************************************/

TDiffRatioEstimator::TDiffRatioEstimator (BOOL AddList)
#ifdef TERMINAL
 : TEVDEstimator ("LRSE (EV)") {
#else
 : TEVDEstimator ("LRSE (EV)", IDM_EVDDiffRatio, IDD_EVDEstimator, AddList) {
#endif
    Data.SetMode (BETA_MODE);
}

TEstimator *TDiffRatioEstimator::Duplicate () {
    return new TDiffRatioEstimator;
}


double TDiffRatioEstimator::RDEEstimate (int i, double n, TUnivariateSample &Sample) {
    double a = sqrt (log ((n-i)/n) / log (i/n)),
           q0 = i / n,
           q1 = pow (q0, a),
           q2 = (n-i) / n;
    double v1 = Sample [(int) floor (n * q2) - 1],
           v2 = Sample [(int) floor (n * q1) - 1],
           v3 = Sample [(int) floor (n * q0) - 1];

    if (v2 == v3) return MYMAXDOUBLE;
    return log ((v1 - v2) / (v2 - v3)) / log (1.0 / a);
}

TEstResult TDiffRatioEstimator::Estimate (int k, TSample *p) {
    TUnivariateSample &Sample = *(TUnivariateSample *) p;
    int n = Sample.GetDataSize ();
    RawParameters [0] = 0.0;
    for (int i = 1; i <= n / 4; i++)
        RawParameters [0] += RDEEstimate (i, n, Sample);
    RawParameters [0] /= (n / 4);

    if (RawParameters [0] == 0.0) RawParameters [0] = 1e-5;
    LeastSquaresEstimate (&Sample);
    if (RawParameters [0] < 0.0)
        RawParameters [1] = max (RawParameters [1], (Sample [n - 1] + RawParameters [2] / RawParameters [0]));

/*
    TUnivariateSample &Sample = *(TUnivariateSample *) p;
    if (Sample.GetDataSize () < 10) return EST_SAMPSIZE;
    int n = Sample.GetDataSize ();
    if (n < 10) return EST_SAMPSIZE;
    const double q1 = 0.1,
                 q2 = 0.464,
                 q3 = 0.774,
                 q4 = 0.918;
    double zaehler1 = Sample [floor(n*q3)-1] - Sample [floor(n*q2)-1],
           nenner1  = Sample [floor(n*q2)-1] - Sample [floor(n*q1)-1],
           zaehler2 = Sample [floor(n*q4)-1] - Sample [floor(n*q3)-1],
           nenner2  = Sample [floor(n*q3)-1] - Sample [floor(n*q2)-1];

    if (nenner1 == 0.0 || nenner2 == 0.0) RawParameters [0] = MYMAXDOUBLE;
    else RawParameters [0] = (0.2 * log (zaehler1 / nenner1) + 0.8 * log (zaehler2 / nenner2)) / log (3.0);
    LeastSquaresEstimate (&Sample);
    if (RawParameters [0] < 0.0)
        RawParameters [1] = max (RawParameters [1], Sample [n - 1] + RawParameters [2] / RawParameters [0]);
*/

    TransformModels (k, &Sample);
    return EST_OK;
}

TEstMode TDiffRatioEstimator::GetEstimatorMode () {
    return FULL_BETA_MODE;
}

TEstimatorData *TDiffRatioEstimator::GetParameter () {
    return &Data;
}

/********************************************************************
        Maximum likelihood estimator for Frechet distribution
********************************************************************/

TFrechetMLEstimator::TFrechetMLEstimator (BOOL AddList)
#ifdef TERMINAL
 : TEVDEstimator ("MLE (EV 1/EV 2)") {
#else
 : TEVDEstimator ("MLE (EV 1/EV 2)", IDM_MLFrechet, IDD_EVDEstimator, AddList) {
#endif
    Data.SetMode (ALPHA_MODE);
}

TEstimator *TFrechetMLEstimator::Duplicate () {
    return new TFrechetMLEstimator;
}

double TFrechetMLEstimator::h0 (double alpha, double sign) {
    double sum = 0.0;
    for (int i = Sample->GetDataSize () - 1; i >= 0; i--)
        sum += exp (-sign * log (sign * (*Sample)[i]) / alpha);
    return sum;
}

double TFrechetMLEstimator::h1 (double alpha, double sign) {
    double sum = 0.0;
    for (int i = Sample->GetDataSize () - 1; i >= 0; i--)
        sum += sign * log (sign * (*Sample)[i]) * exp (-sign * log (sign * (*Sample)[i]) / alpha);
    return sum;
}

double TFrechetMLEstimator::h2 (double alpha, double sign) {
    double sum = 0.0;
    for (int i = Sample->GetDataSize () - 1; i >= 0; i--)
        sum += log (sign * (*Sample)[i]) * log (sign * (*Sample)[i]) * exp (-sign * log (sign * (*Sample)[i]) / alpha);
    return sum;
}

TEstResult TFrechetMLEstimator::DoEstimation (TSample *p) {
    const double EPSILON = 1e-10;
    const double MAXSTEP = 50;
    Sample = (TUnivariateSample *) p;
    int n = Sample->GetDataSize ();
    if (!(( ((*Sample)[0] < 0.0) && ((*Sample)[n-1] < 0.0)) || ( ((*Sample)[0] > 0.0) && ((*Sample)[n-1] > 0.0) ))) {
        RawParameters [0] = RawParameters [2] = 1.0;
        return EST_NEG;
    }

    double alpha0 = 0.0, alpha1, xm = 0.0,
           sign = (*Sample)[0] < 0.0 ? -1.0 : 1.0;
    int i;
    for (i = Sample->GetDataSize () - 1; i >= 0; i--)
        xm += sign * log (sign * (*Sample)[i]) / n;
    for (i = Sample->GetDataSize () - 1; i >= 0; i--)
        alpha0 += (sign * log (sign * (*Sample)[i]) - xm) * (sign * log (sign * (*Sample)[i]) - xm);
    alpha0 = sqrt (alpha0 * 6) / (MY_PI * (n - 1));
    if (!alpha0)
        throw EST_FAILURE;
    int StepCount = 0;
    do {
        alpha1 = alpha0;
        double h0w = h0 (alpha0, sign),
               h1w = h1 (alpha0, sign),
               h2w = h2 (alpha0, sign);
        alpha0 = alpha0 - (alpha0 - xm + h1w / h0w) /
                  (1.0 + (h0w * h2w - h1w * h1w) / (alpha0 * alpha0 * h0w * h0w));
    } while (StepCount++ < MAXSTEP && fabs (alpha0 - alpha1) > EPSILON);
    RawParameters [0] = sign / alpha0;
    RawParameters [2] = exp (-alpha0 * sign * log (h0 (alpha0, sign) / n));
    RawParameters [1] = 0.0;
    TransformModels (k, Sample);
    return EST_OK;
}

TEstResult TFrechetMLEstimator::Estimate (int k, TSample *p) {
    try {
        return DoEstimation (p);
    }
    catch (TEstResult t) {
        Parameters [0] = Parameters [1] = Parameters [2] = 1.0;
        return t;
    }
}

TEstMode TFrechetMLEstimator::GetEstimatorMode () {
    return FULL_ALPHA_MODE;
}

TEstimatorData *TFrechetMLEstimator::GetParameter () {
    return &Data;
}

/********************************************************************
        Maximum likelihood estimator for Gumbel distribution
********************************************************************/

TGumbelMLEstimator::TGumbelMLEstimator (BOOL AddList)
#ifdef TERMINAL
 : TGumbelEstimator ("MLE (EV 0)") {
#else
 : TGumbelEstimator ("MLE (EV 0)", IDM_MLEEVD3, IDD_GaussEstimator, AddList) {
#endif
}

TEstimator *TGumbelMLEstimator::Duplicate () {
    return new TGumbelMLEstimator;
}

double TGumbelMLEstimator::h0 (double alpha) {
    double sum = 0.0;
    for (int i = Sample->GetDataSize () - 1; i >= 0; i--)
//      sum += exp (-log ((*Sample)[i]) / alpha);
        sum += exp (-(*Sample)[i] / alpha);
    return sum;
}

double TGumbelMLEstimator::h1 (double alpha) {
    double sum = 0.0;
    for (int i = Sample->GetDataSize () - 1; i >= 0; i--)
//      sum += log ((*Sample)[i]) * exp (-log ((*Sample)[i]) / alpha);
        sum += (*Sample)[i] * exp (-(*Sample)[i] / alpha);
    return sum;
}

double TGumbelMLEstimator::h2 (double alpha) {
    double sum = 0.0;
    for (int i = Sample->GetDataSize () - 1; i >= 0; i--)
//      sum += log ((*Sample)[i]) * log ((*Sample)[i]) * exp (-log ((*Sample)[i]) / alpha);
        sum += (*Sample)[i] * (*Sample)[i] * exp (-(*Sample)[i] / alpha);
    return sum;
}

TEstResult TGumbelMLEstimator::Estimate (int k, TSample *p) {
    const double EPSILON = 1e-10;
    const double MAXSTEP = 50;
    Sample = (TUnivariateSample *) p;
    int n = Sample->GetDataSize ();

    double alpha0 = 0.0, alpha1, xm = 0.0;
    int i;
    for (i = Sample->GetDataSize () - 1; i >= 0; i--)
        xm += (*Sample)[i] / n;
    for (i = Sample->GetDataSize () - 1; i >= 0; i--)
        alpha0 += ((*Sample)[i] - xm) * ((*Sample)[i] - xm);
    alpha0 = sqrt (alpha0 * 6 / (n - 1)) / MY_PI;
    int StepCount = 0;
    do {
        alpha1 = alpha0;
        double h0w = h0 (alpha0),
               h1w = h1 (alpha0),
               h2w = h2 (alpha0);
        alpha0 = alpha0 - (alpha0 - xm + h1w / h0w) /
                  (1.0 + (h0w * h2w - h1w * h1w) / (alpha0 * alpha0 * h0w * h0w));
    } while (StepCount++ < MAXSTEP && fabs (alpha0 - alpha1) > EPSILON);
    Parameters [1] = alpha0;
    Parameters [0] = -alpha0 * log (h0 (alpha0) / n);
    return EST_OK;
}

/********************************************************************
    Maximum likelihood estimator for extreme value distribution
********************************************************************/


TEVDMLEstimator::TEVDMLEstimator (BOOL AddList)
#ifdef TERMINAL
 : TEVDEstimator ("MLE (EV)") {
#else
 : TEVDEstimator ("MLE (EV)", IDM_MLEVD, IDD_EVDEstimator, AddList) {
#endif
    Data.SetMode (BETA_MODE);
}

TEstimator *TEVDMLEstimator::Duplicate () {
    TEVDMLEstimator *p = new TEVDMLEstimator;
    p->Data = Data;
    return p;
}

const int n = 3;

static double Det (double a [3][3]) {
    double sum = 0.0;
    for (int i = 0; i < 3; i++) {
        sum += a[0][i%3]*a[1][(i+1)%3]*a[2][(i+2)%3];
        sum -= a[0][(i+2)%3]*a[1][(i+1)%3]*a[2][i%3];
    }
    return sum;
}

static int IsPositiveDefinit (double a [3][3]) {
    return a [0][0] > 0.0 &&
           a [0][0] * a [1][1] - a [0][1] * a [1][0] > 0.0 &&
           Det (a) > 0.0;
}

static void Invert (double a [n][n]) {
    int i, j, k, r;
    double maxv, hr;
    double hv [n];
    int p [n];
    for (j = 0; j < n; j++) p [j] = j;
    for (j = 0; j < n; j++) {
        maxv = fabs (a [j][j]);
        r = j;
        for (i = j + 1; i < n; i++)
            if (fabs (a [i][j]) > maxv) {
                maxv = fabs (a [i][j]);
                r = i;
            }
        if (maxv == 0.0)
            throw EST_FAILURE;
        if (r > j) {
            for (k = j; k < n; k++) Swap (a [j][k], a [r][k]);
            Swap (p [j], p [r]);
        }
        hr = 1.0 / a [j][j];
        for (i = 0; i < n; i++)
            a [i][j] *= hr;
        a [j][j] = hr;
        for (k = 0; k < n; k++)
            if (k != j) {
                for (i = 0; i < n; i++)
                    if (i != j)
                        a [i][k] -= a[i][j] * a[j][k];
                a [j][k] *= -hr;
            }
    }
    for (i = 0; i < n; i++) {
        for (k = 0; k < n; k++) hv [p [k]] = a [i][k];
        for (k = 0; k < n; k++) a [i][k] = hv [k];
    }
}


BOOL TEVDMLEstimator::IsRangeValid (double beta, double sigma, double mu) {
    return beta * (xmin - mu) / sigma > -1.0 &&
           beta * (xmax - mu) / sigma > -1.0 &&
           sigma > 0.0;
}

double TEVDMLEstimator::l (double beta, double sigma, double mu) {
    TUnivariateSample &Sample = *(TUnivariateSample *) p;
    int i;
    double v, w, sum = 0.0;
    for (i = 0; i < Sample.GetDataSize (); i++) {
        v = 1.0 + beta * (Sample [i] - mu) / sigma;
        if (v <= 0.0)
            throw EST_FAILURE;
        w = pow (v, -1.0 / beta);
        sum += -w + log (w) - log (v);
    }
    return sum - Sample.GetDataSize () * log (sigma);
}

const double EPSILON = 1.e-7;

void TEVDMLEstimator::CalcDL (double beta, double sigma, double mu,
                              double &dldb, double &dlds, double &dldm) {
    TUnivariateSample &Sample = *(TUnivariateSample *) p;
    dldb = dlds = dldm = 0.0;
    double dhdb,
           dhdm = -beta / sigma,
           dhds, h, hr,
           br = -1.0 / beta,
           brq = br * br,
           logh, power, t;
    for (int i = 0; i < Sample.GetDataSize (); i++) {
        dhdb = (Sample [i] - mu) / sigma;
        dhds = dhdb * dhdm;
        h = 1.0 + beta * dhdb;
        hr = 1.0 / h;
        logh = log (h);
        power = 1.0 - exp (logh * br);
        t = dhdb * hr;
        dldb += (t * br + logh * brq) * power - t;
        t = dhds * hr;
        dlds += t * br * power - t;
        t = dhdm * hr;
        dldm += t * br * power - t;
    }
    dlds -= Sample.GetDataSize () / sigma;
}

double TEVDMLEstimator::dldb (double beta, double sigma, double mu) {
    if (IsRangeValid (beta + EPSILON, sigma, mu))
        return (l (beta + EPSILON, sigma, mu) - l (beta, sigma, mu)) / EPSILON;
    else
        throw EST_FAILURE;
#ifdef _MSC_VER
    return 0.0;
#endif
}

double TEVDMLEstimator::dlds (double beta, double sigma, double mu) {
    if (IsRangeValid (beta, sigma + EPSILON, mu))
        return (l (beta, sigma + EPSILON, mu) - l (beta, sigma, mu)) / EPSILON;
    else
        throw EST_FAILURE;
#ifdef _MSC_VER
    return 0.0;
#endif
}

double TEVDMLEstimator::dldm (double beta, double sigma, double mu) {
    if (IsRangeValid (beta, sigma, mu + EPSILON))
         return (l (beta, sigma, mu + EPSILON) - l (beta, sigma, mu)) / EPSILON;
    else
        throw EST_FAILURE;
#ifdef _MSC_VER
    return 0.0;
#endif
}

void TEVDMLEstimator::CalcDl (double beta, double sigma, double mu, double a [3][3],
                              double &dldbeta, double &dldsigma, double &dldmu) {
    CalcDL (beta, sigma, mu, dldbeta, dldsigma, dldmu);
    double dldb, dlds, dldm;
    CalcDL (beta + EPSILON, sigma, mu, dldb, dlds, dldm);
    a [0][0] = -(dldb - dldbeta) / EPSILON;

    CalcDL (beta, sigma + EPSILON, mu, dldb, dlds, dldm);
    a [0][1] = -(dldb - dldbeta) / EPSILON;
    a [1][1] = -(dlds - dldsigma) / EPSILON;

    CalcDL (beta, sigma, mu + EPSILON, dldb, dlds, dldm);
    a [0][2] = -(dldb - dldbeta) / EPSILON;
    a [1][2] = -(dlds - dldsigma) / EPSILON;
    a [2][2] = -(dldm - dldmu) / EPSILON;

    a [1][0] = a [0][1];
    a [2][0] = a [0][2];
    a [2][1] = a [1][2];
}

BOOL TEVDMLEstimator::Iter (double &b, double &s, double &m) {
    double d, rho = 1.0;
    double a [3][3];
    double stepbeta, stepsigma, stepmu, betas, sigmas, mus;
    CalcDl (b, s, m, a, stepbeta, stepsigma, stepmu);
    if (!IsPositiveDefinit (a)) {
        d = sqrt (stepbeta * stepbeta + stepsigma * stepsigma + stepmu * stepmu);
        if (!d)
            throw EST_FAILURE;
        stepbeta /= d;
        stepsigma /= d;
        stepmu /= d;
    } else {
        Invert (a);
        double beta = stepbeta,
               sigma = stepsigma,
               mu = stepmu;
        stepbeta  = a[0][0] * beta + a[0][1] * sigma + a[0][2] * mu,
        stepsigma = a[1][0] * beta + a[1][1] * sigma + a[1][2] * mu,
        stepmu    = a[2][0] * beta + a[2][1] * sigma + a[2][2] * mu;
    }
    double lv = l (b, s, m), test1 = 0.0;
    do {
        betas = b + rho * stepbeta;
        sigmas = s + rho * stepsigma;
        mus = m + rho * stepmu;
        rho /= 2.0;
    } while ((!IsRangeValid (betas, sigmas, mus) || lv > (test1 = l (betas, sigmas, mus))) &&
               rho * sqrt (stepbeta * stepbeta + stepsigma * stepsigma + stepmu * stepmu) >= 1e-7);
    b = betas; s = sigmas; m = mus;
    if (!IsRangeValid (b, s, m)) throw EST_FAILURE;
#ifdef TERMINAL
    if (test1 - lv < 1e-7) {
//      printf ("Log-Likelihood %g\n", l (b, s, m));
        return 1;
    }
#else
    ProgressField.SetText (form (100, "Current Log-Likelihood: %g", test1));
    if (test1 - lv < 1e-7) {
        ProgressField.SetText ("Iteration terminated because increase of Log-Likelihood is smaller than 1e-7\n");
        return 1;
    }
#endif
    return 0;
}

void TEVDMLEstimator::InitEstimator2 (double &b, double &s, double &m) {
    TUnivariateSample &Sample = *(TUnivariateSample *) p;
    double xm = 0.0;
    int i;
    for (i = 0; i < Sample.GetDataSize (); i++) xm += Sample [i] / Sample.GetDataSize ();
    double sm = 0.0;
    for (i = 0; i < Sample.GetDataSize (); i++) sm += (Sample [i] - xm) * (Sample [i] -xm) / (Sample.GetDataSize () - 1);
    sm = sqrt (sm);
    b = 1e-10;
    s = sm * sqrt (6.0) / 3.1415926;
    m = xm - s * 0.5772;
}

double TEVDMLEstimator::RDEEstimate (int i, double n, TUnivariateSample &Sample) {
    double a = sqrt (log ((n-i)/n) / log (i/n)),
           q0 = i / n,
           q1 = pow (q0, a),
           q2 = (n-i) / n;
    double v1 = Sample [(int) floor (n * q2)],
           v2 = Sample [(int) floor (n * q1)],
           v3 = Sample [(int) floor (n * q0)];

    if (v2 == v3) return MYMAXDOUBLE;
    return log ((v1 - v2) / (v2 - v3)) / log (1.0 / a);
}

void TEVDMLEstimator::InitEstimator () {
    TUnivariateSample &Sample = *(TUnivariateSample *) p;
    int n = Sample.GetDataSize ();
    RawParameters [0] = 0.0;
    for (int i = 1; i < n / 4; i++)
        RawParameters [0] += RDEEstimate (i, n, Sample);
    RawParameters [0] /= (n / 4);

    if (RawParameters [0] == 0.0) RawParameters [0] = 1e-5;
    LeastSquaresEstimate (&Sample);
    if (RawParameters [0] < 0.0)
        RawParameters [1] = max (RawParameters [1], (Sample [n - 1] + RawParameters [2] / RawParameters [0]));
}

TEstMode TEVDMLEstimator::GetEstimatorMode () {
    return FULL_BETA_MODE;
}

TEstResult TEVDMLEstimator::DoEstimation (TSample *q) {
    p = q;
    xmin = DBL_MAX; xmax = -DBL_MAX;
    int i;
    for (i = 0; i < ((TUnivariateSample *) p)->GetDataSize (); i++) {
        if ((*(TUnivariateSample *) p)[i] > xmax) xmax = (*(TUnivariateSample *) p)[i];
        if ((*(TUnivariateSample *) p)[i] < xmin) xmin = (*(TUnivariateSample *) p)[i];
    }
#ifndef TERMINAL
    ProgressField.SetText ("Values returned by initial estimator");
#endif
    InitEstimator ();
    if (!IsRangeValid (RawParameters [0], RawParameters [2], RawParameters [1]))
        InitEstimator2 (RawParameters [0], RawParameters [2], RawParameters [1]);
    TransformModels (k, (TUnivariateSample *) q);
#ifndef TERMINAL
    ShowCurrentEstimation ();
#endif
    BOOL Done = FALSE;
    for (i = 0; i <= 50 && !Done; i++) {
        Done = Iter (RawParameters [0], RawParameters [2], RawParameters [1]);
#ifndef TERMINAL
        TransformModels (k, (TUnivariateSample *) q);
        ShowCurrentEstimation ();
#endif
    }
#ifndef TERMINAL
    if (i > 50) ProgressField.SetText ("Maximum number of iterations performed");
#endif
    TransformModels (k, (TUnivariateSample *) q);
    return EST_OK;
}

TEstResult TEVDMLEstimator::Estimate (int k, TSample *q) {
    LastResult = EST_OK;
    if (((TUnivariateSample *) q)->GetDataSize () < 10) return EST_SAMPSIZE;
    try {
        return DoEstimation (q);
    }
    catch (TEstResult t) {
        Parameters [0] = Parameters [1] = Parameters [2] = 1.0;
        return t;
    }
}

TEstimatorData *TEVDMLEstimator::GetParameter () {
    return &Data;
}

