/*  list.cpp

    Implementation of base classes for list management.
    Copyright (C) 1998 Michael Thomas.
*/

#include <assert.h>
#include <stdlib.h>

#include "list.hpp"

// TObject

TObject::TObject () {
    Observers = NULL;
}

TObject::~TObject () {
    if (Observers) {
        TIterator i (Observers);
        TObject *p;
        while ((p = i.GetNext ()) != NULL)
            p->NotifyDelete (this);
        delete Observers;
    }
}

#ifdef _Windows
#pragma argsused
/*
BOOL TObject::Save (HFILE hFile) {
    return FALSE;
}

#pragma argsused
BOOL TObject::Load (HFILE hFile) {
    return FALSE;
}
*/
#endif

String TObject::Name () {
    return String ("TObject::Name () not overloaded");
}

void TObject::AddObserver (TObject *observer) {
    if (!Observers) Observers = new TList;
    Observers->Append (observer);
}

void TObject::RemoveObserver (TObject *observer) {
    Observers->Remove (observer);
}

void TObject::NotifyDelete (TObject *) {
}

// TStringEl

TStringEl::TStringEl (String s) {
    buf = s;
}

// TList

TList::TList () {
    Begin = new Node;
    End = new Node;
    Begin->Next = End;
    Begin->El = NULL;
    Begin->Prev = NULL;
    End->Prev = Begin;
    End->El = NULL;
    End->Next = NULL;
    ActPtr = NULL;
    entries = 0;
}

TList::~TList () {
    Node *p = Begin, *q;
    while (p) {
	q = p;
        p = p->Next;
        assert (q != NULL);
        delete q;
    }
}

void TList::Append (TListEl *p) {
    InsertBefore (p, End);
}

void TList::Remove (TObject *p, BOOL DelContents) {
    for (TObject *q = GetFirst (); q; q = GetNext ())
        if (p == q) {
            DeleteActive (DelContents);
            break;
        }
}

void TList::InsertBefore (TListEl *p, Node *q) {
    Node * r = new Node;
    r->El = p;
    r->Next = q;
    r->Prev = q->Prev;
    r->Prev->Next = r;
    q->Prev = r;
    entries++;
}

void TList::InsertBeforeActive (TListEl *p) {
    assert (ActPtr != Begin);
    InsertBefore (p, ActPtr);
}

TListEl * TList::GetFirst () {
    ActPtr = Begin->Next;
    return ActPtr->El;
}

TListEl * TList::GetLast () {
    ActPtr = End->Prev;
    return ActPtr->El;
}

TListEl * TList::GetNext () {
    if (ActPtr != End) ActPtr = ActPtr->Next;
    return ActPtr->El;
}

TListEl * TList::GetPrev () {
    if (ActPtr != Begin) ActPtr = ActPtr->Prev;
    return ActPtr->El;
}

TListEl * TList::GetEntry (int Nr) {
    SeekEntry (Nr);
    return GetNext ();
}

TListEl * TList::GetActive () {
    return ActPtr->El;
}

void TList::Advance (int Count) {
    while (ActPtr != Begin && ActPtr != End && Count)
	if (Count > 0)
	    Count--, ActPtr = ActPtr->Next;
	else Count++, ActPtr = ActPtr->Prev;
}

void TList::SeekEntry (int Nr) {
    for (ActPtr = Begin; Nr > 0 && ActPtr != End; Nr--)
        ActPtr = ActPtr->Next;
}

String TList::GetActiveName () {
    return ActPtr->El->Name ();
}

BOOL TList::ActiveMarked () {
    return ActPtr->marked;
}

void TList::ToggleActive () {
    if (ActPtr != Begin && ActPtr != End)
        ActPtr->marked = !ActPtr->marked;
}

void TList::ResetMarks () {
    TListEl *s = GetFirst ();
    while (s) {
        if (ActiveMarked ()) ToggleActive ();
        s = GetNext ();
    }
}

int TList::Entries () {
    return entries;
}

void TList::DeleteAll (BOOL DelContents) {
    Node *p = Begin->Next, *q;
    while (p != End) {
        q = p;
        p = p->Next;
        assert (q->El != NULL);
        if (DelContents) delete q->El;
        assert (q != NULL);
        delete q;
    }
    Begin->Next = End;
    End->Prev = Begin;
    entries = 0;
}

void TList::DeleteMarked (BOOL DelContents) {
    Node *p = Begin->Next, *q;
    while (p != End) {
	q = p;
	p = p->Next;
	if (q->marked) {
	    assert (q->El != NULL);
	    if (DelContents) delete q->El;
	    DeleteNode (q);
	}
    }
}

void TList::DeleteNode (Node *p) {
    assert (p != NULL);
    p->Next->Prev = p->Prev;
    p->Prev->Next = p->Next;
    delete p;
    entries--;
}

void TList::DeleteActive (BOOL DelContents) {
    assert (ActPtr != End);
    assert (ActPtr != Begin);
    if (DelContents) delete ActPtr->El;
    ActPtr = ActPtr->Next;
    DeleteNode (ActPtr->Prev);
}

#ifdef _Windows
/*
BOOL TList::Save (HFILE hFile) {
    TListEl *p = GetFirst ();
    char ch = 1;
    while (p) {
	_lwrite (hFile, &ch, sizeof (ch));
	p->Save (hFile);
	p = GetNext ();
    }
    ch = 0;
    _lwrite (hFile, &ch, sizeof (ch));
    return TRUE;        // Hier noch Pruefung
}

BOOL TList::Load (HFILE hFile) {
    TListEl * p;
    DeleteAll ();
    char ch;
    _lread (hFile, &ch, sizeof (ch));
    while (ch) {
	p = CreateEl ();
	p->Load (hFile);
	_lread (hFile, &ch, sizeof (ch));
	Append (p);
    }
    return TRUE;        // Hier noch Pruefung
}
*/
#endif

TListEl * TList::CreateEl () {
    return NULL;
}

/*********************************************************/
/*			TIterator			 */
/*********************************************************/

TIterator::TIterator (TList *List0) {
    List = List0;
    Restart ();
}

TObject *TIterator::GetNext () {
    if (ActNode == List->End) return NULL;
    TObject *p = ActNode->El;
    ActNode = ActNode->Next;
    return p;
}

void TIterator::Restart () {
    ActNode = List->Begin->Next;
}

BOOL TIterator::IsActiveMarked () {
    return ActNode->Prev->marked;
}

void TIterator::ToggleActive () {
    ActNode->Prev->marked = !ActNode->Prev->marked;
}

// SortedList

TSortedList::TSortedList (TCmpFunc Cmp) {
    cmp = Cmp;
}

void TSortedList::Insert (TListEl * p) {
    Node *q = Begin->Next;
    while (q != End && cmp (q->El, p) < 0) q = q->Next;
    InsertBefore (p, q);
}
