/*  mathfunc.hpp

    Declaration of densities, distribution and quantile functions.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 2000 Xtremes Group, Siegen.
*/

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <float.h>

#include "mathfunc.hpp"
#include "stddef.hpp"
#include "tmatrix.hpp"

const double pow232 = 4294967296.0;
static unsigned int Seed = 1;

void Randomize () {
    Seed = time (NULL);
}

void Randomize (unsigned int seed) {
    Seed = seed;
}

double Random () {
    Seed = Seed * 134775813L + 1;
    return Seed / pow232;
}

unsigned int IRandom () {
    Seed = Seed * 134775813L + 1;
    return Seed;
}
/* The following functions are partly wrappers to provide the standard models
   without location/scale parameter. All functions are prefixed with X to
   avoid name conflicts. */

double XBetaData (double alpha) {
    return ParetoIIQF (Random (), alpha, 1.0);
}

double XBetaDensity (double x, double alpha) {
    return ParetoIIDen (x, alpha, 1.0);
}

double XBetaDF (double x, double alpha) {
    return ParetoIIDF (x, alpha, 1.0);
}

double XBetaQF (double x, double alpha) {
    return ParetoIIQF (x, alpha, 1.0);
}

double UFOXBetaDensity (double alpha, double x) {
    return ParetoIIDen (x, alpha, 1.0);
}

double UFOXBetaDF (double alpha, double x) {
    return ParetoIIDF (x, alpha, 1.0);
}

double UFOXBetaQF (double alpha, double x) {
    return ParetoIIQF (x, alpha, 1.0);
}


double XChiSquareData (double nu) {
    int m = (int) nu;
    double sum = 0.0;
    for (int i = 0; i < m; i++)
	sum += sqr (XGaussianData ());
    return sum;
}

double XChiSquareDensity (double x, double nu) {
    return ChiSquareDensity (x, (int) nu);
}

double XChiSquareQF (double x, double nu) {
    return ChiSquareQF (x, (int) nu);
}

double XChiSquareDF (double x, double nu) {
    return ChiSquareDF (x, (int) nu);
}

double XEVData (double gamma) {
    return vonMisesEXTQF (Random (), gamma, 1.0, 0.0);
}

double XEVDensity (double x, double gamma) {
    return vonMisesEXTDen (x, gamma, 1.0);
}

double XEVDF (double x, double gamma) {
    return vonMisesEXTDF (x, gamma, 1.0);
}

double XEVQF (double x, double gamma) {
    return vonMisesEXTQF (x, gamma, 1.0);
}

double UFOXEVDensity (double gamma, double x) {
    return vonMisesEXTDen (x, gamma, 1.0);
}

double UFOXEVDF (double gamma, double x) {
    return vonMisesEXTDF (x, gamma, 1.0);
}

double UFOXEVQF (double gamma, double x) {
    return vonMisesEXTQF (x, gamma, 1.0);
}

double XExponentialData () {
    return ExpQF (Random (), 1.0);
}

double XExponentialDensity (double x) {
    return ExpDen (x, 1.0);
}

double XExponentialDF (double x) {
    return ExpDF (x, 1.0);
}

double XExponentialQF (double x) {
    return ExpQF (x, 1.0);
}

double XFrechetData (double alpha) {
    return FrechetQF (Random (), alpha, 1.0);
}

double XFrechetDensity (double x, double alpha) {
    return FrechetDen (x, alpha, 1.0);
}

double XFrechetDF (double x, double alpha) {
    return FrechetDF (x, alpha, 1.0);
}

double XFrechetQF (double x, double alpha) {
    return FrechetQF (x, alpha, 1.0);
}

double UFOXFrechetDensity (double alpha, double x) {
    return FrechetDen (x, alpha, 1.0);
}

double UFOXFrechetDF (double alpha, double x) {
    return FrechetDF (x, alpha, 1.0);
}

double UFOXFrechetQF (double alpha, double x) {
    return FrechetQF (x, alpha, 1.0);
}

double XGaussianData () {
    return NormRV ();
}

double XGaussianDensity (double x) {
    return NormDen (x, 0.0, 1.0);
}

double XGaussianDF (double x) {
    return NormDFPol (x, 0.0, 1.0);
}

double XGaussianQF (double x) {
    return NormQF (x, 0.0, 1.0);
}

double XLogNormalData () {
    return LogNormRV (1.0, 1.0);
}

double XLogNormalDensity (double x) {
    return LogNormDen (x, 1.0, 1.0);
}

double XLogNormalDF (double x) {
    return LogNormDFPol (x, 1.0, 1.0);
}

double XLogNormalQF (double x) {
    return LogNormQF (x, 1.0, 1.0);
}

double XGCauchyData (double alpha) {
    return CauchyRV (alpha, FALSE);
}

double XGCauchyDensity (double x, double alpha) {
    return CauchyDen (x, alpha, 1.0, 0.0);
}

double XGCauchyDF (double x, double alpha) {
    return CauchyDF (x, alpha, 0.0);
}

double XGCauchyQF (double x, double alpha) {
    return CauchyQF (x, alpha, 0.0);
}

double UFOXGCauchyDensity (double alpha, double x) {
    return CauchyDen (x, alpha, 1.0, 0.0);
}

double UFOXGCauchyDF (double alpha, double x) {
    return CauchyDF (x, alpha, 0.0);
}

double UFOXGCauchyQF (double alpha, double x) {
    return CauchyQF (x, alpha, 0.0);
}

double XGPData (double gamma) {
    return vonMisesGPDQF (Random (), gamma, 1.0, 0.0);
}

double XGPDensity (double x, double gamma) {
    return vonMisesGPDDen (x, gamma, 1.0, 0.0);
}

double XGPDF (double x, double gamma) {
    return vonMisesGPDDF (x, gamma, 1.0, 0.0);
}

double XGPQF (double x, double gamma) {
    return vonMisesGPDQF (x, gamma, 1.0, 0.0);
}

double UFOXGPDensity (double gamma, double x) {
    return vonMisesGPDDen (x, gamma, 1.0, 0.0);
}

double UFOXGPDF (double gamma, double x) {
    return vonMisesGPDDF (x, gamma, 1.0, 0.0);
}

double UFOXGPQF (double gamma, double x) {
    return vonMisesGPDQF (x, gamma, 1.0, 0.0);
}

double XGumbelData () {
    return GumbelQF (Random (), 1.0);
}

double XGumbelDensity (double x) {
    return GumbelDen (x, 1.0);
}

double XGumbelDF (double x) {
    return GumbelDF (x, 1.0);
}

double XGumbelQF (double x) {
    return GumbelQF (x, 1.0);
}

double XParetoData (double alpha) {
    return ParetoQF (Random (), alpha, 1.0);
}

double XParetoDensity (double x, double alpha) {
    return ParetoDen (x, alpha, 1.0);
}

double XParetoDF (double x, double alpha) {
    return ParetoDF (x, alpha, 1.0);
}

double XParetoQF (double x, double alpha) {
    return ParetoQF (x, alpha, 1.0);
}

double UFOXParetoDensity (double alpha, double x) {
    return ParetoDen (x, alpha, 1.0);
}

double UFOXParetoDF (double alpha, double x) {
    return ParetoDF (x, alpha, 1.0);
}

double UFOXParetoQF (double alpha, double x) {
    return ParetoQF (x, alpha, 1.0);
}

double XTData (double alpha) {
    return XGaussianData () / sqrt (2.0 * XGammaData (alpha / 2.0) / alpha);
}

static double TAlpha = 1.0;

static double TDenWrap (double x) {
    return XTDensity (TAlpha, x);
}

double XTDF (double alpha, double x) {
    TAlpha = alpha;
    if (x > 0.0) return 0.5 + GaussIntegrate (TDenWrap, 0.0, x);
    if (x < 0.0) return 0.5 - GaussIntegrate (TDenWrap, 0.0, -x);
    return 0.5;
}

double XTQF (double alpha, double x0) {
    if (x0 > 0.5) return -XTQF (alpha, 1.0 - x0);
    double x = XGaussianQF (x0);
    double ret = x;
    ret += (pow(x,3)+x)/4/alpha;
    ret += (5*pow(x,5)+16*pow(x,3)+3*x)/96/pow(alpha,2);
    ret += (3*pow(x,7)+19*pow(x,5)+17*pow(x,3)-15*x)/384/pow(alpha,3);
    ret += (79*pow(x,9)+776*pow(x,7)+1482*pow(x,5)-1920*pow(x,3)-945*x)
		    /92160.0/pow(alpha,4);
    return ret;
}

double XTDensity (double alpha, double x) {
    static double c = 0.0,
                  alphac = -1.0;
    if (alpha != alphac) {
	alphac = alpha;
	c = gamma ((1.0 + alpha) / 2.0) /
            (gamma (alpha / 2.0) * sqrt (MY_PI * alpha));
    }
    return c * pow (1.0 + x * x / alpha, (-1.0 - alpha) / 2.0);
}

double XNCTData (double alpha, double beta) {
    alpha *= 0.5;
    return (XGaussianData () + beta) / sqrt (XGammaData (alpha) / alpha);
}

double XNCTDensity (double alpha, double beta, double x) {
    double s = 0.0, s0,
           r = alpha * 0.5,
	   cp = 1.0,
           c = x * beta * sqrt (2.0 / (alpha + x * x)),
	   g = 1.0,
	   u = gamma (r + 1.0) / gamma (r + 0.5);
    int j = 0;
    do {
	s0 = s;
	s += g * cp + u * cp * c;
	cp *= c * c;
	g *= (r + j * 0.5 + 0.5) / ((j + 1) * (j + 2));
	u *= (r + j * 0.5 + 1.0) / ((j + 2) * (j + 3));
	j += 2;
    } while (fabs (s - s0) >= 1e-8);
    return s * XTDensity (alpha, x) * exp (-beta * beta * 0.5);
}

static const int vecsize = 2000;

static double *xvec = NULL, 
              *yvec = NULL,
	      alphag = 0.0,
	      betag = 0.0;

static double nctint (double x) {
    return XNCTDensity (alphag, betag, x);
}

static void buildvec (double alpha, double beta) {
    if (alpha == alphag && beta == betag) return;
    alphag = alpha; betag = beta;
    if (!xvec) xvec = new double [vecsize];
    if (!yvec) yvec = new double [vecsize];
    xvec [0] = -100.0;
    yvec [0] = 0.0;
    for (int i = 1; i < vecsize; i++) {
	xvec [i] = xvec [i - 1] + 200.0 / vecsize;
	yvec [i] = yvec [i - 1] + GaussIntegrate (nctint, xvec [i - 1], xvec [i]);
    }    
}
	
double XNCTDF (double alpha, double beta, double x) {
    buildvec (alpha, beta);
    int i = 0, j = vecsize - 1, k;
    while (j - i > 1) {
	k = (i + j) / 2;
	if (x < xvec [k]) j = k;
	else if (x > xvec [k]) i = k;
	else return yvec [k];
    }
    return yvec [i] *  (x - xvec [i + 1]) * (x - xvec [i + 2])
                    / ((xvec [i] - xvec [i + 1]) * (xvec [i] - xvec [i + 2]))
         + yvec [i + 1] * (x - xvec [i]) * (x - xvec [i + 2])
                    / ((xvec [i + 1] - xvec [i]) * (xvec [i + 1] - xvec [i + 2]))
         + yvec [i + 2] * (x - xvec [i]) * (x - xvec [i + 1])
                    / ((xvec [i + 2] - xvec [i]) * (xvec [i + 2] - xvec [i + 1]));
}

double XNCTQF (double alpha, double beta, double q) {
    return 0.5; // !!!! TODO !!!!
}

double XGammaData (double r) {
    return GammaRV (r);
}

/*
double XGammaDensity (double r, double x) {
    if (x <= 0.0) return 0.0;
    // new 1.3.98
    if (r < 50)
    	if (x < 500)
    	    return pow (x, r - 1) * exp (-x) / gamma (r);
        else
            return 0.0;
    else
        return XGaussianDensity ((x - r) / sqrt (r)) / sqrt (r);
}
*/

double XGammaDensity (double r, double x) {
    if (x <= 0.0) return 0.0;
    if (r < 20)
    	if (x < 500)
    	    return pow (x, r - 1) * exp (-x) / gamma (r);
        else
            return 0.0;
    else if (r < 1000) {
        if (x <= 1.0 || x > 5000) return 0.0;
        int r1g = (int) floor (r - 1),
            xg = (int) floor (x),
            r1gs = 0, xgs = 0;
        double gn = r - 1.0,
               exp1 = exp (1),
               f = 1.0;
        while (r1gs < r1g) {
            while (r1gs < r1g && f < 1e100) {
                f *= x; r1gs++;
            }
            while (xgs < xg && f > 1e-100) {
                f /= exp1; xgs ++;
            }
            while (gn > 1.0 && f > 1e-100) {
                f /= gn; gn -= 1.0;
            }
        }
        while (xgs < xg) {
            f /= exp1; xgs++;
        }
        f /= exp (x - xg);
        while (gn > 1.0) {
            f /= gn; gn -= 1.0;
        }
        f /= (gn * gamma (gn));
        return f * pow (x, r - r1g - 1.0);
    } else {
        double rx = (x - r) / sqrt (r),
               f = 1.0 + ((1 - rx * rx) * (-2.0 * rx) - 2 * rx) / (3.0 * sqrt (r));
        return XGaussianDensity ((x - r) / sqrt (r)) / sqrt (r) * f;
    }
}

double XGammaDF (double r, double x) {
    // Series expansion from Brandt, S.: "Data Analysis"
    const double EPS = 1e-6;
    double h1 = 0.0, rs = r;
    while (rs > 1.0) {
        rs -= 1.0;
        h1 += log (rs);
    }
    h1 += log (gamma (rs));
    // h1 = log (gamma (r));
    if (x <= 0.0) return 0.0;
    if (x < r + 1.0) {
        double h2 = 1.0 / r, s = h2;
        int i = 1;
        do {
            h2 *= x / (r + i);
            s += h2;
        } while (h2 > EPS && ++i < 100);
        return exp (r * log (x) - x - h1) * s;
    } else {
        double a0 = 0.0, b0 = 1.0, a1 = 1.0, b1 = x, c = 1.0, f = 1.0, cn = 0.0;
        BOOL Done = FALSE;
        int i = 1;
        do {
            a0 = (a1 + (i - r) * a0) * f;
            b0 = (b1 + (i - r) * b0) * f;
            a1 = x * a0 + i * a1 * f;
            b1 = x * b0 + i * b1 * f;
            if (b1) {
                f = 1.0 / b1;
                cn = a1 * f;
                Done = fabs ((c - cn) / c) < EPS;
                c = cn;
            }
        } while (!Done && ++i < 100);
        return 1.0 - exp (r * log (x) - x - h1) * cn;
    }
}

double XWeibullData (double alpha) {
    return WeibullQF (Random (), alpha, 1.0);
}

double XWeibullDensity (double x, double alpha) {
    return WeibullDen (x, alpha, 1.0);
}

double XWeibullDF (double x, double alpha) {
    return WeibullDF (x, alpha, 1.0);
}

double XWeibullQF (double x, double alpha) {
    return WeibullQF (x, alpha, 1.0);
}

double UFOXWeibullDensity (double alpha, double x) {
    return WeibullDen (x, alpha, 1.0);
}

double UFOXWeibullDF (double alpha, double x) {
    return WeibullDF (x, alpha, 1.0);
}

double UFOXWeibullQF (double alpha, double x) {
    return WeibullQF (x, alpha, 1.0);
}


double XIndicator (double a, double b, double x) {
    return a <= x && x <= b;
}

double XMax (double a, double b) {
    return a < b ? b : a;
}

double XMin (double a, double b) {
    return a < b ? a : b;
}

double squarefunc (double x) {
    return x * x;
}

double frac (double x) {
    return x > 0 ? x - floor (x) : x - ceil (x);
}


#if !defined(__GNUC__) || __GNUC__ < 3
double round (double x)
{
    return floor (x + 0.5);
}
#endif

double sign (double x) {
    if (x < 0.0) return -1.0;
    if (x > 0.0) return 1.0;
    return 0.0;
}

// Gamma function on |R-{0, -1, -2, ...}. It is GAMMA(x+1)=x*GAMMA(x).

double gamma (double x)
#ifdef __GNUC__
#if ((__GNUC__ == 2)  && (__GNUC_MINOR__ == 96)) || (__GNUC__ >= 3)
    throw ()
#endif
#endif
{

  if(x==0.0) return 0.0; // invalid argument?

  double b[8] = {-0.577191652, 0.988205891, -0.897056937,
		 0.918206857, -0.756704078, 0.482199394,
		 -0.193527818, 0.035868343};

  // else part uses formula 6.1.36 from
  // Abramowitz, Handbook of mathematic formulas.
  // Error is equal or less than 3.0e-7.

  if (x>2.0) return (x-1.0)*gamma(x-1.0);
  else if (x<1.0) return 1.0/x*gamma(x+1.0);
  else {
    int i;
    double temp = 1.0;
    for(i=0;i<8;i++)
      temp+=b[i]*pow(x-1, i+1);
    return temp;
  }
}

/* Pareto */

double ParetoDen (double x, double a, double t, double r) {
//    if (x >= t + r) return a / t * exp (-(a + 1) * log ((x - r) / t));
    if (x >= t + r) {
	x = (x - r) / t;
	double e = (-a - 1.0) * log (x);
	if (e > -500.0) return a / t * exp (e);
//	return a / t * pow ((x - r) / t, -(a + 1.0));
    }
    return 0.0;
}

double ParetoQF (double x, double a, double t, double r) {
    if (x < 0.0  || x > 1.0) return 0.0;
    if (x == 1.0) return MYMAXDOUBLE;
    return t * pow (1.0 - x, -1.0 / a) + r;
}

double ParetoDF (double x, double a, double s, double m) {
    x = (x-m)/s;
    if (x < 1) return 0.0;
    return 1.0 - pow (x, -a);
}

/* Pareto II */

double ParetoIIDen (double x, double a, double t, double m) {
    x -= m;
    x /= t;
    if (x <= -1 || x >= 0.0) return 0.0;
    return (-a/t) * pow (-x, -a - 1.0);
}

double ParetoIIQF (double x, double a, double t, double m) {
    if (x < 0.0 || x > 1.0) return 0.0;
    return m - t * pow (1.0 - x, -1.0 / a);
}

double ParetoIIDF (double x, double a, double s, double m) {
    x = (x-m)/s;
    if (x < -1) return 0.0;
    if (x > 0.0) return 1.0;
    return 1.0 - pow (-x, -a);
}

/* Exponential */

double ExpDen (double x, double s, double m) {
    if (x - m < 0) return 0.0;
    return exp (-(x - m) / s) / s;
}

double ExpQF (double x, double s, double m) {
    if (x < 0.0 || x > 1.0) return 0.0;
    return -s * log (1.0 - x) + m;
}

double ExpDF (double x, double s, double m) {
     x = (x-m)/s;
     if (x < 0) return 0.0;
     else return 1 - exp (-x);
}


/* Cauchy */

double CauchyDen (double x, double ac, double tc, double rc) {
    if (ac == 1.0) return 1.0 / ((1.0 + ((x - rc)/tc) * ((x - rc)/tc)) * tc * MY_PI);
//    double help1 = 1.0 / (exp (log (fabs (x - rc)/tc) * (ac + 1.0)) + 1.0);
    double help1 = 1.0 / (pow (fabs (x - rc) / tc, ac + 1.0) + 1.0);
    double help2 = sin (MY_PI / (1.0 + ac)) / (2.0 * MY_PI * tc);
    return (1.0 + ac) * help2 * help1;
}

double CauchyQF (double x, double t, double r) {
    return t * tan (MY_PI * (x - 0.5)) + r;
}

double CauchyDF (double x, double t, double r) {
    return 1.0 / MY_PI * atan ((x - r) / t) + 0.5;
}

double CauchyInt (double x1, double x2, double a, double t, double r, int n) {
    if (n == 0) n = 2;
    else if (n & 1) n++;
    double delta = (x2 - x1) / n;
    double Inth = CauchyDen (x1, a, t, r);
    for (int j = 1; j <= n; j += 2)
        Inth += 4 * CauchyDen (x1 + j * delta, a, t, r) +
                2 * CauchyDen (x1 + (j + 1) * delta, a, t, r);
    return delta / 3.0 * (Inth - CauchyDen (x2, a, t, r));
}

double CauchyDFStart (double x, double alpha, double sigma, double mu) {
    if (alpha == 1.0) return CauchyDF (x, sigma, mu);
    double h = CauchyInt (0.0, fabs (x - mu), alpha, sigma, 0.0, max ((int) (fabs (x - mu) / 0.1), 2)) + 0.5;
    return x - mu < 0.0 ? 1.0 - h : h;
}

double CauchyDFStep (double x1, double x2, double alpha, double sigma, double mu) {
    if (alpha == 1.0) return CauchyDF (x2, sigma, mu) - CauchyDF (x1, sigma, mu);
    return CauchyInt (x1 - mu, x2 - mu, alpha, sigma, 0.0, max ((int) ((x2 - x1) / 0.1), 2));
}

static double accept (double x, double alpha) {
    double help = pow (x, 1.0 + alpha);
    return ((x <= 1.0) ? 1.0 : help) / (1.0 + help);
}

double CauchyRV (double alpha, BOOL pos) {
    double x, y, z,
           helpa = alpha / (alpha + 1.0);
    for (;;) {
        z = Random (); y = Random ();
        if (z <= helpa) x = z/helpa;
        else x = pow ((1.0 + alpha) * (1.0 - z), -1.0 / alpha);
        if (y <= accept (x, alpha)) {
            if (pos) return x;
            return (Random () < 0.5) ? x : -x;
        }
    }
}

/* Normal */

double NormDen (double x, double m, double s) {
    double xs = (x - m) / s;
    if (fabs (xs) > 37) return 0.0;
    double help = -xs * xs;
    return exp (help / 2.0) / (sqrt (2 * MY_PI) * s);
}

/*
double NormalDF (double x, double m, double s) {
    const double a1 = 0.196854, a2 = 0.115194,
		 a3 = 0.000344, a4 = 0.019527;
    x = (x - m) / s;
    double w = fabs (x);
    if (w > 200.0) w = 0.0;
    else {
	w = 1.0 + w * (a1 + w * (a2 + w * (a3 + w * a4)));
	w *= w; w *= w; // w^4
	w = 1.0 - 0.5 / w;
    }
    return x >= 0.0 ? w : 1.0 - w;
}
*/

double NormDFPol (double x, double m, double s) {
    const double d1 = 0.0498673470, d4 = 0.0000380036,
                 d2 = 0.0211410061, d5 = 0.0000488906,
                 d3 = 0.0032776263, d6 = 0.0000053830;
    x = (x - m) / s;
    double w = min (fabs (x), 1e49);
    w = 1.0 + w * (d1 + w * (d2 + w * (d3 + w * (d4 + w * (d5 + w * d6 )))));
    if (w < 1e16) {    
        w *= w; w *= w; // w^4
        w *= w; w *= w; // w^16
    }
    w = 1.0 - 0.5 / w;
    return x >= 0.0 ? w : 1.0 - w;
}

/*
double NormDFExp (double x, double m, double s, int type) {
    const double d1 = 0.0498673470, d4 = 0.0000380036,
                 d2 = 0.0211410061, d5 = 0.0000488906,
                 d3 = 0.0032776263, d6 = 0.0000053830;
    x = (x - m) / s;
    double w = fabs (x);
    double sum = 0;
    double add = 1;
    if (type == 0) {
       w = 1.0 + w * (d1 + w * (d2 + w * (d3 + w * (d4 + w * (d5 + w * d6 )))));
       w *= w; w *= w; // w^4
       w *= w; w *= w; // w^16
       w = 1.0 - 0.5 / w;
    }
    else {
       int k = type - 1;
       for (int i=1; i<=k; i++) {
           add = -add * (2*i-1)/(w*w);
           sum += add;
       }
       w = 1 - NormDen (w, 0, 1)/w * (1+sum);
    }
    if ( w <= 0.0) w = DBL_MIN;
    if ( w >= 1.0) w = 1-DBL_MIN;
    return x >= 0.0 ? w : 1.0 - w;
}
*/

void NormRV (double &r1, double &r2) {
    double a = sqrt (-2.0 * log (Random ())),
           b = 2.0 * MY_PI * Random ();
    r1 = a * cos (b);
    r2 = a * sin (b);
}

double NormRV () {
    double a = sqrt (-2.0 * log (Random ())),
           b = 2.0 * MY_PI * Random ();
    return a * cos (b);
}

double NormQF (double x, double m, double s) {
    // Abramowitz: Handbook of mathematic formulas.  Formula 26.2.23
    // Error is less than 4.5e-4
    const double c0 = 2.515517, c1 = 0.802853, c2 = 0.010328,
                 d1 = 1.432788, d2 = 0.189269, d3 = 0.001308;
    BOOL f = x >= 0.5;
    if (x == 0.0) return -MYMAXDOUBLE;
    if (x == 1.0) return MYMAXDOUBLE;
    if (f) x = 1.0 - x;
    double t = sqrt (- 2.0 * log (x)),
	   a = (c2 * t + c1) * t + c0,
	   b = ((d3 * t + d2) * t + d1) * t + 1.0;
    return m + s * (f ? t - a / b : a / b - t);
}

/*  LogNormal */

double LogNormDen (double x, double b, double s) {
    return x <= 0.0 ? 0.0 :
           NormDen (log (x / s) / b, 0, 1) / (b * x);
}

double LogNormDFPol (double x, double b, double s) {
    return x <= 0.0 ? 0.0 :
           XGaussianDF (log (x / s) / b);
}

double LogNormRV (double b, double s) {
    return LogNormQF (Random (), b, s);
}

double LogNormQF (double x, double b, double s) {
    return s * exp (b * XGaussianQF (x));
}

// Chi-Square Distribution

double ChiSquareDensity (double x, int nu) {
    if (x <= 0.0) return 0.0;
    int k = nu / 2;
    double val = exp (-x / 2.0);
    if (nu & 1) {
	for (int i = 1; i <= k; i++)
	    val *= x / (2.0 * (k - i + 0.5));
	return val * sqrt (1.0 / (2.0 * x * MY_PI));
    } else {
	for (int i = 1; i < k; i++)
            val *= x / (2 * i);
        return val / 2.0;
    }
}

double ChiSquareDF (double x, int nu) {
    if (x <= 0.0) return 0.0;
    if (x / nu > 50) return 1.0;
    double s;
    if (nu & 1) {       // nu is odd
        double x0 = sqrt (x),
               n0 = 1.0;
        s = nu == 1 ? 0 : x0 / n0;
        for (int r = 2; r <= nu / 2; r++) {
            x0 *= x;
            n0 *= (2 * r - 1);
            s += x0 / n0;
        }
        return 1.0 - 2.0 * ((1.0 - NormDFPol (sqrt (x), 0.0, 1.0)) + NormDen (sqrt (x), 0.0, 1.0) * s);
    } else {
        double x0 = x,
	       n0 = 2.0;
        s = nu == 2 ? 0 : x0 / n0;
        for (int r = 2; r < nu / 2; r++) {
            x0 *= x;
            n0 *= 2 * r;
            s += x0 / n0;
        }
        return 1.0 - sqrt (2.0 * MY_PI) * NormDen (sqrt (x), 0.0, 1.0) * (1.0 + s);
    }
}

double ChiSquareQF (double q, int nu) {
    if (q == 0.0) return 0.0;
    const double EPSILON = 1e-6;
    double x1 = max (NormQF (q, nu, sqrt (2 * nu)), 0.01), x0, step;
    int Schritt = 0;
    do {
        x0 = x1;
        step = (ChiSquareDF (x0, nu) - q) / ChiSquareDensity (x0, nu);
        if (fabs (step) > nu / 2.0) step = step < 0.0 ? -nu / 2.0 : nu / 2.0;
        x1 = x0 - step;
        if (x1 < 0.0) x1 = DBL_MIN;
        Schritt++;
    } while (fabs (x1 - x0) > EPSILON && Schritt < 100);
//    if (Schritt == 100) MessageBox (0, form (50, "%d", q), "", MB_SYSTEMMODAL);
    return x1;
}

// Student's t-distribution

double tDF (double ts, int nu) {
    double t = fabs (ts), theta = atan (t / sqrt (nu)), A;
    if (nu == 1)
        A = 2.0 * theta / MY_PI;
    else if (nu & 1) {
        double v = cos (theta), f = v, s = v;
        v *= v;
        for (int i = 2; i <= nu - 3; i += 2) {
            f *= v * double (i) / double (i + 1);
            s += f;
        }
        A= 2.0 / MY_PI * (theta + sin (theta) * s);
    } else {
        double v = cos (theta), f = 1.0, s = 1.0;
        v *= v;
        for (int i = 2; i <= nu - 2; i += 2) {
            f *= v * double (i - 1) / double (i);
            s += f;
        }
        A = sin (theta) * s;
    }
    if (ts >= 0) return A + (1.0 - A) / 2.0;
    else return (1.0 - A) / 2.0;
}

// sum stable distributions
//
//

#ifdef USE_STABLE

#include "stable1.h"

// Density

double stableden (double x, double alpha, double beta, double mu, double sigma) {
    int n = 1, iparam = 0, ierr;
    double p;
    SPDF (&n, &x, &p, &alpha, &beta, &sigma, &mu, &iparam, &ierr);
    return p;
}

// Distribution function

double stabledf (double x, double alpha, double beta, double mu, double sigma) {
    int n = 1, iparam = 0, ierr;
    double p;
    SCDF (&n, &x, &p, &alpha, &beta, &sigma, &mu, &iparam, &ierr);
    return p;
}

// Quantile function

double stableqf (double q, double alpha, double beta, double mu, double sigma) {
    int n = 1, iparam = 0, ierr;
    double p;
    SQUANT (&n, &q, &p, &alpha, &beta, &sigma, &mu, &iparam, &ierr);
    return p;
}

// Data generation

double stabledata (double alpha, double beta, double mu, double sigma) {
    return stableqf (Random (), alpha, beta, mu, sigma);
}

// Routines with location mu = 0 and scale sigma = 1

double Xstableden (double x, double alpha, double beta) {
    return stableden (x, alpha, beta, 0.0, 1.0);
}

double Xstabledf (double x, double alpha, double beta) {
    return stabledf (x, alpha, beta, 0.0, 1.0);
}

double Xstableqf (double q, double alpha, double beta) {
    return stableqf (q, alpha, beta, 0.0, 1.0);
}

double Xstabledata (double alpha, double beta) {
    return stabledata (alpha, beta, 0.0, 1.0);
}

#endif

/* Chambers, Mallows, Stuck
   Not currently used
******************************

static double d2 (double z) {
    const double p1 = 0.840066852536483239e3,
                 p2 = 0.200011141589964569e2,
                 q1 = 0.168013370507926648e4,
                 q2 = 0.18013370407390023e3,
                 q3 = 1.0;
    if (fabs (z) > 0.1) return (exp (z) - 1.0) / z;
    double zz = z * z,
           pv = p1 + zz * p2;
    return 2.0 * pv / (q1 + zz * (q2 + zz * q3) - z * pv);
}

static double tan2 (double xarg) {
    const double p0 = 0.129221035e3,
                 p1 = -0.887662377e1,
                 p2 = 0.528644456e-1,
                 q0 = 0.164529332e3,
                 q1 = -0.451320561e2,
                 q2 = 1.0,
                 piby4 = 0.785398163;
    double x = fabs (xarg);
    if (x > piby4) return tan (xarg) / xarg;
    x /= piby4;
    double xx = x * x;
    return (p0 + xx * (p1 + xx * p2)) / (piby4 * (q0 + xx * (q1 + xx * q2)));
}

double rstab (double alpha, double bprime) {
    const double piby2 = 1.57079633;
    double u = Random (),
           w = -log (Random ()),
           eps = 1.0 - alpha,
           phiby2 = piby2 * (u - 0.5),
           a = phiby2 * tan2 (phiby2),
           bb = tan2 (eps * phiby2),
           b = eps * phiby2 * bb,
           tau = (eps > -0.99) ?
                     bprime / (tan2  (eps * piby2) * piby2) :
                     bprime * piby2 * eps * (1.0 - eps) * tan2 ((1.0 - eps) * piby2),
           a2 = 1.0 - a * a,
           a2p = 1.0 + a * a,
           b2 = 1.0 - b * b,
           b2p = 1.0 + b * b,
           z = a2p * (b2 + 2.0 * phiby2 * bb * tau) / (w * a2 * b2p),
           logz = log (z),
           d = d2 (eps * logz / (1.0 - eps)) * (logz / (1.0 - eps));
    return (1.0 + eps * d) * 2.0 * ((a - b) * (1.0 + a * b) - phiby2 * tau * bb * (b * a2 - 2.0 * a)) / (a2 * b2p) + tau * d;
}

***********************************/

// von Mises/GPD

double vonMisesGPDDen (double x, double b, double t, double m) {
    x -= m;
    if (x < 0.0) return 0.0;
    if (fabs (b) < 1e-10) return (1.0 / t) * exp (-x / t);
    if (b < 0.0 && x >= -t / b - fabs (t / b) / 1e15)
        return b < -1.0 ? MYMAXDOUBLE : b == - 1.0 ? 1.0 / t : 0.0;
    return (1.0 / t) * pow (1.0 + b * x / t, - 1.0 - 1.0 / b);
}

double vonMisesGPDQF (double x, double b, double t, double m) {
    if (x < 0.0 || x > 1.0) return 0.0;
    if (b > 0.0 && x == 1.0) return MYMAXDOUBLE;
    if (b == 0.0) return -t * log (1.0 - x) + m;
    return t / b * (pow (1.0 - x, -b) - 1.0) + m;
}

double vonMisesGPDDF (double x, double b, double t, double m) {
    x = (x - m) / t;
    if (fabs (b) < 1e-10) {
        if (x > 0.0) return 1.0 - exp (-x);
        return 0.0;
    }
    if (b < 0.0) {
        if (0.0 < x && x < -1.0 / b) return 1.0 - pow (1.0 + b * x, -1.0 / b);
        if (x >= -1.0 / b) return 1.0;
        return 0.0;
    }
    // beta > 0.0
    if (x > 0.0) return 1.0 - pow (1.0 + b * x, -1.0 / b);
    return 0.0;
}

// von Mises/EXT

double vonMisesEXTDen (double x, double b, double s, double m) {
    if (fabs (b) < 1e-10) return GumbelDen (x, s, m);
    double h0 = -1.0 / b, xs = (x - m) / s;
    if (b > 0.0 && xs <= h0 || b < 0.0 && xs >= h0) return 0.0;
    if (b < -1.0 && xs >= s *h0 - s * h0 / 1e13) return MYMAXDOUBLE;
    double h1 = 1.0 + b * xs,
           h2 = pow (h1, h0);
    if (h2 > 100) return 0.0;
    return exp (-h2) * h2 / (h1 * s);
}

double vonMisesEXTQF (double x, double b, double s, double m) {
    if (fabs (b) < 1e-10) return GumbelQF (x, s, m);
    return s * (pow (-log (x), -b) - 1.0) / b + m;
}

double vonMisesEXTDF (double x, double b, double s, double m) {
     if (fabs (b) < 1e-10) return GumbelDF (x, s, m);
     x = (x-m)/s;
     if (b < 0 && x > -1/b) return 1.0;
     if (b > 0 && x < -1/b) return 0.0;
     return exp (- pow (1+b*x,-1/b));
}

// Gumbel

double GumbelDen (double x, double s, double m) {
    double h = exp (-(x - m) / s);
    if (h > 400) return 0.0;
    return exp (-h) * h / s;
}

double GumbelQF (double x, double s, double m) {
    return -s * log (-log (x)) + m;
}

double GumbelDF (double x, double s, double m) {
     x = (x-m)/s;
     return exp (-exp(-x));
}

// Frechet

double FrechetDen (double x, double a, double s, double m) {
    x = (x - m) / s;
    if (x <= MYMINDOUBLE) return 0.0;
    double h = pow (x, -a);
    return h > 50.0 ? 0.0 : exp (-h) * a * h / x / s;
}

double FrechetQF (double x, double a, double s, double m) {
    return s * pow (-log (x), -1.0 / a) + m;
}

double FrechetDF (double x, double a, double s, double m) {
     x = (x-m)/s;
     if (x <= 0.0) return 0.0;
     double e = -a * log (x);
     if (e < -100.0) return 1.0;
     e = -exp (e);
     if (e < -100.0) return 0.0;
     return exp (e);
//     return exp (-pow (x, -a));
}

// Weibull

double WeibullDen (double x, double a, double s, double m) {
    x -= m;
    if (x > 0.0) return 0.0;
    x /= s;
    if (log (-x) / log (10.0) * a > 300.0) return 0.0;
//    if (x > -1e-100) x = -1e-100;
    double h = pow (-x, -a);
//    double f = h / x;
//    double g = a * f;
//    double f = h / s;
//    double g = f / x;
    return h > 100.0 ?  0.0 : exp (-h) * a * h / (s * x);
}

double WeibullQF (double x, double a, double s, double m) {
    return m - s * pow (-log (x), -1.0 / a);
}

double WeibullDF (double x, double a, double s, double m) {
     x = (x - m) / s;
     if (x >= 0) return 1.0;
//     return exp (-pow (-x, -a));
    return FrechetDF (-x, a, 1.0, 0.0);
}


double EVGGDen (double x, double lambda, double beta, double mu, double sigma) {
    x = 1.0 + beta * (x - mu) / sigma;
    if (x <= 0.0) return 0.0;
    return pow (x, -1.0 - lambda / beta) * exp (-pow (x, -1.0 / beta)) / (gamma (lambda) * sigma);
}

double EVGG0Den (double x, double lambda, double mu, double sigma) {
    x = (x - mu) / sigma;
    return exp (-lambda * x) * exp (-exp (-x)) / (gamma (lambda) * sigma);
}

double EVGG1Den (double x, double lambda, double alpha, double mu, double sigma) {
    x = (x - mu) / sigma;
    if (x < 0.0) return 0.0;
    return alpha * pow (x, -1.0 - lambda * alpha) * exp (-pow (x, -alpha)) / (gamma (lambda) * sigma);
}

double EVGG2Den (double x, double lambda, double alpha, double mu, double sigma) {
    x = (x - mu) / sigma;
    if (x > 0.0) return 0.0;
    return -alpha * pow (-x, -1.0 - lambda * alpha) * exp (-pow (-x, -alpha)) / (gamma (lambda) * sigma);
}


double HuesslerReissDF (double lambda, double gamma1, double gamma2, double x, double y) {
    if (gamma1 != 0.0)
        x = XGumbelQF (XEVDF (x, gamma1));
    if (gamma2 != 0.0)
        y = XGumbelQF (XEVDF (y, gamma2));
    if (lambda == 0.0) return GumbelDF (x, 1.0, 0.0) * GumbelDF (y, 1.0, 0.0);
    double u1 = NormDFPol (1.0 / lambda + lambda * (x - y) / 2.0, 0.0, 1.0),
           u2 = NormDFPol (1.0 / lambda + lambda * (y - x) / 2.0, 0.0, 1.0);
    return exp (-u1 * exp (-y) - u2 * exp (-x));
}

double HuesslerReissSF (double lambda, double gamma1, double gamma2, double x, double y) {
    if (gamma1 != 0.0)
        x = XGumbelQF (XEVDF (x, gamma1));
    if (gamma2 != 0.0)
        y = XGumbelQF (XEVDF (y, gamma2));
    return 1.0 - GumbelDF (x, 1.0, 0.0) - GumbelDF (y, 1.0, 0.0) + HuesslerReissDF (lambda, 0.0, 0.0, x, y);
}

double HuesslerReissDen (double lambda, double gamma1, double gamma2, double x, double y) {
    // transformation of marginal distributions to Gumbel
    double g1 = XEVDensity (x, gamma1),
           g2 = XEVDensity (y, gamma2);
    if (g1 == 0.0 || g2 == 0.0) return 0.0;
    double h1 = XGumbelQF (XEVDF (x, gamma1)),
           h2 = XGumbelQF (XEVDF (y, gamma2));
    // calculate standard form
    if (lambda == 0.0) return GumbelDen (h1, 1.0, 0.0) * GumbelDen (h2, 1.0, 0.0);
    double u1 = NormDFPol (1.0 / lambda + lambda * (h1 - h2) / 2.0, 0.0, 1.0),
           u2 = NormDFPol (1.0 / lambda + lambda * (h2 - h1) / 2.0, 0.0, 1.0),
           u3 = NormDen (1.0 / lambda + lambda * (h1 - h1) / 2.0, 0.0, 1.0),
           h =  HuesslerReissDF (lambda, 0.0, 0.0, h1, h2) * (
                u1 * u2 * exp (-h1 - h2) + lambda * u3 * exp (-h1) / 2.0);
    // return transformed version
    return h * g1 * g2 / (XGumbelDensity (h1) * XGumbelDensity (h2));
}

double HuesslerReissCondDF (double lambda, double x, double y) {
    if (lambda == 0.0) return GumbelDF (y, 1.0, 0.0);
    double u1 = NormDFPol (1.0 / lambda + lambda * (y - x) / 2.0, 0.0, 1.0);
    return HuesslerReissDF (lambda, 0.0, 0.0, x, y) * u1 / GumbelDF (x, 1.0, 0.0);
}

double HuesslerReissCondDen (double lambda, double x, double y) {
    return HuesslerReissDen (lambda, 0.0, 0.0, x, y) / GumbelDen (x, 1.0, 0.0);
}

double HuesslerReissCondQF (double lambda, double x, double q) {
    double y1 = GumbelQF (q, 1.0, 0.0), y0;
    if (lambda > 1) y1 = x;
    int step = 0;
    do {
        y0 = y1;
        y1 -= (HuesslerReissCondDF (lambda, x, y1) - q) / HuesslerReissCondDen (lambda, x, y1);
        step++;

        if (y1 > 10) y1 = 0;
        if (y1 < -10) y1 = 0;

    } while (step < 50 && fabs (y1 - y0) > 1e-6);
//
//    if (step == 50) MessageBox (NULL, "Generation failed", "", MB_SYSTEMMODAL);
//
    return y1;
}


double MarshallOlkinDF (double rho, double x, double y) {
    return exp ((1.0 - rho) * (x + y) + rho * min (x, y));
}

double MarshallOlkinSF (double rho, double x, double y) {
    return 1.0 - WeibullDF (x, -1.0, 1.0, 0.0) - WeibullDF (y, -1.0, 1.0, 0.0) + MarshallOlkinDF (rho, x, y);
}

double GumbelMcFaddenEVDF (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to Weibull
    x = XWeibullQF (XEVDF (x, gamma1), -1.0);
    y = XWeibullQF (XEVDF (y, gamma2), -1.0);
    //
    return exp (-pow (pow (-x, lambda) + pow (-y, lambda), 1.0 / lambda));
}

double GumbelMcFaddenEVSF (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to Weibull
    double xs = XWeibullQF (XEVDF (x, gamma1), -1.0),
    	   ys = XWeibullQF (XEVDF (y, gamma2), -1.0);
    return 1.0 - WeibullDF (xs, -1.0, 1.0, 0.0) - WeibullDF (ys, -1.0, 1.0, 0.0) + GumbelMcFaddenEVDF (lambda, gamma1, gamma2, x, y);
}

double GumbelMcFaddenEVDen (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to Weibull
    double g1 = XEVDensity (x, gamma1),
           g2 = XEVDensity (y, gamma2);
    if (g1 == 0.0 || g2 == 0.0) return 0.0;
    double xs = XWeibullQF (XEVDF (x, gamma1), -1.0),
           ys = XWeibullQF (XEVDF (y, gamma2), -1.0);
    //
    double h0 = pow (-xs, lambda) + pow (-ys, lambda),
           h1 = pow (h0, 1.0 / lambda),
           d = exp (-h1) * pow (xs * ys, lambda - 1) * (pow (h0, 2.0 / lambda - 2.0) +
                                                  (lambda - 1.0) * pow (h0, 1.0 / lambda - 2.0));
    //
    return d * g1 * g2 / (XWeibullDensity (xs, -1.0) * XWeibullDensity (ys, -1.0));
}

double BiGaussDF (double rho, double x, double y) {
    if (x <= -4.0) return 0.0;
    double f = 1.0 / sqrt (1.0 - rho * rho), f0, sum = 0.0, f1 = 0.0,
           step = (x + 4.0) / max (2.0, floor ((x + 4.0) * 25.0)),
           q = step / 2.0;
    for (double xs = -4.0; xs <= x; xs += step) {
        f0 = f1;
        f1 = NormDen (xs, 0.0, 1.0) * NormDFPol (f * (y - rho * xs), 0.0, 1.0);
        sum += (f0 + f1);
    }
    return sum * q;
}

double BiGaussSF (double rho, double x, double y) {
    return 1.0 - XGaussianDF (x) - XGaussianDF (y) + BiGaussDF (rho, x, y);
}

double BiGaussDen (double rho, double x, double y) {
    double u = 2.0 * MY_PI * sqrt (1.0 - rho * rho),
	   e = x * x - 2.0 * rho * x * y + y * y;
    e /= (1.0 - rho * rho);
    return exp (e / (-2.0)) / u;
}

static double HRGPw1 (double lambda, double x, double y) {
    double h = 1.0 / lambda + lambda * log (y / x) / 2.0,
	   h1 = lambda * x / (2.0 * y),
	   h2 = -lambda * x * x / (2.0 * y * y);
    double phi = XGaussianDensity (h);
    return -y / (x * x) * (2.0 * phi * h1 + y / x * (-phi * h * h1 * h1 + phi * h2));
}

static double HuesslerReissGPDF (double lambda, double x, double y) {
    if (lambda == 0.0) return max (1 + x + y, 0.0);
    if (x >= 0.0) x = -1e-300;
    if (y >= 0.0) y = -1e-300;
    return max (0.0, 1.0 + XGaussianDF (1.0 / lambda + lambda / 2.0 * log (y / x)) * y
                         + XGaussianDF (1.0 / lambda + lambda / 2.0 * log (x / y)) * x);
}

double HuesslerReissGPDen (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to uniform on [-1, 0]
    double xs = XGPDF (x, gamma1) - 1.0,
           ys = XGPDF (y, gamma2) - 1.0;
    // check support
    if (HuesslerReissGPDF (lambda, xs, ys) == 0.0 || xs >= 0.0 || ys >= 0.0) return 0.0;
    double h = HRGPw1 (lambda, xs, ys) + HRGPw1 (lambda, ys, xs);
    //
    return h * XGPDensity (x, gamma1) * XGPDensity (y, gamma2);
}

double HuesslerReissGPDF (double lambda, double gamma1, double gamma2, double x, double y) {
    x = XGPDF (x, gamma1) - 1.0;
    y = XGPDF (y, gamma2) - 1.0;
    return HuesslerReissGPDF (lambda, x, y);
}

double HuesslerReissGPSF (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to uniform on [-1, 0]
    x = XGPDF (x, gamma1) - 1.0;
    y = XGPDF (y, gamma2) - 1.0;
    //
    return 1.0 - max (min (1.0, 1.0 + x), 0.0) - max (min (1.0 + y, 1.0), 0.0) + HuesslerReissGPDF (lambda, x, y);
}

double GumbelMcFaddenGPDF (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to uniform on [-1, 0]
    x = XGPDF (x, gamma1) - 1.0,
    y = XGPDF (y, gamma2) - 1.0;
    //
    if (lambda== 0.0) return max (1.0 + x + y, 0.0);
    return max (0.0, 1.0 - pow (pow (-x, lambda) + pow (-y, lambda), 1.0 / lambda));
}

double GumbelMcFaddenGPSF (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to uniform on [-1, 0]
    double xs = XGPDF (x, gamma1) - 1.0,
           ys = XGPDF (y, gamma2) - 1.0;
    //
    return 1.0 - max (min (1.0, 1.0 + xs), 0.0) - max (min (1.0 + ys, 1.0), 0.0)
           + GumbelMcFaddenGPDF (lambda, gamma1, gamma2, x, y);
}

double GumbelMcFaddenGPDen (double lambda, double gamma1, double gamma2, double x, double y) {
    // transform to uniform on [-1, 0]
    double xs = XGPDF (x, gamma1) - 1.0,
           ys = XGPDF (y, gamma2) - 1.0;
    //
    double h = (lambda - 1.0) * pow (pow (-xs, lambda) + pow (-ys, lambda), 1.0 / lambda - 2) * pow (xs * ys, lambda - 1.0);
    //
    return h * XGPDensity (x, gamma1) * XGPDensity (y, gamma2);
}

double MarshallOlkinGPDF (double lambda, double x, double y) {
    if (lambda == 0.0) return max (1.0 + x + y, 0.0);
    return max (0.0, 1.0 + (1.0 - lambda) * (x + y) + lambda * min (x, y));
}

double MarshallOlkinGPSF (double lambda, double x, double y) {
    return 1.0 - max (min (1.0, 1.0 + x), 0.0) - max (min (1.0 + y, 1.0), 0.0) + MarshallOlkinGPDF (lambda, x, y);
}

/********************************************************************
                        Discrete distributions
********************************************************************/

double NegativeBinomial (double r, double p, int k) {
    double prod = 1.0;
    int i, flr = (int) floor (r), minkr = min (k, flr);
    for (i = 1; i <= minkr; i++) 
        prod *= (r + i - 1) * (1.0 - p) * p/ i;
    for (; i <= k; i++)
        prod *= (r + i - 1) * (1.0 - p) / i;
    for (; i <= flr; i++)
        prod *= p;
    return prod * pow (p, r - floor (r));
}

double Binomial (int n, double p, int k) {
    int binK = min (k, n - k), i = 1, j = 1, l = 1;
    double prod = 1.0;
    while (i <= binK) {
    	prod *= double (n -i + 1) / i;
        i++;
        while (prod > 1.0 && j <= k) {
            prod *= p;
            j++;
        }
        while (prod > 1.0 && l <= n - k) {
            prod *= (1.0 -p);
            l++;
        }
    }
    while (j <= k) {
        prod *= p;
        j++;
    }
    while (l <= n - k) {
        prod *= (1.0 - p);
        l++;
    }
    return prod;
}

int PoissonRV (double lambda) {
    double l = exp (-lambda), r = Random ();
    int n = 1;
    while (r >= l) {
        r *= Random ();
        n++;
    }
    return n - 1;
}

double GammaRV (double alpha) {
    double Z, E, X, Y, U, V, W;
    if (alpha == 1.0) return ExpQF (Random (), 1.0, 0.0);
    if (alpha < 1.0) {
	// Johnk (see Devroye, p. 418)
        do {
            U = Random ();
            V = Random ();
            X = pow (U, 1.0 / alpha);
            Y = pow (V, 1.0 / (1.0 - alpha));
        } while (X + Y > 1.0);
        E = -log (Random ());
        return E * X / (X + Y);
    }
    double b = alpha - 1.0,
           c = 3.0 * alpha - 0.75;
    BOOL Accept = FALSE;
    do {
        U = Random ();
        V = Random ();
        W = U * (1.0 - U);
        Y = sqrt (c / W) * (U - 0.5);
        X = b + Y;
        if (X >= 0.0) {
            Z = 64.0 * W * W * W * V * V;
            Accept = (Z <= 1.0 - 2.0 * Y * Y / X) ||
                     (log (Z) <= 2.0 * (b * log (X / b) -Y ));
        }
    } while (!Accept);
    return X;
}

double NegBinRV (double r, double p) {
    double y = (1.0 - p) / p * GammaRV (r);
    return PoissonRV (y);
}

/*******************************************************************/
/*              Polynomial regression                              */
/*******************************************************************/

BOOL MatInvert (double **a, double **b, int n) {
    int nx = 0, ny = 0;
    double pivot, factor;
    int mx[100], my[100], i;
    for (i = 0; i < n; i++)
        for (int j = 0; j < n; j++)
            b[i][j] = a[i][j];
    for (i = 0; i < n; i++)
        mx[i] = my[i] = -1;
    for (i = 0; i < n; i++) {
        pivot = 0.0;
        for (int ix = 0; ix < n; ix++)
            if (mx[ix] == -1)
                for (int iy = 0; iy < n; iy++)
                    if (my[iy] == -1)
                        if (fabs (b[ix][iy]) > fabs (pivot)) {
                            pivot = b[ix][iy];
                            nx = ix; ny = iy;
                        }
        if (pivot == 0.0) return FALSE;
        mx[nx] = ny; my[ny]=nx;
        for (int j = 0; j < n; j++)
            if (j != nx) {
                factor = b[j][ny] / pivot;
                for (int k = 0; k < n; k++)
                    b[j][k] -= b[nx][k] * factor;
                b[j][ny] = -factor;
            }
        for (int k = 0; k < n; k++)
            b[nx][k] /= pivot;
        b[nx][ny] = 1.0 / pivot;
    }
    for (i = 0; i < n; i++) {
        int j = mx[i];
        if (i != j) {
            for (int k = 0; k < n; k++)
                Swap (b[i][k], b[j][k]);
            mx[i] = mx[j];
            mx[j] = j;
        }
        j = my[i];
        if (j != i) {
            for (int k = 0; k < n; k++)
                Swap (b[k][i], b[k][j]);
            my[i] = my[j]; my[j] = j;
        }
    }

/*
    // Test
    double s1 = 0.0, s2 = 0.0;
    for (i = 0; i < n; i++)
        for (int j = 0; j <n; j++) {
            double h = 0.0;
            for (int k = 0; k < n; k++)
                h += a [i][k] * b[k][j];
            if (i == j) s1 += fabs (h - 1.0);
            else s2 += abs (h);
        }
    MessageBox (0, form (100, "%g %g", s1, s2), "", MB_SYSTEMMODAL);
*/
    return TRUE;
}

inline double intpow (double x, int m) { return m ? pow (x, m) : 1; }

double *PolynomialRegression (int m, int n, double *x, double *y, BOOL *Marked) {
    double **a, **b;
    a = new double * [m + 1];
    b = new double * [m + 1];
    int i;
    for (i = 0; i <= m; i++) {
        a [i] = new double [m + 1];
        b [i] = new double [m + 1];
    }
    double *v = new double [2 * m + 1];
    if (Marked) {
        v [0] = 0.0;
        for (i = 0; i < n; i++)
            v [0] += !Marked[i];
    } else v [0] = n;
    for (i = 1; i <= 2 * m; i++) {
        v [i] = 0.0;
        for (int j = 0; j < n; j++)
            if ((Marked == NULL) || !Marked [j])
                v [i] += intpow (x [j], i);
    }
    double *ys = new double [m + 1];
    for (i = 0; i <= m; i++) {
        ys [i] = 0.0;
        for (int j = 0; j < n; j++)
            if ((Marked == NULL) || !Marked [j])
                ys [i] += y [j] * intpow (x [j], i);
    }
    for (i = 0; i <= m; i++)
        for (int j = 0; j <= m; j++)
            a [i][j] = v [i + j];
    // !!!! check for errors
    MatInvert (a, b, m + 1);
    double *c = new double [m + 1];
    for (i = 0; i <= m; i++) {
        c [i] = 0.0;
        for (int j = 0; j <= m; j++)
            c [i] += b [i][j] * ys [j];
    }
    delete [] v;
    for (i = 0; i <= m; i++) {
        delete [] a[i];
        delete [] b[i];
    }
    delete [] a;
    delete [] b;
    delete [] ys;
    return c;
}

double EvaluatePoly (double *a, int m, double x) {
    double v = a [m];
    for (int i = m - 1; i >= 0; i--) v = v * x + a [i];
    return v;
}

// Gaussian integration

static double integrate (double (*f) (double), double a, double b) {
    double w [] = { 0.066671344308688,  0.149451349050581,
                    0.219086362515982,  0.269266719309996,
                    0.295524334714753,  0.295524334714753,
                    0.269266719309996,  0.219086362515982,
                    0.149451349050581,  0.066671344308688 },
           x [] = {-0.973906528517172, -0.865063366688985,
                   -0.679409568299024, -0.433395394129247,
                   -0.148874338981631,  0.148874338981631,
                    0.433395394129247,  0.679409568299024,
                    0.865063366688985,  0.973906528517172 },
           sum = 0.0;
    for (int i = 0; i < 10; i++)    
        sum += (b - a) * 0.5 * w [i] * f ((b + a) * 0.5 + (b - a) * 0.5 * x [i]);
    return sum;
}

static double improve (double (*f) (double), double a, double b, double prev, double err, int level) {
    double m = 0.5 * (a + b),
           v1 = integrate (f, a, m),
           v2 = integrate (f, m, b),
           d = prev - v1 - v2;
    if (fabs (d) < 1e6 * err || level == 9)
        return v1 + v2;
    else
        return improve (f, a, m, v1, 0.5 * err, level + 1) +
               improve (f, m, b, v2, 0.5 * err, level + 1);
}

double GaussIntegrate (double (*f) (double), double a, double b) {
    double p = integrate (f, a, b);
    return improve (f, a, b, p, p / 1e12, 1);
}

// statistical procedures

double mean (double *x, int n) {
    if (n < 1) return 0.0;
    double sum = 0.0;
    for (int i = 0; i < n; i++)
    	sum += x [i];
    return sum / n;
}

double variance (double *x, int n) {
    if (n < 2) return 0.0;
    double m = mean (x, n),
           sum = 0.0;
    for (int i = 0; i < n; i++)
        sum += sqr (x [i] - m);
    return sum / (n - 1);
}

double covariance (double *x, double *y, int n) {
    double mx = mean (x, n),
           my = mean (y, n),
           sum = 0.0;
    for (int i = 0; i < n; i++)
	sum += (x [i] - mx) * (y [i] - my);
    return sum / n;
}

// two sample tests

double umpstat (double *x, int n, double *y, int m, double sigma) {
    // two sample ump with known variance
    return (mean (x, n) - mean (y, m)) / (sigma * sqrt (1.0 / n + 1.0 / m));
}

double ump (double *x, int n, double *y, int m, double sigma) {
    // two sample ump with known variance
    return 2.0 - 2.0 * XGaussianDF (fabs (umpstat (x, n, y, m, sigma)));
}

double umpss (double *x, int n, double *y, int m, double sigma) {
    // two sample ump with known variance
    return 1.0 - XGaussianDF ((mean (x, n) - mean (y, m)) / (sigma * sqrt (1.0 / n + 1.0 / m)));
}

double tteststat (double *x, int n, double *y, int m) {
    if (m < 2 || n < 2) return 0.0;
    double sx = variance (x, n),
           sy = variance (y, m),
           sp = ((n - 1) * sx + (m - 1) * sy) / (m + n - 2);
    return (mean (x, n) - mean (y, m)) / (sp * sqrt (1.0 / n + 1.0 / m));
}

double ttest (double *x, int n, double *y, int m) {
    return 2.0 - 2.0 *  XTDF (m + n - 2, fabs (tteststat (x, n, y, m)));
}

double ttestss (double *x, int n, double *y, int m) {
    return 1.0 - XTDF (m + n - 2, tteststat (x, n, y, m));
}

double welchtest (double *x, int n, double *y, int m) {
    if (m < 2 || n < 2) return 0.0;
    double sxn = variance (x, n) / n,
           sym = variance (y, m) / m,
           s = sqrt (sxn + sym);
    int deg = (int) floor (sqr (sxn + sym) / (sqr (sxn) / (n - 1) + sqr (sym) / (m - 1)));
    return 2.0 - 2.0 * XTDF (deg, fabs ((mean (x, n) - mean (y, m)) / s));
}

double welchtestss (double *x, int n, double *y, int m) {
    if (m < 2 || n < 2) return 0.0;
    double sxn = variance (x, n) / n,
           sym = variance (y, m) / m,
           s = sqrt (sxn + sym);
    int deg = (int) floor (sqr (sxn + sym) / (sqr (sxn) / (n - 1) + sqr (sym) / (m - 1)));
    return 1.0 - 1.0 * XTDF (deg, (mean (x, n) - mean (y, m)) / s);
}

double wilcoxontest (double *x, int n, double *y, int m) {
    double *z = new double [n + m],
    	   a = 0.0;
    int *r = new int [n + m], i;
    for (i = 0; i < n + m; i++) r [i] = i + 1;
    memcpy (z, y, m * sizeof (double));
    memcpy (z + m, x, n * sizeof (double));
    sort (0, n + m - 1, z, r);
    for (i = 0; i < m; i++)
        a += r [i];
    a = (a - (double) m * (m + n + 1) / 2.0) / sqrt ((double) n * (double) m * (n + m + 1) / 12.0);
    delete [] z;
    delete [] r;
    return 2.0 - 2.0 * XGaussianDF (fabs (a));
}

//multi t-Student Density

double BiStudentDensity(double alpha,double x,double y, TMatrix m) {
TMatrix i = inv(m);
double h = ((i(1,1)*x+i(1,2)*y)*x+(i(2,1)*x+i(2,2)*y)*y)/alpha;
return (gamma(0.5*(2+alpha)))/( (alpha*MY_PI)*gamma(0.5*alpha)*sqrt(det(m)) )*pow((1+h),(-0.5*(2+alpha)));
}

double TriStudentDensity(double alpha,double x,double y, double z, TMatrix m) {
TMatrix i = inv(m);
double h = ((i(1,1)*x+i(1,2)*y+i(1,3)*z)*x+(i(2,1)*x+i(2,2)*y+i(2,3))*y+(i(3,1)*x+i(3,2)*y+i(3,3)*z)*z)/alpha;
return (gamma(0.5*(3+alpha)))/( (alpha*MY_PI)*gamma(0.5*alpha)*sqrt(det(m)) )*pow((1+h),(-0.5*(3+alpha)));
}

double MultiStudentDensity(double alpha,int dim, double * vec, TMatrix m) {
if (dim==m.Columns()==m.Rows())
    {
    TMatrix i = inv(m);
    double h=0.0;
    for (int j=1;j<=dim;j++)
        {
      	double hsumme=0.0;
   	for (int k=1;k<dim;k++){
            hsumme=i(j,k)*vec[k-1];
        }
	h=h+hsumme*vec[j-1];
        }
   h=h/alpha;
   return (gamma(0.5*(dim+alpha)))/( (alpha*MY_PI)*gamma(0.5*alpha)*sqrt(det(m)) )*pow((1+h),(-0.5*(dim+alpha)));
   }
else
   {
   return 0.0;
   }
}


double digamma(double x)
{
    return ((gamma(x+1e-5)-gamma(x))/1e-5)/gamma(x);
}
