/*  mathfunc.hpp

    Declaration of densities, distribution and quantile functions.

    This file is part of Xtremes for Windows,
    Copyright (C) 1994 Xtremes Group, Siegen.
*/

#ifndef MATHFUNC_HPP
#define MATHFUNC_HPP

#include <stdlib.h>

#include "boolean.hpp"
#include "tmatrix.hpp"
#include "stddef.hpp"

void Randomize ();
void Randomize (unsigned int seed);
double Random ();
unsigned int IRandom ();

/* The following functions are partly wrappers to provide the standard models
   without location/scale parameter. All functions are prefixed with X to
   avoid name conflicts.

   The UFO* versions swap shape parameter and argument */

double XBetaData (double alpha);
double XBetaDensity (double x, double alpha);
double XBetaDF (double x, double alpha);
double XBetaQF (double x, double alpha);
double UFOXBetaDensity (double alpha, double x);
double UFOXBetaDF (double alpha, double x);
double UFOXBetaQF (double alpha, double x);

double XChiSquareData (double nu);
double XChiSquareDensity (double x, double nu);
double XChiSquareQF (double x, double nu);
double XChiSquareDF (double x, double nu);

double XEVData (double gamma);
double XEVDensity (double x, double gamma);
double XEVDF (double x, double gamma);
double XEVQF (double x, double gamma);
double UFOXEVDensity (double gamma, double x);
double UFOXEVDF (double gamma, double x);
double UFOXEVQF (double gamma, double x);

double XExponentialData ();
double XExponentialDensity (double x);
double XExponentialDF (double x);
double XExponentialQF (double x);

double XFrechetData (double alpha);
double XFrechetDensity (double x, double alpha);
double XFrechetDF (double x, double alpha);
double XFrechetQF (double x, double alpha);
double UFOXFrechetDensity (double alpha, double x);
double UFOXFrechetDF (double alpha, double x);
double UFOXFrechetQF (double alpha, double x);

double XGaussianData ();
double XGaussianDensity (double x);
double XGaussianDF (double x);
double XGaussianQF (double x);

double XLogNormalData ();
double XLogNormalDensity (double x);
double XLogNormalDF (double x);
double XLogNormalQF (double x);

double XGCauchyData (double alpha);
double XGCauchyDensity (double x, double alpha);
double XGCauchyDF (double x, double alpha);
double XGCauchyQF (double x, double alpha);
double UFOXGCauchyDensity (double alpha, double x);
double UFOXGCauchyDF (double alpha, double x);
double UFOXGCauchyQF (double alpha, double x);

double XGPData (double gamma);
double XGPDensity (double x, double gamma);
double XGPDF (double x, double gamma);
double XGPQF (double x, double gamma);
double UFOXGPDensity (double gamma, double x);
double UFOXGPDF (double gamma, double x);
double UFOXGPQF (double gamma, double x);

double XGumbelData ();
double XGumbelDensity (double x);
double XGumbelDF (double x);
double XGumbelQF (double x);

double XParetoData (double alpha);
double XParetoDensity (double x, double alpha);
double XParetoDF (double x, double alpha);
double XParetoQF (double x, double alpha);
double UFOXParetoDensity (double alpha, double x);
double UFOXParetoDF (double alpha, double x);
double UFOXParetoQF (double alpha, double x);

double XTData (double n);
double XTDF (double alpha, double x);
double XTQF (double alpha, double x);
double XTDensity (double alpha, double x);

double XNCTData (double alpha, double beta);
double XNCTDensity (double alpha, double beta, double x);
double XNCTDF (double alpha, double beta, double x);
double XNCTQF (double alpha, double beta, double x);

double XGammaData (double r);
double XGammaDensity (double r, double x);
double XGammaDF (double r, double x);

double XWeibullData (double alpha);
double XWeibullDensity (double x, double alpha);
double XWeibullDF (double x, double alpha);
double XWeibullQF (double x, double alpha);
double UFOXWeibullDensity (double alpha, double x);
double UFOXWeibullDF (double alpha, double x);
double UFOXWeibullQF (double alpha, double x);

double XIndicator (double a, double b, double x);
double XMin (double a, double b);
double XMax (double a, double b);

/***************/

BOOL MatInvert (double **a, double **b, int n);
double *PolynomialRegression (int m, int n, double *x, double *y, BOOL *Marked = NULL);	// allocates array
double EvaluatePoly (double *a, int m, double x);

double frac (double x);
#if !defined(__GNUC__) || __GNUC__ < 3
double round (double x);
#endif
double sign (double x);

#ifdef __GNUC__
  extern "C"
#endif
double gamma (double x)
#ifdef __GNUC__
#if ((__GNUC__ == 2)  && (__GNUC_MINOR__ == 96)) || (__GNUC__ >= 3)
    throw ()
#endif
#endif
    ;	// gamma function

double ParetoDen (double x, double a, double t, double m = 0.0);
double ParetoQF (double x, double a, double t, double m = 0.0);
double ParetoDF (double x, double a, double s, double m =0.0);

double ParetoIIDen (double x, double a, double t, double m = 0.0);
double ParetoIIQF (double x, double a, double t, double m = 0.0);
double ParetoIIDF (double x, double a, double s, double m = 0.0);

double ExpDen (double x, double s, double m = 0.0);
double ExpQF (double x, double s, double m = 0.0);
double ExpDF (double x, double s, double m = 0.0);

double CauchyDen (double x, double ac, double tc, double rc);
double CauchyQF (double x, double t, double r);
double CauchyDF (double x, double t, double r);
double CauchyInt (double x1, double x2, double a, double t, double r, int n);
double CauchyDFStart (double x, double alpha, double sigma, double mu);
double CauchyDFStep (double x1, double x2, double alpha, double sigma, double mu);
double CauchyRV (double alpha, BOOL pos);

double NormDen (double x, double m, double s);
// double NormalDF (double x, double m, double s);
void NormRV (double &r1, double &r2);
double NormRV ();
double NormDFPol (double x, double m, double s);
// double NormDFExp (double x, double m, double s, int type);
double NormQF (double x, double m, double s);

double LogNormDen (double x, double b, double s);
double LogNormDFPol (double x, double b, double s);
double LogNormRV (double b, double s);
double LogNormQF (double x, double b, double s);

double ChiSquareDensity (double x, int nu);
double ChiSquareQF (double x, int nu);
double ChiSquareDF (double x, int nu);

double squarefunc (double x);

double NegativeBinomial (double r, double p, int k);
double Binomial (int n, double p, int k);

int PoissonRV (double lambda);
double GammaRV (double alpha);
double NegBinRV (double r, double p);

double tDF (double x, int nu);

double rstab (double alpha, double primeb);

double vonMisesGPDDen (double x, double b, double t, double m);
double vonMisesGPDQF (double x, double b, double t, double m);
double vonMisesGPDDF (double x, double b, double t, double m);

double vonMisesEXTDen (double x, double b, double s, double m = 0.0);
double vonMisesEXTQF (double x, double b, double s, double m = 0.0);
double vonMisesEXTDF (double x, double b, double s, double m = 0.0);

double GumbelDen (double x, double s, double m = 0.0);
double GumbelQF (double x, double s, double m = 0.0);
double GumbelDF (double x, double s, double m = 0.0);

double FrechetDen (double x, double a, double s, double m = 0.0);
double FrechetQF (double x, double a, double s, double m = 0.0);
double FrechetDF (double x, double a, double s, double m = 0.0);

double WeibullDen (double x, double a, double s, double m = 0.0);
double WeibullQF (double x, double a, double s, double m = 0.0);
double WeibullDF (double x, double a, double s, double m = 0.0);

double EVGGDen (double x, double lambda, double beta, double mu, double sigma);
double EVGG0Den (double x, double lambda, double mu, double sigma);
double EVGG1Den (double x, double lambda, double alpha, double mu, double sigma);
double EVGG2Den (double x, double lambda, double alpha, double mu, double sigma);

double HuesslerReissDF (double lambda, double gamma1, double gamma2, double x, double y);
double HuesslerReissSF (double lambda, double gamma1, double gamma2, double x, double y);
double HuesslerReissDen (double lambda, double gamma1, double gamma2, double x, double y);
double HuesslerReissCondDF (double theta, double x, double y);	// H_theta (y | x)
double HuesslerReissCondQF (double theta, double x, double y);  // H^{-1}_theta (q | x)

double MarshallOlkinDF (double rho, double x, double y);
double MarshallOlkinSF (double rho, double x, double y);

double GumbelMcFaddenEVDF (double lambda, double gamma1, double gamma2, double x, double y);
double GumbelMcFaddenEVSF (double lambda, double gamma1, double gamma2, double x, double y);
double GumbelMcFaddenEVDen (double lambda, double gamma1, double gamma2, double x, double y);

double HuesslerReissGPDen (double lambda, double gamma1, double gamma2, double x, double y);
double HuesslerReissGPDF (double lambda, double gamma1, double gamma2, double x, double y);
double HuesslerReissGPSF (double lambda, double gamma1, double gamma2, double x, double y);

double GumbelMcFaddenGPDF (double lambda, double gamma1, double gamma2, double x, double y);
double GumbelMcFaddenGPSF (double lambda, double gamma1, double gamma2, double x, double y);
double GumbelMcFaddenGPDen (double lambda, double gamma1, double gamma2, double x, double y);

double MarshallOlkinGPDF (double lambda, double x, double y);
double MarshallOlkinGPSF (double lambda, double x, double y);

double BiGaussDF (double rho, double x, double y);
double BiGaussSF (double rho, double x, double y);
double BiGaussDen (double rho, double x, double y);

double BiStudentDensity (double alpha,double x,double y,TMatrix m);
double TriStudentDensity (double alpha,double x,double y,double z,TMatrix m);
double MultiStudentDensity (double alpha,int dim,double * vec,TMatrix m);

// Gaussian integration

double GaussIntegrate (double (*f) (double), double a, double b);

// statistical procedures

double mean (double *x, int n);
double variance (double *x, int n);
double covariance (double *x, double *y, int n);

// tests (return p-value)

double ump (double *x, int n, double *y, int m, double sigma);
double ttest (double *x, int n, double *y, int m);
double welchtest (double *x, int n, double *y, int m);
double umpss (double *x, int n, double *y, int m, double sigma);
double ttestss (double *x, int n, double *y, int m);
double welchtestss (double *x, int n, double *y, int m);

double wilcoxontest (double *x, int n, double *y, int m);

// Stable distributions

#ifdef USE_STABLE

double stableden (double x, double alpha, double beta, double mu, double sigma);
double stabledf (double x, double alpha, double beta, double mu, double sigma);
double stableqf (double q, double alpha, double beta, double mu, double sigma);
double stabledata (double alpha, double beta, double mu, double sigma);

double Xstableden (double x, double alpha, double beta);
double Xstabledf (double x, double alpha, double beta);
double Xstableqf (double q, double alpha, double beta);
double Xstabledata (double alpha, double beta);

#endif

double digamma(double x);

#endif
