/*  mlegpd.cpp

    Implemenation of Maximum Likelihood estimator for GP model.

    This file is part of Xtremes for Windows,
    Copyright (C) 1994-95 Xtremes Group, Siegen.
*/

// b,m,s

#include <math.h>
#include "allest.hpp"
#include "mathfunc.hpp"
#include "stddef.hpp"

#ifndef TERMINAL
#include "resource.h"
#endif


TMLEEstimator::TMLEEstimator (BOOL AddList)
#ifdef TERMINAL
 : TUnivariateGPDEstimator ("MLE(GP)") {
#else
 : TUnivariateGPDEstimator ("MLE(GP)", IDM_MLEGPD, IDD_Estimator, AddList) {
#endif
    Data.SetMode (BETA_MODE);
}

const double LOG2 = 0.6931472;

static double PEst (TUnivariateSample *p, int size, int k) {
    int x2, m = k >> 2;
    double h1, h2;
    do {
        x2 = size - (m << 1);
        h1 = (*p)[size - m] - (*p)[x2];
        h2 = (*p)[x2] - (*p)[size - (m << 2)];
        m++;
    } while ((h1 == 0.0 || h2 == 0.0) && m <= size / 4);
    return (h2 == 0.0) ? MYMAXDOUBLE : log (h1 / h2) / LOG2;
 }

inline double betadach (TUnivariateSample *p, int n, int k) {
    return PEst (p, n, (k << 2));
}

const double EPS = 1e-6;

inline int rnd (double x) {
    return (int) floor (x + 1 - EPS);
}

double MPEst (TUnivariateSample *p, int n, int k) {
    double b = PEst (p, n, k);
    int m = k / 4;
    if (k < 4) return b;
    int l = max ((int) floor (log (k) / LOG2) - 1,1);
    double *c = new double [l];
    if (b < -0.5) b = -(1.0 + b);
    double Pwr2b = pow (2.0, b + 1.0);
    int i;
    for (i = 0; i < l - 1; i++)
        c [i] = (b != 0.0 ?
                 (Pwr2b - 1.0) / (Pwr2b * 0.5 - 1.0) *
                   (1.0 - pow (2.0, -(i+1) * b)) * pow (2.0, -i) :
                 (i+1) * pow (2.0, -i)
                ) / 4.0;
    c [l-1]   = b != 0.0 ?
                 pow (2.0, -l) / (Pwr2b * 0.5 - 1.0) *
                   (Pwr2b - 1.0 - pow (2.0, (1-l)*b)) :
                 (l+1) * pow (2.0, -l);
    for (i = 0, b = 0.0; i < l; i++)
        b += c[i]* betadach (p, n, rnd (m * pow (2, -i)));
    delete c;
    return b;
}

static double MomLogEst (TUnivariateSample *p, int n, int k) {
    double M1 = 0.0, M2 = 0.0, l, nk = ((*p)[n-k]);
    while ((*p)[n-1]*nk<=0.0 && k>8) {
        k--;
        nk = ((*p)[n-k]); }
    if (k<=8) return 1;
    for (int i = 1; i < k; i++) {
        l = log ((*p) [n-i]/nk);
        M1 += l;
        M2 += l * l;
    }
    M1 /= (k-1); M2 /= (k-1);
    double g = M2 == 0.0 ?
                         0 :
                 M1 + 1.0 - 0.5 / (1.0 - M1 * M1 / M2);
    return g;
}

static double MomEst (TUnivariateSample *p, int n, int k) {
        double s= (*p)[n-k-1];
        double M1 = 0.0, M2 = 0.0;
        for (int i=n-k; i<n; i++) {
            M1+=((*p)[i]-s);
            M2+=((*p)[i]-s)*((*p)[i]-s);
        }
        M1/=k; M2/=k;
        double nenner = M2 - M1*M1;
        double zaehler = M1*M1;
        if (nenner == 0.0) return MYMAXDOUBLE;
        else return 0.5* (1-zaehler / nenner);
//      return 0.0;
}

void TMLEEstimator::FindStart (double &bstart, double &tstart, int type,
                               TUnivariateSample *p, double s,
                               int k ,int n) {
    if (type == 1) bstart = MPEst(p, n, k);
    else { if (type == 0) bstart = MomLogEst(p, n, k);
           else bstart = MomEst(p, n, k); }
    double tq = 0.0;
    double help;
    if (bstart == 0.0) bstart = 0.01;
    for (int i=2; i<=k; i++) {
        help = (double) (i-1)/k;
        help = pow (help,bstart);
        tq+=bstart*(help*(*p)[n-i]-s)/(1-help);
    }
    tq/=(k-1);
    tstart = tq;
}

int TMLEEstimator::FindZero (double &alpha, double &sigma, double &mu,
                             TUnivariateSample *p, double s, int k, int n) {
    double beta, theta, rho, bstep, tstep;
    double g1, g2, gs1, gs2, gs3, gs4;
    double help, det;
    int step = 0;
    BOOL stop = FALSE;
    beta = bstart; theta = tstart;
    bstep = 1; tstep=1;
    while ((fabs (bstep)>1e-8 || fabs (tstep)>1e-8) && step <500 && !stop) {
          if (beta < 0.0 && theta+beta*(*p)[n-1]<=0.0) {
                  if (beta==0.0 || (*p)[n-1]==0.0) theta = 0.01;
                  else theta = -beta*(*p)[n-1]+fabs(beta*(*p)[n-1])*0.01;}
          else if (theta+beta*s<=0.0){
                  if (beta==0.0 || s==0.0) theta = 0.01;
                  else theta = -beta*s+fabs(beta*s)*0.01;}
          g1 = 0.0; g2 = 0.0;
          gs1 = 0.0; gs2 = 0.0; gs3 = 0.0; gs4 = 0.0;
          for (int i=1; i<=k; i++) {
              help = theta+beta*(*p)[n-i];
              g1+=1/help;
              g2+=log(help);
              gs1+=1/(help*help);
              gs2+=(*p)[n-i]/(help*help);
              gs3+=(*p)[n-i]/help;
              gs4+=(*p)[n-i]*(*p)[n-i]/(help*help);
          }
          help = theta+beta*s;
          gs4 = -1/beta*s*s/(help*help)+(1/beta+1)*gs4/k; // Reihenfolge nicht
          gs4 += 2/pow(beta,3)*(log(help)-g2/k);
          gs4 += -2/(beta*beta)*(s/help-gs3/k);
          gs2 = -1/beta*s/(help*help)+(1/beta+1)*gs2/k;
          gs2 += -1/(beta*beta)*(1/help-g1/k);
          gs1 = -1/(beta*help*help)+(1/beta+1)*gs1/k;
          g2 = -1/(beta*beta)*(log(help)-g2/k);
          g2 += 1/beta*s/help-(1/beta+1)*gs3/k;
          g1 = 1/(beta*help)-(1/beta+1)*g1/k;
          gs3 = gs2;                                     // verndern !!!
          det = gs1*gs4-gs2*gs3;
          if (!(det==0.0)) {
             tstep = - (gs4*g1-gs2*g2)/det;
             bstep = - (-gs3*g1+gs1*g2)/det;
             beta = beta + bstep;
             theta = theta + tstep;
          }
          else  stop = TRUE;
          if (fabs(g1)>1e100 || fabs(g2)>1e100) stop = TRUE;
          if ((fabs(beta-bstart)>1e10) || (fabs(theta-tstart)> 1e10)) stop = TRUE;
          step++;
    }
    if (fabs (bstep)>1e-5 || fabs (tstep)>1e-5 || stop == TRUE) return 0;
    rho = (double) (k+1)/n;
    sigma = pow (rho,beta)*(theta+beta*s);
    mu = (sigma-theta)/beta;
    alpha = beta;
    return 1;
}

TEstMode TMLEEstimator::GetEstimatorMode () {
    return FULL_BETA_MODE;
}

TEstResult TMLEEstimator::Estimate (int k, TSample *q) {
    TUnivariateSample *p = (TUnivariateSample *) q;
    int n = p->GetDataSize ();
    if (k <= 8) return EST_EXT;
    if (k > n) return EST_PAR;
    k = k-1;
    double s = (*p)[n-k-1];
    double alpha, sigma, mu;
    FindStart (bstart, tstart, 1, p, s, k, n);
    int result = 0;
    for (int estno = 1; (estno >= -1) && (result == 0); estno--) {   // MPEst, MomLogEst, MomEst
        FindStart (bstart, tstart, estno, p, s, k, n);
        result = FindZero (alpha, sigma, mu, p, s, k, n);
    }
    if (result == 0) return EST_POZ;
    RawParameters [0] = alpha;
    RawParameters [2] = sigma;
    RawParameters [1] = mu;
    TransformModels (k, p);
    return EST_OK;
}

TEstimatorData *TMLEEstimator::GetParameter () {
    return &Data;
}

TEstimator *TMLEEstimator::Duplicate () {
    TMLEEstimator *p = new TMLEEstimator;
    p->Data = Data;
    return p;
}

int TMLEEstimator::GetMinK () {
    return 8;
}

double TMLEEstimator::GetExponent () {
    return 0.33333333;
}

void TMLEEstimator::CalcEstimator (Array<double> &alpha, int &kmax, TSample *q) {
    TUnivariateSample &p = *(TUnivariateSample *) q;
    kmax = p.GetDataSize ();
    for (int k = GetMinK (); k <= kmax; k++)
        if (Estimate (k, q) == EST_OK)
            alpha [k] = RawParameters [0];
        else
            alpha [k] = 0.0;
}
