/*  normest.cpp

    Estimators for Gaussian, stable and t distributions.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 2000 Xtremes Group, Siegen.
*/

#include <math.h>
#include <stdlib.h>

#include "allest.hpp"
#include "mathfunc.hpp"
#include "stddef.hpp"

#ifndef	TERMINAL
#include "resource.h"
#endif

/* Maximum-Likelihood-Estimator	*/

TNormalMLEstimator::TNormalMLEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TNormalEstimator ("MLE (Gaussian)", IDM_NormalML, IDD_GaussEstimator, AddList) {
#else
 : TNormalEstimator ("MLE (Gaussian)") {
#endif
}

TEstimator *TNormalMLEstimator::Duplicate () {
    return new TNormalMLEstimator;
}

TEstResult TNormalMLEstimator::Estimate	(int k,	TSample	*p) {
    TUnivariateSample &Data = *(TUnivariateSample *) p;
    int	n = Data.GetDataSize ();
    double xm =	0.0, s = 0.0;
    int i;
    for	(i = 0; i < n; i++)
	xm += Data [i] / n;
    for	(i = 0;	i < n; i++)
	s += (Data [i] - xm) * (Data [i] - xm) / n;
    Parameters [1] = sqrt (s);
    Parameters [0] = xm;
    return EST_OK;
}


/********************************************************************
		       Quick Estimate
********************************************************************/


TNormalQuickEstimator::TNormalQuickEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TNormalEstimator ("Quick (Gaussian)", IDM_GaussianQuick, IDD_GaussQuick, AddList),
   qField (IDD_Q, GlobalData.NormalQuickQ, 0.51, 1.0),
   SymmetricField (IDD_Symmetric, GlobalData.NormalQuickSymmetric) {
    *this << qField << SymmetricField;
#else
 : TNormalEstimator ("Quick (Gaussian)") {
#endif
}

TEstimator *TNormalQuickEstimator::Duplicate () {
    TNormalQuickEstimator *p = new TNormalQuickEstimator;
    p->Data = Data;
    return p;
}

TEstResult TNormalQuickEstimator::Estimate (int k, TSample *p) {
    TUnivariateSample &s = *(TUnivariateSample *) p;
    int	n = s.GetDataSize ();
    k = (int) (n * Data.q);
    if (k < 0) k = 0;
    if (k >= n) k = n - 1;
    if (Data.Symmetric) {
        double *x = new double [n];
        for (int i = 0; i < n; i++)
            x [i] = fabs (s [i]);
        qsort (x, n, sizeof (double), realcomp);
        Parameters [1] = x [k] / XGaussianQF (0.5 + Data.q * 0.5);
        delete [] x;
    } else
        Parameters [1] = s [k] / XGaussianQF (Data.q);
    Parameters [0] = 0.0;
    return EST_OK;
}

TEstimatorData *TNormalQuickEstimator::GetParameter () {
    return &Data;
}

#ifndef TERMINAL

void TNormalQuickEstimator::NotifyActionEvent (TDialogMember *m) {
    if (m->GetID () == IDD_Estimate || m->GetID () == IDD_Symmetric)
        PerformEstimation ();
    inherited::NotifyActionEvent (m);
}

void TNormalQuickEstimator::EvaluateAdditionalParameters () {
    if (qField.IsCorrect ()) {
        Data.q = GlobalData.NormalQuickQ = qField.GetValue ();
        Data.Symmetric = GlobalData.NormalQuickSymmetric = SymmetricField.Selected ();
    }
}
#endif


/********************************************************************
		       Least-Squares Estimate
********************************************************************/

TNormalLSEstimator::TNormalLSEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TNormalEstimator ("Least Square (Gaussian)",	IDM_LSGaussian,	IDD_GaussEstimator, AddList) {
#else
 : TNormalEstimator ("Least Square (Gaussian)")	{
#endif
}

TEstimator *TNormalLSEstimator::Duplicate () {
    return new TNormalLSEstimator;
}

TEstResult TNormalLSEstimator::Estimate	(int l,	TSample	*p) {
    TUnivariateSample &Data = *(TUnivariateSample *) p;
    int	k = Data.GetDataSize ();
    double xq =	0.0, yq	= 0;
    int i;
    for	(i = k / 10; i < k - k / 10; i++) {
	yq += Data [i];
	xq += NormQF (double (i) / double (k + 1), 0.0,	1.0);
    }
    xq /= (0.8 * k); yq	/= (0.8	* k);
    double zaehler = 0.0, nenner = 0.0,	x_xq, y_yq;
    for	(i = k / 10; i < k - k / 10; i++) {
	x_xq = (NormQF (double (i) / double (k + 1), 0.0, 1.0) - xq);
	y_yq = Data[i] - yq;
	zaehler	+= y_yq	* y_yq;
	nenner += x_xq * y_yq;
    }
    if (nenner == 0.0) Parameters [1] =	MYMAXDOUBLE;
    else Parameters [1]	= zaehler / nenner;
    Parameters [0] = yq	- Parameters [1] * xq;
    return EST_OK;
}

#ifdef USE_STABLE

/********************************************************************
		ML estimator for stable distribution
********************************************************************/

#include "stable1.h"

TStableMLEstimator::TStableMLEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TStableEstimator ("MLE (Stable)", IDM_StableML, IDD_GaussEstimator, AddList) {
#else
 : TStableEstimator ("MLE (Stable)") {
#endif
}

TEstimator *TStableMLEstimator::Duplicate () {
    return new TStableMLEstimator;
}

TEstResult TStableMLEstimator::Estimate	(int k,	TSample	*p) {
    TUnivariateSample &Data = *(TUnivariateSample *) p;
    int	n = Data.GetDataSize (),
        method = 1,			// ML
        iparam = 0,			// S0
        ierr = 0;
    double *x = new double [n];
    if (!x) return EST_FAILURE;
    for (int i = 0; i < n; i++)
        x [i] = Data.GetRawData (i);
    SFIT (&n, x, Parameters, &method, &iparam, &ierr);
    Swap (Parameters [2], Parameters [3]);
    delete [] x;
    return ierr? EST_FAILURE : EST_OK;
}

/********************************************************************
           Koutrovelis-Kogon-Williams estimator
********************************************************************/


TStableKKWEstimator::TStableKKWEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TStableEstimator ("McCulloch Quantile Estimator (Stable)", IDM_StableQuantile, IDD_GaussEstimator, AddList) {
#else
 : TStableEstimator ("McCulloch Quantile Estimator (Stable)") {
#endif
}

TEstimator *TStableKKWEstimator::Duplicate () {
    return new TStableKKWEstimator;
}

TEstResult TStableKKWEstimator::Estimate (int k, TSample *p) {
    TUnivariateSample &Data = *(TUnivariateSample *) p;
    int	n = Data.GetDataSize (),
        method = 2,			// Quantile Estimator
        iparam = 0,			// S0
        ierr = 0;
    double *x = new double [n];
    if (!x) return EST_FAILURE;
    for (int i = 0; i < n; i++)
        x [i] = Data.GetRawData (i);
    SFIT (&n, x, Parameters, &method, &iparam, &ierr);
    Swap (Parameters [2], Parameters [3]);
    delete [] x;
    return ierr? EST_FAILURE : EST_OK;
}

#endif

/********************************************************************
 	           MLE for t distribution
********************************************************************/

TTMLEstimator::TTMLEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TTEstimator ("MLE (Student)", IDM_MLET, IDD_GaussEstimator, AddList) {
#else
 : TTEstimator ("MLE (Student)") {
#endif
}

TEstimator *TTMLEstimator::Duplicate () {
    return new TTMLEstimator;
}

TEstResult TTMLEstimator::Estimate (int k, TSample *q) {
    p = q;
    TUnivariateSample &Data = *(TUnivariateSample *) q;
    int	n = Data.GetDataSize ();
    double *x = new double [2 * n];
    for (int i = 0; i < n; i++) {
        x [i] = Data [i];
        x [n + i] = -Data [i];
    }
    TUnivariateSample *s = new TUnivariateSample ();
    s->SetData (2 * n, x, FALSE);
    HillGP1->Estimate (n / 6, s);
    Parameters [0] = HillGP1->GetEstimatedParameterValue (0);
    Parameters [1] = s->SampleQf (0.8) / XTQF (Parameters [0], 0.8);
    delete s;
    BOOL Success = SearchMax (2);
    return Success ? EST_OK : EST_FAILURE;
}

double TTMLEstimator::SearchL (double *para) {
    double sum = 0.0;
    int n = p->GetDataSize ();
    for (int i = 0; i < n; i++)
        sum += log (XTDensity (para [0], (*(TUnivariateSample *) p) [i] / para [1]));
    return sum - n * log (para [1]);
}

BOOL TTMLEstimator::SearchBounds (double *para) {
    return para [0] > 0.0 && para [1] > 0.0;
}

TNCTMLEstimator::TNCTMLEstimator (BOOL AddList)
#ifndef	TERMINAL
 : TNCTEstimator ("MLE (Non-central Student)", IDM_MLENCT, IDD_GaussEstimator, AddList) {
#else
 : TNCTEstimator ("MLE (Non-central Student)") {
#endif
}

TEstimator *TNCTMLEstimator::Duplicate () {
    return new TNCTMLEstimator;
}

TEstResult TNCTMLEstimator::Estimate (int k, TSample *q) {
    p = q;
    TUnivariateSample &Data = *(TUnivariateSample *) q;
    int	n = Data.GetDataSize ();
    double *x = new double [2 * n];
    for (int i = 0; i < n; i++) {
        x [i] = Data [i];
        x [n + i] = -Data [i];
    }
    TUnivariateSample *s = new TUnivariateSample ();
    s->SetData (2 * n, x, FALSE);
    HillGP1->Estimate (n / 6, s);
    Parameters [0] = HillGP1->GetEstimatedParameterValue (0);
    Parameters [1] = 0.0;
    Parameters [2] = s->SampleQf (0.8) / XTQF (Parameters [0], 0.8);
    delete s;
//    BOOL Success = GradientMax (Parameters [0], Parameters [1]);
    BOOL Success = SearchMax (3);
    return Success ? EST_OK : EST_FAILURE;
}

double TNCTMLEstimator::SearchL (double *para) {
    double sum = 0.0;
    int n = p->GetDataSize ();
    for (int i = 0; i < n; i++)
        sum += log (XNCTDensity (para [0], para [1], (*(TUnivariateSample *) p) [i] / para [2]));
    return sum - n * log (para [2]);
}

BOOL TNCTMLEstimator::SearchBounds (double *para) {
    return para [0] > 0.0 && para [2] > 0.0;
}

#ifndef TERMINAL

#include "garchmle.hpp"


void OpenGARCHDialog (TTimeSeriesSample *p, TWindow *parent) {
    double alpha, beta, gamma;
    int n = p->GetDataSize ();
    double *x = new double [n];
    for (int i = 0; i < n; i++)
        x [i] = p->GetData (i, 1);
    int r = garchmle (n, x, alpha, beta, gamma);
    delete [] x;
    String s =form (100, "alpha0 = %g, alpha1 = %g, beta1 = %g",
                           gamma, alpha, beta);
    if (r)
        s = s & "\nMLE failed - displaying pre-estimator.";
    InfoBox (parent, s, "MLE(GARCH(1,1))");
}

#endif
