/*  pcode.cpp
 *
 *  Implementation of virtual machine for StatPascal.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998, 2000 Xtremes Group, Siegen.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>

#include "mathfunc.hpp"
#include "pcode.hpp"
#include "stddef.hpp"
#include "tmatrix.hpp"

#ifndef TERMINAL
  #include "cancel.hpp"
  #include "wintypes.hpp"
  #include "resource.h"
#endif

typedef char *TSetCast;

// Visual C++ does not like the empty parameters given
// in macro calls; turn off annoying warning

#ifdef _MSC_VER
#pragma warning (disable: 4003) 
#endif

/*********************************************************/
/*		Management of file handles		 */
/*********************************************************/

TFileDescriptor::TFileDescriptor () {
    InputPtr = InputBuf;
    *InputPtr = '\0';
}

TFileDescriptor::~TFileDescriptor () {
}

void TFileDescriptor::Reset () {
}

void TFileDescriptor::Rewrite () {
}

void TFileDescriptor::OutputBin (void *, int) {
}

void TFileDescriptor::InputBin (void *, int) {
}

void TFileDescriptor::OutputText (String) {
}

void TFileDescriptor::FillInputBuf () {
    puts ("TFileDesc::FillInputBuf called");
}

double TFileDescriptor::GetDouble () {
    while (*InputPtr == ' ') InputPtr++;
    while (!*InputPtr || *InputPtr == '\n') FillInputBuf ();
    double v = strtod (InputPtr, &InputPtr);
    return v;
}

int TFileDescriptor::GetInt () {
    while (*InputPtr == ' ') InputPtr++;
    while (!*InputPtr || *InputPtr == '\n') FillInputBuf ();
    int n = (int) strtol (InputPtr, &InputPtr, 10);
    return n;
}

int TFileDescriptor::GetChar () {
    while (!*InputPtr || *InputPtr == '\n') FillInputBuf ();
    return *InputPtr++;
}

String TFileDescriptor::GetString () {
    if (!*InputPtr || *InputPtr == '\n') FillInputBuf ();
    char *s = InputPtr;
    while (*InputPtr && *InputPtr != '\n') InputPtr++;
    if (*InputPtr == '\n') *InputPtr++ = '\0';
    return s;
}

void TFileDescriptor::GetLF () {
    if (!*InputPtr)	// throw away one line
	FillInputBuf ();
    while (*InputPtr != '\n' && *InputPtr) InputPtr++;   // skip to end of line
}

int TFileDescriptor::Eof () {
    return 0;
}

void TFileDescriptor::FFlush () {
}

void TFileDescriptor::Page () {
}

void TFileDescriptor::SetTextColor (int) {
}

void TFileDescriptor::SetTextBackground (int) {
}

void TFileDescriptor::GotoXY (int x, int y) {
}

int TFileDescriptor::GetKey () {
    return 0;
}

TBinFileDescriptor::TBinFileDescriptor (String s0) {
    f = NULL;
    s = s0;
}

TBinFileDescriptor::~TBinFileDescriptor () {
    if (f) fclose (f);
}

void TBinFileDescriptor::Reset () {
    f = fopen (s, "rb");
}

void TBinFileDescriptor::Rewrite () {
    f = fopen (s, "wb");
}

void TBinFileDescriptor::OutputBin (void *ptr, int size) {
    fwrite (ptr, 1, size, f);
}

void TBinFileDescriptor::InputBin (void *ptr, int size) {
    fread (ptr, 1, size, f);
}

int TBinFileDescriptor::Eof () {
    return feof (f);
}

void TBinFileDescriptor::FFlush () {
    fflush (f);
}


TTextFileDescriptor::TTextFileDescriptor (String s0) {
    f = NULL;
    s = s0;
}

TTextFileDescriptor::~TTextFileDescriptor () {
    if (f) fclose (f);
}

void TTextFileDescriptor::Reset () {
    f = fopen (s, "rt");
}

void TTextFileDescriptor::Rewrite () {
    f = fopen (s, "wt");
}

void TTextFileDescriptor::OutputText (String s) {
    fputs (s, f);
}

void  TTextFileDescriptor::FillInputBuf () {
    fgets (InputBuf, INPUTBUFSIZE - 1, f);
//    int l = strlen (InputBuf);
//    if (l && InputBuf [l - 1] == '\n') InputBuf [l - 1] = '\0';
    InputPtr = InputBuf;
}

int TTextFileDescriptor::Eof () {
    return feof (f);
}

void TTextFileDescriptor::FFlush () {
    fflush (f);
}


void TConsoleFileDescriptor::OutputText (String s) {
    printf (s);
}

void TConsoleFileDescriptor::FillInputBuf () {
    fgets (InputBuf, INPUTBUFSIZE - 1, stdin);
//    int l = strlen (InputBuf);
//    if (l && InputBuf [l - 1] == '\n') InputBuf [l - 1] = '\0';
    InputPtr = InputBuf;
}

int TConsoleFileDescriptor::Eof () {
    return feof (stdin);
}



TFileList::TFileList (int N0) {
    typedef TFileDescriptor *TPFileDescriptor;
    N = N0;
    p = new TPFileDescriptor [N];
    isFree = new BOOL [N];
    for (int i = 0; i < N; i++)
        isFree [i] = TRUE;
}

TFileList::~TFileList () {
    CloseAll ();
    delete [] p;
    delete [] isFree;
}

int TFileList::FindFree () {
    for (int i = 2; i < N; i++)
	if (isFree [i]) {
	    isFree [i] = FALSE;
	    return i;
	}
    return -1;
}

int TFileList::OpenBin (String fn) {
    int f = FindFree ();
    if (f != -1)
	p [f] = new TBinFileDescriptor (fn);
    return f;
}

int TFileList::OpenText (String fn) {
    int f = FindFree ();
    if (f != -1) 
        p [f] = new TTextFileDescriptor (fn);
    return f;
}

TFileDescriptor *TFileList::GetFileDescriptor (int f) {
    return p [f];
}

void TFileList::Close (int f) {
    if (isFree [f]) return;
    delete p [f];
    p [f] = NULL;
    isFree [f] = TRUE;
}

void TFileList::CloseAll () {
    for (int i = 0; i < N; i++)
	Close (i);
}

void TFileList::SetStandardOutput (TFileDescriptor *fd) {
    p [1] = fd;
    isFree [1] = FALSE;
}

void TFileList::SetStandardInput (TFileDescriptor *fd) {
    p [0] = fd;
    isFree [0] = FALSE;
}

/*********************************************************/
/*			Heap				 */
/*********************************************************/

int TPCodeInterpreter::Allocate (int n) {
    n = n | 15;
    int f = freeptr, *g = &freeptr;
    BOOL Found = FALSE;
    while (f && !Found) 
	if (store [f - 1].s.l1 > n) {
	    *g += (n + 1);
	    store [*g - 1].s.l1 = store [f - 1].s.l1 - (n + 1);
	    store [*g].s.l1 = store [f].s.l1;
	    store [f - 1].s.l1 = n;
	    Found = TRUE;
	} else if (store [f - 1].intval == n) {
	    *g = store [f].s.l1;
	    Found = TRUE;
	} else {
	    g = &(store [f].s.l1);
	    f = store [f].s.l1;
        }
    if (!Found) {

        if (heapptr + n >= sp) Compactify ();
	if (heapptr + n >= sp) return 0;

        store [heapptr].s.l1 = n;
        heapptr += (n + 1);
	f = heapptr - n;
    }
    int i = 1;
    while (store [i].s.l2 != -1) i++;
    store [i].s.l1 = f;
    store [i].s.l2 = n;
    store [f - 1].s.l2 = i;	// store descriptor within block
    return i;
}

void TPCodeInterpreter::Deallocate (int i) {
    int p = store [i].s.l1;
    store [i].s.l2 = -1;
    int *g = &freeptr, prev = 0;
    while (*g && p > *g) {
    	prev = *g;
	g = &(store [*g].intval);
    }
    store [p].intval = *g;
    *g = p;

    // try to merge with previous block
    if (prev && prev + store [prev - 1].s.l1 + 1 == p) {
        store [prev] = store [p];
    	store [prev - 1].s.l1 += (1 + store [p - 1].s.l1);
	p = prev;
    }
    // try to merge with next block
    if (p + store [p - 1].s.l1 + 1 == store [p].s.l1) {
	int next = store [p].s.l1;
	store [p] = store [next];
	store [p - 1].s.l1 += (1 + store [next - 1].s.l1);
    }
    // check end of used heap
    if (p + store [p - 1].s.l1 == heapptr) {
	heapptr = p - 1;
        int *g = &freeptr;
        while (*g < heapptr)
            g = &(store [*g].intval);
        *g = 0;
    }
}

void TPCodeInterpreter::Compactify () {
    while (freeptr) {
	int dest = freeptr - 1,
            src = dest + store [freeptr - 1].s.l1 + 1,
	    stop = store [freeptr].s.l1 ? store [freeptr].s.l1 - 1 : heapptr;
	while (src != stop) {
	    int l = store [src].s.l1;
	    store [store [src].s.l2].s.l1 -= (src - dest);	// adjust descriptor pointer
	    for (int i = 0; i <= l; i++)
		store [dest++] = store [src++];
	}
	if (src == heapptr) {
	    freeptr = 0;
	    heapptr = dest;
	} else {
	    freeptr = dest + 1;
	    store [freeptr] = store [src + 1];
	    store [freeptr - 1].s.l1 = store [src].s.l1 + (src - dest);
	}
    }
}

/*********************************************************/
/*                   P-code interpreter                  */
/*********************************************************/

//inline double sign (double x) { return x < 0 ? -1 : (x > 0 ? 1 : 0); }
//inline double frac (double x) { return x - floor (x); }

void sort (TStorage *p, int r) {
    // r is last index (number of elements - 1)
    TStorage *i = p, *j = p + r;
    if (r <= 2) {	// compare three elements directly
	if (r == 2) {
	    if (i->doubleval > (i + 1)->doubleval)
		Swap (*i, *(i + 1));
	    if ((i + 1)->doubleval > j->doubleval)
		Swap (*(i + 1), *j);
	    j--;
	}
	if (i->doubleval > j->doubleval)
	    Swap (*i, *j);
	return;
    }
    double x = p [r / 2].doubleval;
    do {
         while (i->doubleval < x) i++;
         while (x < j->doubleval) j--;
         if (i <= j) {
                Swap (*i, *j);
                i++; j--;
         }
     } while (i <= j);
    if (j > p) sort (p, j - p);
    if (i < p + r) sort(i, p + r - i);
}

inline int odd (int n) {
    return n & 1;
}

TPCode::~TPCode () {
    if (code == LOADSTR) delete para.stringpara;
}

TPCodeInterpreter::TPCodeInterpreter () {
    freeptr = pc = 0;
    code = NULL;
    store = NULL;
    FileList = NULL;
    argc = 0;
}

TPCodeInterpreter::~TPCodeInterpreter () {
    DeleteMemory ();
    if (FileList) delete FileList;
}

BOOL TPCodeInterpreter::CreateMemory (int MemSize0, int PointerTableSize0, int CalcStackSize0, int CodeSize0) {
    BOOL f = CreateDataMemory (MemSize0, PointerTableSize0, CalcStackSize0);
    if (code) delete [] code;
    CodeSize = CodeSize0;
    code = new TPCode [CodeSize];
    return code != NULL && f;
}

BOOL TPCodeInterpreter::CreateDataMemory (int MemSize0, int PointerTableSize0, int CalcStackSize0) {
    PointerTableSize = PointerTableSize0;
    CalcStackSize = CalcStackSize0;    
    MemSize = MemSize0;
    if (store) delete [] store;
    store = new TStorage [MemSize];
    return store != NULL;
}

void TPCodeInterpreter::DeleteDataMemory () {
    if (store) delete [] store;
    store = NULL;
}

void TPCodeInterpreter::SetFileList (TFileList *l) {
    FileList = l;
}

void TPCodeInterpreter::SetProgramArguments (int argc0, char **argv0) {
    argc = argc0;
    argv = argv0;
}

void TPCodeInterpreter::DeleteMemory () {
    if (store) delete [] store;
    if (code) delete [] code;
    store = NULL;
    code = NULL;
}

void TPCodeInterpreter::AddListingLine (String s) {
    puts (s);
}

String TPCodeInterpreter::ListingLine (int i) {
    char *t [] = { "load", "loadt", "fload", "loadmem", "loadval", "loadptr",
                   "storemem", "storeval", "inc", "dec", "incptr", "copymem",
		   "loadstr",
		   "popind", "setind", "loadind", "vloadind", "storeind", "vstoreind", 
	           "loadls", "loop", "mvspind", "addind", 
		   "check", "isma", "rsp",
		   "alloc", "dispose", "avail",
		   "jump", "jz", "jnz", "call", "iterinit", "iterdo", "iterloop", "push",
		   "enter", "leave", "addcsp",
		   "dup",
		   "fadd", "fsub", "fmul", "fdiv", "fdivc", "fpow", "ffac", "fneg", "ineg",
		   "feq", "fne", "flt", "fgt", "fle", "fge",
		   "ieq", "ine", "ilt", "igt", "ile", "ige",
		   "seq", "sne", "slt", "sgt", "sle", "sge",
		   "cif", "cfi",
		   "iadd", "isub", "isub", "imul", "idiv", "idivc", "imod", "ipow",
		   "imodc", "and", "or", "xor", "not",
		   "cmpjeq", "cmprange",

                   "cmake", "cadd", "csub", "cmul", "cdiv", "cdivc", 

		   "setcreate", "setand", "setor", "setneg", "setinc", "setempty",

		   "vinit", "vfree", "vload", "vstore", "vpush", "vreal", "vint", "vcif",
		   "vcval", "vcvec", "vunit",
	           "vprintdbl", "vprintint", "vprintbool", 
		   "mprintdbl", "mprintint", "mprintbool", "mprintchar",
		   "vrev", "vcumsum", "vsum", "vfsrt", "vfrank", "vfmax", "vfmin", "vimax", "vimin",
		   "vfadd", "vfsub", "vfmul", "vfdiv", "vfdivc", "vfpow", 
		   "viadd", "visub", "vimul", "vidiv", "vidivc", "vimod", "vimodc", "vidivf", "vidivfc", "vipow",
	           "viand", "vior", "vixor", "vinot", "vineg", "vfneg",

	           "mfmul", "mimul",
		   "mtrans", "minv", "munit", "msqrt", "mreal",
		   "mcbind", "mrbind",

		   "vfeq", "vfne", "vflt", "vfgt", "vfle", "vfge",
		   "vieq", "vine", "vilt", "vigt", "vile", "vige",
		   "vmax2", "vmin2", "vimax2", "vimin2",
		   "vsize", "vssize", "vcheck", "vindint", "vindintc", "vindbool", 
	       	   "vsindint", "vsindintc", "vsindbool", "vlvindint",

		   "vsin", "varcsin", "vcos", "varccos", "vtan", "varctan",
		   "vcosh", "vsinh", "vtanh", "vlog", "vlogc",  "vexp", "vsqrt", "vsqrtc", "vsign", "vfrac",
		   "vround", "vtrunc", "vmax", "vmin", "visqr", "vfsqr", "vexist", "vforall",
		   "viabs", "vfabs", "viodd",

		   "isqr", "fsqr", "iabs", "imax", "imin", "iodd",
		   "fabs", "fsin", "fasin", "fcos", "facos", "ftan",
		   "fatan", "fcosh", "fsinh", "ftanh", "flog", "fexp",
		   "fsqrt", "fsign", "ffrac", "fint", "ftrunc",
		   "fmax", "fmin", "fchr", "frandom",

		   "cfs",
		   "printbool", "printchar", "printint", "printdbl", "printstr", "printc",
		   "printlf",
//		   "printbool1", "printchar1", "printint1", "printdbl1", "printstr1",
//		   "printdbl2",
		   "getchar", "getint", "getdbl", "getstr", "getlf", "scan",
		   "page", "txtcol", "txtbck", "gotoxy", "keyp",

		   "fassign", "ftassign", "freset", "frewrite", "fwrite", "fread", "fclose", "feof",
		   "fflush", 

		   "system", "sysdate", "systime", "argc", "argv",

		   "halt",

		   "cts", "cs", "ccs", "cms", "mcms", "msc", "msf",
		   "ldt", "lds", "ldd", "lda", "ldu", "cname", "cdata", "rdata",
		   "gp1hill", "gp1hilla", "gp1me", "gp1mea", "gpmle", "gpmlea", 
                   "gpmoment", "gpmomenta", "gpdrees", "gpdreesa",
                   "ev1mle", "ev1mlea", "evmle", "evmlea", "evmde", "evmdea", "evrde", "evrdea",
		   "gp0mle", "gp0mlea", "ev0mle", "ev0mlea",
		   "gp1hillcen", "gpmlecen",
                   "tmle", "tmlea", "nctmle", "nctmlea",
		   "gaussmle", "gaussmlea", "gausslse", "gausslsea", "gaussmde", "gaussmdea",
                   "gaussmhde", "gaussmdhea", 
		   "sdf", "sqf", "sden", "smcs", "sread", "plot", "boxplot", "plot3d", "cont3d", "clwin",
                   "mbox", "vbox", "cbox",
		   "exec", "play",

		   "fparetorv", "fgaussrv", "fnormalrv", "fcauchyrv", "ffrechetrv",
		   "fbetarv", "fweibullrv", "fgumbelrv", "fexprv",
		   "fgpdrv", "fevdrv", "fgammarv", "fchirv", "ftrv",
		   "vfparetorv", "vfgaussrv", "vfnormalrv", "vfcauchyrv", "vffrechetrv",
		   "vfbetarv", "vfweibullrv", "vfgumbelrv", "vfexprv",
		   "vfgpdrv", "vfevdrv", "vfgammarv", "vfchirv", "vftrv",
		   "fparetoden", "fparetoqf", "fparetodf",
		   "fgaussden", "fgaussqf", "fgaussdf",
		   "fgaussdenv", "fgaussqfv", "fgaussdfv",
		   "fcauchyden",
		   "ffrechetden", "ffrechetqf", "ffrechetdf",
		   "fbetaden", "fbetaqf", "fbetadf",
		   "fweibullden", "fweibullqf", "fweibulldf",
		   "fexpden", "fexpqf", "fexpdf",
		   "fgumbelden", "fgumbelqf", "fgumbeldf",
		   "fgpdden", "fgpdqf", "fgpddf",
		   "fevdden", "fevdqf", "fevddf",
		   "fchiden", "fchiqf", "fchidf",
		   "fgammaden", "fgammadf", 
		   "ftden", "ftdf", "ftqf",
		   "fnctden", "fnctdf", "fnctqf",
		   "fhrden", "fhrdf", "fhrsf", "fhrgpden", "fhrgpdf", "fhrgpsf",
		   "fgmfden", "fgmfdf", "fgmfsf", "fgmfgpden", "fgmfgpdf", "fgmfgpsf",
		   "fgamma", "fruinsim", "fruinsimb",

                   "funiform",

		   "tst", "tswelch", "tswilcoxon", 

		   "fmean", "fvar", "fmon", "fmedian", "fsmooth",
                   "reg", "sca", 
		   "setcol", "setlin", "setsty", "setopt", "seps", 
                   "setcs", "setct", "setcm", "setcf", "setlf", "setlb",

                   "sgv", "ggv", "sgs", "ggs",

		   "lex", "ufoeval", "ufodef", "ufoerr" };
    TPCode &c = code [i];
    String s = form (10, "%3d:   ", i) & form (20, "%-12s", t [(int) c.code]);
    switch (c.code) {
	case FLOAD:
	    s = s & form (30, "%g", c.para.doublepara);
	    break;
        case LOADMEM:
	case STOREMEM:
	case PUSH:
	case JUMP:
	case JZ:
	case JNZ:
	case ADDCSP:
	case DUP:
	case LOAD:
	case RSP:
	case ISUBD:
	case LOOP:
	case STOREIND:
	case LOADIND:
	case ADDIND:
	case INCPTR:
	case VCVAL:
	    s = s & form (30, "%ld", c.para.intp.intpara);
	    break;
	case LOADSTR:
	    s = s & *c.para.stringpara;
	    break;
	case ENTER:
	case LEAVE:
	case LOADT:
	case INC:
	case DEC:
	case LOADPTR:
	case LOADVAL:
	case STOREVAL:
	case CHECK:
	case ISMA:
	case GOTOXY:
	case CMPJEQ:
	case CMPRANGE:
	case SETIND:
	    s = s & form (40, "%ld, %ld", c.para.intp.intpara, c.para.intp.intpara2);
	    break;
    }
    return s;
}

void TPCodeInterpreter::Listing () {
    for (int i = 0; i < ProgramSize; i++) {
	AddListingLine (ListingLine (i));
    }
}

void TPCodeInterpreter::OutCode (TCode c) {
    code [pc++].code = c;
}

void TPCodeInterpreter::OutCode (TCode c, int p, int p2) {
    code [pc].code = c;
    code [pc].para.intp.intpara = p;
    code [pc++].para.intp.intpara2 = p2;
}

void TPCodeInterpreter::OutCode (TCode c, double p) {
    code [pc].code = c;
    code [pc++].para.doublepara = p;
}

void TPCodeInterpreter::OutCode (TCode c, String s) {
    code [pc].code = c;
    code [pc++].para.stringpara = new String (s);
}

int TPCodeInterpreter::GetPos () {
    return pc;
}

void TPCodeInterpreter::Fixup (int FixupPos, TCode c, int d) {
    code [FixupPos].code = c;
    code [FixupPos].para.intp.intpara = d;
}

TRunTimeErrorType TPCodeInterpreter::ExtendedCode (TPCode &c) {
    return RT_OPCODE;
}

void TPCodeInterpreter::Init () {
    pc = 0;
    freeptr = 0;
    heapptr = PointerTableSize + CalcStackSize;
    sp = MemSize - 1;
    loopsp = -1;
    csp = PointerTableSize;
    for (int i = 0; i < PointerTableSize; i++)
	store [i].s.l2 = -1;
}

void TPCodeInterpreter::Save (FILE *f) {
    int a = ProgramSize;
    fwrite (&a, sizeof (a), 1, f);
    for (int i = 0; i < ProgramSize; i++) {
	fwrite (&code [i], sizeof (code [i]), 1, f);
	if (code [i].code == LOADSTR) StringToFile (f, *(code [i].para.stringpara));
    }
}

void TPCodeInterpreter::Load (FILE *f) {
    int a;
    fread (&a, sizeof (a), 1, f);
    ProgramSize = (int) a;
    CreateMemory (1, 1, 1, ProgramSize);		// ???? Verarbeitet Xtremes Datengroesse korrekt?
    for (int i = 0; i < ProgramSize; i++) {
	fread (&code [i], sizeof (code [i]), 1, f);
	if (code [i].code == LOADSTR)
	    code [i].para.stringpara = new String (StringFromFile (f));
    }
}

/**********************************************************
		I/O Routines
**********************************************************/

void TPCodeInterpreter::WriteString (int fd, String s) {
    FileList->GetFileDescriptor (fd)->OutputText (s);
}

void TPCodeInterpreter::WriteChar (int fd, int c) {
    char s [2];
    s [0] = (char) c; s [1] = '\0';
    FileList->GetFileDescriptor (fd)->OutputText (s);
}

void TPCodeInterpreter::WriteLinefeed (int fd) {
    WriteChar (fd, '\n');
}

/**********************************************************
			Interpreter

Layout of vectors/matrices on calculator stack is:

  | -----------------------|
  |        		   |
  |        Data            |
  |			   |
  |------------------------|
  |  nRows   |   nCols     |
  |------------------------|
  |  Length  | Member Size |
  |------------------------|
  |                        |   <- csp

For vector types: nCols = 1

**********************************************************/

void TPCodeInterpreter::PushStringAsVector (String s) {
    int k = s.Len ();
    const char *t = s;
    for (int i = 0; i < k; i++)
	IPush (t [i]);
    IPush (k, 1);
    IPush (k + 1, 1);
}

String TPCodeInterpreter::PopStringFromVector () {
    int k, dummy;
    IPop ();
    IPop (k, dummy);
    char * s = new char [k + 1];
    s [k] = '\0';
    for (int i = k - 1; i >= 0; i--)
	s [i] = (char) IPop ();
    return String (s, FALSE);
}

void TPCodeInterpreter::PopDoubleVector (double * &a, int &n) {
    n = VectorLength () - 1;
    a = (double *) VectorStart ();
    ReleaseCalcStack (n + 2);
}

void TPCodeInterpreter::PopDoubleMatrix (double * &a, int &n, int &m) {
    n = VectorRows ();
    m = VectorCols ();
    int k = VectorLength ();
    a = (double *) VectorStart ();
    ReleaseCalcStack (k + 1);
}

void TPCodeInterpreter::LoadVectorOnCalculatorStack (TStorage VectorData) {
    // loads a vector on the calculator stack
    // VectorData contains a the address of an entry
    // in the pointer table and the length of the vector
    // (i.e., the usual representation of a vector variable) 

    if (!VectorData.s.l1) {	// load the empty vector
	puts ("Loading empty vector");
	IPush (0, 1);
	IPush (1, 0);		// we do not know member type size, set to zero
    } else {
    	TStorage *ptr1 = store + store [VectorData.s.l1].s.l1;	// Pointer to  data
    	int l = VectorData.s.l2, k = l;        // Length of data
    	TStorage *ptr2 = CalcStackFirstFree ();
    	ReserveCalcStack (l + 1);
    	memcpy (ptr2, ptr1, k * sizeof (TStorage));
    	ptr2 += k;
    	ptr2->s.l1 = l;
	if (l == 1)	// empty vector?
	    ptr2->s.l2 = 0;
	else
    	    ptr2->s.l2 = (l - 1) / ((ptr2 - 1)->s.l1 * (ptr2 - 1)->s.l2);
    }
}

void TPCodeInterpreter::CPush (dcomplex a) {
    store [csp++].doubleval = std::real (a); 
    store [csp++].doubleval = std::imag (a); 
}

dcomplex TPCodeInterpreter::CPop () {
    double im = DPop (), re = DPop ();
    return dcomplex (re, im);
}

TRunTimeErrorType TPCodeInterpreter::Execute () {
    double a, b, c;
    int i, k, l, m;
    String s1, s2;
    TRunTimeErrorType ext;
    TStorage *ptr, *ptr1, *ptr2, *ptr3;

    #if !defined TERMINAL && defined _WINDOWS
	int statementcount = 0;
	TCancelBox *CancelBox = new TCancelBox (PlotWindowManager.GetFramePtr (), IDD_SPBreakBox);
	CancelBox->OpenDialog (PlotWindowManager.GetFramePtr ());
//	CancelBox->DisableParent ();
    #endif

    for (;;) {

//	String dummy = ListingLine (pc);
//	printf ("%s\n", (const char *) dummy);

	switch (code [pc].code) {
	    case LOAD:
		IPush (code [pc].para.intp.intpara);
		break;
	    case LOADT:
		IPush (code [pc].para.intp.intpara, code [pc].para.intp.intpara2);
		break;
	    case FLOAD:
                DPush (code [pc].para.doublepara);
                break;
            case LOADMEM:
		k = (int) store [csp - 1].intval;
	        l = (int) code [pc].para.intp.intpara;
	        m = csp - 1; 
	   	csp += (l - 1);
		while (l--) store [m++] = store [k++];
                break;
            case LOADVAL:
		Push (store [(int) code [pc].para.intp.intpara + DisplayBP [(int) code [pc].para.intp.intpara2]]);
		break;
	    case LOADPTR:
		IPush ((int) code [pc].para.intp.intpara + DisplayBP [(int) code [pc].para.intp.intpara2]);
		break;
            case STOREMEM:
		l = (int) code [pc].para.intp.intpara;
		k = (int) store [csp - l - 1].intval;
		m = csp - l;
		csp -= (l + 1);
		while (l--) store [k++] = store [m++];
                break;
            case STOREVAL:
		store [(int) code [pc].para.intp.intpara + DisplayBP [(int) code [pc].para.intp.intpara2]] = Pop ();
		break;
	    case INC:
		store [(int) code [pc].para.intp.intpara + DisplayBP [(int) code [pc].para.intp.intpara2]].intval++;
                break;
            case DEC:
		store [(int) code [pc].para.intp.intpara + DisplayBP [(int) code [pc].para.intp.intpara2]].intval--;
                break;
	    case INCPTR:
		store [IPop ()].intval += code [pc].para.intp.intpara;
		break;
	    case COPYMEM:
	        k = IPop (); l = IPop (); m = IPop ();
		memcpy (store + (int) m, store + (int) l, (int) k * sizeof (TStorage));
		break;
	    case LOADSTR:
                PushStringAsVector (*code [pc].para.stringpara);
                break;
	
	    case POPIND:
		indexptr = IPop ();
		break;
	    case SETIND:
		indexptr = code [pc].para.intp.intpara + DisplayBP [(int) code [pc].para.intp.intpara2];
		break;
	    case LOADIND:
		l = code [pc].para.intp.intpara;
		for (i = 0; i < l; i++)
		    store [csp++] = store [indexptr++];
		break;
	    case VLOADIND:
	        LoadVectorOnCalculatorStack (store [indexptr]);
		indexptr++;
		break;

	    case STOREIND:
		l = code [pc].para.intp.intpara;
		for (i = 0; i < l; i++)
		    store [csp - l + i] = store [indexptr--];
		csp -= l;
		break;
	    case VSTOREIND:
		l = VectorLength ();
		ptr = VectorStart ();
                k = indexptr--;         	// addr
                if (store [k].s.l2 != l) {
                    if (store [k].s.l1) Deallocate (store [k].s.l1);
                    if ((store [k].s.l1 = Allocate (l)) == 0) return RT_HEAP_FULL;
                    store [k].s.l2 = l;
                }
                m = store [k].s.l1;
		memcpy (store + store [m].s.l1, ptr, l * sizeof (TStorage));
                ReleaseCalcStack (l + 1);
                break;
	    case LOADLS:
		loopstack [++loopsp] = IPop ();
		break;
	    case LOOP:
		if (--loopstack [loopsp] > 0) 
		    pc = code [pc].para.intp.intpara - 1;
		else
		    loopsp--;
		break;
	    case MVSPIND:
		indexptr = sp + 1;
		break;
	    case ADDIND:
		indexptr += code [pc].para.intp.intpara;
		break;


	    case CHECK:
		if (store [csp - 1].intval < code [pc].para.intp.intpara ||
		    store [csp - 1].intval > code [pc].para.intp.intpara2)
                    return RT_RANGECHECK;
                break;
	    case ISMA:
		k = IPop ();
		store [csp - 1].intval += (k - code [pc].para.intp.intpara) * code [pc].para.intp.intpara2;
		break;
	    case ALLOC:
		k = IPop ();
		l = Allocate (k);
		if (l == 0) return RT_HEAP_FULL;
		IPush (l);
		break;
  	    case DISPOSE:
		Deallocate (IPop ());
		break;
	    case AVAIL:
		Compactify ();
		IPush (sp - heapptr);
                break;
	    case RSP:
		k = code [pc].para.intp.intpara;
		csp = (int) store [DisplayBP [(int) k] - 3].intval;
		break;
	    case ENTER:
		k = code [pc].para.intp.intpara2;
		store [sp--].intval = DisplayBP [(int) k];
		sp--;
		store [sp--].intval = csp;
		store [sp--].intval = 0; // scsp;
		store [sp--].intval = code [pc].para.intp.intpara2;
		DisplayBP [(int) k] = sp + 6;
		sp -= (int) code [pc].para.intp.intpara;
		if (sp <= csp) return RT_STACKOVERFLOW;
		break;
	    case LEAVE: {
		k = code [pc].para.intp.intpara2;
		sp = DisplayBP [(int) k];
		int h = pc;
		pc = (int) store [sp].intval;
		sp += (int) code [h].para.intp.intpara;
		DisplayBP [(int) k] = (int) store [DisplayBP [(int) k] - 1].intval;
		break; }
	    case ADDCSP:
		csp += (int) code [pc].para.intp.intpara;
                break;
            case DUP:
		store [csp] = store [csp - 1 - (int) code [pc].para.intp.intpara];
		csp++;
                break;
            case JNZ:
		if (IPop ()) pc = (int) code [pc].para.intp.intpara - 1;
                break;
            case JZ:
                if (IPop ()) break;
                // fall through
            case JUMP:
		pc = (int) code [pc].para.intp.intpara - 1;
		break;
            case CALL:
		store [sp--].intval = pc;
		pc = IPop () - 1;
                break;
	
            /* 
		Stack layout for implicit loops:

                |                      |
                | vector data	       |
                |                      |
		| rows  | cols         |
		| len   | member size  |
                | address of function  |
                | counter for length   |
                |                      |  <- csp
	    */ 

	    case ITERINIT:
		l = store [csp - 3].intval;	// number of rows in vector on stack
		IPush (l);			// create counter 
		break;
	    case ITERDO:
		if ((l = store [csp - 1].intval) > 0) {
		    store [sp--] = store [csp - 4 - l];
		    store [sp--].intval = pc;			// simulate function call
		    pc = store [csp - 2].intval - 1;
		} else csp++;
		break;
	    case ITERLOOP:
		if ((l = store [csp - 2].intval) > 0) {		// remove result of call from stack and move to vector
		    store [csp - 5 - l] = store [csp - 1];
		    csp--;
		    store [csp - 1].intval--;
		    pc -= 2;					// repeat loop
		} else 
		    csp -= 3;
		break;


            case PUSH:
		l = code [pc].para.intp.intpara;
		while (l--) store [sp--] = Pop ();
                break;
	    case IADD:
		store [csp - 2].intval += store [csp - 1].intval;
		csp--;
		break;
	    case ISUB:
		store [csp - 2].intval -= store [csp - 1].intval;
		csp--;
		break;
	    case ISUBD:
		store [csp - 1].intval -= code [pc].para.intp.intpara;
		break;
            case IMUL:
		store [csp - 2].intval *= store [csp - 1].intval;
                csp--;
		break;
	    case IDIVC:
		if (!store [csp - 1].intval) return RT_DIVZERO;
                // fall-through
            case IDIV:
                store [csp - 2].intval /= store [csp - 1].intval;
                csp--;
                break;
            case IAND:
                store [csp - 2].intval &= store [csp - 1].intval;
                csp--;
		break;
	    case IOR:
		store [csp - 2].intval |= store [csp - 1].intval;
                csp--;
                break;
            case IXOR:
                k = IPop (); l = IPop (); IPush (l ^ k);
		break;
	    case INOT:
		IPush (!IPop ());
                break;
	    case IMODC:
                if (!store [csp - 1].intval) return RT_DIVZERO;
		// fall-through
	    case IMOD:
		store [csp - 2].intval %= store [csp - 1].intval;
                csp--;
		break;
	    case IPOW:
		b = IPop (); a = IPop ();
		if (a == 0 && b < 0) return RT_ARGUMENT;
		c = pow (a, b);
		if (fabs (c) > INT_MAX) return RT_OVERLOW;
		IPush ((int) c);
		break;
	    case CMPJEQ:
		if (store [csp - 1].intval == code [pc].para.intp.intpara)
		    pc = code [pc].para.intp.intpara2 - 1;
		break;
	    case CMPRANGE:
		if (store [csp - 1].intval >= code [pc].para.intp.intpara && 
		    store [csp - 1].intval <= code [pc].para.intp.intpara2)
		    IPush (1);
		else IPush (0);
		break;

            case CMAKE:
                // nothing to do; just leave stack unchanged
                break;
            case CADD:
                CPush (CPop () + CPop ());
                break;
            case CSUB: {
                dcomplex a = CPop (), b = CPop ();
                CPush (b - a); }
                break;
            case CMUL: 
                CPush (CPop () * CPop ());
                break;
            case CDIVC: {
                dcomplex a = CPop (), b = CPop ();
                if (a != 0.0) CPush (b / a);
                else return RT_DIVZERO; }
                break;
            case CDIV: {
                dcomplex a = CPop (), b = CPop ();
                CPush (b / a); }
                break;
 


                

// set operations
// It is assumed that no space is reserved between consecutive store members

#define N_SET_BYTES  (int)((sizeof (double) * N_SET_CELLS))

	    case SETCREATE: 
		memset (store + csp, 0, N_SET_BYTES);
		csp += N_SET_CELLS;
		break;
	    case SETAND: {
		TSetCast p1 = (TSetCast) (store + csp - N_SET_CELLS),
			 p2 = (TSetCast) (store + csp - 2 * N_SET_CELLS);
		for (i = 0; i < N_SET_BYTES; i++)
		    p2 [i] &= p1 [i];
		csp -= N_SET_CELLS; }
		break;
	    case SETOR: {
		TSetCast p1 = (TSetCast) (store + csp - N_SET_CELLS),
			 p2 = (TSetCast) (store + csp - 2 * N_SET_CELLS);
		for (i = 0; i < N_SET_BYTES; i++)
		    p2 [i] |= p1 [i];
		csp -= N_SET_CELLS; }
		break;
	    case SETNEG: {
		TSetCast p1 = (TSetCast) (store + csp - N_SET_CELLS);
		for (i = 0; i < N_SET_BYTES; i++)
		    p1 [i] = (char) ~p1 [i]; }
		break;
	    case SETINC: {
		k = IPop ();
		TSetCast p1 = (TSetCast) (store + csp - N_SET_CELLS);
		p1 [k / 8] |= (char) (1 << (k % 8)); }
		break;
	    case SETEMPTY: {
		csp -= N_SET_CELLS;
		TSetCast p1 = (TSetCast) (store + csp);
		BOOL Found = FALSE;
		for (i = 0; i < N_SET_BYTES && !Found; i++)
		    Found = (p1 [i] != 0);
		IPush (Found); }
		break;
		
// Vector operations

	    case VINIT:
		store [indexptr].s.l1 = store [indexptr].s.l2 = 0;
		indexptr++;
		break;
	    case VFREE:
		if (store [indexptr].s.l1) Deallocate (store [indexptr].s.l1);
		indexptr++;
		break;
	    case VLOAD:
		ptr = CalcStackTop ();
		ReleaseCalcStack (1);
		LoadVectorOnCalculatorStack (*ptr);
		break;
	    case VSTORE: 
		l = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (l + 1);
		k = IPop ();
		if (store [k].s.l2 != l) {
		    if (store [k].s.l1) Deallocate (store [k].s.l1);
		    if ((store [k].s.l1 = Allocate (l)) == 0) return RT_HEAP_FULL;
		    store [k].s.l2 = l;
		}
		m = store [k].s.l1;
		ptr1 = store + store [m].s.l1;
		memcpy (ptr1, ptr, l * sizeof (TStorage));
		break;
	    case VPUSH: 
		k = VectorLength ();
		ptr = VectorStart ();
		if ((l = store [sp].s.l1 = Allocate (k)) == 0) return RT_HEAP_FULL;
		store [sp].s.l2 = k;
		sp--;
		l = store [l].s.l1;
		memcpy (store + l, ptr, k * sizeof (TStorage));
		ReleaseCalcStack (k + 1);
		break;
	    case VREAL: 
		k = IPop ();
		b = DPop ();
		a = DPop ();
		if (k < 1) return RT_ARGUMENT;
	 	if (k == 1) DPush (a);
		else
		    for (i = 0; i < k; i++)
		    	DPush (a + i * (b - a) / (k - 1));
		IPush (k, 1);
		IPush (k + 1, 1);
		break;
	    case VINT: 
		l = IPop ();
		k = IPop ();
		if (k > l) {
		    IPush (0, 1);
		    IPush (1, 1);
		} else {
//		    ptr = CalcStackFirstFree ();
//		    while (m <= l)
//			(ptr++)->intval = m++;
//		    ReserveCalcStack (l - k + 1);
//
//		    Loop with IPush is faster?!
//
		    for (m = k; m <= l; m++) IPush (m);
		    IPush (l - k + 1, 1);
		    IPush (l - k + 2, 1);
		}
		break;
	    case VCIF:
		k = VectorLength ();
		ptr = VectorStart ();
		while (--k) {
		    ptr->doubleval = ptr->intval;
		    ptr++;
		}
		break;
	    case VCVAL: {
		k = code [pc].para.intp.intpara;
		TStorage p1 = *(CalcStackTop () - 1 - k),
                         p2 = *(CalcStackTop () - k );
		for (int i = 0; i < k; i++)
		    *(CalcStackTop () - k - 1 + i) = *(CalcStackTop () - k + 1 + i);
		p1.s.l1++;	// inc nRows
		p2.s.l1 += k;	// inc Length
		*(CalcStackTop () - 1) = p1;
		*CalcStackTop () = p2; }
		break;
	    case VCVEC: 
		k = VectorLength ();
		(CalcStackTop () - 1)->s.l1 += (CalcStackTop () - 2 - k)->s.l1;
		CalcStackTop ()->s.l1 += (CalcStackTop () - 1 - k)->s.l1 - 1;
		memmove (CalcStackTop () - 2 - k, CalcStackTop () - k, (k + 1) * sizeof (TStorage));
		ReleaseCalcStack (2);
		break;
	    case VUNIT:
		l = IPop ();
		k = IPop ();
		for (i = 1; i <= l; i++)
		    DPush (i == k);
		IPush (l, 1);
		IPush (l + 1, 1);
		break;
		

	    case MPRINTDBL: {
		int col = IPop (),
		    dig = IPop (),
		    len = IPop (),
	    	    dummy, n;
		ptr = VectorStart ();
		IPop (k, dummy);
		IPop (n, m);
		ReleaseCalcStack (k - 1);
		l = IPop ();
		char buf [30];
		for (i = 0; i < n; i++) {
		    for (int j = 0; j < m; j++) {
    	                if (dig != -1) sprintf (buf, "%*.*f", len, dig, ptr [i * m + j].doubleval);
    		        else if (len != -1) sprintf (buf, "%*g", len, ptr [i * m + j].doubleval);
    		        else sprintf (buf, "%g ", ptr [i * m + j].doubleval);
    		        WriteString (l, buf); 
    		        if (col > 0 && (j + 1) % col == 0 && j != m - 1) WriteLinefeed (l);
		    }
		    WriteLinefeed (l);
		}}
		break;
	    case MPRINTINT: {
		IPop ();
		int col = IPop (),
		    len = IPop (),
	    	    dummy, n;
		ptr = VectorStart ();
		IPop (k, dummy);
		IPop (n, m);
		ReleaseCalcStack (k - 1);
		l = IPop ();
		char buf [30];
		for (i = 0; i < n; i++) {
		    for (int j = 0; j < m; j++) {
    	                if (len != -1) sprintf (buf, "%*d", len,  ptr [i * m + j].intval);
    		        else sprintf (buf, "%d ", ptr [i * m + j].intval);
    		        WriteString (l, buf); 
    		        if (col > 0 && (j + 1) % col == 0 && j != m - 1) WriteLinefeed (l);
		    }
		    WriteLinefeed (l);
		}}
		break;
	    case VPRINTDBL: {
		int col = IPop (),
		    dig = IPop (),
		    len = IPop ();
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
		l = IPop ();
		char *buf = new char [30 + len];
		for (i = 0; i < k - 1; i++) {
	            if (dig != -1) sprintf (buf, "%*.*f", len, dig, ptr->doubleval);
		    else if (len != -1) sprintf (buf, "%*g", len, ptr->doubleval);
		    else sprintf (buf, "%g ", ptr->doubleval);
		    ptr++;
		    WriteString (l, buf); 
		    if (col > 0 && (i + 1) % col == 0 && i != k - 1) WriteLinefeed (l);
		}
		delete [] buf; }
		break;

	    case VPRINTINT: {
                IPop ();	// throw away 
                int dig = IPop (),
		    len = IPop ();
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
		l = IPop ();
		char *buf = new char [30 + len];
		for (i = 0; i < k - 1; i++) {
		    if (len == -1) sprintf (buf, "%d ", ptr->intval);
		    else sprintf (buf, "%*d", len, ptr->intval);
		    ptr++;
		    WriteString (l, buf);
		    if (dig > 0 && (i + 1) % dig == 0 && i != k - 1) WriteLinefeed (l);
		}
		delete [] buf; }
		break;

	    case VPRINTBOOL: {
                IPop ();	// throw away columns
                int dig = IPop (),
		    len = IPop ();
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
		l = IPop ();

		for (i = 0; i < k - 1; i++) {
		    if (len != -1)
		        for (int j = 4 + !ptr->intval; j < len; j++)
		            WriteChar (l, (int) ' ');
		    WriteString ((int) l, form (10, "%s ", ptr->intval ? "true" : "false"));
		    if (dig > 0 && (i + 1) % dig == 0 && i != k - 1) WriteLinefeed (l);
		    ptr++;
		}}
		break;

	    case VREV: 
		k = VectorMemberSize ();
		l = VectorLength ();;
		ptr1 = VectorStart ();
		ptr2 = ptr1 + l - k - 1;
		if (l == 1) 
		    while (ptr1 < ptr2)
			Swap (*ptr1++, *ptr2--);
		else
		    while (ptr1 < ptr2) {
		        for (i = 0; i < k; i++) {
			    Swap (*ptr1++, *ptr2++);
		        }
		        ptr2 -= (k + k);
		    }
		break;
	    case VCUMSUM: 
		k = VectorLength ();
		ptr = VectorStart ();
		for (i = 1; i < k; i++) {
		    ptr++;
		    ptr->doubleval += (ptr - 1)->doubleval;
		}
		break;
	    case VSUM: 
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
		a = 0.0;
		while (--k) a += (ptr++)->doubleval;
		DPush (a);
		break;
	    case VFSRT:
		k = VectorLength () - 1;
		if (!k) break;
		sort (VectorStart (), k - 1);
		break;
	    case VFRANK: {
		k = VectorLength () - 1;
		ptr = VectorStart ();
		if (!k) break;
		int *l = new int [k];
		double *x = new double [k];
		for (i = 0; i < k; i++) {
		    l [i] = i + 1;
		    x [i] = (ptr++)->doubleval;
		}
		sort (0, k - 1, x, l);
		ptr = VectorStart ();
		for (i = 0; i < k; i++) 
		    (ptr++)->intval = l [i];
		delete [] x;
		delete [] l; }
		break;


	    case VINOT:
		k = VectorLength ();
		ptr = VectorStart ();
		while (--k) {
		    ptr->intval = !ptr->intval;
		    ptr++;
		}
		break;
	    case VINEG:
		k = VectorLength ();
		ptr = VectorStart ();
		while (--k) {
		    ptr->intval = -ptr->intval;
		    ptr++;
		}
		break;
	    case VFNEG:
		k = VectorLength ();
		ptr = VectorStart ();
		while (--k) {
		    ptr->doubleval = -ptr->doubleval;
		    ptr++;
		}
		break;



	    case VFADD: case VFSUB: case VFMUL: case VFDIV: case VFDIVC: case VFPOW: 
	    case VIADD: case VISUB: case VIMUL: case VIDIV: case VIDIVC: case VIMOD: 
	    case VIMODC: case VIDIVF: case VIDIVFC: case VIPOW:
	    case VIAND: case VIOR: case VIXOR: case VIMAX2: case VIMIN2:
	    case VFEQ: case VFNE: case VFLT: case VFGT: case VFLE: case VFGE:
	    case VIEQ: case VINE: case VILT: case VIGT: case VILE: case VIGE:
	    case VMAX2: case VMIN2:
		{

		TStorage *p1 = VectorStart ();
		int l1 = VectorLength ();
		TStorage *p2 = p1 - 1, *dest, *src;
		int l2 = p2->s.l1;
		p2 -= l2;
		l1--; l2--;

		if (l1 > l2) {
		    dest = p1; src = p2;
		    m = l1; l = l2;
		} else {
		    dest = p2; src = p1;
		    m = l2; l = l1;
		}
		if (l1 == 0 || l2 == 0) return RT_VECTOREMPTY;


		#define VECOP(name,op,resfield,srcfield,cast,check1,check2) \
		    case name: \
			if (dest == p2) \
			    for (i = 0; i < m; i++) { \
				check1; \
				dest->resfield = cast dest->srcfield op cast src [i % l].srcfield; \
				dest++; \
			    } \
			else  \
			    for (i = 0; i < m; i++) { \
				check2; \
				dest->resfield = cast src [i % l].srcfield op cast dest->srcfield; \
				dest++; \
			    } \
			break;


		switch (code [pc].code) {
		    VECOP (VFADD, +, doubleval, doubleval,,,)
		    VECOP (VFSUB, -, doubleval, doubleval,,,)
		    VECOP (VFMUL, *, doubleval, doubleval,,,)
		    VECOP (VFDIV, /, doubleval, doubleval,,,)
		    VECOP (VFDIVC, /, doubleval, doubleval, (double), if (!src [i % l].doubleval) return RT_DIVZERO, if (!dest->doubleval) return RT_DIVZERO)

		    VECOP (VIAND, &, intval, intval,,,)
		    VECOP (VIOR,  |, intval, intval,,,)
		    VECOP (VIXOR, ^, intval, intval,,,)
		    VECOP (VIADD, +, intval, intval,,,)
		    VECOP (VISUB, -, intval, intval,,,)
		    VECOP (VIMUL, *, intval, intval,,,)
		    VECOP (VIDIV, /, intval, intval,,,)
 		    VECOP (VIMOD, %, intval, intval,,,)
		    VECOP (VIDIVF, /, doubleval, intval, (double),,)
 
		    VECOP (VIDIVC, /, intval, intval, (int), if (!src [i % l].intval) return RT_DIVZERO, if (!dest->intval) return RT_DIVZERO)
		    VECOP (VIMODC, /, intval, intval, (int) , if (!src [i % l].intval) return RT_DIVZERO, if (!dest->intval) return RT_DIVZERO)
		    VECOP (VIDIVFC, /, doubleval, intval, (int) , if (!src [i % l].intval) return RT_DIVZERO, if (!dest->intval) return RT_DIVZERO)

		    VECOP (VFEQ, ==, intval, doubleval,,,)	
		    VECOP (VFNE, !=, intval, doubleval,,,)	
		    VECOP (VFLT,  <, intval, doubleval,,,)	
		    VECOP (VFGT,  >, intval, doubleval,,,)	
		    VECOP (VFLE, <=, intval, doubleval,,,)	
		    VECOP (VFGE, >=, intval, doubleval,,,)	

		    VECOP (VIEQ, ==, intval, intval,,,)
		    VECOP (VINE, !=, intval, intval,,,)
		    VECOP (VILT,  <, intval, intval,,,)
		    VECOP (VIGT,  >, intval, intval,,,)
		    VECOP (VILE, <=, intval, intval,,,)
		    VECOP (VIGE, >=, intval, intval,,,)

		    case VFPOW:
			for (i = 0; i < m; i++) {
			    a = dest->doubleval; b = src [i % l].doubleval;
			    if (dest == p1) Swap (a, b);
                            if (a < 0.0 && frac (b) != 0.0) return RT_ARGUMENT;
		            if (a == 0.0 && b < 0.0) return RT_ARGUMENT;
			    (dest++)->doubleval = pow (a, b);
			}
			break;
		    case VIPOW:
			for (i = 0; i < m; i++) {
			    a = dest->intval; b = src [i % l].intval;
			    if (dest == p1) Swap (a, b);
		            if (a == 0.0 && b < 0.0) return RT_ARGUMENT;
			    c = pow (a, b);
			    if (fabs (c) > INT_MAX) return RT_OVERLOW;
			    (dest++)->intval = (int) c;
			}
			break;
                    case VMAX2:
                        for (i = 0; i < m; i++) {
                            if (dest->doubleval < src [i % l].doubleval)
				dest->doubleval = src [i % l].doubleval;
			    dest++;
			}
		  	break;
                    case VMIN2:
                        for (i = 0; i < m; i++) {
                            if (dest->doubleval > src [i % l].doubleval)
				dest->doubleval = src [i % l].doubleval;
			    dest++;
			}
		  	break;
                    case VIMAX2:
                        for (i = 0; i < m; i++) {
                            if (dest->intval < src [i % l].intval)
				dest->intval = src [i % l].intval;
			    dest++;
			}
		  	break;
                    case VIMIN2:
                        for (i = 0; i < m; i++) {
                            if (dest->intval > src [i % l].intval)
				dest->intval = src [i % l].intval;
			    dest++;
			}
		  	break;

		}

		if (m != l2)  
		    memmove (p2, p1, (m + 2) * sizeof (TStorage));
		ReleaseCalcStack (l + 2);
		}
		break;


	    case MFMUL: {
		int n;
		double *ptr1, *ptr2;
		PopDoubleMatrix (ptr2, m, n);
		PopDoubleMatrix (ptr1, k, l);
		if (l != m) return RT_MATRIXOP;		
		double *res = new double [k * n];
		for (i = 0; i < k; i++)
		    for (int j = 0; j < n; j++) {
			res [i * n + j] = 0.0;
			for (int h = 0; h < l; h++)
			    res [i * n + j] += ptr1 [i * l + h] * ptr2 [h * n + j];
		    }
		for (i = 0; i < k * n; i++) DPush (res [i]);
	   	IPush (k, n);
		IPush (k * n + 1, 1);
		delete [] res; }
		break;

	    case MUNIT:
		k = IPop ();
		for (i = 0; i < k; i++)
		    for (l = 0; l < k; l++)
			DPush (i == l);
		IPush (k, k);
		IPush (k * k + 1, 1);
		break;		

	    case MTRANS: {
		m = VectorRows ();
		int n = VectorCols ();
		l = VectorMemberSize ();
		ptr = VectorStart ();
		ptr1 = ptr2 = new TStorage [n * m];
		for (int j  = 0; j < n; j++)
		    for (i = 0; i < m; i++)
			*ptr1++ = ptr [i * n + j];
		memcpy (ptr, ptr2, m * n * l * sizeof (TStorage));
		delete [] ptr2;
		ReleaseCalcStack (2);
		IPush (n, m);
		IPush (m * n * l + 1, l); }
		break; 
	    case MINV: {
		double *p;
		PopDoubleMatrix (p, l, m);
		TMatrix a (l, m, p),
		        b = inv (a);
		for (int i = 1; i <= l; i++)
		    for (int j = 1; j <= m; j++) 
			DPush (b (i, j));
		IPush (l, m);
		IPush (l * m + 1, 1);
		}
		break;
	    case MSQRT: {
		double *p;
		PopDoubleMatrix (p, l, m);
		TMatrix a (l, m, p),
		        b = sqrt (a);
		for (int i = 1; i <= l; i++)
		    for (int j = 1; j <= m; j++) 
			DPush (b (i, j));
		IPush (l, m);
		IPush (l * m + 1, 1);
		}
		break;
	    case MREAL: {
		// does not work for elements with size > 1 !!!!
		int n = IPop (),		// columns
		    m = IPop ();		// rows
		ptr = VectorStart ();
		l = VectorLength ();
		ReleaseCalcStack (l + 1 - m * n);
		if (l - 1 != m * n) 
		    for (int i = l - 1; i < m * n; i++)
		        ptr [i] = ptr [i % (l - 1)];
		IPush (m, n);
		IPush (m * n + 1, 1);
		}
		break;
	    case MCBIND: {
		int s = VectorMemberSize (),
		    l2 = VectorLength (),
		    m = VectorRows (),
		    n2 = VectorCols ();
		ptr2 = VectorStart ();
		ReleaseCalcStack (l2 + 1);
		if (s != VectorMemberSize () || m != VectorRows ()) return RT_MATRIXOP;
		int l1 = VectorLength (),
		    n1 = VectorCols ();
		TStorage *ptr4 = ptr1 = VectorStart ();
		ReleaseCalcStack (l1 + 1);
		ptr3 = ptr = new TStorage [(n1 + n2) * m * s];
		for (int i = 0; i < m; i++) {
		    memcpy (ptr3, ptr1, n1 * s * sizeof (TStorage));
		    ptr3 += n1 * s;
		    ptr1 += n1 * s;
		    memcpy (ptr3, ptr2, n2 * s * sizeof (TStorage));
		    ptr3 += n2 * s;
		    ptr2 += n2 * s;
		}
		memcpy (ptr4, ptr, (n1 + n2) * m * s * sizeof (TStorage));
		ReserveCalcStack ((n1 + n2) * m * s);
		IPush (m, n1 + n2);
		IPush ((n1 + n2) * m * s + 1, s);
		delete [] ptr;
		}
		break;
	    case MRBIND: {
		int s = VectorMemberSize (),
		    l2 = VectorLength (),
		    m2 = VectorRows (),
		    n = VectorCols ();
		ptr = VectorStart ();
		ReleaseCalcStack (l2 + 1);
		if (s != VectorMemberSize () || n != VectorCols ()) return RT_MATRIXOP;
		int m1 = VectorRows (),
		    l1 = VectorLength ();
		memcpy (ptr - 2, ptr, m2 * n * s * sizeof (TStorage));
		ReserveCalcStack (m2 * n * s - 2);
		IPush (m1 + m2, n);
		IPush (l1 + l2 - 1, s);
		}
		break;
		
	    case VSSIZE:
		k = VectorLength ();
		l = VectorRows ();
		ReleaseCalcStack (k + 1);
		IPush (l);
		break;
	    case VSIZE:
		ptr = store + store [CalcStackTop ()->s.l1].s.l1 + CalcStackTop ()->s.l2 - 1;
		CalcStackTop ()->intval = ptr->s.l1;
		break;
	    case VCHECK:
		k = CalcStackTop ()->intval;
		if (k <= 0 || k > (CalcStackTop () - 1)->s.l2) return RT_RANGECHECK;
		break;
	    case VLVINDINT: {
		int size = IPop (),
		    vec = CalcStackTop ()->s.l1,
                    length = CalcStackTop ()->s.l2 - 1;
		ReleaseCalcStack (1);
		int ind = IPop ();
//	        printf ("Size = %d, Index = %d, Vector Length = %d\n", size, ind, length);
		if (ind < 1 || ind > length) return RT_RANGECHECK;
		vec = store [vec].s.l1;
		IPush (vec + (ind - 1) * size); }
		break;
	    case VINDINT:
	    case VINDINTC: {
		int size = IPop ();	// size of vector base type
		if (size == 1) {	
		    // replace index vector on stack with vector data
		    m = IPop ();
		    ptr1 = store + store [m].s.l1 - 1; // pointer before vector data
		    l = store [m].s.l2;
		    k = VectorLength ();	// length of index vector;
		    ptr = VectorStart ();	// start of index vector;
		    if (code [pc].code == VINDINT)
		        while (--k) {
			    *ptr = *(ptr1 + ptr->intval);
			    ptr++;
		        }
		    else
			while (--k) {
			    if (ptr->intval < 1 || ptr->intval >= l) return RT_RANGECHECK;
                            *ptr = *(ptr1 + ptr->intval);
                            ptr++;
                        }
		} else {
			// TODO
		    int n = 0,
			v = store [csp - 1].intval;
		    m = store [csp - 2].intval;
		    int maxsize = store [csp - 1].s.l2;
		    v = store [v].intval;
		    int i;
		    for (i = 0; i < m; i++) {
			l = store [csp - 2 - m + i].intval - 1;
		        if (l * size > maxsize) return RT_RANGECHECK;
			for (int j = 0; j < size; j++)
			    store [csp + n++] = store [v + l * size + j];
		    }
		    for (i = 0; i < n; i++) 
			store [csp - 2 - m + i] = store [csp + i];
		    store [csp - 2 - m + n].intval = n;
		    csp = csp - m + n - 1;
		}}
		break;

	    case VINDBOOL: {
	        int size = IPop ();
		if (size == 1) {
		    // replace boolean index vector with vector data
		    k = IPop ();
		    ptr1 = store + store [k].s.l1;	// points to vector data on heap
		    l = store [k].s.l2;		        // length of vector
		    m = k = VectorLength () - 1;	
		    ptr2 = ptr3 = ptr = VectorStart ();	// length and start of boolean index vector
		    if (k > l) return RT_RANGECHECK;
		    while (m--) {
			if (ptr++->intval) 
			    *ptr2++ = *ptr1;
			ptr1++;
		    }
		    l = ptr2 - ptr3;			// length of resulting vector
		    ptr->s.l1 = l;
		    *ptr2++ = *ptr++;
		    ptr->s.l1 = l + 1;
		    *ptr2 = *ptr;
		    ReleaseCalcStack (k - l);		    
		} else {
			// TODO
		    int n = 0,
			v = store [csp - 1].intval;
		    m = store [csp - 2].intval;
		    if (m * size > store [csp - 1].s.l2) return RT_RANGECHECK;
		    v = store [v].intval;
		    int i;
		    for (i = 0; i < m; i++)
			if (store [csp - 2 - m + i].intval)
			    for (int j = 0; j < size; j++)
				store [csp + n++] = store [v + i * size + j];
		    for (i = 0; i < n; i++)
			store [csp - 2 - m + i] = store [csp + i];
		    store [csp - 2 - m + n].intval = n;
		    csp = csp - m + n - 1;
		}}
		break;

	    case VSINDINT: 
	    case VSINDINTC: {
		int size = IPop ();
		k = VectorLength ();		// data vector
		ptr = VectorStart () - size;
		ReleaseCalcStack (k + 1);
		l = VectorLength ();		// index vector
		ptr1 = VectorStart ();

		if (size == 1)
		    // replace index vector on stack
		    if (code [pc].code == VSINDINTC)
		        while (--l) {
			    if (ptr1->intval >= k || ptr1->intval < 1) return RT_RANGECHECK;
			    *ptr1 = ptr [ptr1->intval];
			    ptr1++;
		        }
		    else
		        while (--l) {
			    *ptr1 = ptr [ptr1->intval];
			    ptr1++;
		        }

/*
		// TODO
		else {
		    n = 0;
		    for (l = 0; l < k; l++) {
			o = store [j + l].intval - 1;
			if (o < 0 || o * size >= m) return RT_RANGECHECK;
			for (int p = 0; p < size; p++)
			    store [csp + n++] = store [i + o * size + p];
		    }
		    for (int p = 0; p < n; p++)
			store [j + p] = store [csp + p];
		    store [j + n].intval = n;
		    csp = j + n + 1;
		}
*/
		}
		break;
	    case VSINDBOOL: {
		int size = IPop ();

		k = VectorLength ();
		ptr1 = VectorStart ();	// index vector
		ReleaseCalcStack (k + 1);
		l = VectorLength ();
		ptr = ptr2 = ptr3 = VectorStart ();	// data vector

		if (size == 1) 
		    while (--k) {
			if ((ptr1++)->intval)
			    *ptr2++ = *ptr;
			ptr++;
		    }
		else
		    while (--k)
			if ((ptr1++)->intval)
			    for (i = 0; i < size; i++)
				*ptr2++ = *ptr++;
			else
			    ptr += size;

		ptr2->s.l1 = (ptr2 - ptr3) / size;
		(ptr2++)->s.l2 = 1;
		ptr2->s.l1 = ptr2 - ptr3;
		ptr2->s.l2 = size;
		ReleaseCalcStack (l - (ptr2 - ptr3));
		}
		break;

	    // vector functions f: R -> R

	    #define REALVECFUNC(name,call,check) \
    		case name: \
		    k = VectorLength (); \
                    ptr = VectorStart (); \
                    while (--k) { \
			check; \
                        ptr->doubleval = call (ptr->doubleval); \
		        ptr++; \
                    } \
                    break; 
	    #define INTVECFUNC(name,call,check) \
    		case name: \
		    k = VectorLength (); \
                    ptr = VectorStart (); \
                    while (--k) { \
			check; \
                        ptr->intval = call (ptr->intval); \
		        ptr++; \
                    } \
                    break; 

	    REALVECFUNC (VSIN, sin, )
	    REALVECFUNC (VARCSIN, asin, )
	    REALVECFUNC (VCOS, cos, )
	    REALVECFUNC (VARCCOS, acos, )
	    REALVECFUNC (VTAN, tan, )
	    REALVECFUNC (VARCTAN, atan, )
	    REALVECFUNC (VCOSH, cosh, )
	    REALVECFUNC (VSINH, sinh, )
	    REALVECFUNC (VTANH, tanh, )
	    REALVECFUNC (VEXP, exp, )
	    REALVECFUNC (VFRAC, frac, )
	    REALVECFUNC (VSIGN, sign, )
	    REALVECFUNC (VFSQR, square, )
	    REALVECFUNC (VFABS, fabs, )
	    REALVECFUNC (VLOGC, log, if (ptr->doubleval <= 0.0) return RT_ARGUMENT)
	    REALVECFUNC (VLOG, log, )
	    REALVECFUNC (VSQRTC, sqrt, if (ptr->doubleval < 0.0) return RT_ARGUMENT)
	    REALVECFUNC (VSQRT, sqrt, )
	    INTVECFUNC  (VISQR, square, )
	    INTVECFUNC  (VIABS, abs, )
            INTVECFUNC  (VIODD, odd, )

            case VMAXV:
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
                a = -MYMAXDOUBLE;
		while (--k) {
		    if (ptr->doubleval > a) a = ptr->doubleval;
		    ptr++;
		}
                DPush (a);
                break;
            case VMINV:
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
                a = MYMAXDOUBLE;
		while (--k) {
		    if (ptr->doubleval < a) a = ptr->doubleval;
		    ptr++;
		}
                DPush (a);
                break;
            case VIMAX:
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
                l = -INT_MAX;
		while (--k) {
		    if (ptr->intval > l) l = ptr->intval;
		    ptr++;
		}
                IPush (l);
                break;
            case VIMIN:
		k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
                l = INT_MAX;
		while (--k) {
		    if (ptr->intval < l) l = ptr->intval;
		    ptr++;
		}
                IPush (l);
                break;
            case VEXIST:
                k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
                l = 0;
                while (--k && !l)
		    l = (ptr++)->intval;
                IPush (!!l);
                break;
            case VFORALL:
                k = VectorLength ();
		ptr = VectorStart ();
		ReleaseCalcStack (k + 1);
                l = 1;
		while (--k && l)
		    l = (ptr++)->intval;
                IPush (!!l);
                break;

	    case CIF:
		store [csp - 1].doubleval = (double) store [csp - 1].intval;
		break;
	    case CFS: {
		char buf [80];
		sprintf (buf, "%g", DPop ());
		PushStringAsVector (buf); }
		break;
	    case FADD:
		store [csp - 2].doubleval += store [csp - 1].doubleval;
		csp--;
		break;
	    case FSUB:
		store [csp - 2].doubleval -= store [csp - 1].doubleval;
		csp--;
		break;
	    case FMUL:
		store [csp - 2].doubleval *= store [csp - 1].doubleval;
		csp--;
		break;
	    case FDIVC:
                if (!store [csp - 1].doubleval) return RT_DIVZERO;
		// fall-through
            case FDIV:
		store [csp - 2].doubleval /= store [csp - 1].doubleval;
                csp--;
                break;
            case FPOW:
		b = DPop (); a = DPop ();
                if (a < 0.0 && frac (b) != 0.0) return RT_ARGUMENT;
		if (a == 0.0 && b < 0.0) return RT_ARGUMENT;
                DPush (pow (a, b));
                break;
	    case FFAC:
		l = IPop (); b = 1.0;
                while (l > 1) b *= l--;
                DPush (b);
                break;
            case FNEG:
		store [csp - 1].doubleval *= -1;
                break;
            case INEG:
                store [csp - 1].intval *= -1;
                break;
            case FEQ:
                store [csp - 2].intval =
			store [csp - 2].doubleval == store [csp - 1].doubleval;
                csp--;
                break;
            case FNE:
                store [csp - 2].intval =
                        store [csp - 2].doubleval != store [csp - 1].doubleval;
                csp--;
                break;
            case FLT:
                store [csp - 2].intval =
                        store [csp - 2].doubleval < store [csp - 1].doubleval;
		csp--;
		break;
            case FGT:
                store [csp - 2].intval =
                        store [csp - 2].doubleval > store [csp - 1].doubleval;
		csp--;
		break;
            case FLE:
                store [csp - 2].intval =
			store [csp - 2].doubleval <= store [csp - 1].doubleval;
                csp--;
		break;
            case FGE:
		store [csp - 2].intval =
                        store [csp - 2].doubleval >= store [csp - 1].doubleval;
                csp--;
                break;
            case IEQ:
                store [csp - 2].intval =
                        store [csp - 2].intval == store [csp - 1].intval;
                csp--;
                break;
            case INE:
                store [csp - 2].intval =
			store [csp - 2].intval != store [csp - 1].intval;
		csp--;
                break;
            case ILT:
                store [csp - 2].intval =
                        store [csp - 2].intval < store [csp - 1].intval;
		csp--;
                break;
            case IGT:
                store [csp - 2].intval =
                        store [csp - 2].intval > store [csp - 1].intval;
                csp--;
                break;
	    case ILE:
                store [csp - 2].intval =
			store [csp - 2].intval <= store [csp - 1].intval;
                csp--;
                break;
            case IGE:
		store [csp - 2].intval =
                        store [csp - 2].intval >= store [csp - 1].intval;
                csp--;
                break;
	    case SEQ:
		IPush (PopStringFromVector () == PopStringFromVector ());
                break;
            case SNE:
		s1 = PopStringFromVector (); s2 = PopStringFromVector ();
                IPush (s1 != s2);
		break;
            case SLT:
		s1 = PopStringFromVector (); s2 = PopStringFromVector ();
                IPush (s1 > s2);
                break;
            case SGT:
		s1 = PopStringFromVector (); s2 = PopStringFromVector ();
                IPush (s1 < s2);
                break;
            case SGE:
		s1 = PopStringFromVector (); s2 = PopStringFromVector ();
                IPush (s1 <= s2);
                break;
            case SLE:
		s1 = PopStringFromVector (); s2 = PopStringFromVector ();
                IPush (s1 >= s2);
                break;

	    case IMIN:
		IPush (min (IPop (), IPop ()));
		break;
	    case IMAX:
		IPush (max (IPop (), IPop ()));
		break;
	    case ISQR:
		IPush (sqr (IPop ())); break;
	    case FSQR:
		DPush (sqr (DPop ())); break;
	    case IABS:
                IPush (labs (IPop ())); break;
	    case IODD:
		IPush (odd (IPop ())); break;
	    case FABS:
		DPush (fabs (DPop ())); break;
            case FSIN:
                DPush (sin (DPop ())); break;
            case FASIN:
		a = DPop ();
                if (a < -1.0 || a > 1.0) return RT_ARGUMENT;
		DPush (asin (a));
                break;
	    case FCOS:
                DPush (cos (DPop ())); break;
            case FACOS:
                a = DPop ();
		if (a < -1.0 || a > 1.0) return RT_ARGUMENT;
                DPush (acos (a)); break;
	    case FTAN:
                DPush (tan (DPop ())); break;
            case FATAN:
                DPush (atan (DPop ())); break;
            case FCOSH:
                DPush (cosh (DPop ())); break;
            case FSINH:
                DPush (sinh (DPop ())); break;
            case FTANH:
		DPush (tanh (DPop ())); break;
	    case FLOG:
                a = DPop ();
                if (a <= 0.0) {
		    printf ("Attempt to evaluate log %g\n", a);
		    return RT_ARGUMENT;
                }
                DPush (log (a)); break;
            case FEXP:
                DPush (exp (DPop ())); break;
	    case FSQRT:
                a = DPop ();
                if (a < 0.0) return RT_ARGUMENT;
                DPush (sqrt (a)); break;
            case FSIGN:
                DPush (sign (DPop ())); break;
            case FFRAC:
                DPush (frac (DPop ())); break;
            case FINT:
                IPush (int (floor (DPop () + 0.5))); break;
            case FTRUNC:
                IPush (int (floor (DPop ()))); break;
	    case FMAX:
                DPush (max (DPop (), DPop ())); break;
            case FMIN:
                DPush (min (DPop (), DPop ())); break;
	    case FRANDOM:
		DPush (Random ());
		break;
	    case FCHR:
		c = DPop (); b = DPop (); a = DPop ();
		DPush (c <= b && c >= a); break;

	    case PRINTBOOL: {
                IPop (); IPop ();
                int len = IPop ();
		k = IPop (); l = IPop ();
		if (len != -1) 
		    for (i = 4 + !k; i < len; i++)
		        WriteChar ((int) l, (int) ' ');
		WriteString ((int) l, k  ? "true" : "false"); }
		break;
	    case PRINTCHAR: {
                IPop (); IPop ();
                int len = IPop ();
		k = IPop (); l = IPop ();
		if (len != -1)
		    for (i = 1; i < len; i++)
		        WriteChar ((int) l, (int) ' ');
		WriteChar ((int) l, (int) k); }
		break;
	    case PRINTINT: {
                IPop (); IPop ();
                int len = IPop ();
		k = IPop (); l = IPop ();
                char *buf = new char [20 + len];
		if (len == -1) sprintf (buf, "%d", k);
		else sprintf (buf, "%*d", len, k);
		WriteString ((int) l, buf);
		delete [] buf; }
		break;
	    case PRINTDBL: {
                IPop ();
                int dig = IPop (),
		    len = IPop ();
		a = DPop (); l = IPop ();
		char *buf = new char [30 + len];
	        if (dig != -1) sprintf (buf, "%*.*f", len, dig, a);
		else if (len != -1) sprintf (buf, "%*g", len, a);
		else sprintf (buf, "%g", a);
		WriteString ((int) l, buf);
		delete [] buf; }
		break;
	    case PRINTSTR: {
                IPop (); IPop ();
                int len = IPop ();
		s1 = PopStringFromVector ();
		k = IPop ();
		if (len != -1)
		    for (i = s1.Len (); i < len; i++)
		        WriteChar ((int) k, (int) ' ');
		WriteString ((int) k, s1); }
		break;
            case PRINTC: {
                IPop ();
                int dig = IPop (),
		    len = IPop ();
                dcomplex a = CPop (); l = IPop ();
                char *buf = new char [60 + 2 * len];
                if (dig != -1) sprintf (buf, "(%*.*f, *.*f)", len, dig, std::real (a), len, dig, std::imag (a));
                else if (len != -1) sprintf (buf, "(%*g, %*g)", len, std::real (a), len, std::imag (a));
                else sprintf (buf, "(%g, %g)", std::real (a), std::imag (a));
                WriteString ((int) l, buf);
                delete [] buf; }
                break;
	    case PRINTLF:
		WriteLinefeed ((int) IPop ());
		break;

	    case GETINT:
		IPush (FileList->GetFileDescriptor ((int) IPop ())->GetInt ());
		break;
	    case GETDBL:
		DPush (FileList->GetFileDescriptor ((int) IPop ())->GetDouble ());
		break;
	    case GETCHAR:
		IPush (FileList->GetFileDescriptor ((int) IPop ())->GetChar ());
		break;
	    case GETLF:	
		FileList->GetFileDescriptor ((int) IPop ())->GetLF ();
		break;
	    case PAGE:
		FileList->GetFileDescriptor ((int) IPop ())->Page ();
		break;
            case TXTCOL:
		k = IPop (); l = IPop ();
                FileList->GetFileDescriptor (l)->SetTextColor (k);
                break;
            case TXTBCK:
		k = IPop (); l = IPop ();
                FileList->GetFileDescriptor (l)->SetTextBackground (k);
                break;
	    case GOTOXY:
		k = IPop (); l = IPop (); m = IPop ();
		FileList->GetFileDescriptor ((int) m)->GotoXY ((int) l, (int) k);
		break;
	    case KEYP:
		IPush (FileList->GetFileDescriptor (IPop ())->GetKey ());
		break;
	    case GETSTR:
		PushStringAsVector (FileList->GetFileDescriptor ((int) IPop ())->GetString ());
		break;
	    case SCAN: {
		void *p [60];
		l = IPop ();
		for (i = 0; i < l; i++)
		    p [i] = store + IPop ();
		String format = PopStringFromVector (),
                       src = PopStringFromVector ();
		sscanf (src, format, p [0], p [1], p[2], p [3], p [4], p [5], p [6], p [7], p[8], p [9],
				     p [10], p [11], p[12], p [13], p [14], p [15], p [16], p [17], p[18], p [19],
				     p [20], p [21], p[22], p [23], p [24], p [25], p [26], p [27], p[28], p [29],
				     p [30], p [31], p[32], p [33], p [34], p [35], p [36], p [37], p[38], p [39],
				     p [40], p [41], p[42], p [43], p [44], p [45], p [46], p [47], p[48], p [49],
				     p [50], p [51], p[52], p [53], p [54], p [55], p [56], p [57], p[58], p [59]);
		}
		break;

	    case FASSIGN: 
		s1 = PopStringFromVector ();
		store [IPop ()].intval = FileList->OpenBin (s1);
		break;
	    case FTASSIGN:
		s1 = PopStringFromVector ();
		store [IPop ()].intval = FileList->OpenText (s1);
		break;

	    case FRESET:
		FileList->GetFileDescriptor ((int) IPop ())->Reset ();
		break;
	    case FREWRITE:
		FileList->GetFileDescriptor ((int) IPop ())->Rewrite ();
		break;
	    case FCLOSE:
		FileList->Close ((int) IPop ());
		break;
	    case FWRITE:
		m = IPop (); l = IPop (); k = IPop ();
		FileList->GetFileDescriptor ((int) k)->OutputBin ((void *) (store + (int) l), (int) m * sizeof (TStorage));
		break;
	    case FREAD:
		m = IPop (); l = IPop (); k = IPop ();
		FileList->GetFileDescriptor ((int) k)->InputBin ((void *) (store + (int) l), (int) m * sizeof (TStorage));
		break;
	    case FEOF:
		IPush (FileList->GetFileDescriptor ((int) IPop ())->Eof ());
		break;
	    case FFLUSH:
		FileList->GetFileDescriptor ((int) IPop ())->FFlush ();
		break;
	    case SYSTEM:
		IPush (system (PopStringFromVector ()));
		break;
	    case SYSDATE: {
		time_t t;
		time (&t);
		struct tm *tm_time = localtime (&t);
		store [IPop ()].intval = 1900 + tm_time->tm_year;
		store [IPop ()].intval = tm_time->tm_mon + 1;
		store [IPop ()].intval = tm_time->tm_mday; }
		break;
	    case SYSTIME: {
		time_t t;
		time (&t);
		struct tm *tm_time = localtime (&t);
		store [IPop ()].intval = tm_time->tm_sec;
		store [IPop ()].intval = tm_time->tm_min;
		store [IPop ()].intval = tm_time->tm_hour; }
		break;

	    case ARGC:
		IPush (argc);
		break;

	    case ARGV:
		k = IPop ();
		if (k < 1 || k > argc) return RT_ARGUMENT;
		PushStringAsVector (String (argv [k - 1]));
		break;

	    case SPHALT:
                #if !defined TERMINAL && defined _WINDOWS
	            CancelBox->CloseBox ();
//	            delete CancelBox;
                #endif
		return RT_OK;
	    default:
		if ((ext = ExtendedCode (code [pc])) != RT_OK)
		    return ext;
		break;
	}
	pc++;

    	#if !defined TERMINAL && defined _WINDOWS
	    statementcount++;
	    if (statementcount == 1000) {
	        CancelBox->CheckBoxMessages ();
	        if (CancelBox->IsCancelled ()) {
		    CancelBox->CloseBox ();
		    return RT_USERBREAK;
	        }
		statementcount = 0;
	    }
	#endif
    }
}

TRunTimeErrorType TPCodeInterpreter::Run () {
    Init ();
    return Execute ();
}

void TPCodeInterpreter::PushValue (double a) {
    store [sp--].doubleval = a;
}

void TPCodeInterpreter::PushValue (int m) {
    store [sp--].intval = m;
}

void TPCodeInterpreter::PushVector (int n, double *v) {
    int m = store [sp].s.l1 = Allocate (n + 1);
    store [sp--].s.l2 = n + 1;
    for (int i = 0; i < n; i++)
        store [store [m].s.l1 + i].doubleval = v [i];
    store [store [m].s.l1 + n].s.l1 = n;
    store [store [m].s.l1 + n].s.l2 = 1;
}

void TPCodeInterpreter::PushMatrix (int lines, int cols, double *v) {
    // !!!! Funktioniert so nicht mehr!!!!
    int size = lines * cols,
        m = store [sp].s.l1 = Allocate (size + 2);
    store [sp--].s.l2 = size + 2;
    for (int i = 0; i < size; i++)
	store [store [m].s.l1 + i].doubleval = v [i];
    store [store [m].s.l1 + size].intval = cols;
    store [store [m].s.l1 + size + 1].intval = lines;
}

double TPCodeInterpreter::GetDataValue (int Offset) {
    return store [DisplayBP [0] - 8 - Offset].doubleval;
}

int TPCodeInterpreter::GetIntValue (int Offset) {
    return store [DisplayBP [0] - 8 - Offset].intval;
}

void TPCodeInterpreter::GetVector (int Offset, int &n, double *&v) {
    n = store [DisplayBP [0] - 8 - Offset].s.l2 - 1;
    int m = store [DisplayBP [0] - 8 - Offset].s.l1;
    v = new double [n];
    for (int i = 0; i < n; i++)
        v [i] = store [store [m].s.l1 + i].doubleval;
}

void TPCodeInterpreter::GetMatrix (int Offset, int &lines, int &cols, double *&v) {
    int m = store [DisplayBP [0] - 8 - Offset].s.l1,
        size = store [DisplayBP [0] - 8 - Offset].s.l2;
    lines = store [store [m].s.l1 + size + 1].intval;
    cols = store [store [m].s.l1 + size].intval;
    v = new double [lines * cols];
    for (int i = 0; i < lines * cols; i++)
	v [i] = store [store [m].s.l1 + i].doubleval;
}
