/*  point.cpp

    Class definition for points and rectangles with real coordinates

    This file is part of Xtremes for Windows,
    Copyright (C) 1994 Xtremes Group, Siegen.
*/

#include <math.h>

#include "point.hpp"
#include "stddef.hpp"

TDblPoint operator + (TDblPoint a, TDblPoint b) {
    return TDblPoint (a.x + b.x, a.y + b.y);
}

TDblPoint operator - (TDblPoint a, TDblPoint b) {
    return TDblPoint (a.x - b.x, a.y - b.y);
}

BOOL operator == (TDblPoint a, TDblPoint b) {
    return a.x == b.x && a.y == b.y;
}

BOOL operator != (TDblPoint a, TDblPoint b) {
    return a.x != b.x || a.y != b.y;
}

TDblRectangle::TDblRectangle (double x0s, double y0s, double x1s, double y1s) {
    x0 = x0s; y0 = y0s; x1 = x1s; y1 = y1s;
    if (x0 > x1) Swap (x0, x1);
    if (y0 > y1) Swap (y0, y1);
}

TDblRectangle::TDblRectangle (TDblPoint p, TDblPoint q) {
    x0 = p.x; y0 = p.y;
    x1 = q.x; y1 = q.y;
    if (x0 > x1) Swap (x0, x1);
    if (y0 > y1) Swap (y0, y1);
}

TDblRectangle TDblRectangle::operator * (double v) {
    TDblRectangle r = *this;
    r.x0 -= (r.x1 - r.x0) * (v - 1.0);
    r.x1 += (r.x1 - r.x0) * (v - 1.0);
    r.y0 -= (r.y1 - r.y0) * (v - 1.0);
    r.y1 += (r.y1 - r.y0) * (v - 1.0);
    return r;
}

BOOL TDblRectangle::Within (TDblPoint p) {
    return p.x >= x0 && p.x <= x1 && p.y >= y0 && p.y <= y1;
}



TPoint3::TPoint3 (double x0, double y0, double z0) {
    x = x0; y = y0; z = z0;
}

TPoint3 operator * (double a, TPoint3 x) {
    TPoint3 y;
    y.x = a * x.x;
    y.y = a * x.y;
    y.z = a * x.z;
    return y;
}

TPoint3 operator + (TPoint3 a, TPoint3 b) {
    TPoint3 c;
    c.x = a.x + b.x;
    c.y = a.y + b.y;
    c.z = a.z + b.z;
    return c;
}

TPoint3 operator - (TPoint3 a, TPoint3 b) {
    TPoint3 c;
    c.x = a.x - b.x;
    c.y = a.y - b.y;
    c.z = a.z - b.z;
    return c;
}

double & TPoint3::operator [] (int i) {
    assert (0 <= i);
    assert (i < 3);
    if (i == 0) return x;
    if (i == 1) return y;
    return z;
}

double operator * (TPoint3 a, TPoint3 b) {
    return a.x * b.x + a.y * b.y + a.z * b.z;
}

inline double sqr (double x) { return x*x; }

double norm (TPoint3 p) {
    return sqrt (sqr (p.x) + sqr (p.y) + sqr (p.z));
}

double dist (TPoint3 a, TPoint3 b) {
    return norm (a - b);
}

BOOL operator < (TPoint3 a, TPoint3 b) {
    return a.x < b.x && a.y < b.y && a.z < b.z;
}

TPoint3 cross (TPoint3 p, TPoint3 q) {
    return TPoint3 (p [1] * q [2] - p [2] * q [1],
		    p [2] * q [0] - p [0] * q [2],
		    p [0] * q [1] - p [1] * q [0]);
}

TMatrix3::TMatrix3 () {
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            p [i][j] = 0.0;
}

TPoint3 TMatrix3::operator * (TPoint3 r) {
    TPoint3 q;
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
	    q [i] += p [i][j] * r [j];
    return q;
}

TMatrix3 TMatrix3::operator * (TMatrix3 r) {
    TMatrix3 q;
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            for (int k = 0; k < 3; k++)
                q.p [i][j] += p [i][k] * r.p [k][j];
    return q;
}

TMatrix3 RotateZ (double a) {
    TMatrix3 q;
    q.p [0][0] = cos (a);
    q.p [1][0] = sin (a);
    q.p [0][1] = -q.p [1][0];
    q.p [1][1] = q.p [0][0];
    q.p [2][2] = 1.0;
    return q;
}

TMatrix3 RotateY (double a) {
    TMatrix3 q;
    q.p [0][0] = cos (a);
    q.p [0][2] = sin (a);
    q.p [2][0] = -q.p [0][2];
    q.p [2][2] = q.p [0][0];
    q.p [1][1] = 1.0;
    return q;
}

TMatrix3 RotateX (double a) {
    TMatrix3 q;
    q.p [1][1] = cos (a);
    q.p [2][1] = sin (a);
    q.p [1][2] = -q.p [2][1];
    q.p [2][2] = q.p [1][1];
    q.p [0][0] = 1.0;
    return q;
}


