/*  risk.cpp

    Simulation of risk processes.

    This file is part of Xtremes for Windows,
    Copyright 1991, 2000 Xtremes Group, Siegen.
*/

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>

#include "risk.hpp"
#include "mathfunc.hpp"
#include "samplist.hpp"
#include "stddef.hpp"

#ifndef	TERMINAL
#include "xplfunc.hpp"
#include "ufodlg.hpp"
#include "cancel.hpp"
#include "stddlg.hpp"

#include "nyi.hpp"

#include "resource.h"
#include "privmsg.hpp"

TRiskProcess::TRiskProcess () {
    lambda = GlobalData.RiskLambda;
    rho	= GlobalData.RiskRho;
    u =	GlobalData.RiskU;
    T =	GlobalData.RiskT;
    fn = GlobalData.fnRisk;
    beta = GlobalData.RiskBeta;
    polalpha = GlobalData.PolyaProcAlpha;
    polsigma = GlobalData.PolyaProcSigma;
    PolyaLundbergSim = GlobalData.RuinProcSel;
    CurrT = 0.0;

    n =	Allocated = 0;
    t =	us = NULL;
    Done = FALSE;
    First = TRUE;
    S =	0.0;
}

String TRiskProcess::Name () {
    return String ("Risk process sample");
}

void TRiskProcess::DrawNext (double &t,	double &u) {
    CurrT += ExpQF (Random (), 1.0 / lambda);
    t =	CurrT;
    u =	DrawEvent ();
}

double TRiskProcess::GetU (double t) {
    return u + GetExpectedEvent	() * meanvalue * t + rho * GetExpectedEvent () * meanvalue * pow (t, beta);
}

void TRiskProcess::DrawCurveTo (TCoordinateWindow *w, double t0, double t1, double delta) {
    for	(int i = 1; i <= 10; i++) {
	double t = t0 +	i / 10.0 * (t1 - t0),
	       v = GetU	(t) - delta;
	w->LineTo (TDblPoint (t, v));
    }
}

void TRiskProcess::Plot	(TCoordinateWindow *w, BOOL BackColor, TColor c) {
    if (w->IsAdjustable ()) {
	w->NewXCoordinates (0, T);
	w->NewYCoordinates (0, 4 * u);
    }
    inherited::Plot (w, BackColor, c);
}

void TRiskProcess::TransferData () {
    SetPoints (n, t, us, TRUE);    
/*
    w->MoveTo (TDblPoint (0, GetU (0)));
    if (n) {
	DrawCurveTo (w, 0.0, t[0], 0.0);
	w->LineTo (TDblPoint (t[0], us[0]));
    }
    for	(int i = 1; i <	n; i++)	{
	DrawCurveTo (w, t[i-1], t[i], GetU	(t[i-1]) - us[i-1]);
	w->LineTo (TDblPoint (t[i], us[i]));
    }
    w->EndLine ();
*/
}

void TRiskProcess::Entry (double dataT,	double dataU) {
    if (n == Allocated)	{
	Allocated += 20;
	t = (double *) realloc (t, sizeof (double) * Allocated);
	us = (double *)	realloc	(us, sizeof (double) * Allocated);
    }
    t [n] = dataT;
    us [n] = dataU;
    n++;
}

void TRiskProcess::GeneratePoint () {
    if (First) {
    	Entry (0, u);
        First = FALSE;
    }
    double dataT, s0;
    DrawNext (dataT, s0);
    S += s0;
    double v = GetU (dataT) - S;
    if (!Done)
	Entry (dataT, v);
    if ((dataT > T || v	< 0.0) && !Done) {
	Done = TRUE;
	if (n) {
	    TMultivariateSample	*s = new TMultivariateSample;
	    s->SetData (2, fn, n);
	    s->SetColumn (0, t,	"Time");
	    s->SetColumn (1, us, "Reserve");

	    TList l;
	    l.Append (new TStringEl ("Path of Risk Process"));
	    l.Append (new TStringEl ("********************************"));
	    l.Append (new TStringEl ("Claim size distr.:    " & ClaimSizeName ()));
	    l.Append (new TStringEl ("Intensity lambda:     " & form (30, "%g", lambda)));
	    l.Append (new TStringEl ("Safety exponent beta: " & form (30, "%g", beta)));
	    l.Append (new TStringEl ("Safety loading rho:   " & form (30, "%g", rho)));
	    l.Append (new TStringEl ("Initial reserve s:    " & form (30, "%g", u)));
	    l.Append (new TStringEl ("Time horizon T:       " & form (30, "%g", T)));
	    if (v < 0.0) l.Append (new TStringEl ("Ruin time:            " & form (30, "%g", dataT)));

            s->SetDescription (&l);
	    s->SaveData	(0);
	    SampleList.InsertSample (s);
	} else ErrorBox (PlotWindowManager.GetFramePtr (), "There was no observation below the choosen maximum.\n\nNo data set generated.");
    }
}

BOOL TRiskProcess::IncParameter	() {
    GeneratePoint ();
    TransferData ();
    return TRUE;
}

BOOL TRiskProcess::IncParameterMany () {
    for	(int i = 0; i <	20; i++)
	GeneratePoint ();
    TransferData ();
    return TRUE;
}


TExpRiskProcess::TExpRiskProcess () {
    mu = GlobalData.mue;
    sigma = GlobalData.sigmae;
}

double TExpRiskProcess::DrawEvent () {
    return ExpQF (Random (), sigma, mu);
}

double TExpRiskProcess::GetExpectedEvent () {
    return mu + sigma;
}

String TExpRiskProcess::ClaimSizeName () {
    return form (100, "Exponential (mu = %g, sigma = %g)", mu, sigma);
}

BOOL TExpRiskProcess::EditParameter (TWindow *parent) {
    TCalcDoubleField fmu (IDD_Mu, mu, -DBL_MAX, DBL_MAX),
		 fsigma	(IDD_Sigma, sigma, DBL_MIN, DBL_MAX),
		 flambda (IDD_Lambda, lambda, DBL_MIN, DBL_MAX),
		 frho (IDD_Rho,	rho, -DBL_MAX, DBL_MAX),
		 fu (IDD_U, u, -DBL_MAX, DBL_MAX),
		 ft (IDD_TimeHorizont, T, 0, DBL_MAX),
		 fbeta (IDD_Beta, beta,	0, DBL_MAX);
//    TEditField	 fname	 (IDD_Filename,	fn, MAXFILENAME);
    TWriteFileField fname (IDD_FileName, fn);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TCalcDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TDialog d (IDD_ExpRiskProcessSample, 1, IDD_Visualize);

//    d.SetHelpNumber (GetHelpNumber ());
//    d.SetTitel (GetBoxTitel ());
    d << fmu << fsigma << flambda << frho	<< fu << ft << fname <<	fbeta << a << pa << ps;
    unsigned ch	= d.Handle (parent);
    if (ch != IDCANCEL)	{
	GlobalData.mue = mu = fmu.GetValue ();
	GlobalData.sigmae = sigma = fsigma.GetValue ();
	GlobalData.RiskLambda =	lambda = flambda.GetValue ();
	GlobalData.RiskRho = rho = frho.GetValue ();
	GlobalData.RiskU = u = fu.GetValue ();
	GlobalData.RiskT = T = ft.GetValue ();
	GlobalData.fnRisk = fn = fname.ReadBuf ();
	GlobalData.RiskBeta = beta = fbeta.GetValue ();
	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();
	if (a.Selector ()) {
	    meanvalue = pa.GetValue () * ps.GetValue ();
	    lambda = ps.GetValue () * GammaRV (pa.GetValue ());
	} else meanvalue = lambda;
    }
    return ch != IDCANCEL;
}

TParetoRiskProcess::TParetoRiskProcess () {
    alpha = max (2.0, GlobalData.alphap);
//    mu = GlobalData.mup;
    mu = 0.0;
    sigma = GlobalData.sigmap;
}

double TParetoRiskProcess::DrawEvent ()	{
    return ParetoQF (Random (),	alpha, sigma, mu);
}

double TParetoRiskProcess::GetExpectedEvent () {
    return mu +	sigma *	alpha /	(alpha - 1.0);
}

String TParetoRiskProcess::ClaimSizeName () {
    return form (100, "Pareto (alpha = %g, mu = %g, sigma = %g)", alpha, mu, sigma);
}

BOOL TParetoRiskProcess::EditParameter (TWindow *parent) {
    TDoubleField falpha	(IDD_Alpha, alpha, 1.00001, DBL_MAX),
		 fmu (IDD_Mu, mu, -DBL_MAX, DBL_MAX),
		 fsigma	(IDD_Sigma, sigma, DBL_MIN, DBL_MAX),
		 flambda (IDD_Lambda, lambda, DBL_MIN, DBL_MAX),
		 frho (IDD_Rho,	rho, -DBL_MAX, DBL_MAX),
		 fu (IDD_U, u, -DBL_MAX, DBL_MAX),
		 ft (IDD_TimeHorizont, T, 0, DBL_MAX),
		 fbeta (IDD_Beta, beta,	0, DBL_MAX);
    TWriteFileField	 fname	 (IDD_FileName,	fn);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TDialog d (IDD_ParetoRiskProcessSample, 1, IDD_Visualize);
//    d.SetHelpNumber (GetHelpNumber ());
//    d.SetTitel (GetBoxTitel ());
    d << fmu <<	falpha << fsigma << flambda << frho << fu << ft	<< fname << fbeta << a << pa << ps;
    unsigned ch	= d.Handle (parent);
    if (ch != IDCANCEL)	{
	GlobalData.alphap = alpha = falpha.GetValue ();
	GlobalData.mup = mu = fmu.GetValue ();
	GlobalData.sigmap = sigma = fsigma.GetValue ();
	GlobalData.RiskLambda =	lambda = flambda.GetValue ();
	GlobalData.RiskRho = rho = frho.GetValue ();
	GlobalData.RiskU = u = fu.GetValue ();
	GlobalData.RiskT = T = ft.GetValue ();
	GlobalData.fnRisk = fn = fname.ReadBuf ();
	GlobalData.RiskBeta = beta = fbeta.GetValue ();
	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();
	if (a.Selector ()) {
	    meanvalue = pa.GetValue () * ps.GetValue ();
	    lambda = ps.GetValue () * GammaRV (pa.GetValue ());
	} else meanvalue = lambda;
    }
    return ch != IDCANCEL;
}

TGPRiskProcess::TGPRiskProcess () {
    gamma = GlobalData.betagp;
    mu = GlobalData.mugp;
    sigma = GlobalData.sigmagp;
}

double TGPRiskProcess::DrawEvent () {
    return vonMisesGPDQF (Random (), gamma, sigma, mu);
}

double TGPRiskProcess::GetExpectedEvent	() {
    return mu +	sigma /	(1.0 - gamma);
}

String TGPRiskProcess::ClaimSizeName () {
    return form (100, "GP (gamma = %g, mu = %g, sigma = %g)", gamma, mu, sigma);
}

BOOL TGPRiskProcess::EditParameter (TWindow *parent) {
    TDoubleField fgamma	(IDD_Gamma, min	(gamma,	0.5), -DBL_MAX, 0.999999),
		 fmu (IDD_Mu, mu, -DBL_MAX, DBL_MAX),
		 fsigma	(IDD_Sigma, sigma, DBL_MIN, DBL_MAX),
		 flambda (IDD_Lambda, lambda, DBL_MIN, DBL_MAX),
		 frho (IDD_Rho,	rho, -DBL_MAX, DBL_MAX),
		 fu (IDD_U, u, -DBL_MAX, DBL_MAX),
		 ft (IDD_TimeHorizont, T, 0, DBL_MAX),
		 fbeta (IDD_Beta, beta,	0, DBL_MAX);
    TWriteFileField	 fname	 (IDD_FileName,	fn);
    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TDialog d (IDD_GPRiskProcessSample,	1, IDD_Visualize);
    d << fmu <<	fgamma << fsigma << flambda << frho << fu << ft	<< fname << fbeta << a << pa << ps;
    unsigned ch	= d.Handle (parent);
    if (ch != IDCANCEL)	{
	GlobalData.betagp = gamma = fgamma.GetValue ();
	GlobalData.mugp	= mu = fmu.GetValue ();
	GlobalData.sigmagp = sigma = fsigma.GetValue ();
	GlobalData.RiskLambda =	lambda = flambda.GetValue ();
	GlobalData.RiskRho = rho = frho.GetValue ();
	GlobalData.RiskU = u = fu.GetValue ();
	GlobalData.RiskT = T = ft.GetValue ();
	GlobalData.fnRisk = fn = fname.ReadBuf ();
	GlobalData.RiskBeta = beta = fbeta.GetValue ();
	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();
	if (a.Selector ()) {
	    meanvalue = pa.GetValue () * ps.GetValue ();
	    lambda = ps.GetValue () * GammaRV (pa.GetValue ());
	} else meanvalue = lambda;
    }
    return ch != IDCANCEL;
}

#endif

/*******************************************************************/
/*		Simulation of ruin propability			   */
/*******************************************************************/

#ifndef	TERMINAL

class TRuinSimLowerBound : public TFunction {
    public:
	TRuinSimLowerBound (double alpha, double sigma,	double mu, double rho);
	virtual	String Name ();
	virtual	double f (double);
	virtual	void GetSupport	(double	&x0, double &x1);
    private:
	double alpha, sigma, mu, rho, c;
};

TRuinSimLowerBound::TRuinSimLowerBound (double alpha0, double sigma0, double mu0, double rho0) {
    alpha = alpha0;
    sigma = sigma0;
    mu = mu0;
    rho	= rho0;
    c =	sigma /	((mu + sigma * alpha / (alpha -	1.0)) *	(alpha - 1.0) *	(1.0 + rho));
}

String TRuinSimLowerBound::Name	() {
    return "Lower Bound	for ruin propability";
}

double TRuinSimLowerBound::f (double x)	{
    return c * pow ((x - mu) / sigma, -alpha + 1.0);
}

void TRuinSimLowerBound::GetSupport (double &x0, double	&x1) {
    x0 = 1.0; x1 = DBL_MAX;
}

class TRuinSimAsymptoticBound :	public TFunction {
    public:
	TRuinSimAsymptoticBound	(double	alpha, double sigma, double mu,	double rho);
	virtual	String Name ();
	virtual	double f (double);
	virtual	void GetSupport	(double	&x0, double &x1);
    private:
	double alpha, sigma, mu, rho, c;
};

TRuinSimAsymptoticBound::TRuinSimAsymptoticBound (double alpha0, double	sigma0,	double mu0, double rho0) {
    alpha = alpha0;
    sigma = sigma0;
    mu = mu0;
    rho	= rho0;
    c =	sigma /	((mu + sigma * alpha / (alpha -	1.0)) *	(alpha - 1.0) *	rho);
}

String TRuinSimAsymptoticBound::Name ()	{
    return "Asymptotic ruin propability";
}

double TRuinSimAsymptoticBound::f (double x) {
    return c * pow ((x - mu) / sigma, -alpha + 1.0);
}

void TRuinSimAsymptoticBound::GetSupport (double &x0, double &x1) {
    x0 = 1.0; x1 = DBL_MAX;
}

#endif


TRuinSimulation::TRuinSimulation (int NrSim0, double rho0, double beta0, double	lambda0, double polalpha0, double polsigma0, BOOL Poly0, double	maxT0) {
    NrSim = NrSim0;
    rho	= rho0;
    beta = beta0;
    lambda = lambda0;
    polalpha = polalpha0;
    polsigma = polsigma0;
    PolyaLundbergSim = Poly0;
    maxT = maxT0;
    u =	t = p =	NULL;
    if (PolyaLundbergSim) meanvalue = polalpha * polsigma;
    else meanvalue = lambda;
}

TRuinSimulation::~TRuinSimulation () {
    if (u) delete [] u;
    if (t) delete [] t;
    if (p) delete [] p;
}

void TRuinSimulation::SetU (double umin, double	umax, int n) {
    u =	new double [n];
    t =	new double [n];
    p =	new double [n];
    for	(int i = 0; i <	n; i++)	{
	if (n >	1) u [i] = umin	+ (double) i / (n - 1) * (umax - umin);
	else u [i] = umin;
	p [i] =	0.0;
    }
    un = n;
}

void TRuinSimulation::SingleSimulation () {
    for	(int i = 0; i <	un; i++) t [i] = maxT;
    tc = 0.0, s	= 0.0;
    RuinPtr = 0;
    if (PolyaLundbergSim) lambda = polsigma * GammaRV (polalpha);
    while (tc <= maxT && RuinPtr != un)	{
	tc += ExpQF (Random (),	1.0 / lambda);
	s += DrawEvent ();
	CheckRuin ();
    }
}

void TRuinSimulation::CheckRuin	() {
    double sc =	tc * meanvalue * ExpectedEvent () + rho * meanvalue * pow (tc, beta) * ExpectedEvent ();
    while (RuinPtr < un	&& u [RuinPtr ]	+ sc < s)
	t [RuinPtr++] =	tc;
}

#ifndef	TERMINAL
void TRuinSimulation::Simulation (TWindow *parent) {
    int	i, j;
    TCancelBox *CancelBox = new	TCancelBox (parent, IDD_CancelBox);
    CancelBox->OpenDialog (parent);
    CancelBox->SetTitel	("Simulate ruin time");
    for	(i = 0;	i < NrSim && !CancelBox->IsCancelled (); i++) {
	CancelBox->CheckBoxMessages ();
	SingleSimulation ();
	for (j = 0; j <	un; j++)
	    p [j] += (t	[j] < maxT);
	CancelBox->SetText (form (10, "%4.1f", i / double (NrSim) * 100.0));
    }
    CancelBox->CloseBox	();
    for	(j = 0;	j < un;	j++)
	p [j] /= i;
}
#endif

TDblPoint TRuinSimulation::GetValue (int i) {
    assert (0 <= i && i	< un);
    return TDblPoint (u	[i], p [i]);
}


TExponentialRuinSimulation::TExponentialRuinSimulation (double mu0, double sigma0, int NrSim, double rho, double beta,	double lambda, double polalpha, double polsigma, BOOL PolyaLundbergSim, double maxT)
 : TRuinSimulation (NrSim, rho,	beta, lambda, polalpha, polsigma, PolyaLundbergSim, maxT) {
     mu	= mu0;
     sigma = sigma0;
}

double TExponentialRuinSimulation::DrawEvent ()	{
    return mu + sigma *	ExpQF (Random (), 1.0);
}

double TExponentialRuinSimulation::ExpectedEvent () {
    return mu + sigma;
}


TParetoRuinSimulation::TParetoRuinSimulation (double alpha0, double mu0, double	sigma0,	int NrSim,
					      double rho, double beta, double lambda, double polalpha, double polsigma, BOOL PolyaLundbergSim, double maxT)
 : TRuinSimulation (NrSim, rho,	beta, lambda, polalpha, polsigma, PolyaLundbergSim, maxT) {
     alpha = alpha0;
     mu	= mu0;
     sigma = sigma0;
}

double TParetoRuinSimulation::DrawEvent	() {
    return ParetoQF (Random (),	alpha, sigma, mu);
}

double TParetoRuinSimulation::ExpectedEvent () {
    return mu +	sigma *	alpha /	(alpha - 1.0);
}


TGPRuinSimulation::TGPRuinSimulation (double gamma0, double mu0, double	sigma0,	int NrSim,
				      double rho, double beta, double lambda, double polalpha, double polsigma, BOOL PolyaLundbergSim, double maxT)
 : TRuinSimulation (NrSim, rho,	beta, lambda, polalpha, polsigma, PolyaLundbergSim, maxT) {
     gamma = gamma0;
     mu	= mu0;
     sigma = sigma0;
}

double TGPRuinSimulation::DrawEvent () {
    return vonMisesGPDQF (Random (), gamma, sigma, mu);
}

double TGPRuinSimulation::ExpectedEvent	() {
    return mu +	sigma /	(1.0 - gamma);
}


#ifndef	TERMINAL

void PerformExpRuinSimulation (TWindow *parent)	{
    TDialog d (IDD_SimulateExpRuinProp);
    TDoubleField flambda (IDD_Lambda, GlobalData.RiskLambda, DBL_MIN,	DBL_MAX),
		 frho (IDD_Rho,	GlobalData.RiskRho, 0, DBL_MAX),
		 fbeta (IDD_Beta, GlobalData.RiskBeta, 0, DBL_MAX),
		 fsigma	(IDD_Sigma, GlobalData.sigmae, DBL_MIN, DBL_MAX),
		 fmu (IDD_Mu, GlobalData.mue, -DBL_MAX, DBL_MAX),
		 fumin (IDD_Umin, GlobalData.RuinUmin, 0.0, DBL_MAX),
		 fumax (IDD_Umax, GlobalData.RuinUmax, 0.0, DBL_MAX),
		 fmaxt (IDD_TimeHorizont, GlobalData.RuinTime, 0.0, DBL_MAX);
    TIntField	 fustep	(IDD_Steps, GlobalData.RuinSteps, 2, INT_MAX),
		 fNrSims (IDD_NrSims, GlobalData.RuinNrSims, 1,	INT_MAX);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TEditField	 fName (IDD_FileName, GlobalData.RuinName, 80);
    d << flambda << frho << fbeta << fsigma << fmu <<	fumin << fumax
      << fustep	<< fNrSims << fName << fmaxt << a << pa << ps;
    if (d.Handle (parent) == IDOK) {
	GlobalData.RiskLambda =	flambda.GetValue ();
	GlobalData.RiskRho = frho.GetValue ();
	GlobalData.RiskBeta = fbeta.GetValue ();
	GlobalData.sigmae = fsigma.GetValue ();
	GlobalData.mue = fmu.GetValue ();
	GlobalData.RuinUmin = fumin.GetValue ();
	GlobalData.RuinUmax = fumax.GetValue ();
	GlobalData.RuinTime = fmaxt.GetValue ();
	int USteps = GlobalData.RuinSteps = fustep.GetValue ();
	GlobalData.RuinNrSims =	fNrSims.GetValue ();
	GlobalData.RuinName = fName.ReadBuf ();

	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();

	TExponentialRuinSimulation d	(fmu.GetValue (), fsigma.GetValue (),
				 fNrSims.GetValue (), frho.GetValue (),	fbeta.GetValue (), flambda.GetValue (),
                                 pa.GetValue (), ps.GetValue (), a.Selector (), 
				 fmaxt.GetValue ());
	d.SetU (fumin.GetValue (), fumax.GetValue (), fustep.GetValue ());
	d.Simulation (parent);
	TPlotWindow *w = PlotWindowManager.GetOpenWindow ("Ruin Probability Simulation");
	double *x = new	double [USteps],
	       *y = new	double [USteps];
	TDblPoint q;
	for (int i = 0;	i < USteps; i++) {
	    q =	d.GetValue (i);
	    x [i] = q.x;
	    y [i] = q.y;
	}
	w->InsertFunction (new TXPLFunction (USteps, x,	y, "Ruin sim", FALSE));

	TMultivariateSample *s = new TMultivariateSample;
	s->SetData (2, fName.ReadBuf (), USteps);
	s->SetColumn (0, x, "Reserve");
	s->SetColumn (1, y, "Ruin prop");
	s->SaveData (0);
	SampleList.InsertSample	(s);
    }
}

void PerformParetoRuinSimulation (TWindow *parent) {
    TDialog d (IDD_SimulateParetoRuinProp);
    TDoubleField flambda (IDD_Lambda, GlobalData.RiskLambda, DBL_MIN,	DBL_MAX),
		 frho (IDD_Rho,	GlobalData.RiskRho, 0, DBL_MAX),
		 falpha	(IDD_Alpha, max	(GlobalData.alphap, 2.0), 1.00001, DBL_MAX),
		 fbeta (IDD_Beta, GlobalData.RiskBeta, 0, DBL_MAX),
		 fsigma	(IDD_Sigma, GlobalData.sigmap, DBL_MIN, DBL_MAX),
		 fmu (IDD_Mu, GlobalData.mup, -DBL_MAX, DBL_MAX),
		 fumin (IDD_Umin, GlobalData.RuinUmin, 0.0, DBL_MAX),
		 fumax (IDD_Umax, GlobalData.RuinUmax, 0.0, DBL_MAX),
		 fmaxt (IDD_TimeHorizont, GlobalData.RuinTime, 0.0, DBL_MAX);
    TIntField	 fustep	(IDD_Steps, GlobalData.RuinSteps, 2, INT_MAX),
		 fNrSims (IDD_NrSims, GlobalData.RuinNrSims, 1,	INT_MAX);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TEditField	 fName (IDD_FileName, GlobalData.RuinName, 80);
    d << flambda << frho << fbeta << falpha << fsigma << fmu <<	fumin << fumax
      << fustep	<< fNrSims << fName << fmaxt << a << pa << ps;
    if (d.Handle (parent) == IDOK) {
	GlobalData.RiskLambda =	flambda.GetValue ();
	GlobalData.RiskRho = frho.GetValue ();
	GlobalData.RiskBeta = fbeta.GetValue ();
	GlobalData.alphap = falpha.GetValue ();
	GlobalData.sigmap = fsigma.GetValue ();
	GlobalData.mup = fmu.GetValue ();
	GlobalData.RuinUmin = fumin.GetValue ();
	GlobalData.RuinUmax = fumax.GetValue ();
	GlobalData.RuinTime = fmaxt.GetValue ();
	int USteps = GlobalData.RuinSteps = fustep.GetValue ();
	GlobalData.RuinNrSims =	fNrSims.GetValue ();
	GlobalData.RuinName = fName.ReadBuf ();

	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();

	TParetoRuinSimulation d	(falpha.GetValue (), fmu.GetValue (), fsigma.GetValue (),
				 fNrSims.GetValue (), frho.GetValue (),	fbeta.GetValue (), flambda.GetValue (),
				 pa.GetValue (), ps.GetValue (), a.Selector (),
				 fmaxt.GetValue ());
	d.SetU (fumin.GetValue (), fumax.GetValue (), fustep.GetValue ());
	d.Simulation (parent);
	TPlotWindow *w = PlotWindowManager.GetOpenWindow ("Ruin Probability Simulation");
	double *x = new	double [USteps],
	       *y = new	double [USteps];
	TDblPoint q;
	for (int i = 0;	i < USteps; i++) {
	    q =	d.GetValue (i);
	    x [i] = q.x;
	    y [i] = q.y;
	}
	w->InsertFunction (new TXPLFunction (USteps, x,	y, "Ruin sim", FALSE));

	TMultivariateSample *s = new TMultivariateSample;
	s->SetData (2, fName.ReadBuf (), USteps);
	s->SetColumn (0, x, "Reserve");
	s->SetColumn (1, y, "Ruin prop");
	s->SaveData (0);
	SampleList.InsertSample	(s);
    }
}

void PerformGPRuinSimulation (TWindow *parent) {
    TDialog d (IDD_SimulateGPRuinProp);
    TDoubleField flambda (IDD_Lambda, GlobalData.RiskLambda, DBL_MIN,	DBL_MAX),
		 frho (IDD_Rho,	GlobalData.RiskRho, 0, DBL_MAX),
		 fgamma	(IDD_Gamma, min	(GlobalData.betagp, 0.5), -DBL_MAX, 0.9999999),
		 fbeta (IDD_Beta, GlobalData.RiskBeta, 0, DBL_MAX),
		 fsigma	(IDD_Sigma, GlobalData.sigmae, DBL_MIN, DBL_MAX),
		 fmu (IDD_Mu, GlobalData.mue, -DBL_MAX, DBL_MAX),
		 fumin (IDD_Umin, GlobalData.RuinUmin, 0.0, DBL_MAX),
		 fumax (IDD_Umax, GlobalData.RuinUmax, 0.0, DBL_MAX),
		 fmaxt (IDD_TimeHorizont, GlobalData.RuinTime, 0.0, DBL_MAX);
    TIntField	 fustep	(IDD_Steps, GlobalData.RuinSteps, 2, INT_MAX),
		 fNrSims (IDD_NrSims, GlobalData.RuinNrSims, 1,	INT_MAX);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TEditField	 fName (IDD_FileName, GlobalData.RuinName, 80);
    d << flambda << frho << fbeta << fgamma << fsigma << fmu <<	fumin << fumax
      << fustep	<< fNrSims << fName << fmaxt << a << pa << ps;
    if (d.Handle (parent) == IDOK) {
	GlobalData.RiskLambda =	flambda.GetValue ();
	GlobalData.RiskRho = frho.GetValue ();
	GlobalData.RiskBeta = fbeta.GetValue ();
	GlobalData.sigmagp = fsigma.GetValue ();
	GlobalData.mugp = fmu.GetValue ();
        GlobalData.betagp = fgamma.GetValue ();
	GlobalData.RuinUmin = fumin.GetValue ();
	GlobalData.RuinUmax = fumax.GetValue ();
	GlobalData.RuinTime = fmaxt.GetValue ();
	int USteps = GlobalData.RuinSteps = fustep.GetValue ();
	GlobalData.RuinNrSims =	fNrSims.GetValue ();
	GlobalData.RuinName = fName.ReadBuf ();

	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();

	TGPRuinSimulation d	(fgamma.GetValue (), fmu.GetValue (), fsigma.GetValue (),
				 fNrSims.GetValue (), frho.GetValue (),	fbeta.GetValue (), flambda.GetValue (),
				 pa.GetValue (), ps.GetValue (), a.Selector (),
				 fmaxt.GetValue ());
	d.SetU (fumin.GetValue (), fumax.GetValue (), fustep.GetValue ());
	d.Simulation (parent);
	TPlotWindow *w = PlotWindowManager.GetOpenWindow ("Ruin Probability Simulation");
	double *x = new	double [USteps],
	       *y = new	double [USteps];
	TDblPoint q;
	for (int i = 0;	i < USteps; i++) {
	    q =	d.GetValue (i);
	    x [i] = q.x;
	    y [i] = q.y;
	}
	w->InsertFunction (new TXPLFunction (USteps, x,	y, "Ruin sim", FALSE));

	TMultivariateSample *s = new TMultivariateSample;
	s->SetData (2, fName.ReadBuf (), USteps);
	s->SetColumn (0, x, "Reserve");
	s->SetColumn (1, y, "Ruin prop");
	s->SaveData (0);
	SampleList.InsertSample	(s);
    }
}

/*******************************************************************/
/*		   Simulation of ruin times			   */
/*******************************************************************/

void PerformParetoRuinTimeSimulation (TWindow *parent) {
    TDialog d (IDD_SimulateParetoRuinTime);
    TDoubleField flambda (IDD_Lambda, GlobalData.RiskLambda, DBL_MIN,	DBL_MAX),
		 frho (IDD_Rho,	GlobalData.RiskRho, 0, DBL_MAX),
		 fbeta (IDD_Beta, GlobalData.RiskBeta, 0, DBL_MAX),
		 falpha	(IDD_Alpha, max	(GlobalData.alphap, 2.0), 1.00001, DBL_MAX),
		 fsigma	(IDD_Sigma, GlobalData.sigmap, DBL_MIN, DBL_MAX),
		 fmu (IDD_Mu, GlobalData.mup, -DBL_MAX, DBL_MAX),
		 fumin (IDD_Umin, GlobalData.RuinUmin, 0.0, DBL_MAX),
		 fmaxt (IDD_TimeHorizont, GlobalData.RuinTime, 0.0, DBL_MAX);
    TIntField	 fNrSims (IDD_NrSims, GlobalData.RuinNrSims, 1,	INT_MAX);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TEditField	 fName (IDD_FileName, GlobalData.RuinName, 80);
    d << flambda << frho << fbeta << falpha << fsigma << fmu <<	fumin
      << fNrSims << fName << fmaxt << a << pa << ps;
    if (d.Handle (parent) == IDOK) {
	double lambda = GlobalData.RiskLambda =	flambda.GetValue ();
	double rho = GlobalData.RiskRho = frho.GetValue ();
	double beta = GlobalData.RiskBeta = fbeta.GetValue ();
	double alpha = GlobalData.alphap = falpha.GetValue ();
	double sigma = GlobalData.sigmap = fsigma.GetValue ();
	double mu = GlobalData.mup = fmu.GetValue ();
	double u = GlobalData.RuinUmin = fumin.GetValue ();
	double T = GlobalData.RuinTime = fmaxt.GetValue ();
	int NrSims = GlobalData.RuinNrSims = fNrSims.GetValue ();
	GlobalData.RuinName = fName.ReadBuf ();

	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();

	TParetoRuinSimulation d	(falpha.GetValue (), fmu.GetValue (), fsigma.GetValue (),
				 fNrSims.GetValue (), frho.GetValue (),	fbeta.GetValue (), flambda.GetValue (),
				 pa.GetValue (), ps.GetValue (), a.Selector (),
				 fmaxt.GetValue ());
	d.SetU (fumin.GetValue (), fumin.GetValue (), 1);
	double *x = new	double [NrSims];
	TCancelBox *CancelBox =	new TCancelBox (parent, IDD_CancelBox);
	CancelBox->OpenDialog (parent);
	CancelBox->SetTitel ("Simulate ruin time");
	int i;
	for (i = 0; i <	NrSims && !CancelBox->IsCancelled (); i++) {
	    CancelBox->CheckBoxMessages	();
	    d.SingleSimulation ();
	    x [i] = d.GetTime (0);
	    CancelBox->SetText (form (10, "%4.1f", i / double (NrSims) * 100.0));
	}
	CancelBox->CloseBox ();
	TList s;
	s.Append (new TStringEl	("Ruin time simulation"));
	s.Append (new TStringEl ("********************************"));
	s.Append (new TStringEl (form (100, "Claim size distr.:    Pareto (alpha = %g, mu = %g, sigma = %g)", alpha, mu, sigma)));
	s.Append (new TStringEl ("Intensity lambda:     " & form (30, "%g", lambda)));
	s.Append (new TStringEl ("Safety exponent beta: " & form (30, "%g", beta)));
	s.Append (new TStringEl ("Safety loading rho:   " & form (30, "%g", rho)));
	s.Append (new TStringEl ("Initial reserve s:    " & form (30, "%g", u)));
	s.Append (new TStringEl ("Time horizon T:       " & form (30, "%g", T)));

	TUnivariateSample *p = new TUnivariateSample;
	p->SetName ("Ruin time simulation");
	p->SetDescription (&s);
	p->SetData (i, x, fName.ReadBuf	());
	s.DeleteAll (FALSE);
	delete [] x;
    }
}

void PerformExpRuinTimeSimulation (TWindow *parent) {
    TDialog d (IDD_SimulateExpRuinTime);
    TDoubleField flambda (IDD_Lambda, GlobalData.RiskLambda, DBL_MIN,	DBL_MAX),
		 frho (IDD_Rho,	GlobalData.RiskRho, 0, DBL_MAX),
		 fbeta (IDD_Beta, GlobalData.RiskBeta, 0, DBL_MAX),
		 fmu (IDD_Mu, GlobalData.mue, -DBL_MAX, DBL_MAX),
		 fsigma (IDD_Sigma, GlobalData.sigmae, DBL_MIN, DBL_MAX),
		 fumin (IDD_Umin, GlobalData.RuinUmin, 0.0, DBL_MAX),
		 fmaxt (IDD_TimeHorizont, GlobalData.RuinTime, 0.0, DBL_MAX);
    TIntField	 fNrSims (IDD_NrSims, GlobalData.RuinNrSims, 1,	INT_MAX);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TEditField	 fName (IDD_FileName, GlobalData.RuinName, 80);
    d << flambda << frho << fbeta << fmu << fsigma << fumin
      << fNrSims << fName << fmaxt << a << pa << ps;
    if (d.Handle (parent) == IDOK) {
	GlobalData.RiskLambda =	flambda.GetValue ();
	GlobalData.RiskRho = frho.GetValue ();
	GlobalData.RiskBeta = fbeta.GetValue ();
	GlobalData.mue = fmu.GetValue ();
	GlobalData.sigmae = fsigma.GetValue ();
	GlobalData.RuinUmin = fumin.GetValue ();
	GlobalData.RuinTime = fmaxt.GetValue ();
	int NrSims = GlobalData.RuinNrSims = fNrSims.GetValue ();
	GlobalData.RuinName = fName.ReadBuf ();

	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();

	TExponentialRuinSimulation d (fmu.GetValue (), fsigma.GetValue (), fNrSims.GetValue	(), frho.GetValue (),
			      fbeta.GetValue (), flambda.GetValue (),
			      pa.GetValue (), ps.GetValue (), a.Selector (),
			      fmaxt.GetValue ());
	d.SetU (fumin.GetValue (), fumin.GetValue (), 1);
	double *x = new	double [NrSims];
	TCancelBox *CancelBox =	new TCancelBox (parent, IDD_CancelBox);
	CancelBox->OpenDialog (parent);
	CancelBox->SetTitel ("Simulate ruin time");
	int i;
	for (i = 0; i <	NrSims && !CancelBox->IsCancelled (); i++) {
	    CancelBox->CheckBoxMessages	();
	    d.SingleSimulation ();
	    x [i] = d.GetTime (0);
	    CancelBox->SetText (form (10, "%4.1f", i / double (NrSims) * 100.0));
	}
	CancelBox->CloseBox ();
	TList s;
	s.Append (new TStringEl	("Ruin time simulation"));
	TUnivariateSample *p = new TUnivariateSample;
	p->SetName ("Ruin time simulation");
	p->SetDescription (&s);
	p->SetData (i, x, fName.ReadBuf	());
	s.DeleteAll (FALSE);
	delete [] x;
    }
}

void PerformGPRuinTimeSimulation (TWindow *parent) {
    TDialog d (IDD_SimulateGPRuinTime);
    TDoubleField flambda (IDD_Lambda, GlobalData.RiskLambda, DBL_MIN,	DBL_MAX),
		 frho (IDD_Rho,	GlobalData.RiskRho, 0, DBL_MAX),
		 fbeta (IDD_Beta, GlobalData.RiskBeta, 0, DBL_MAX),
		 falpha	(IDD_Gamma, min	(GlobalData.betagp, 0.5), -DBL_MAX, 0.9999999),
		 fsigma	(IDD_Sigma, GlobalData.sigmagp,	DBL_MIN, DBL_MAX),
		 fmu (IDD_Mu, GlobalData.mugp, -DBL_MAX, DBL_MAX),
		 fumin (IDD_Umin, GlobalData.RuinUmin, 0.0, DBL_MAX),
		 fmaxt (IDD_TimeHorizont, GlobalData.RuinTime, 0.0, DBL_MAX);
    TIntField	 fNrSims (IDD_NrSims, GlobalData.RuinNrSims, 1,	INT_MAX);

    TAlterField  a (GlobalData.RuinProcSel, 2, IDD_PoissonClaim, IDD_PLClaim);
    TDoubleField pa (IDD_PolayAlpha, GlobalData.PolyaProcAlpha, DBL_MIN, DBL_MAX),
		 ps (IDD_PolyaSigma, GlobalData.PolyaProcSigma, DBL_MIN, DBL_MAX);

    TEditField	 fName (IDD_FileName, GlobalData.RuinName, 80);
    d << flambda << frho << fbeta << falpha << fsigma << fmu <<	fumin
      << fNrSims << fName << fmaxt << a << pa << ps;
    if (d.Handle (parent) == IDOK) {
	GlobalData.RiskLambda =	flambda.GetValue ();
	GlobalData.RiskRho = frho.GetValue ();
	GlobalData.RiskBeta = fbeta.GetValue ();
	GlobalData.betagp = falpha.GetValue ();
	GlobalData.sigmagp = fsigma.GetValue ();
	GlobalData.mugp	= fmu.GetValue ();
	GlobalData.RuinUmin = fumin.GetValue ();
	GlobalData.RuinTime = fmaxt.GetValue ();
	int NrSims = GlobalData.RuinNrSims = fNrSims.GetValue ();
	GlobalData.RuinName = fName.ReadBuf ();

	GlobalData.RuinProcSel = a.Selector ();
	GlobalData.PolyaProcAlpha = pa.GetValue ();
	GlobalData.PolyaProcSigma = ps.GetValue ();

	TGPRuinSimulation d (falpha.GetValue (), fmu.GetValue (), fsigma.GetValue (),
				 fNrSims.GetValue (), frho.GetValue (),	fbeta.GetValue (), flambda.GetValue (),
				 pa.GetValue (), ps.GetValue (), a.Selector (),
				 fmaxt.GetValue ());
	d.SetU (fumin.GetValue (), fumin.GetValue (), 1);
	double *x = new	double [NrSims];
	TCancelBox *CancelBox =	new TCancelBox (parent, IDD_CancelBox);
	CancelBox->OpenDialog (parent);
	CancelBox->SetTitel ("Simulate ruin time");
	int i;
	for (i = 0; i <	NrSims && !CancelBox->IsCancelled (); i++) {
	    CancelBox->CheckBoxMessages	();
	    d.SingleSimulation ();
	    x [i] = d.GetTime (0);
	    CancelBox->SetText (form (10, "%4.1f", i / double (NrSims) * 100.0));
	}
	CancelBox->CloseBox ();
	TList s;
	s.Append (new TStringEl	("Ruin time simulation"));
	TUnivariateSample *p = new TUnivariateSample;
	p->SetName ("Ruin time simulation");
	p->SetDescription (&s);
	p->SetData (i, x, fName.ReadBuf	());
	s.DeleteAll (FALSE);
	delete [] x;
    }
}

#endif


double SimulateSingleGPRuinTime	(double	*p) {
    //TGPRuinSimulation::TGPRuinSimulation (double gamma0, double mu0, double	sigma0,	int NrSim,
    //				      double rho, double beta, double lambda, double polalpha,
    //				      double polsigma, BOOL PolyaLundbergSim, double maxT)
    // p [0] - s0	p [1] = rho	p [2] = beta	p [3] = lambda
    // p [4] - T 	p [5] = gamma	p [6] = mu	p [7] = sigma
/*
    TGPRuinSimulation d	(p [5],	p [6], p [7], 1, p [1],	p[2], p	[3], 0.0, 0.0, FALSE, p [4]);
    d.SetU (p [0], p [0], 1);
    d.SingleSimulation ();
    return d.GetTime (0);
*/
    double e = p [3] * (p [6] + p [7] / (1.0 - p [5])),
	   s = p [0],
	   t = 0.0,
	   c = 0.0;
    while (c <= s && t < p [4]) {
	c += vonMisesGPDQF (Random (), p [5], p [7], p [6]);
	t -= log (1.0 - Random ()) / p [3];
	s = p [0] + e * (t + p [1] * pow (t, p [2]));
    }
    return min (t, p [4]);
}

