/*  sampbase.hpp

    Base class for all sample types.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 1999 Xtremes Group, Siegen.
*/

#ifndef SAMPBASE_HPP
#define SAMPBASE_HPP

#include <float.h>

#include "list.hpp"
#include "textfile.hpp"

#ifndef TERMINAL
#include "function.hpp"
#include "display.hpp"
#else
typedef int *TWindow;
#endif

enum TSampleType { UNIVARIATE_SAMPLE,
                   TIMESERIES_SAMPLE,
                   GROUPED_SAMPLE,
                   CENSORED_SAMPLE,
                   FREQUENCY_SAMPLE,
                   MULTIVARIATE_SAMPLE,
				   COPULA_SAMPLE,
                   UNIVARIATE_REQUIRED,
                   BIVARIATE_BASE_REQUIRED,
                   UNIVARIATE_GROUPED_REQUIRED };

extern char *SampleHeaders [];

extern int MAXSAMPLESIZE;
#define ACADEMICSIZE 1000

// Comparison function for quicksort

#ifndef _USERENTRY
#define _USERENTRY
#endif

extern int _USERENTRY realcomp (const void *a, const void *b);

/*******************************************************************/
/*                 Base class for all data set types               */
/*******************************************************************/

class TSample : public TListEl {
    public:
	TSample ();
	virtual ~TSample ();
	virtual TSampleType GetSampleType () = 0;
	virtual String Name ();                            // with filename
	String GetName ()               { return name; }   // without filename
	String GetFileName ()           { return filename; }
	void SetName (String t);
	void SetFileName (String s);
	TList *GetDescription ();
	void SetDescription (TList *);  // takes over entries in list

	BOOL ReadSampleSize (TTextFileReader &f, int &EntrySize);
	BOOL LoadData (TTextFileReader &);
	BOOL ReadDescription (TTextFileReader &);
	String GetReadMessage ();
	BOOL SaveData (TWindow *parent);

	BOOL IsDeleted ()               { return Deleted; }
	void SetDelete (BOOL f)         { Deleted = f; }
	String ModifyName (char ch);

	// methods to access data
	virtual int GetDimension () = 0;
	virtual int GetDataSize () = 0;
	virtual double GetData (int i, int j = 0) = 0;
	virtual String GetColumnName (int j) = 0;

	virtual void SetContamination (double d);       // for simulations
	virtual void SetSampleSize (int s, int k);

	virtual void PMLDistributeDataSet (String filename, String ParaName, int PlotMode, int UseClaimSize, double t, int EstimatorNr);
	virtual void PMLFunctionalParameterTrafo (String filename);
	virtual BOOL PMLAddPriorities (String filename, int m, double *p);
	virtual void PMLAddDegrees (String filename);
	virtual void PMLSortPML (String filename);
	virtual void PMLNearestNeighbors (String filename, int k, int est, double gamma, double mu);
	virtual void PriceReturnTrafo (String filename, int TreatMondays);
	virtual void Symmetrize (String Filename, BOOL Mean, double val);
	virtual void DateTransformation (String filename);
        virtual void OrderSample (String filename, int Column);
	virtual void TruncateSample (String filename, double threshold, int Column);
	virtual void CalculateHydrograph (int r0, double c, int r1, double * &x, double * &y, int &count);
	virtual void AnalysePartialDurationSeries (double u, int k, String fn, double tmin, double tmax, int tpoints);

	void NotApplicableMessage (TWindow *parent);
#ifndef TERMINAL
	virtual BOOL Create (TWindow *parent);          // creates, saves and registers data set
	virtual BOOL CreateMemory (TWindow *parent);    // only creates data set
	virtual void ListData (TMDIFrame *) = 0;
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void Edit (TMDIFrame *);

	// Empirical plots

	virtual void PlotHistogram (TWindow *parent);
	virtual void PlotSampleQF (TWindow *parent);
	virtual void PlotSampleDF (TWindow *parent);
	virtual void PlotKernelDensity (TWindow *parent);
	virtual void PlotBoundedKernelDensity (TWindow *parent);
	virtual void PlotMeanExcessFunction (TWindow *parent);
	virtual void PlotMedianExcessFunction (TWindow *parent);
	virtual void PlotHazardFunction (TWindow *parent);
	virtual void PlotSampleACF (TWindow *parent);
	virtual void PlotSampleAutoCorrelation (TWindow *parent);
	virtual void PlotSampleDepFunction (TWindow *parent);
	virtual void PlotDTail1 (TWindow *parent);
	virtual void PlotDTail2 (TWindow *parent);
	virtual void PlotDTail3 (TWindow *parent);
	virtual void PlotDTail4 (TWindow *parent);
	virtual void PlotCTail1 (TWindow *parent);
	virtual void PlotCTail2 (TWindow *parent);
	virtual void PlotCTail3 (TWindow *parent);
	virtual void PlotCTail4 (TWindow *parent);
	virtual void PlotCParaC (TWindow *parent);
	virtual void PlotDParaC (TWindow *parent);
	virtual void PlotChi (TWindow *parent);
	virtual void PlotChiBarE (TWindow *parent);
	virtual void PlotChiBarP (TWindow *parent);
	virtual void PlotTimeTailDep(TWindow *parent);


    virtual void PlotSampleVar (TWindow *parent, int type, int distr);
	virtual void PlotPath (TWindow *parent);
	virtual void ScatterPlot2D (TMDIFrame *);
	virtual void ScatterPlot3D (TMDIFrame *);
    virtual void BoxPlot (TMDIFrame *);
    void PlotFunction (TWindow *parent, TFunction *f, String WindowName);

	virtual void ShowFunctionalParameters (TWindow *parent);
    virtual void GaussianTwoSampleTest (TWindow *parent);
    virtual void WilcoxonTest (TWindow *parent);
	// Transformations to other data sets

	virtual void TransformSample (TWindow *parent);
	virtual void ConverseSample (TWindow *parent);
	virtual void LogScaleSample (TWindow *parent);
	virtual void TruncateSample (TWindow *parent);
	virtual void ExtractMaxima (TWindow *parent);
	virtual void SaveBlocksSum (TWindow *parent);
	virtual void SaveMovingMaxima (TWindow *parent);
	virtual void SaveMovingSum (TWindow *parent);
	virtual void DeclusterSample (TWindow *parent, int mingap, double threshold);
	virtual void UserTransformation (TWindow *parent);
	virtual void OrderSample (TWindow *parent);
	virtual void FillMissing (TWindow *parent);
	virtual void CumulateSample (TWindow *parent);
	virtual void SelectComponent (TWindow *parent);
	virtual void PriceReturnTrafo (TWindow *parent);
	virtual void PMLAddPriorities (TWindow *parent);
	virtual void PMLAddDegrees (TWindow *parent);
	virtual void PMLSortPML (TWindow *parent);
	virtual void PMLFunctionalParameterTrafo (TWindow *parent);
	virtual void PMLDistributeDataSet (TWindow *parent);
	virtual void PMLNearestNeighbors (TWindow *parent);
	virtual void Symmetrize (TWindow *parent);
	virtual void DateTransformation (TWindow *parent);
	virtual void CalculateHydrograph (TWindow *parent);
	virtual void AnalysePartialDurationSeries (TWindow *parent);
        virtual void FrequenciesToRelative (TWindow *parent);

#endif
	// Sample density, df, qf and hazard function

	virtual double SampleQf (double x);
	virtual double SampleDf (double x);
	virtual double SampleDen (double x, double beta, int KNo);
	virtual double SampleHazard (double t, double beta);
	virtual double SampleMean ();
	virtual double SampleVariance ();
	virtual double SampleSkewness ();
        virtual double MeanClusterSize (int mingap, double threshold);


    protected:
#ifndef TERMINAL
	virtual unsigned EditParameter (TWindow *parent)= 0;
	virtual BOOL CreateData (TWindow *parent) = 0;
	virtual void CreateInteractive (TWindow *parent);
#endif

	virtual BOOL GetMemory () = 0;
	virtual void SortData ();
	virtual void CreateDescription ();      // called after CreateData ()
	void AddDescriptionLine (String s);
	virtual String GetHeader () = 0;

	String CutPath (String);

	virtual BOOL WriteData (FILE *) = 0;
	virtual BOOL ReadData (TTextFileReader &) = 0;
	BOOL WriteHeader (FILE *);
	BOOL ReadHeader (TTextFileReader &);

	TList Description;
	String name, filename;
	String ReadMessage;
    private:
	BOOL Deleted;
};

#ifndef TERMINAL

// handle generic sample methods

class TSampleListWindow : public TDisplayWindow {
    typedef TDisplayWindow inherited;
    public:
	TSampleListWindow (TSample *s, TMDIFrame *, String Titel = "");
    protected:
	virtual void NotifyMenuEvent (int id);
	virtual	void NotifyEvent (int ev);
        TSample *s;
};

#endif

#endif

