/*  sample.hpp

    Class definition for sample types.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 1999 Xtremes Group, Siegen.
*/

#ifndef SAMPLE_HPP
#define SAMPLE_HPP

#include <float.h>

#include "sampbase.hpp"
#include "point.hpp"

#ifndef TERMINAL
#include "distr.hpp"
#include "global.hpp"
#include "listdlg.hpp"
#include "mdi.hpp"
//#include "visual.hpp"
#endif

#define LOADERRNUM "Error in number"
#define INSUFFMEM "Insufficient memory to load data set"
#define ACADEMIC "The academic version does not support more that 1000 data points"

#define MAXKERN 4

/*******************************************************************/
/*			Univariate data sets			   */
/*******************************************************************/

class TUnivariateBaseSample : public TSample {
    public:
	TUnivariateBaseSample ();
	virtual ~TUnivariateBaseSample ();

	double operator [] (int);
	double GetRawData (int Nr) { return RawDataPtr [Nr]; }

	virtual int GetDataSize ();
        virtual int GetDimension ();
        virtual double GetData (int i, int j = 0);
        virtual String GetColumnName (int j);

#ifndef TERMINAL
	virtual void ListData (TMDIFrame *);
#endif
    protected:
	int DataSize, BlockSize, AllocatedSize;
	double *DataPtr, *RawDataPtr;	// sorted, unsorted data
	virtual BOOL GetMemory ();
        void Sort (int l, int r);
	virtual void SortData ();
	BOOL Append (double);
	virtual BOOL ReadData (TTextFileReader &);
	virtual BOOL WriteData (FILE *);
};

class TUnivariateSample : public TUnivariateBaseSample {
    public:
	TUnivariateSample ();

	virtual TSampleType GetSampleType ();
	void SetData (int n, double *, BOOL MakeCopy = true);
	void SetData (int n, double *, String fn);

	virtual double SampleQf (double x);
	virtual double SampleDf (double x);
	virtual double SampleDen (double x, double beta, int KNo);
	virtual double SampleHazard (double t, double beta);
	virtual double SampleMean ();
	virtual double SampleVariance ();
	virtual double SampleSkewness ();

        // supporting routines for kernel density estimation
	double DKern (double z, int KNo);       // Available kernels (MAXKERN is defined in global.hpp)
	double bw1 (double * y, int n, int KNo, double t, double a);
	double bw2 (double *y, int n, int KNo, double a);
	double bw3 (double *y, int n, int KNo, double sl, double int1, double int2);
	double OptBandW (int KNo, double sl = 0.02, double int1 = 0.0, double int2 = 1.0);
	double BoundDenNone  (int Kernel, double t, double bandwidth);
	double BoundDenLeft  (int Kernel, double t, double bandwidth, double beta0);
	double BoundDenRight (int Kernel, double t, double bandwidth, double beta1);
	double BoundDenBoth  (int Kernel, double t, double bandwidth, double beta0, double beta1);
	BOOL IsEstimated (int Kernel);
	double GetBandWidth (int Kernel);
	void SetBandWidth (int Kernel, double);

#ifndef TERMINAL
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void ShowFunctionalParameters (TWindow *parent);

	virtual void PlotSampleQF (TWindow *parent);
	virtual void PlotSampleDF (TWindow *parent);
	virtual void PlotKernelDensity (TWindow *parent);
	virtual void PlotBoundedKernelDensity (TWindow *parent);
	virtual void PlotMeanExcessFunction (TWindow *parent);
	virtual void PlotMedianExcessFunction (TWindow *parent);
	virtual void PlotHazardFunction (TWindow *parent);
	virtual void PlotPath (TWindow *parent);
    virtual void BoxPlot (TMDIFrame *);
	virtual void PlotTimeTailDep(TWindow *parent);

	virtual void UserTransformation (TWindow *parent);
	virtual void TransformSample (TWindow *parent);
	virtual void TruncateSample (TWindow *parent);
	virtual void ConverseSample (TWindow *parent);
	virtual void ExtractMaxima (TWindow *parent);
	virtual void SaveBlocksSum (TWindow *parent);
	virtual void SaveMovingMaxima (TWindow *parent);
	virtual void SaveMovingSum (TWindow *parent);
	virtual void OrderSample (String, int);
	virtual void OrderSample (TWindow *parent);
	virtual void CumulateSample (TWindow *parent);
	virtual void Symmetrize (TWindow *parent);
#endif
	virtual void Symmetrize (String Filename, BOOL Mean, double val);
	virtual void TruncateSample (String filename, double threshold, int Column);

	virtual void SetSampleSize (int s, int k);


    protected:
#ifndef TERMINAL
	virtual unsigned EditParameter (TWindow *parent);
	virtual BOOL CreateData (TWindow *parent);
#endif
	virtual String GetHeader ();
    private:
	double BandWidth [MAXKERN];
	BOOL Estimated [MAXKERN];
#ifndef TERMINAL
	TSample *ConvertToTimeSeries (TWindow *parent);
        TSample *ConvertToFrequency (TWindow *parent);
#endif
};


#ifndef TERMINAL
class TIIDSample : public TUnivariateSample {
    public:
	TIIDSample ();
	virtual ~TIIDSample ();
	virtual double DrawRV () = 0;
	virtual TFunction *GetDF () = 0;
    protected:
	virtual BOOL CreateData (TWindow *parent);
	virtual void CreateInteractive (TWindow *parent);
};
#endif

/*******************************************************************/
/*			Frequency data sets 	   		   */
/*******************************************************************/

class TAnimatorWindow;

class TFrequencySample : public TSample {
    public:
	TFrequencySample ();
	virtual ~TFrequencySample ();
	int GetPoint (int);
	int GetFrequency (int);
	virtual TSampleType GetSampleType ();
	BOOL AddPoint (int p);

        virtual int GetDimension ();
        virtual int GetDataSize ();
	virtual double GetData (int i, int j = 0);
        virtual String GetColumnName (int j);

        void SetData (int n, int *t, int *c, String fn); // consumes data pointers
#ifndef TERMINAL
	virtual void ListData (TMDIFrame *);
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void PlotHistogram (TWindow *parent);
	virtual void ShowFunctionalParameters (TWindow *parent);
        virtual void FrequenciesToRelative (TWindow *parent);
#endif
    protected:
#ifndef TERMINAL
	TSample *ConvertToUnivariate (TWindow *parent);
	virtual unsigned EditParameter (TWindow *parent);
	virtual BOOL CreateData (TWindow *parent);
        virtual TAnimatorWindow *GetAnimatorWindow ();
#endif
	virtual String GetHeader ();
	virtual BOOL GetMemory ();
	virtual BOOL ReadData (TTextFileReader &);
	virtual BOOL WriteData (FILE *f);
	BOOL Append (int, int);
	int DataSize, Allocated, AllocatedSize;
	int *PointData, *FrequencyData;
};

class TFrequencyIIDSample : public TFrequencySample {
    public:
//	TFrequencySample ();
//	virtual ~TFrequencySample ();
	virtual double DrawRV () = 0;
#ifndef TERMINAL
	virtual TFunction *GetDF () = 0;
#endif
    protected:
#ifndef TERMINAL
	virtual BOOL CreateData (TWindow *parent);
	virtual void CreateInteractive (TWindow *parent);
#endif
	int TotalSize;
};

/*******************************************************************/
/*		Bivariate data sets and time series		   */
/*******************************************************************/

class TBivariateBaseSample : public TSample {
    public:
	TBivariateBaseSample ();
	virtual ~TBivariateBaseSample ();
	int GetRegisteredSize ();
	void SetRegisteredSize (int n)		{ RegisteredSize = n; }
	TDblPoint operator [] (int);

	virtual int GetDimension ();
	virtual int GetDataSize ();
	virtual double GetData (int i, int j = 0);
	virtual String GetColumnName (int j);

	void SetData (int n, double *point, double *data, String fn);
	void SetValues (double threshold);		// for extracting
	virtual void TruncateSample (String filename, double threshold, int Column);


	virtual TBivariateBaseSample *DuplicateType () = 0;

#ifndef TERMINAL
	virtual void ListData (TMDIFrame *);
	virtual unsigned EditParameter (TWindow *parent);
	virtual BOOL CreateData (TWindow *parent);
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void TransformSample (TWindow *parent);
	virtual void ConverseSample (TWindow *parent);
	virtual void LogScaleSample (TWindow *parent);
	virtual void UserTransformation (TWindow *parent);
	virtual void TruncateSample (TWindow *parent);
	virtual void ScatterPlot2D (TMDIFrame *);
	virtual void ShowFunctionalParameters (TWindow *parent);
        virtual void BoxPlot (TMDIFrame *);
#endif
    protected:
#ifndef TERMINAL
	TSample *ConvertToGrouped (TWindow *parent);
	TSample *ConvertToUnivariate (TWindow *parent);
	TSample *ConvertToCopulaScale (TWindow *parent);
#endif
	int DataSize, RegisteredSize, AllocatedSize;
	BOOL Append (double, double);
	TDblPoint *DataPtr;
	virtual BOOL GetMemory ();
	virtual BOOL ReadData (TTextFileReader &);
	virtual BOOL WriteData (FILE *f);
	double threshold;
};

class TTimeSeriesSample : public TBivariateBaseSample {
    public:
	virtual TSampleType GetSampleType ();
	virtual TBivariateBaseSample *DuplicateType ();
	double MeanClusterSize (int mingap, double threshold);
	virtual void CalculateHydrograph (int r0, double c, int r1, double * &x, double * &y, int &count);
	virtual void AnalysePartialDurationSeries (double u, int k, String fn, double tmin, double tmax, int tpoints);
#ifndef TERMINAL
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void DeclusterSample (TWindow *parent, int mingap, double threshold);
	virtual void PlotSampleACF (TWindow *parent);
	virtual void PlotSampleAutoCorrelation (TWindow *parent);
        virtual void PlotSampleVar (TWindow *parent, int type, int distr);
	virtual void CalculateHydrograph (TWindow *parent);
	virtual void AnalysePartialDurationSeries (TWindow *parent);
#endif
    protected:
#ifndef TERMINAL
	TSample *ConvertToMultivariate (TWindow *parent);
#endif
	virtual String GetHeader ();
    private:
	// Hydrograph
	int SearchMaxIndex (int t, double val, int sign0, int maxgap);
	void Entry (double x, double y, double * &xp, double * &yp, int &count, int &allocated);
	void TestEntry (double x, double y, double * &xp, double * &yp, int postgap, int &gapcount, int &gapallocated);
	void DetectCluster (int sign0, int maxgap, double c, int postgap, double * &xgap, double * &ygap, int &gapcount, int &gapallocated);

	// Partial Duration Series
	int *nc, nPeriods, l;
	double **x, *gamma, *sigma, *z;
	void CumulateData ();
        void FreeMemory ();
	void Entry (double *z, int &Nr, int c);
	void Estimate (int i, int k, double u);
	void CalcEstimates (int k, double u);
	void StoreEstimates (String fn);
	double f (double theta, double u);
	void MakePlots (double u, double tmin, double tmax, int tpoints);
};

/*******************************************************************/
/*			Multivariate data sets 	   		   */
/*******************************************************************/

class TMultivariateSample : public TSample {
    public:
	TMultivariateSample ();
	virtual ~TMultivariateSample ();
	double operator () (int, int);	// (line, column)

	virtual int GetDimension ();
	virtual int GetDataSize ();
	virtual double GetData (int i, int j = 0);
	virtual String GetColumnName (int j);
    virtual BOOL LoadUntypedData (TTextFileReader &);

	void SetData (int Dimension, String fn, int Size);
	void SetData (int Dimension, int Size);
	void SetFilename (String fn, String des);
	void SetColumn (int c, double *data, String Header);
	void SetColumn (int c, double *data, int n, String Header);

	virtual TSampleType GetSampleType ();

	virtual void PMLDistributeDataSet (String filename, String ParaName, int PlotMode, int UseClaimSize, double t, int EstimatorNr);
	virtual void PMLFunctionalParameterTrafo (String filename);
	virtual BOOL PMLAddPriorities (String filename, int m, double *p);
	virtual void PMLNearestNeighbors (String filename, int k, int est, double gamma, double mu);
	virtual void PriceReturnTrafo (String filename, int TreatMondays);
	virtual void DateTransformation (String filename);
	virtual void PMLAddDegrees (String filename);
	virtual void PMLSortPML (String filename);
	virtual void TruncateSample (String filename, double threshold, int Column);

	virtual void OrderSample (String filename, int Column);

#ifndef TERMINAL
	virtual void ListData (TMDIFrame *);
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void ShowFunctionalParameters (TWindow *parent);
	virtual void FillMissing (TWindow *parent);
	virtual void SelectComponent (TWindow *parent);

	virtual void UserTransformation (TWindow *parent);
	virtual void TransformSample (TWindow *parent);
	virtual void TruncateSample (TWindow *parent);
	virtual void ConverseSample (TWindow *parent);
	virtual void LogScaleSample (TWindow *parent);
	virtual void ExtractMaxima (TWindow *parent);
	virtual void SaveBlocksSum (TWindow *parent);
	virtual void SaveMovingMaxima (TWindow *parent);
	virtual void SaveMovingSum (TWindow *parent);
	virtual void OrderSample (TWindow *parent);
	virtual void CumulateSample (TWindow *parent);
	virtual void PriceReturnTrafo (TWindow *parent);
	virtual void PMLFunctionalParameterTrafo (TWindow *parent);
	virtual void PMLDistributeDataSet (TWindow *parent);
	virtual void PMLAddDegrees (TWindow *parent);
    virtual void PMLSortPML (TWindow *parent);
    virtual void DateTransformation (TWindow *parent);
    virtual void PMLNearestNeighbors (TWindow *parent);

    virtual void GaussianTwoSampleTest (TWindow *parent);
    virtual void WilcoxonTest (TWindow *parent);
#endif
    protected:
#ifndef TERMINAL
	TSample *ConvertToUnivariate (TWindow *parent);
	TSample *ConvertToTimeSeries (TWindow *parent);
	TSample *ConvertToGrouped (TWindow *parent);
	TSample *ConvertToCopulaScale (TWindow *parent);
	BOOL ConvertBivariateBase (TWindow *parent, TBivariateBaseSample *p);
	virtual unsigned EditParameter (TWindow *parent);
	virtual BOOL CreateData (TWindow *parent);
	virtual void ScatterPlot2D (TMDIFrame *);
	virtual void ScatterPlot3D (TMDIFrame *);
    virtual void BoxPlot (TMDIFrame *);
    virtual void PlotSampleDepFunction (TWindow *parent);
	virtual void PlotDTail1 (TWindow *parent);
	virtual void PlotDTail2 (TWindow *parent);
	virtual void PlotDTail3 (TWindow *parent);
	virtual void PlotDTail4 (TWindow *parent);
	virtual void PlotCTail1 (TWindow *parent);
	virtual void PlotCTail2 (TWindow *parent);
	virtual void PlotCTail3 (TWindow *parent);
	virtual void PlotCTail4 (TWindow *parent);
	virtual void PlotCParaC (TWindow *parent);
	virtual void PlotDParaC (TWindow *parent);
	virtual void PlotChi (TWindow *parent);
	virtual void PlotChiBarE (TWindow *parent);
	virtual void PlotChiBarP (TWindow *parent);

	
	virtual void PMLAddPriorities (TWindow *parent);

#endif

	// methods for price-return transformation
	void Entry (int day, int i);
        void TradingDays ();
	void FillGap ();
	void RemoveGap ();
        void FillDate ();
	int startday, maxday;
	double *plog, **x, *y;

	void Sort (double *, int *, int l, int r);
	void AllocatePointers ();
	BOOL Append (double *);
	virtual String GetHeader ();
	int DataSize, Dimension, AllocatedSize;
	double **DataPtr;
	String *ColumnNamePtr;
	virtual BOOL GetMemory ();
	virtual BOOL ReadData (TTextFileReader &);
	virtual BOOL WriteData (FILE *f);

        BOOL CheckDates ();
        BOOL CheckReturns ();
        
	int Count, Allocated;
	int PMLEstCount, PMLEstAllocated;
	double *PMLgroupNr, *PMLgroup, *PMLbeta, *PMLmu, *PMLsigma, *PMLu, *PMLd;
	void PMLEntry (double center, double claimmean, double claimvar, double degmean, double degvar);
	void PMLDistributeMakeDataSet (String filename, int group, double s0, double s1, int PlotMode, double t, int EstimatorNr, BOOL UseClaimSize);
	void PMLDistributeEntry (double val);
	void PMLEstimatorEntry (double groupnr, double t, double b, double m, double s, double u, double d);
};

/*******************************************************************/
/*			Grouped data sets			   */
/*******************************************************************/

class TGroupedSample : public TSample {
    public:
	TGroupedSample (TUnivariateSample * = NULL);
	virtual ~TGroupedSample ();
	virtual TSampleType GetSampleType ();
	int GetGroupCount () 		{ return GroupCount; }
	double GetFrequency (int);
	double GetPartition (int);
	double EmpQF (double q);
	virtual double SampleQf (double x)	{ return EmpQF (x); }

        virtual int GetDimension ();
        virtual int GetDataSize ();
        virtual double GetData (int i, int j = 0);
        virtual String GetColumnName (int j);

	void SetData (int n, double *par, double *count, String fn);

	virtual double SampleHazard (double t, double beta);

#ifndef TERMINAL
	virtual void ListData (TMDIFrame *);
	virtual TSample *ConvertTo (TSampleType st, TWindow *parent);
	virtual void ConverseSample (TWindow *parent);
	virtual void PlotHistogram (TWindow *parent);
	virtual void PlotSampleQF (TWindow *parent);
	virtual void PlotSampleDF (TWindow *parent);
	virtual void PlotMeanExcessFunction (TWindow *parent);
	virtual void PlotMedianExcessFunction (TWindow *parent);
	virtual void PlotHazardFunction (TWindow *parent);
#endif
    protected:
	virtual BOOL ReadData (TTextFileReader &);
	virtual BOOL WriteData (FILE *f);
	virtual BOOL GetMemory ();
	virtual String GetHeader ();

#ifndef TERMINAL
	virtual unsigned EditParameter (TWindow *parent);
	virtual BOOL CreateData (TWindow *parent);
#endif
    private:
	BOOL Append (double freq, double par);
	int GroupCount, AllocatedSize;
	double *Frequencies, *Partition;
	TUnivariateSample *p;
    	static double *Tab;
        static int TabCount;
};

#ifndef TERMINAL
class TSmoothedSample : public TUnivariateSample {
    public:
	TSmoothedSample (TGroupedSample *p);
	virtual unsigned EditParameter (TWindow *parent);
	virtual BOOL CreateData (TWindow *parent);
    protected:
	TGroupedSample *p;
};
#endif

#endif
