/*  scanner.cpp
 *
 *  Implementation of StatPascal scanner.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998 Xtremes Group, Siegen.
 */

#include <math.h>
#include <stdlib.h>
#include <ctype.h>

#include "errors.hpp"
#include "list.hpp"
#include "scanner.hpp"

/*********************************************************/
/*                   Scanner                             */
/*********************************************************/

char *TScanner::Names [] = {
    "begin", "end", "if", "then", "else", "for", "to", "downto", "do", "const", "type", "var",
    "while", "repeat", "until", 
    "record", "array", "set", "vector", "matrix", "of",
    "procedure", "function", "return", "case", "range", "nil",
    "forward", "unit", "interface", "implementation", "uses", "label", "goto",

    "estimator", "returning", "program", "component", "inports", "outports",

    "in", "and", "or", "xor", "not", "div", "mod"
};

TScanner::TScanner () {
    buffer = NULL;
    f = NULL;
    validFlag = TRUE;
}

TScanner::TScanner (String fn0) {
    buffer = new char [30000];	// !!!! An Dateigroesse anpassen
    fn = fn0;
    f = fopen (fn, "rt");
    validFlag = !!f;
    if (f) {
        int l = fread (buffer, 1, 29999, f);
        buffer [l] = '\0';
        fclose (f);
        fpos = startpos = buffer;
        Line = 1;
    }
}

TScanner::~TScanner () {
    if (buffer) delete [] buffer;
}

BOOL TScanner::Try (TToken c) {
    if (GetToken () == c) {
	GetNextToken ();
	return TRUE;
    }
    return FALSE;
}

void TScanner::SetProgram (const char *buf, String FileName) {
    startpos = fpos = buf;
    Line = 1; fn = FileName;
}

void TScanner::Restart () {
    fpos = startpos;
    Line = 1;
}

void TScanner::SetTokenInc (TToken c) {
    cToken = c;
    fpos++;
}

void TScanner::ParseNumber () {
    BOOL Integer = TRUE;
    double z = 0.0, d = 1.0;
    int n = 0;
    int s = 0, e = 0;
    while (isdigit (*fpos)) {
	z = z * 10.0 + *fpos - '0';
	n = n * 10 + *fpos++ - '0';
    }
    if (*fpos == '.' && *(fpos + 1) != '.') {
	Integer = FALSE;
	fpos++;
	while (isdigit (*fpos))
	    z = z + (d /= 10.0) * (*fpos++ - '0');
    }
    if (*fpos == 'e' || *fpos == 'E') {
	Integer = FALSE;
	fpos++;
	if (*fpos == '+' || *fpos == '-') s = *fpos++ == '-';
	while (isdigit (*fpos)) e = e * 10 + *fpos++ - '0';
	if (s) e = -e;
    }
    if (e) z *= pow (10.0, e);
    if (Integer && fabs (z - n) < 1e-3) {
	iValue = n;
	cToken = INTCONST;
    } else {
	cToken = DBLCONST;
	dValue = z;
    }
}

void TScanner::ParseIdentifier () {
    char buf [32], *p = buf;
    while (isalpha (*fpos) || isdigit (*fpos) || *fpos == '_') {
	if (p - buf < 30)
            if (*fpos >= 'A' && *fpos <= 'Z')
                *p++ = (char) (*fpos | 32);    // make small letter
            else *p++ = (char) (*fpos);
	fpos++;
    }
    *p = '\0';
    sValue = buf;
    TToken f;
    for (f = BEGIN; f < ADD && strcmp (buf, Names [(int) f]);
	 f = (TToken) ((int) f + 1));
    cToken = (f == ADD) ? IDENTIFIER : f;
}

void TScanner::ParseString (char TermChar) {
    char buf [128], *p = buf;
    fpos++;
    while (*fpos && *fpos != TermChar) {
	if (p - buf < 126)
	    if (*fpos == '\\' && *(fpos + 1) == 'n') { *p++ = '\n'; fpos++; }
	    else
	    *p++ = *fpos;
	fpos++;
    }
    *p = '\0';
    if (*fpos == TermChar) fpos++;
    if (p - buf == 1 && TermChar == '\'') {
	cToken = CHARCONST;
	iValue = *buf;
    } else {
	cToken = STRCONST;
	sValue = buf;
    }
}

void TScanner::ParseOperator () {
    switch (*fpos++) {
	case '*':
	    if (*fpos == '*') { fpos++; cToken = POW; }
	    else cToken = MUL;
	    break;
	case '.':
	    if (*fpos == '.') { fpos++; cToken = TWOPOINT; }
	    else cToken = ENDPOINT;
	    break;
	case '<':
	    if (*fpos == '=') { fpos++; cToken = LESSEQUAL; }
	    else if (*fpos == '>') { fpos++; cToken = NOTEQUAL; }
	    else cToken = LESSTHAN;
	    break;
	case '>':
	    if (*fpos == '=') { fpos++; cToken = GREATEREQUAL; }
	    else cToken = GREATERTHAN;
	    break;
	case '=':
	    if (*fpos == '=') fpos++;
	    cToken = EQUAL;
	    break;
	case '!':
	    if (*fpos == '=') {
		fpos++;
		cToken = NOTEQUAL;
	    } else cToken = FAC;
	    break;
	case ':':
	    if (*fpos == '=') {
		fpos++;
		cToken = DEFINE;
	    }
	    else cToken = COLON;
	    break;
    }
}

BOOL TScanner::SkipWhiteSpace () {
    BOOL Skipped = FALSE;
    while (*fpos == 13 || *fpos == 10 || *fpos == ' ' || *fpos == '\t') {
	if (*fpos == 10) Line++;
	fpos++;
	Skipped = TRUE;
    }
    return Skipped;
}

BOOL TScanner::SkipComment () {
    if (*fpos == '(' && *(fpos + 1) == '*') {
	fpos += 2;
	BOOL Done = FALSE;
	while (!Done && *fpos) {
	    if (*fpos == 10) Line++;
	    if (*fpos == '*' && *(fpos + 1) == ')') {
		fpos += 2; Done = TRUE;
	    } else fpos++;
	}
	return TRUE;
    }
    if (*fpos == '{') {
	fpos++;
	BOOL Done = FALSE;
	while (!Done && *fpos) {
	    if (*fpos == '\n') Line++;
	    Done = *fpos++ == '}';
	}
	return TRUE;
    }	
    if (*fpos == '%') {
	while (*fpos && *fpos != 10) fpos++;
	if (*fpos == 10) { fpos++; Line++; }
	return TRUE;
    }
    return FALSE;
}


void TScanner::GetNextToken () {
    while (SkipWhiteSpace () || SkipComment ());
    if (isalpha (*fpos)) ParseIdentifier ();
    else if (isdigit (*fpos) || (*fpos == '.' && isdigit (*(fpos + 1)))) ParseNumber ();
    else switch (*fpos) {
	case '+' : SetTokenInc (ADD); break;
	case '-' : SetTokenInc (SUB); break;
	case '/' : SetTokenInc (DIV); break;
	case '^' : SetTokenInc (DEREF); break;
	case '(' : SetTokenInc (BRACKETOPEN); break;
        case ')' : SetTokenInc (BRACKETCLOSE); break;
        case '[' : SetTokenInc (SQUAREBRACKETOPEN); break;
        case ']' : SetTokenInc (SQUAREBRACKETCLOSE); break;
	case ',' : SetTokenInc (COMMA); break;
//      case '?' : SetTokenInc (QUESTION); break;
	case ';' : SetTokenInc (SEMICOLON); break;
	case '<' :
	case '>' :
	case '=' :
	case ':' :
	case '*' :
	case '.' :
	case '!' : ParseOperator (); break;
	case '"' :
	case '\'': ParseString (*fpos); break;
	case '\0': cToken = TERMINATOR; break;
	default  : cToken = ERROR_TOKEN; fpos++;
    }
}
