/*  symbols.cpp
 *
 *  Symbol table of StatPascal compiler.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998, 2000 Xtremes Group, Siegen.
 */

#include <stdlib.h>
#include <limits.h>

#include "errors.hpp"
#include "symbols.hpp"
#include "compiler.hpp"

/********************************************************************
                        Handling of data types
*********************************************************************/

TType::TType (String Name) {
    name = Name;
}

TType::~TType () {
}

String TType::Name () {
    return name;
}

void TType::DeleteType () {
    delete this;
}

int TType::GetSize () {
    return 1;
}

BOOL TType::IsEnumerated () {
    return FALSE;
}

BOOL TType::IsSubrange () {
    return FALSE;
}

BOOL TType::IsArray () {
    return FALSE;
}

BOOL TType::IsRecord () {
    return FALSE;
}

BOOL TType::IsRoutine () {
    return FALSE;
} 

BOOL TType::IsReference () {
    return FALSE;
} 

BOOL TType::IsPointer () {
    return FALSE;
}

BOOL TType::IsSet () {
    return FALSE;
}

BOOL TType::IsVector () {
    return FALSE;
}

BOOL TType::IsMatrix () {
    return FALSE;
}

BOOL TType::ContainsVector () {
    return FALSE;
}

TType *TType::GetBaseType () {
    return NULL;
}

TType *TType::GetIndexType () {
    return NULL;
}

void TType::GetIndexRange (int &, int &) {
}

void TType::CodeConstructor (int line, TCompiler *) {
}

void TType::CodeDestructor (int line, TCompiler *) {
}

void TType::CodeLoad (int line, TCompiler *c) {
    c->OutCode (line, LOADIND, GetSize ());
}
    
void TType::CodeSave (int line, TCompiler *c) {
    c->OutCode (line, STOREIND, GetSize ());
}

void TType::CodePush (int line, TCompiler *c) {
    c->OutCode (line, PUSH, GetSize ());
}

TEnumerated::TEnumerated (String s) 
 : TType (s) {
}

String TEnumerated::Name () {
    String s = inherited::Name ();
    if (s != String ("")) return s;
    return "Enumeration " & form (30, "(%d, %d)", a, b);
}

BOOL TEnumerated::IsEnumerated () {
    return TRUE;
}

void TEnumerated::SetIndexRange (int a0, int b0) {
    a = a0; b = b0;
}

void TEnumerated::GetIndexRange (int &a0, int &b0) {
    a0 = a; b0 = b;
}


TSubrange::TSubrange (TType *t, int a, int b) {
    BaseType = t;
    SetIndexRange (a, b);
}

String TSubrange::Name () {
    int a, b;
    GetIndexRange (a, b);
    return "Subrange " & form (30, "%ld..%ld of ", a, b) & BaseType->Name ();
}

BOOL TSubrange::IsSubrange () {
    return TRUE;
}

TType *TSubrange::GetBaseType () {
    return BaseType;
}


TSet::TSet (TType *BaseType0) {
    BaseType = BaseType0;
}

String TSet::Name () {
    return "Set of " & BaseType->Name ();
}

BOOL TSet::IsSet () {
    return TRUE;
}

TType *TSet::GetBaseType () {
    return BaseType;
}

int TSet::GetSize () {
    return N_SET_CELLS;
}

TComplex::TComplex () {
}

String TComplex::Name () {
    return "complex";
}

int TComplex::GetSize () {
    return 2;
}

TArray::TArray (TType *IndexType0, TType *BaseType0) {
    IndexType = IndexType0;
    BaseType = BaseType0;
}

BOOL TArray::IsArray () {
    return TRUE;
}

TType *TArray::GetBaseType () {
    return BaseType;
}

TType *TArray::GetIndexType () {
    return IndexType;
}

void TArray::DeleteType () {
    BaseType->DeleteType ();
    IndexType->DeleteType ();
    delete this;
}

int TArray::GetSize () {
    int a, b;
    IndexType->GetIndexRange (a, b);
    return (int) (b - a + 1) * BaseType->GetSize ();
}

String TArray::Name () {
    return "array [" & IndexType->Name () &  "] of " & BaseType->Name ();
}

void TArray::CodeConstructor (int line, TCompiler *c) {
    int a, b;
    IndexType->GetIndexRange (a, b);
    c->OutCode (line, LOAD, b - a + 1);
    c->OutCode (line, LOADLS);
    int pos = c->GetPos ();
    BaseType->CodeConstructor (line, c);
    c->OutCode (line, LOOP, pos);
}

void TArray::CodeDestructor (int line, TCompiler *c) {
    int a, b;
    IndexType->GetIndexRange (a, b);
    c->OutCode (line, LOAD, b - a + 1);
    c->OutCode (line, LOADLS);
    int pos = c->GetPos ();
    BaseType->CodeDestructor (line, c);
    c->OutCode (line, LOOP, pos);
}

BOOL TArray::ContainsVector () {
    return BaseType->ContainsVector ();
}

void TArray::CodeLoad (int line, TCompiler *c) {
    int a, b;
    IndexType->GetIndexRange (a, b);
    if (ContainsVector ()) {
	Compiler->OutCode (line, LOAD, b - a + 1);
	Compiler->OutCode (line, LOADLS);
	int pos = Compiler->GetPos ();
	BaseType->CodeLoad (line, c);
	Compiler->OutCode (line, LOOP, pos);
    } else
	inherited::CodeLoad (line, c);
}

void TArray::CodeSave (int line, TCompiler *c) {
    int a, b;
    IndexType->GetIndexRange (a, b);
    if (ContainsVector ()) {
        Compiler->OutCode (line, LOAD, b - a + 1);
        Compiler->OutCode (line, LOADLS);
        int pos = Compiler->GetPos ();
        BaseType->CodeSave (line, c);
        Compiler->OutCode (line, LOOP, pos);
    } else
        inherited::CodeSave (line, c);
}

void TArray::CodePush (int line, TCompiler *c) {
    int a, b;
    IndexType->GetIndexRange (a, b);
    if (ContainsVector ()) {
        Compiler->OutCode (line, LOAD, b - a + 1);
        Compiler->OutCode (line, LOADLS);
        int pos = Compiler->GetPos ();
        BaseType->CodePush (line, c);
        Compiler->OutCode (line, LOOP, pos);
    } else
        inherited::CodePush (line, c);
}

TRecord::TRecordEl::TRecordEl (String Name, TType *t) {
    name = Name; type = t;
}

String TRecord::TRecordEl::Name () {
    return name;
}

TType *TRecord::TRecordEl::GetType () {
    return type;
}

TRecord::TRecord () {
}

BOOL TRecord::IsRecord () {
    return TRUE;
}

void TRecord::DeleteType () {
}

int TRecord::GetSize () {
    int size = 0;
    for (TRecordEl *p = (TRecordEl *) ComponentList.GetFirst (); p; p = (TRecordEl *) ComponentList.GetNext ())
	size += p->GetType ()->GetSize ();
    return size;
}

String TRecord::Name () {
    return "record ";
}

void TRecord::AppendComponent (String Name, TType *t)  {
    ComponentList.Append (new TRecordEl (Name, t));
}

BOOL TRecord::GetComponent (String Name, int &Offset, TType *&t, int SourceLine) {
    Offset = 0;
    TRecordEl *p;
    for (p = (TRecordEl *) ComponentList.GetFirst (); p; p = (TRecordEl *) ComponentList.GetNext ())
	if (p->Name () == Name) break;
	else Offset += p->GetType ()->GetSize ();
    if (!p && SourceLine) ErrorMessage (COMPONENT_NOT_FOUND, SourceLine);
    else t = p->GetType ();
    return p != NULL;
}

void TRecord::CodeConstructor (int line, TCompiler *c) {
    int add = 0;
    for (TRecordEl *p = (TRecordEl *) ComponentList.GetFirst (); p; p = (TRecordEl *) ComponentList.GetNext ()) {
    	TType *t = p->GetType ();
	if (t->ContainsVector ()) {
	    if (add) {
		if (add) Compiler->OutCode (line, ADDIND, add);
		add = 0;
	    }
	    t->CodeConstructor (line, c);
	} else add += t->GetSize ();
    }
    if (add) Compiler->OutCode (line, ADDIND, add);
}

void TRecord::CodeDestructor (int line, TCompiler *c) {
    int add = 0;
    for (TRecordEl *p = (TRecordEl *) ComponentList.GetFirst (); p; p = (TRecordEl *) ComponentList.GetNext ()) {
    	TType *t = p->GetType ();
	if (t->ContainsVector ()) {
	    if (add) {
		Compiler->OutCode (line, ADDIND, add);
		add = 0;
	    }
	    t->CodeDestructor (line, c);
	} else add += t->GetSize ();
    }
    if (add) Compiler->OutCode (line, ADDIND, add);
}

BOOL TRecord::ContainsVector () {
    for (TRecordEl *p = (TRecordEl *) ComponentList.GetFirst (); p; p = (TRecordEl *) ComponentList.GetNext ())
	if (p->GetType ()->ContainsVector ())
	    return TRUE;
    return FALSE;
}

void TRecord::CodeLoad (int line, TCompiler *c) {
    if (ContainsVector ()) 
       	for (TRecordEl *p = (TRecordEl *) ComponentList.GetFirst (); p; p = (TRecordEl *) ComponentList.GetNext ())
	    p->GetType ()->CodeLoad (line, c);
    else
	inherited::CodeLoad (line, c);
}

void TRecord::CodeSave (int line, TCompiler *c) {
    if (ContainsVector ()) 
       	for (TRecordEl *p = (TRecordEl *) ComponentList.GetLast (); p; p = (TRecordEl *) ComponentList.GetPrev ())
	    p->GetType ()->CodeSave (line, c);
    else
	inherited::CodeSave (line, c);
}

void TRecord::CodePush (int line, TCompiler *c) {
    if (ContainsVector ()) 
       	for (TRecordEl *p = (TRecordEl *) ComponentList.GetLast (); p; p = (TRecordEl *) ComponentList.GetPrev ())
	    p->GetType ()->CodePush (line, c);
    else
	inherited::CodePush (line, c);
}

TReference::TReference (TType *b) {
    BaseType = b;
}

TType *TReference::GetBaseType () {
    return BaseType;
}

BOOL TReference::IsReference () {
    return TRUE;
}

String TReference::Name () {
    return "Reference to " & BaseType->Name ();
}


TVector::TVector (TType *b) {
    BaseType = b;
}

TType *TVector::GetBaseType () {
    return BaseType;
}

BOOL TVector::IsVector () {
    return TRUE;
}

String TVector::Name () {
    return "vector of " & BaseType->Name ();
}

BOOL TVector::ContainsVector () {
    return TRUE;
}

void TVector::CodeLoad (int line, TCompiler *c) {
    c->OutCode (line, VLOADIND);
}

void TVector::CodeSave (int line, TCompiler *c) {
    c->OutCode (line, VSTOREIND);
}

void TVector::CodePush (int line, TCompiler *c) {
    c->OutCode (line, VPUSH);
}

void TVector::CodeConstructor (int line, TCompiler *c) {
    c->OutCode (line, VINIT);
}

void TVector::CodeDestructor (int line, TCompiler *c) {
    c->OutCode (line, VFREE);
}


TSPMatrix::TSPMatrix (TType *b) {
    BaseType = b;
}

TType *TSPMatrix::GetBaseType () {
    return BaseType;
}

BOOL TSPMatrix::IsMatrix () {
    return TRUE;
}

String TSPMatrix::Name () {
    return "matrix of " & BaseType->Name ();
}

BOOL TSPMatrix::ContainsVector () {
    return TRUE;
}

void TSPMatrix::CodeLoad (int line, TCompiler *c) {
    c->OutCode (line, VLOADIND);
}

void TSPMatrix::CodeSave (int line, TCompiler *c) {
    c->OutCode (line, VSTOREIND);
}

void TSPMatrix::CodePush (int line, TCompiler *c) {
    c->OutCode (line, VPUSH);
}

void TSPMatrix::CodeConstructor (int line, TCompiler *c) {
    c->OutCode (line, VINIT);
}

void TSPMatrix::CodeDestructor (int line, TCompiler *c) {
    c->OutCode (line, VFREE);
}


TPointer::TPointer (TType *b) {
    BaseType = b;
}

TType *TPointer::GetBaseType () {
    return BaseType;
}

BOOL TPointer::IsPointer () {
    return TRUE;
}

void TPointer::SetBaseName (String s) {
    BaseName = s;
}

String TPointer::Name () {
    return "Pointer to " & BaseType->Name ();
}


TRoutine::TRoutine () {
    ReturnType = NULL;
}

//TRoutine::~TRoutine () {
//    ParaTypes.DeleteAll (FALSE);
//}

BOOL TRoutine::IsRoutine () {
    return TRUE;
}

String TRoutine::Name () {
    String s = "routine (";
    for (TType *t = GetFirstParaType (); t; t = GetNextParaType ())
	s = s & t->Name () & " ";
    s = s & " ): " &  ReturnType->Name ();
    return s;
}

String TRoutine::GetFirstParaName () {
    TStringEl *s = (TStringEl *) ParaNames.GetFirst ();
    if (s) return s->Name ();
    return "";
}

String TRoutine::GetNextParaName () {
    TStringEl *s = (TStringEl *) ParaNames.GetNext ();
    if (s) return s->Name ();
    return "";
}

void TRoutine::AppendParaType (TType *t, String s) {
    ParaTypes.Append (t); ParaNames.Append (new TStringEl (s));
}

/*********************************************************/
/*                   Symbol table                        */
/*********************************************************/

TTypeListEl::TTypeListEl (TType *p0, String name0) {
    name = name0;
    p = p0;
};

String TTypeListEl::Name () {
    return name;
}


TSymbol::TSymbol (String name0, TType *type0) {
    name = name0;
    type = type0;
    Temporary = Forward = parameter = constant = Label = FALSE;
    UseAddress = NULL;
    UseCount = UseAlloc = UseReq = 0;
}

String TSymbol::Name () {
    return name;
}

void TSymbol::SetUseAddress (int a) {
    if (UseCount == UseAlloc) {
	UseAlloc += 10;
	UseAddress = (int *) realloc (UseAddress, sizeof (int) * UseAlloc);
    }
    UseAddress [UseCount++] = a;
}

int TSymbol::GetFirstUseAddress () {
    UseReq = 0;
    return GetNextUseAddress ();
}

int TSymbol::GetNextUseAddress () {
    if (UseReq < UseCount)
	return UseAddress [UseReq++];
    return 0;
}

void TSymbol::SetForward () {
    Forward = TRUE;
}

void TSymbol::SetForwardResolved () {
    Forward = FALSE;
}

BOOL TSymbol::IsForward () {
    return Forward;
}

void TSymbol::SetFixupAddress (int a) {
    FixupAddress = a;
}

int TSymbol::GetFixupAddress () {
    return FixupAddress;
}

void TSymbol::SetTemporary () {
    Temporary = TRUE;
}

void TSymbol::HideIfTemp () {
   if (Temporary) {
   	printf ("Hiding %s\n", (const char *) name);
   	name = "$" & name;
	Temporary = FALSE;
   }
}

void TSymbol::SetStringVal (String s) {
    StringConstVal = s;
}


TSymbolTable::TSymbolTable () {
}

void TSymbolTable::DeleteAllSymbols () {
    for (int i = 0; i <= StaticLevel; i++) {
        Symbols [i].DeleteAll (TRUE);
        TypeSymbols [i].DeleteAll (TRUE);
    }
}

TSymbol *TSymbolTable::NewSymbol (String name, TType *type, int Line) {
    for (TSymbol *s = (TSymbol *) Symbols [StaticLevel].GetFirst ();
         s; s = (TSymbol *) Symbols [StaticLevel].GetNext ())
	if (name == s->Name ())
	    ErrorMessage (IDENTIFIER_ALREADY_DECLARED, Line, "'" & name & "' is already declared in the current block");
    TSymbol *p = new TSymbol (name, type);
    p->SetLevel (StaticLevel);
    Symbols [StaticLevel].Append (p);
    return p;
}

void TSymbolTable::InsertType (TTypeListEl *p) {
    TypeSymbols [StaticLevel].Append (p);
}

TType *TSymbolTable::SearchType (String name, int Line, int StartLevel) {
    for (int i = (StartLevel == -1) ? StaticLevel : StartLevel; i >= 0; i--)
	for (TListEl *p = TypeSymbols [i].GetFirst (); p; p = TypeSymbols [i].GetNext ())
	    if (p->Name () == name) return ((TTypeListEl *) p)->GetType ();
    return TFORWARDPTR;
}

void TSymbolTable::AssignStackPositions (int Line) {
    const int FrameSize = 6;
    StackOffsetsNumLocals = StackOffsetsNumPara [StaticLevel] = 0;
    TSymbol *p;
    for (p = (TSymbol *) Symbols [StaticLevel].GetFirst (); p; p = (TSymbol *) Symbols [StaticLevel].GetNext ())
	 if (p->IsParameter ())
	     StackOffsetsNumPara [StaticLevel] += p->GetSize ();
    int NumPara = StackOffsetsNumPara [StaticLevel];
#ifdef TEST
    printf ("Local parameter: %d positions\n", NumPara);
#endif
    for (p = (TSymbol *) Symbols [StaticLevel].GetFirst (); p; p = (TSymbol *) Symbols [StaticLevel].GetNext ())
	if (p->IsParameter ()) {
	    NumPara -= p->GetSize ();
	    p->SetAddress (NumPara + 1);
	}
    for (p = (TSymbol *) Symbols [StaticLevel].GetFirst (); p; p = (TSymbol *) Symbols [StaticLevel].GetNext ())
	if (!p->IsConstant () && !p->IsParameter ()) {
            StackOffsetsNumLocals += p->GetSize ();
            p->SetAddress (1 - (StackOffsetsNumLocals + FrameSize));
        }
}

void TSymbolTable::CodeConstructors (int line, TCompiler *c) {
    // is a local variable with a vector declared?
    BOOL Required = FALSE;
    TSymbol *p;
    for (p = (TSymbol *) Symbols [StaticLevel].GetLast (); p; p = (TSymbol *) Symbols [StaticLevel].GetPrev ())
	if (!p->IsConstant () && !p->IsParameter ()) 
	    Required = Required || p->GetType ()->ContainsVector ();
    if (!Required) return;

    c->OutCode (line, MVSPIND);
    int add = 0;
    for (p = (TSymbol *) Symbols [StaticLevel].GetLast (); p; p = (TSymbol *) Symbols [StaticLevel].GetPrev ())
	if (!p->IsConstant () && !p->IsParameter ()) {
	    TType *t = p->GetType ();
	    if (t->ContainsVector ()) {
		if (add) c->OutCode (line, ADDIND, add);
		add = 0;
		t->CodeConstructor (line, c);
	    } else
		add += t->GetSize ();
	}
}

void TSymbolTable::CodeDestructors (int line, TCompiler *c) {
    // is a local variable with a vector declared?
    BOOL Required = FALSE;
    TSymbol *p;
    for (p = (TSymbol *) Symbols [StaticLevel].GetLast (); p; p = (TSymbol *) Symbols [StaticLevel].GetPrev ())
	if (!p->IsConstant () && !p->IsParameter ()) 
	    Required = Required || p->GetType ()->ContainsVector ();

    int add = 0;
    if (Required) {
        c->OutCode (line, MVSPIND);
        for (p = (TSymbol *) Symbols [StaticLevel].GetLast (); p; p = (TSymbol *) Symbols [StaticLevel].GetPrev ())
    	    if (!p->IsConstant () && !p->IsParameter ()) {
    	        TType *t = p->GetType ();
    	        if (t->ContainsVector ()) {
    		    if (add) c->OutCode (line, ADDIND, add);
    		    add = 0;
    		    t->CodeDestructor (line, c);
    	        } else
    		    add += t->GetSize ();
    	    }
        add = 0;
    }

    // is a parameter with a vector declared?
    Required = FALSE;
    for (p = (TSymbol *) Symbols [StaticLevel].GetLast (); p; p = (TSymbol *) Symbols [StaticLevel].GetPrev ())
	if (p->IsParameter () && p->GetType ()->ContainsVector ())   // Reference parameters have their own type TReference which 
	    Required = TRUE;				  	     // responds 'false' upon this request, so no destructor is called for them
    if (Required) {	    
        c->OutCode (line, SETIND, 1, StaticLevel);
    	for (p = (TSymbol *) Symbols [StaticLevel].GetLast (); p; p = (TSymbol *) Symbols [StaticLevel].GetPrev ())
	    if (p->IsParameter ()) {
	        TType *t = p->GetType ();
	        if (t->ContainsVector ()) {
		    if (add) c->OutCode (line, ADDIND, add);
		    add = 0;
		    t->CodeDestructor (line, c);
	        } else
		    add += t->GetSize ();
	    }
    }
}

void TSymbolTable::PrintLevel (int n) {
#ifdef TEST
    printf ("Symbols declared at level %d:\n", n);
    TSymbol *p;
    for (p = (TSymbol *) Symbols [StaticLevel].GetFirst (); p; p = (TSymbol *) Symbols [StaticLevel].GetNext ()) {
        String s = p->GetType ()->Name ();
	if (!p->IsConstant ())
	    printf ("%d %-20s %s\n", p->GetAddress (), (const char *) p->Name (), (const char *) s);
    }
    puts ("Functions:");
    for (p = (TSymbol *) Symbols [StaticLevel].GetFirst (); p; p = (TSymbol *) Symbols [StaticLevel].GetNext ())
        if (p->GetType ()->IsRoutine () && p->IsConstant ()) {
            printf (p->Name ());
            printf (":");
	    printf (p->GetType ()->Name ());
            printf (" @ %d\n", p->GetIntVal ());
        }
    putchar ('\n');
#endif    
}

void TSymbolTable::AssignParameters (TRoutine *f) {
    for (TListEl *q = Symbols [StaticLevel].GetFirst (); q; q = Symbols [StaticLevel].GetNext ())
        if (((TSymbol *) q)->IsParameter ())
            f->AppendParaType (((TSymbol *) q)->GetType (), ((TSymbol *) q)->Name ());
}

void TSymbolTable::GetVariableTypes (int Level, TList *l) {
    for (TListEl *q = Symbols [Level].GetFirst (); q; q = Symbols [Level].GetNext ())
        if (!((TSymbol *) q)->IsParameter () && !((TSymbol *) q)->IsConstant () && ((TSymbol *) q)->GetType () != TTEXT)
            l->Append (((TSymbol *) q)->GetType ());
}

TSymbol *TSymbolTable::Search (String Name, int Line, BOOL Error, int StartLevel) {
    for (int i = (StartLevel == -1) ? StaticLevel : StartLevel; i >= 0; i--)
	for (TSymbol *s = (TSymbol *) Symbols [i].GetFirst (); s; s = (TSymbol *) Symbols [i].GetNext ())
	    if (s->Name () == Name) return s;
    if (Error) ErrorMessage (IDENTIFIER_NOT_FOUND, Line);
    return NULL;
}

void TSymbolTable::FillPointerTypes (int Line) {
    TList MissingTypes;
    for (TTypeListEl *t = (TTypeListEl *) TypeSymbols[StaticLevel].GetFirst (); t;
         t = (TTypeListEl *) TypeSymbols[StaticLevel].GetNext ()) 
	if (t->GetType ()->IsPointer () && ((TPointer *) t->GetType  ())->GetBaseType () == TFORWARDPTR)
	    MissingTypes.Append (t->GetType ());
    for (TType *u = (TType *) MissingTypes.GetFirst (); u; u = (TType *) MissingTypes.GetNext ()) {
	((TPointer *) u)->SetBaseType (SearchType (((TPointer *) u)->GetBaseName (), Line));
#ifdef TEST
	printf ("Resolving type ^%s\n", (const char *) ((TPointer *) u)->GetBaseName ());
#endif
    }
    MissingTypes.DeleteAll (FALSE);
}

void TSymbolTable::EnterLevel (int n) {
    StaticLevel = n;
}

void TSymbolTable::LeaveLevel (int n) {
    // check if all used labels are defined
    for (TSymbol *p = (TSymbol *) Symbols [StaticLevel].GetFirst (); p; p = (TSymbol *) Symbols [StaticLevel].GetNext ())
	if (p->IsLabel () && !p->GetIntVal () && p->GetFirstUseAddress ())
	    ErrorMessage (INVALID_USE_OF_SYMBOL, Scanner->GetLine (), "Label '" & p->Name () & "' undefined");
    Symbols [StaticLevel].DeleteAll (TRUE);
    for (TListEl *q = TypeSymbols [StaticLevel].GetLast (); q; q = TypeSymbols [StaticLevel].GetPrev ()) {
#ifdef TEST
	String s = ((TType *) q)->Name ();
	printf ("Deleting ");
	puts (s);
#endif
	delete q;
    }
    TypeSymbols [StaticLevel].DeleteAll (FALSE);
#ifdef TEST
    printf ("Leaving level %d\n\n", StaticLevel);
#endif
    StaticLevel--;
}

void TSymbolTable::HideTemporaries () {
    for (TSymbol *s = (TSymbol *) Symbols [StaticLevel].GetFirst ();
         s; s = (TSymbol *) Symbols [StaticLevel].GetNext ())
	s->HideIfTemp ();
}

void TSymbolTable::BuildTypes () {
    InsertType (new TTypeListEl (TINT = new TEnumerated ("integer"), "integer"));
    InsertType (new TTypeListEl (TREAL = new TType ("real"), "real"));
    InsertType (new TTypeListEl (TCOMPLEX = new TComplex, "complex"));
    InsertType (new TTypeListEl (TFILE = new TType ("file"), "file"));
    InsertType (new TTypeListEl (TTEXT = new TType ("text"), "text"));
    InsertType (new TTypeListEl (TCHAR = new TEnumerated ("char"), "char"));
    InsertType (new TTypeListEl (TBOOL = new TEnumerated ("boolean"), "boolean"));
    InsertType (new TTypeListEl (TSTRING = new TVector (TCHAR), "string"));

    InsertType (new TTypeListEl (TREALVECT = new TVector (TREAL), "realvector"));
    InsertType (new TTypeListEl (TREALMATR = new TSPMatrix (TREAL), "realmatrix"));
    InsertType (new TTypeListEl (TREAL2 = new TArray (new TSubrange (TINT, 1, 2), TREAL), "real2"));
    InsertType (new TTypeListEl (TREAL3 = new TArray (new TSubrange (TINT, 1, 3), TREAL), "real3"));
    TVOID = new TType ("void");
    TEMPTYSET = new TSet (TVOID);
    TFORWARDPTR = new TType ("undefined pointer");
    TBOOLVECT = new TVector (TBOOL);
    TINTVECT = new TVector (TINT);
    TBOOLMATR = new TSPMatrix (TBOOL);
    TINTMATR = new TSPMatrix (TINT);
    TINT->SetIndexRange (-INT_MAX, INT_MAX);
    TCHAR->SetIndexRange (0, 255);
    TBOOL->SetIndexRange (0, 1);
    TCARDINAL = new TSubrange (TINT, 0, INT_MAX);
    TGENERICPOINTER = new TPointer (TFORWARDPTR);	// pointer to invalid type to prohibit dereferencing
    TERROR = new TType ("Error");
    TSymbol *s = NewSymbol ("false", TBOOL, 0);
    s->SetConstant (TRUE);
    s->SetIntVal (0);
    s = NewSymbol ("true", TBOOL, 0);
    s->SetConstant (TRUE);
    s->SetIntVal (1);
    NewSymbol ("input", TTEXT, 0);
    NewSymbol ("output", TTEXT, 0);
}
