/*  symbols.hpp
 *
 *  Symbol table of StatPascal compiler.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998, 2000 Xtremes Group, Siegen.
 */

#ifndef SYMBOLS_HPP
#define SYMBOLS_HPP

#include "list.hpp"
#include "string.hpp"

#include "maxstat.hpp"

/********************************************************************
			Handling of data types
*********************************************************************/

class TCompiler;

class TType : public TObject {
    public:
	TType () {}
	TType (String Name);
	virtual ~TType ();
	virtual String Name ();

	virtual void DeleteType ();
	virtual int GetSize ();

	virtual BOOL IsEnumerated ();
	virtual BOOL IsSubrange ();
	virtual BOOL IsArray ();
	virtual BOOL IsRecord ();
	virtual BOOL IsReference ();
	virtual BOOL IsPointer ();
	virtual BOOL IsRoutine ();
	virtual BOOL IsSet ();
	virtual BOOL IsVector ();
	virtual BOOL IsMatrix ();
	virtual BOOL ContainsVector ();

	virtual TType *GetBaseType ();
	virtual TType *GetIndexType ();		// Arrays only
	virtual void GetIndexRange (int &, int &);

	virtual void CodeConstructor (int line, TCompiler *);
	virtual void CodeDestructor (int line, TCompiler *);
	virtual void CodeLoad (int line, TCompiler *);
	virtual void CodeSave (int line, TCompiler *);
	virtual void CodePush (int line, TCompiler *);

    private:
	String name;
};

class TEnumerated : public TType {
    typedef TType inherited;
    public:
	TEnumerated (String s = "");
        virtual String Name ();
	virtual BOOL IsEnumerated ();
	void SetIndexRange (int a, int b);
	void GetIndexRange (int &, int &b);
    private:
	int a, b;
};

class TSubrange : public TEnumerated {
    public:
	TSubrange (TType *BaseType, int a, int b);
  	virtual String Name ();
	virtual BOOL IsSubrange ();
	virtual TType *GetBaseType ();
    private:
	TType *BaseType;
};

class TSet : public TType {
    typedef TType inherited;
    public:
	TSet (TType *BaseType);
	virtual String Name ();
	virtual BOOL IsSet ();
	virtual TType *GetBaseType ();
	virtual int GetSize ();
    private:
	TType *BaseType;
};

class TComplex : public TType {
    typedef TType inherited;
    public:
	TComplex ();
	virtual String Name ();
	virtual int GetSize ();
};

class TArray : public TType {
    typedef TType inherited;
    public:
        TArray (TType *IndexType, TType *BaseType);
        virtual BOOL IsArray ();
        virtual TType *GetBaseType ();
        virtual TType *GetIndexType ();
        virtual void DeleteType ();
        virtual int GetSize ();
	virtual String Name ();

	virtual void CodeConstructor (int line, TCompiler *);
	virtual void CodeDestructor (int line, TCompiler *);
	virtual BOOL ContainsVector ();
        virtual void CodeLoad (int line, TCompiler *);
        virtual void CodeSave (int line, TCompiler *);
        virtual void CodePush (int line, TCompiler *);

    private:
        TType *IndexType, *BaseType;
};

class TRecord : public TType {
    typedef TType inherited;
    public:
	TRecord ();
	virtual BOOL IsRecord ();
	virtual void DeleteType ();
	virtual int GetSize ();
	virtual String Name ();
	void AppendComponent (String Name, TType *);
	BOOL GetComponent (String Name, int &Offset, TType *&t, int SourceLine);
		// displays error if not found and SourceLine <> 0

	virtual void CodeConstructor (int line, TCompiler *);
	virtual void CodeDestructor (int line, TCompiler *);
        virtual BOOL ContainsVector ();
        virtual void CodeLoad (int line, TCompiler *);
        virtual void CodeSave (int line, TCompiler *);
        virtual void CodePush (int line, TCompiler *);

    private:
        class TRecordEl : public TListEl {
            public:
        	TRecordEl (String Name, TType *);
        	virtual String Name ();
        	TType *GetType ();
            private:
        	String name;
        	TType *type;
        };
 	TList ComponentList;
};

class TReference : public TType {
    public:
   	TReference (TType *BaseType);
	virtual TType *GetBaseType ();
	virtual BOOL IsReference ();
	virtual String Name ();
    private:
  	TType *BaseType;
}; 

class TVector : public TType {
    public:
   	TVector (TType *BaseType);
	virtual TType *GetBaseType ();
	virtual BOOL IsVector ();
	virtual String Name ();

        virtual BOOL ContainsVector ();
	virtual void CodeConstructor (int line, TCompiler *);
	virtual void CodeDestructor (int line, TCompiler *);
        virtual void CodeLoad (int line, TCompiler *);
        virtual void CodeSave (int line, TCompiler *);
        virtual void CodePush (int line, TCompiler *);
    private:
  	TType *BaseType;
}; 

class TSPMatrix : public TType {
    public:
   	TSPMatrix (TType *BaseType);
	virtual TType *GetBaseType ();
	virtual BOOL IsMatrix ();
	virtual String Name ();

        virtual BOOL ContainsVector ();
	virtual void CodeConstructor (int line, TCompiler *);
	virtual void CodeDestructor (int line, TCompiler *);
        virtual void CodeLoad (int line, TCompiler *);
        virtual void CodeSave (int line, TCompiler *);
        virtual void CodePush (int line, TCompiler *);
    private:
  	TType *BaseType;
};

class TPointer : public TType {
    public:
	TPointer (TType *BaseType);
	virtual TType *GetBaseType ();
	virtual BOOL IsPointer ();
	virtual String Name ();
	void SetBaseName (String s);
	String GetBaseName ()			{ return BaseName; }
	void SetBaseType (TType *t)		{ BaseType = t; }
    private:
	String BaseName;
  	TType *BaseType;
};

class TRoutine : public TType {
    public:
        TRoutine ();
//        ~TRoutine ();
        virtual BOOL IsRoutine ();
        virtual String Name ();

	TType *GetReturnType ()		{ return ReturnType; }
	TType *GetFirstParaType ()      { return (TType *) ParaTypes.GetFirst (); }
	TType *GetNextParaType ()       { return (TType *) ParaTypes.GetNext (); }
	String GetFirstParaName ();
	String GetNextParaName ();

	void SetReturnType (TType * t)	{ ReturnType = t; }
	void AppendParaType (TType *t, String s);
	int GetNumberOfParameters ()	{ return ParaTypes.Entries (); }
    private:
        TList ParaTypes;
	TList ParaNames;
        TType *ReturnType;
};

/********************************************************************
                                Symbol tables
********************************************************************/

class TTypeListEl : public TListEl {
    public:
	TTypeListEl (TType *p, String Name);
	TType *GetType () { return p; }
	virtual String Name ();
    private:
	String name;
	TType *p;
};

#define FORWARD_ADDR	0

class TSymbol : public TListEl {
    public:
        TSymbol (String name0, TType *type0);
	virtual String Name ();
	int GetSize ()			{ return type->GetSize (); }

        BOOL IsRoutine ()		{ return type->IsRoutine (); }

	// labels
	void SetUseAddress (int a);
	int GetFirstUseAddress ();
	int GetNextUseAddress ();		// 0 add end of list
	void SetLabel (BOOL t)			{ Label = t; }
	BOOL IsLabel ()				{ return Label; }
	void SetStatementStackLevel (int l) 	{ ssl = l; }
	int GetStatementStackLevel ()		{ return ssl; }

	// forward declarations
	void SetForward ();
	void SetForwardResolved ();
	BOOL IsForward ();
	void SetFixupAddress (int a);
	int GetFixupAddress ();

	// temporary declarations (implementation part of unit)
	void SetTemporary ();
	void HideIfTemp ();		// sets invalid name, but retains entry in symbol table 
        
        void SetConstant (BOOL t)       { constant = t; }
        BOOL IsConstant ()      	{ return constant; }
        
        void SetLevel (int n)           { level = n; }
        int GetLevel ()         	{ return level; }
        
        void SetAddress (int a)         { address = a; }
        int GetAddress ()       	{ return address; }
        
	void SetType (TType *t)         { type = t; }
	TType *GetType ()       	{ return type; }

	void SetParameter (BOOL p)	{ parameter = p; }
	BOOL IsParameter ()		{ return parameter; }

	int GetIntVal ()        { return constval.intval; }
	double GetDoubleVal ()  { return constval.doubleval; }
	String GetStringVal ()  { return StringConstVal; }

	void SetIntVal (int a)          { constval.intval = a; }
	void SetDoubleVal (double a)    { constval.doubleval = a; }
	void SetStringVal (String s);

    private:
        String name;
        BOOL constant, parameter;
        int address, level;
        TType *type;
        union ConstValues {
            int intval;
            double doubleval;
        } constval;
        String StringConstVal;
	// Label
	BOOL Label;
	int *UseAddress;
	int UseCount, UseAlloc, UseReq;
	// forward
	BOOL Forward;
	int FixupAddress;
	int ssl;	// static stack level of label decl, required for jumps
	//
	BOOL Temporary;
};


class TSymbolTable {
    public:
        TSymbolTable ();
        void AssignStackPositions (int Line);
	void FillPointerTypes (int Line);

        void InsertType (TTypeListEl *p);
        TType *SearchType (String Name, int Line, int StartLevel = -1);
         	// limit search to certain level (e.g., input/output is searched only on top level)

        TSymbol *NewSymbol (String name, TType *type, int Line);
        TSymbol *Search (String Name, int Line, BOOL Error = TRUE, int StartLevel = -1);
        
        void DeleteAllSymbols ();
        void AssignParameters (TRoutine *f);
        void GetVariableTypes (int Level, TList *l);       // needed for ports
        int GetNumStackSize ()          { return StackOffsetsNumLocals; }
        int GetNumParaSize (int Level)          { return StackOffsetsNumPara [Level]; }
        void EnterLevel (int n);
        void LeaveLevel (int n);
	void HideTemporaries ();

	void CodeConstructors (int line, TCompiler *);
	void CodeDestructors (int line, TCompiler *);

        void PrintLevel (int n);

        void BuildTypes ();
        TType *TVOID, *TREAL, *TCOMPLEX, *TSTRING, *TFORWARDPTR, *TEMPTYSET, *TTEXT, *TFILE, 
              *TBOOLVECT, *TINTVECT, *TREALVECT, *TBOOLMATR, *TINTMATR, *TREALMATR, 
              *TREAL2, *TREAL3, *TCARDINAL, *TGENERICPOINTER, *TERROR;
	TEnumerated *TINT, *TCHAR, *TBOOL;

    private:
        TList Symbols [MAXSTATIC], TypeSymbols [MAXSTATIC];
        int StackOffsetsNumLocals,
            StackOffsetsNumPara [MAXSTATIC],
            StaticLevel;
};

#endif
