/*  matrix.cpp

    Implementation of matrix operations.

    This file is part of Xtremes for Windows,
    Copyright (C) 1991, 1996 Xtremes Group, Siegen.
*/

#include <assert.h>
#include <iostream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tmatrix.hpp"
#include "stddef.hpp"

TMatrix::TMatrix () {
    p = NULL;
    n = m = 0;
}

TMatrix::TMatrix (int n0, int m0) {
    n = n0;
    m = m0;
    p = new double [n * m];
}

TMatrix::TMatrix (int n0, int m0, double *a) {
    n = n0; m = m0;
    p = new double [n * m];
    memcpy (p, a, n * m * sizeof (double));
}

TMatrix::TMatrix (const TMatrix &A) {
    n = A.n; m = A.n;
    p = new double [n * m];
    memcpy (p, A.p, n * m * sizeof (double));
}

TMatrix::~TMatrix () {
    if (p) delete [] p;
}

TMatrix & TMatrix::operator = (const TMatrix &A) {
    n = A.n; m = A.n;
    p = new double [n * m];
    memcpy (p, A.p, n * m * sizeof (double));
    return *this;
}

double &TMatrix::operator () (int l, int c) {
    assert (0 < l && l <= n);
    assert (0 < c && c <= m);
    return p [(l - 1) * m + c - 1];
}

int TMatrix::Rows () {
    return n;
}

int TMatrix::Columns () {
    return m;
}

TMatrix operator + (TMatrix a, TMatrix b) {
    int n = a.Rows (), m = a.Columns ();
    assert (n == b.Rows ());
    assert (m == b.Columns ());
    TMatrix c (n, m);
    for (int i = 1; i <= n; i++)
	for (int j = 1; j <= m; j++)
	    c (i, j) = a (i, j) + b (i, j);
    return c;
}

TMatrix operator - (TMatrix a, TMatrix b) {
    int n = a.Rows (), m = a.Columns ();
    assert (n == b.Rows ());
    assert (m == b.Columns ());
    TMatrix c (n, m);
    for (int i = 1; i <= n; i++)
	for (int j = 1; j <= m; j++)
	    c (i, j) = a (i, j) - b (i, j);
    return c;
}

TMatrix operator * (double a, TMatrix b) {
    int n = b.Rows (), m = b.Columns ();
    TMatrix c (n, m);
    for (int i = 1; i <= n; i++)
	for (int j = 1; j <= m; j++)
	    c (i, j) = a * b (i, j);
    return c;
}

TMatrix operator * (TMatrix a, TMatrix b) {
    int n = a.Rows (), m = a.Columns (), o = b.Rows ();
    assert (a.Columns () == b.Rows ());
    TMatrix c (n, o);
    for (int i = 1; i <= n; i++)
	for (int j = 1; j <= o; j++) {
	    double d = 0.0;
	    for (int k = 1; k <= m; k++)
		d += a (i, k) * b (k, j);
	    c (i, j) = d;
	}
    return c;
}

TMatrix operator ~ (TMatrix A) {
    int n = A.Rows (), m = A.Columns ();
    TMatrix B (m, n);
    for (int i = 1; i <= m; i++)
	for (int j = 1; j <= n; j++)
	    B (i, j) = A (j, i);
    return B;
}

double maxNorm (TMatrix a) {
    double c = 0.0;
    for (int i = 1; i <= a.Rows (); i++) {
	double d = 0.0;
	for (int j = 1; j <= a.Columns (); j++)
	    d += fabs (a (i, j));
	if (d > c) c = d;
    }
    return c;
}

std::ostream & operator << (std::ostream &out, TMatrix c) {
    for (int i = 1; i <= c.Rows (); i++) {
	for (int j = 1; j <= c.Columns (); j++)
	    out << c (i, j) << ' ';
	out << "\n";
    }
    return out;
}

TMatrix GetE (int n) {
    TMatrix A (n, n);
    for (int i = 1; i <= n; i++)
	for (int j = 1; j <= n; j++)
	    A (i, j) = (i == j);
    return A;
}

TMatrix sqrt (TMatrix A) {
    assert (A.Rows () == A.Columns ());
    int n = A.Rows ();
    double sigma = 0.9 / maxNorm (A);
    TMatrix E = GetE (n), B = E - sigma * A, X1 = B, X0;
    double q = maxNorm (B);
    int steps = 0;
    do {
	X0 = X1;
	X1 = 0.5 * (X1 * X1 + B);
    } while (q / (1.0 - q) * maxNorm (X1 - X0) > 1e-9 && steps++ < 500);
    return (1.0 / sqrt (sigma)) * (E - X1);
}

TMatrix inv (TMatrix A) {
    assert (A.Rows () == A.Columns ());
    int n = A.Rows ();
    TMatrix B (n, n);
    int nx = 0, ny = 0;
    double pivot, factor;
    int mx[100], my[100], i;
    for (i = 1; i <= n; i++)
	for (int j = 1; j <= n; j++)
	    B (i, j) = A (i, j);
    for (i = 1; i <= n; i++)
	mx[i] = my[i] = -1;
    for (i = 1; i <= n; i++) {
	pivot = 0.0;
	for (int ix = 1; ix <= n; ix++)
	    if (mx[ix] == -1)
		for (int iy = 1; iy <= n; iy++)
		    if (my[iy] == -1)
			if (fabs (B (ix, iy)) > fabs (pivot)) {
			    pivot = B (ix, iy);
			    nx = ix; ny = iy;
			}
	if (pivot == 0.0) abort (); // !!!!;
	mx[nx] = ny; my[ny]=nx;
	for (int j = 1; j <= n; j++)
	    if (j != nx) {
		factor = B (j, ny) / pivot;
		for (int k = 1; k <= n; k++)
		    B (j, k) -= B (nx, k) * factor;
		B (j, ny) = -factor;
	    }
	for (int k = 1; k <= n; k++)
	    B (nx, k) /= pivot;
	B (nx, ny) = 1.0 / pivot;
    }
    for (i = 1; i <= n; i++) {
	int j = mx[i];
	if (i != j) {
	    for (int k = 1; k <= n; k++)
		Swap (B (i, k), B (j, k));
	    mx[i] = mx[j];
	    mx[j] = j;
	}
	j = my[i];
        if (j != i) {
	    for (int k = 1; k <= n; k++)
		Swap (B (k, i), B (k, j));
	    my[i] = my[j]; my[j] = j;
	}
    }
    return B;
}


int sgn(int c)
{
if ((c % 2)== 0)
   return 1;
else
   return -1;
}



TMatrix chol(TMatrix mat)
{
if (mat.Rows()==mat.Columns())
    {
    //generate new Matrix
    int r=mat.Rows(), i;
    double * d = new double[r*r];
    for (i=0;i<r*r;i++)
   	d[i]=0.0;
    TMatrix ch(r,r,d);
    for (i=1;i<=r;i++){
        double summe=0.0;
        for (int k=1;k<i;k++){
      	    summe=summe+ch(i,k)*ch(i,k);
        }
    ch(i,i)=sqrt(mat(i,i)-summe);
    for (int j=i+1;j<=r;j++){
         double summe=0.0;
         for (int k=1;k<i;k++){
             summe=summe+ch(j,k)*ch(i,k);
         }
         ch(j,i)=(mat(j,i)-summe)/ch(i,i);
    }
    }
    return ch;
}
else
   {
   TMatrix ch(1,1);
   return ch;
   }
}

TMatrix invert(TMatrix m)
{
if (m.Rows()==m.Columns())
	{
        int zeilenver=0;
        //Hilfsmatrix erstellen
        int z=m.Rows();
        TMatrix mat(z,z);
        TMatrix em=GetE(z);
	int i;
        for (i=1;i<=z;i++)
        	for (int j=1;j<=z;j++)
                	mat(i,j)=m(i,j);
        //Determinate berechnen
        for (i=1;i<z;i++)
        	{
                if(mat(i,i)==0)
                	{
                        zeilenver++;
                        //Zeile feststellen, die getauscht werden soll
                        int h=i+1;
                        while ((mat(h,i)==0) && (h<=z))
                        	h++;
                        if (h>z)
                        	return GetE(1);
                        else
                        	{
	                        //Zeilenvertauschung
                                for (int j=1;j<=z;j++)
                                	{
                                        double tmp = mat(i,j);
                                        mat(i,j) = mat(h,j);
                                        mat(h,j) = tmp;
                                        double tmp1 = em(i,j);
                                        em(i,j) = em(h,j);
                                        em(h,j) = tmp1;
                                        }
                                }
                        }
                        //Alle Elemente unter dem Pivot-Element lschen
                        //durch Zeilenoperationen

                        for (int ii = i+1;ii<=z;ii++)
                        	{
                                double faktor = mat(ii,i)/mat(i,i);
                                mat(ii,i)=0.0;
                                for (int j=i+1;j<z+1;j++)
                                	{
                                	mat(ii,j)=mat(ii,j)-faktor*mat(i,j);
                                        em(ii,j)=em(ii,j)-faktor*em(i,j);
                                        }
                                }
                }
                // Obere Dreiecksmatrix wurde erzeugt
                // auf Diagonalmatrix bringen
                for (i=z;i>1;i--)
                	{
                        for (int ii=z-1;ii>=1;ii--)
                        	{
                                double faktor = mat(ii,i)/mat(i,i);
                                mat(ii,i)=0.0;
                                for (int j=z;j>=i-1;j--)
                                	{
					mat(ii,j)=mat(ii,j)-faktor*mat(i,j);
                                        em(ii,j)=em(ii,j)-faktor*em(i,j);
                                        }
                                }
                        }
                //Diagonalmatrix auf Einheitmatrix normieren
                for (i=1;i<=z;i++)
                	{
                        double faktor = 1/mat(i,i);
                        for (int j=1;j<=z;j++)
                        	em(i,j)=em(i,j)*faktor;
                        }
                return em;
        }
else
	return GetE(1);
}

double det(TMatrix m)
{
if (m.Rows()==m.Columns())
	{
        int zeilenver=0;
        //Hilfsmatrix erstellen
        int z=m.Rows();
        TMatrix mat(z,z);
	int i;
        for (i=1;i<=z;i++)
        	for (int j=1;j<=z;j++)
                	mat(i,j)=m(i,j);
        //Determinate berechnen
        for (i=1;i<z;i++)
        	{
                if(mat(i,i)==0)
                	{
                        zeilenver++;
                        //Zeile feststellen, die getauscht werden soll
                        int h=i+1;
                        while ((mat(h,i)==0) && (h<=z))
                        	h++;
                        if (h>z)
                        	return 0.0;
                        else
                        	{
	                        //Zeilenvertauschung
                                for (int j=1;j<=z;j++)
                                	{
                                        double tmp = mat(i,j);
                                        mat(i,j) = mat(h,j);
                                        mat(h,j) = tmp;
                                        }
                                }
                        }
                        //Alle Elemente unter dem Pivot-Element lschen
                        //durch Zeilenoperationen

                        for (int ii = i+1;ii<=z;ii++)
                        	{
                                double faktor = mat(ii,i)/mat(i,i);
                                mat(ii,i)=0.0;
                                for (int j=i+1;j<z+1;j++)
                                	mat(ii,j)=mat(ii,j)-faktor*mat(i,j);
                                }
                }
                double d = 1.0;
                for (i = 1; i <=z; i++)

                	d = d * mat(i,i);
                return sgn(zeilenver)*d;

        }
else
	return 0.0;
}

/*

int main () {
    double f [][2] = { {5, 3},
		       {3, 2} };

    cout << "Here we go...\n";
    TMatrix b (2, 2, (double *) f);

    cout << b;
    return 0;
}

*/
