/*  xpl.cpp
 *
 *  Adds Xtremes specific routines (access to data sets,
 *  distribution, estimators) to the StatPascal compiler.
 *
 *  This file is part of StatPascal.
 *  Copyright (C) 1998, 2002 Xtremes Group, Siegen.
 */

#include <math.h>
#include <fstream>
#include <stdlib.h>
#include <limits.h>

#include "errors.hpp"
#include "mathfunc.hpp"
#include "stddef.hpp"
#include "xpl.hpp"
#include "risk.hpp"
#include "sample.hpp"

#ifndef TERMINAL
    #include "boxplot.hpp"
    #include "censor.hpp"
    #include "wintypes.hpp"
    #include "xplfunc.hpp"
    #include "distr3d.hpp"
    #include "contour.hpp"
    #include "scatter.hpp"
    #include "wave.hpp"
#else
    #include <stdlib.h>
#endif

// Global Objects

class TGlobalVarStorage : public TObject {
    public:
        TGlobalVarStorage (String s);
        virtual String Name ();
    private:
        String s;
};

class TGlobalVectorVarStorage : public TGlobalVarStorage {
    public:
	TGlobalVectorVarStorage (String s, double *p, int n);	// claims ownership
        virtual ~TGlobalVectorVarStorage ();
        double *GetValue ();
        int GetSize ();
    private:
        double *p;
        int n;
};

TGlobalVarStorage::TGlobalVarStorage (String s0) {
    s = s0;
}

String TGlobalVarStorage::Name () {
    return s;
}

TGlobalVectorVarStorage::TGlobalVectorVarStorage (String s, double *p0, int n0)
 : TGlobalVarStorage (s) {
    p = p0;
    n = n0;
}

TGlobalVectorVarStorage::~TGlobalVectorVarStorage () {
    delete [] p;
}

double *TGlobalVectorVarStorage::GetValue () {
    return p;
}

int TGlobalVectorVarStorage::GetSize () {
    return n;
}

TList GlobalVarList;

/*
 *  Xtremes specific extensions to the virtual machine
 */

TXPLPCodeInterpreter::TXPLPCodeInterpreter ()   {
    ActSample = NULL;
#ifndef TERMINAL
    w = NULL;
    PlotFont = TFont (FT_Times, FS_Normal, 10);
#endif
}

TXPLPCodeInterpreter::~TXPLPCodeInterpreter () {
}

String TXPLPCodeInterpreter::GetRunTimeErrorMessage (TRunTimeErrorType t) {
    String Msg;
    if (t != RT_OK) {
        switch (t) {
	    case RT_OPCODE:		Msg = "Invalid code generated by compiler"; break;
            case RT_DIVZERO:            Msg = "Division by zero"; break;
            case RT_RANGECHECK:         Msg = "Range check error"; break;
            case RT_STACKOVERFLOW:      Msg = "Stack overflow"; break;
	    case RT_HEAP_FULL:		Msg = "Heap overflow"; break;
            case RT_ARGUMENT:           Msg = "Invalid argument in function call"; break;
            case RT_NOSAMPLE:           Msg = "Active sample required to execute operation"; break;
	    case RT_MATRIXOP:		Msg = "Invalid matrix operation"; break;
	    case RT_VECTOREMPTY:	Msg = "Empty vector used in arithmetic operation"; break;
	    case RT_OVERLOW:		Msg = "Numeric overflow"; break;
	    case RT_USERBREAK:		Msg = "User break"; break;
	    default:			Msg = "No error"; break;
        }
        Msg = Msg & form (50, " at address %d", GetPC ());
    }
    return Msg;
}

#ifndef TERMINAL

void TXPLPCodeInterpreter::ExecuteProcedure () {
//    int Count = (int) IPop ();	!!!!
    int Count = 3;
    double *parameter = new double [10];
    for (int i = 0; i < Count; i++)
        parameter [Count - 1 - i] = DPop ();
    String name = PopStringFromVector ();

    #define CheckCurve(ident, p1, p2, p3, IDSWindow, object) \
        if (name == String (ident)) { \
            if (Count < 3) parameter [2] = p1; \
            if (Count < 2) parameter [1] = p2; \
            if (!Count)    parameter [0] = p3; \
            PlotWindowManager.GetOpenWindow (IDSWindow)->InsertFunction ( \
                new object (parameter [0], parameter [2], parameter [1], 1.0)); \
        }

    CheckCurve ("paretodensity",
                GlobalData.mup, GlobalData.sigmap, GlobalData.alphap,
                DENSITY, TParetoDensity)
    CheckCurve ("paretodf",
                GlobalData.mup, GlobalData.sigmap, GlobalData.alphap,
                DISTRIBUTION, TParetoDistribution)
    CheckCurve ("paretoqf",
                GlobalData.mup, GlobalData.sigmap, GlobalData.alphap,
                QUANTILE, TParetoQuantile)

    CheckCurve ("frechetdensity",
                GlobalData.muf, GlobalData.sigmaf, GlobalData.alphaf,
                DENSITY, TFrechetDensity)
    CheckCurve ("frechetdf",
                GlobalData.muf, GlobalData.sigmaf, GlobalData.alphaf,
                DISTRIBUTION, TFrechetDistribution)
    CheckCurve ("frechetqf",
                GlobalData.muf, GlobalData.sigmaf, GlobalData.alphaf,
                QUANTILE, TFrechetQuantile)

    CheckCurve ("betadensity",
                GlobalData.mupii, GlobalData.sigmapii, GlobalData.alphapii,
                DENSITY, TBetaDensity)
    CheckCurve ("betadf",
                GlobalData.mupii, GlobalData.sigmapii, GlobalData.alphapii,
                DISTRIBUTION, TBetaDistribution)
    CheckCurve ("betaqf",
                GlobalData.mupii, GlobalData.sigmapii, GlobalData.alphapii,
                QUANTILE, TBetaQuantile)

    CheckCurve ("weibulldensity",
                GlobalData.muw, GlobalData.sigmaw, GlobalData.alphaw,
                DENSITY, TWeibullDensity)
    CheckCurve ("weibulldf",
                GlobalData.muw, GlobalData.sigmaw, GlobalData.alphaw,
                DISTRIBUTION, TWeibullDistribution)
    CheckCurve ("weibullqf",
                GlobalData.muw, GlobalData.sigmaw, GlobalData.alphaw,
                QUANTILE, TWeibullQuantile)

    CheckCurve ("evdensity",
                GlobalData.muge, GlobalData.sigmage, GlobalData.betage,
                DENSITY, TEVDDensity)
    CheckCurve ("evdf",
                GlobalData.muge, GlobalData.sigmage, GlobalData.betage,
                DISTRIBUTION, TEVDDistribution)
    CheckCurve ("evqf",
                GlobalData.muge, GlobalData.sigmage, GlobalData.betage,
                QUANTILE, TEVDQuantile)

    CheckCurve ("gpdensity",
                GlobalData.mugp, GlobalData.sigmagp, GlobalData.betagp,
                DENSITY, TGPDDensity)
    CheckCurve ("gpdf",
                GlobalData.mugp, GlobalData.sigmagp, GlobalData.betagp,
                DISTRIBUTION, TGPDDistribution)
    CheckCurve ("gpqf",
                GlobalData.mugp, GlobalData.sigmagp, GlobalData.betagp,
                QUANTILE, TGPDQuantile)

    #undef CheckCurve
    #define CheckCurve(ident, p1, p2, IDSWindow, object) \
        if (name == String (ident)) { \
            if (Count < 2) parameter [1] = p1; \
            if (!Count)    parameter [0] = p2; \
            PlotWindowManager.GetOpenWindow (IDSWindow)->InsertFunction ( \
                new object (parameter [0], parameter [1], 0.0)); \
        }

    CheckCurve ("gumbeldensity",
                GlobalData.mug, GlobalData.sigmag,
                DENSITY, TGumbelDensity)
    CheckCurve ("gumbeldf",
                GlobalData.mug, GlobalData.sigmag,
                DISTRIBUTION, TGumbelDistribution)
    CheckCurve ("gumbelqf",
                GlobalData.mug, GlobalData.sigmag,
                QUANTILE, TGumbelQuantile)

    CheckCurve ("exponentialdensity",
                GlobalData.mue, GlobalData.sigmae,
                DENSITY, TExponentialDensity)
    CheckCurve ("exponentialdf",
                GlobalData.mue, GlobalData.sigmae,
                DISTRIBUTION, TExponentialDistribution)
    CheckCurve ("exponentialqf",
                GlobalData.mue, GlobalData.sigmae,
                QUANTILE, TExponentialQuantile)

    if (name == String ("readdata")) {
        SampleList.LoadSample (PlotWindowManager.GetFramePtr ());
        ActSample = SampleList.GetActiveSample (PlotWindowManager.GetFramePtr ());
    } else if (name == String ("choosedata")) {
        SampleList.ChooseActiveSample (PlotWindowManager.GetFramePtr ());
	ActSample = SampleList.GetActiveSample (PlotWindowManager.GetFramePtr ());
    }
}

#endif

TSample *TXPLPCodeInterpreter::CreateTempSample () {
    TUnivariateSample *p = new TUnivariateSample ();
    int n;
    double *a;
    PopDoubleVector (a, n);
    p->SetData (n, a);
    return p;
}

TRunTimeErrorType TXPLPCodeInterpreter::CallPOTEstimator (TEstimator *p, BOOL VectorGiven, int NrOfParameters) {
    int Error = IPop (),
	sa = IPop (),
        ma = IPop (), aa = 0, k;
    if (NrOfParameters == 3) aa = IPop ();
    k = IPop ();
    TSample *s;
    if (VectorGiven) s = CreateTempSample ();
    else {
	s = ActSample;
	if (!ActSample) return RT_NOSAMPLE;
    }
    store [Error].intval = p->Estimate (k, s);
    if (NrOfParameters == 3) store [aa].doubleval = p->GetEstimatedParameterValue (0);
    store [ma].doubleval = p->GetEstimatedParameterValue (1);
    store [sa].doubleval = p->GetEstimatedParameterValue (2);
    if (VectorGiven) delete s;
    return RT_OK;
}

TRunTimeErrorType TXPLPCodeInterpreter::CallOtherEstimator (TEstimator *p, BOOL VectorGiven, int NrOfParameters) {
    int Error = IPop (),
	sa = IPop (),
        ma = IPop (), aa = 0, k = 0;
    if (NrOfParameters == 3) aa = IPop ();
    TSample *s;
    if (VectorGiven) s = CreateTempSample ();
    else {
	s = ActSample;
	if (!ActSample) return RT_NOSAMPLE;
    }
    store [Error].intval = p->Estimate (k, s);
    int value = 0;
    if (NrOfParameters == 3) store [aa].doubleval = p->GetEstimatedParameterValue (value++);
    store [ma].doubleval = p->GetEstimatedParameterValue (value++);
    store [sa].doubleval = p->GetEstimatedParameterValue (value++);
    if (VectorGiven) delete s;
    return RT_OK;
}

TRunTimeErrorType TXPLPCodeInterpreter::ExtendedCode (TPCode &p) {
    int k, l, m, n, o;
    double a, b, c, d, e;
    double *x, *y;
    TSampleType t;
    TStorage *ptr;
    TRunTimeErrorType rt = RT_OK;
    switch (p.code) {

	case FPARETORV:
	    DPush (ParetoQF (Random (), DPop (), 1.0));
	    break;
	case FNORMALRV:
	case FGAUSSRV:
	    NormRV (a, b);
            DPush (a);
            break;
        case FCAUCHYRV:
            DPush (CauchyRV (DPop (), FALSE));
            break;
        case FFRECHETRV:
            DPush (FrechetQF (Random (), DPop (), 1.0));
            break;
        case FBETARV:
            DPush (ParetoIIQF (Random (), DPop (), 1.0));
            break;
        case FWEIBULLRV:
	    DPush (WeibullQF (Random (), DPop (), 1.0));
            break;
        case FGUMBELRV:
            DPush (GumbelQF (Random (), 1.0));
            break;
        case FEXPONENTIALRV:
            DPush (ExpQF (Random (), 1.0));
            break;
        case FGPDRV:
            DPush (vonMisesGPDQF (Random (), DPop (), 1.0, 0.0));
            break;
        case FEVDRV:
            DPush (vonMisesEXTQF (Random (), DPop (), 1.0, 0.0));
	    break;
	case FGAMMARV:
	    DPush (XGammaData (DPop ()));
	    break;
	case FCHIRV:
	    DPush (XChiSquareData (IPop ()));
	    break;
	case FTRV:
	    DPush (XTData (DPop ()));
	    break;

	#define RANDOMVECTOR(code,generate) \
	    case code: \
		k = n = IPop (); \
		a = DPop (); \
		while (k--) \
		    DPush (generate); \
		IPush (n, 1); \
		IPush (n + 1, 1); \
		break;

	RANDOMVECTOR (VFPARETORV, ParetoQF (Random (), a, 1.0))
	RANDOMVECTOR (VFCAUCHYRV, CauchyRV (a, FALSE))
	RANDOMVECTOR (VFFRECHETRV, FrechetQF (Random (), a, 1.0))
	RANDOMVECTOR (VFBETARV, ParetoIIQF (Random (), a, 1.0))
	RANDOMVECTOR (VFWEIBULLRV, WeibullQF (Random (), a, 1.0))
	RANDOMVECTOR (VFGPDRV, vonMisesGPDQF (Random (), a, 1.0, 0.0))
	RANDOMVECTOR (VFEVDRV, vonMisesEXTQF (Random (), a, 1.0, 0.0))	
	RANDOMVECTOR (VFGAMMARV, XGammaData (a))
        RANDOMVECTOR (VFTRV, XTData (a))

	case VFNORMALRV:
	case VFGAUSSRV:
            n = IPop ();
            for (k = 0; k < n; k++) {
                NormRV (a, b);
                DPush (a);
            }
            IPush (n, 1);
	    IPush (n + 1, 1);
            break;
        case VFGUMBELRV:
            n = IPop ();
            for (k = 0; k < n; k++)
                DPush (GumbelQF (Random (), 1.0));
            IPush (n, 1);
	    IPush (n + 1, 1);
            break;
        case VFEXPONENTIALRV:
            n = IPop ();
            for (k = 0; k < n; k++)
                DPush (ExpQF (Random (), 1.0));
            IPush (n, 1);
	    IPush (n + 1, 1);
            break;
	case VFCHIRV:
            n = IPop (); m = IPop ();
            for (k = 0; k < n; k++)
	        DPush (XChiSquareData (m));
            IPush (n, 1);
	    IPush (n + 1, 1);
	    break;


	case FPARETODEN:
            a = DPop (); b = DPop ();
	    DPush (ParetoDen (a, b, 1.0, 0.0));
            break;
        case FPARETOQF:
            a = DPop (); b = DPop ();
            if (b < 0.0 || b >= 1.0) return RT_ARGUMENT;
	    DPush (ParetoQF (a, b, 1.0, 0.0));
            break;
        case FPARETODF:
            a = DPop (); b = DPop ();
            DPush (ParetoDF (a, b, 1.0, 0.0));
            break;
        case FGAUSSDEN:
            DPush (NormDen (DPop (), 0.0, 1.0));
            break;
        case FGAUSSDENV:
            k = VectorLength ();
	    ptr = VectorStart ();
	    while (--k) {
		ptr->doubleval = NormDen (ptr->doubleval, 0.0, 1.0);
		ptr++;
	    }
	    break;
        case FGAUSSQF:
            b = DPop ();
            if (b <= 0.0 || b >= 1.0) return RT_ARGUMENT;
            DPush (NormQF (b, 0.0, 1.0));
            break;
        case FGAUSSQFV:
            k = VectorLength ();
	    ptr = VectorStart ();
	    while (--k) {
                ptr->doubleval = NormQF (ptr->doubleval, 0.0, 1.0);
                ptr++;
            }
	    break;
        case FGAUSSDF:
            DPush (NormDFPol (DPop (), 0.0, 1.0));
            break;
        case FGAUSSDFV:
            k = VectorLength ();
            ptr = VectorStart ();
            while (--k) {
                ptr->doubleval = NormDFPol (ptr->doubleval, 0.0, 1.0);
                ptr++;
            }
	    break;
        case FCAUCHYDEN:
            a = DPop (); b = DPop ();
            DPush (CauchyDen (a, b, 1.0, 0.0));
            break;
	case FFRECHETDEN:
            a = DPop (); b = DPop ();
            DPush (FrechetDen (a, b, 1.0, 0.0));
            break;
        case FFRECHETDF:
            a = DPop (); b = DPop ();
            DPush (FrechetDF (a, b, 1.0, 0.0));
            break;
        case FFRECHETQF:
            a = DPop (); b = DPop ();
            if (a < 0.0 || a >= 1.0) return RT_ARGUMENT;
            DPush (FrechetQF (a, b, 1.0, 0.0));
            break;
        case FBETADEN:
            a = DPop (); b = DPop ();
            DPush (ParetoIIDen (a, b, 1.0, 0.0));
            break;
        case FBETADF:
            a = DPop (); b = DPop ();
            DPush (ParetoIIDF (a, b, 1.0, 0.0));
            break;
        case FBETAQF:
            a = DPop (); b = DPop ();
            if (a < 0.0 || a > 1.0) return RT_ARGUMENT;
	    DPush (ParetoIIQF (a, b, 1.0, 0.0));
            break;
        case FWEIBULLDEN:
            a = DPop (); b = DPop ();
            DPush (WeibullDen (a, b, 1.0, 0.0));
	    break;
        case FWEIBULLDF:
            a = DPop (); b = DPop ();
            DPush (WeibullDF (a, b, 1.0, 0.0));
            break;
        case FWEIBULLQF:
            a = DPop (); b = DPop ();
            if (a < 0.0 || a > 1.0) return RT_ARGUMENT;
            DPush (WeibullQF (a, b, 1.0, 0.0));
            break;
        case FEXPONENTIALDEN:
            DPush (ExpDen (DPop (), 1.0, 0.0));
            break;
        case FEXPONENTIALDF:
            DPush (ExpDF (DPop (), 1.0, 0.0));
            break;
        case FEXPONENTIALQF:
            b = DPop ();
            if (b < 0.0 || b >= 1.0) return RT_ARGUMENT;
	    DPush (ExpQF (b, 1.0, 0.0));
            break;
        case FGUMBELDEN:
            DPush (GumbelDen (DPop (), 1.0, 0.0));
            break;
        case FGUMBELDF:
            DPush (GumbelDF (DPop (), 1.0, 0.0));
            break;
        case FGUMBELQF:
            b = DPop ();
	    if (b <= 0.0 || b >= 1.0) return RT_ARGUMENT;
            DPush (GumbelQF (b, 1.0, 0.0));
            break;
        case FGPDDEN:
            a = DPop (); b = DPop ();
            DPush (vonMisesGPDDen (a, b, 1.0, 0.0));
            break;
        case FGPDDF:
            a = DPop (); b = DPop ();
            DPush (vonMisesGPDDF (a, b, 1.0, 0.0));
            break;
        case FGPDQF:
            a = DPop (); b = DPop ();
            if (a < 0.0 || a > 1.0 || (b > 0.0 && a >= 1.0)) return RT_ARGUMENT;
	    DPush (vonMisesGPDQF (a, b, 1.0, 0.0));
            break;
        case FEVDDEN:
            a = DPop (); b = DPop ();
            DPush (vonMisesEXTDen (a, b, 1.0, 0.0));
            break;
        case FEVDDF:
            a = DPop (); b = DPop ();
            DPush (vonMisesEXTDF (a, b, 1.0, 0.0));
            break;
        case FEVDQF:
            a = DPop (); b = DPop ();
            if (a < 0.0 || a > 1.0 || (b >= 0.0 && a >= 1.0)) return RT_ARGUMENT;
            DPush (vonMisesEXTQF (a, b, 1.0, 0.0));
            break;
	case FCHIDEN:
	    k = (int) DPop (); a = DPop ();
	    DPush (ChiSquareDensity (a, (int) k));
	    break;
	case FCHIQF:
	    k = (int) DPop (); a = DPop ();
            DPush (ChiSquareQF (a, (int) k));
            break;
        case FCHIDF:
	    k = (int) DPop (); a = DPop ();
            DPush (ChiSquareDF (a, (int) k));
	    break;
	case FGAMMADEN:
	    b = DPop (); a = DPop ();
	    DPush (XGammaDensity (a, b));
	    break;
	case FGAMMADF:
	    b = DPop (); a = DPop ();
	    DPush (XGammaDF (a, b));
	    break;
	case FTDEN:
	    b = DPop (); a = DPop ();
	    DPush (XTDensity (a, b));
	    break;
	case FTDF:
	    b = DPop (); a = DPop ();
	    DPush (XTDF (a, b));
	    break;
	case FTQF:
	    b = DPop (); a = DPop ();
	    DPush (XTQF (a, b));
	    break;
	case FNCTDEN:
	    c = DPop (); b = DPop (); a = DPop ();
            DPush (XNCTDensity (a, b, c));
	    break;
	case FNCTDF:
	    c = DPop (); b = DPop (); a = DPop ();
            DPush (XNCTDF (a, b, c));
	    break;
	case FNCTQF:
	    c = DPop (); b = DPop (); a = DPop ();
            DPush (XNCTQF (a, b, c));
	    break;
	case FHRDEN:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (HuesslerReissDen (a, b, c, d, e));
	    break;
	case FHRDF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (HuesslerReissDF (a, b, c, d, e));
	    break;
	case FHRSF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (HuesslerReissSF (a, b, c, d, e));
	    break;
	case FGMFDEN:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (GumbelMcFaddenEVDen (a, b, c, d, e));
	    break;
	case FGMFDF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (GumbelMcFaddenEVDF (a, b, c, d, e));
	    break;
	case FGMFSF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (GumbelMcFaddenEVSF (a, b, c, d, e));
	    break;
	case FHRGPDEN:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (HuesslerReissGPDen (a, b, c, d, e));
	    break;
	case FHRGPDF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (HuesslerReissGPDF (a, b, c, d, e));
	    break;
	case FHRGPSF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (HuesslerReissGPSF (a, b, c, d, e));
	    break;
	case FGMFGPDEN:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (GumbelMcFaddenGPDen (a, b, c, d, e));
	    break;
	case FGMFGPDF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (GumbelMcFaddenGPDF (a, b, c, d, e));
	    break;
	case FGMFGPSF:
	    e = DPop (); d = DPop (); c = DPop (); b = DPop (); a = DPop ();
	    DPush (GumbelMcFaddenGPSF (a, b, c, d, e));
	    break;
	case FGAMMA:
	    DPush (gamma (DPop ()));
	    break;

	case EXEC:
#ifndef TERMINAL
	    ExecuteProcedure ();
#endif
	    break;
	case PLAY: {
	    String s = PopStringFromVector ();
#ifndef TERMINAL
	    WavePlayer.PlayFile (s, PlotWindowManager.GetFramePtr ());
#endif
	    }
            break;
        case LDS:
            if (!ActSample) return RT_NOSAMPLE;
            IPush (((TUnivariateSample *) ActSample)->GetDataSize ());
            break;
        case LDD:
            if (!ActSample) return RT_NOSAMPLE;
            IPush (((TUnivariateSample *) ActSample)->GetDimension ());
            break;

        case SDF:
            if (!ActSample) return RT_NOSAMPLE;
            DPush (ActSample->SampleDf (DPop ()));
            break;
        case SQF:
            if (!ActSample) return RT_NOSAMPLE;
	    DPush (ActSample->SampleQf (DPop ()));
            break;
        case SDEN: {
            if (!ActSample) return RT_NOSAMPLE;
            k = IPop ();
            double b = DPop (), t = DPop ();
            DPush (ActSample->SampleDen (t, b, (int) k)); }
            break;
        case SMCS:
            if (!ActSample) return RT_NOSAMPLE;
            a = DPop (); k = IPop ();
            DPush (ActSample->MeanClusterSize ((int) k, a));
            break;
	case SREAD: {
            BOOL Result = SampleList.LoadSample (PopStringFromVector ());
            if (Result == TRUE)
#ifdef TERMINAL
                ActSample = SampleList.GetActiveSample ();
#else
                ActSample = SampleList.GetActiveSample (PlotWindowManager.GetFramePtr ());
#endif
            IPush ((int) Result);
            }
            break;
	case MBOX:
#ifndef TERMINAL
	    InfoBox (PlotWindowManager.GetFramePtr (), PopStringFromVector (), "");
#else
	    puts (PopStringFromVector ());
#endif
            break;
	case VBOX: {
            int k = IPop (), i;
            double *f = new double [k];
            for (i = 0; i < k; i++)
                f [k - 1 - i] = DPop ();
            String a = PopStringFromVector (),
            	   b = PopStringFromVector ();
            int Nr, *p;
	    GetStringPartition (a, p, Nr);
#ifndef TERMINAL
            TDialog d (IDD_XPLParaBox1 - 1 + Nr);
            d.SetTitel (b);
            TCalcDoubleField **e = new TCalcDoubleField * [Nr];
            TStaticField **s = new TStaticField * [Nr];
            for (i = 0; i < Nr; i++) {
                e [i] = new TCalcDoubleField (IDD_EditValue1 + i, i < k ? f [i] : 0.0, -DBL_MAX, DBL_MAX);
                s [i] = new TStaticField (IDD_EditName1 + i, a (p [i] + 1, p [i + 1] - p [i] - 1));
                d << (*e [i]) << (*s [i]);
            }
            if (d.Handle (PlotWindowManager.GetFramePtr ()) == IDOK) {
                for (i = 0; i < Nr; i++)
                    DPush (e [i]->GetValue ());
                IPush (Nr);
            } else
                IPush (0);
            for (i = 0; i < Nr; i++) {
                delete e [i];
                delete s [i];
            }
            delete [] e;
            delete [] s;
#else
	    for (i = 0; i < Nr; i++) {
                printf ("%s (%g): ", (const char *) a (p [i] + 1, p [i + 1] - p [i] - 1), i < k ? f [i] : 0.0);
                char buf [80];
                fgets (buf, 79, stdin);
                if (*buf == '\n') DPush (i < k ? f [i] : 0.0);
                else DPush (atof (buf));
            }
            IPush (Nr);
#endif
            delete [] f;
            delete [] p;
            }
            break;
	case CBOX: {
            String a = PopStringFromVector (),
            	   b = PopStringFromVector ();
            int Nr, i;
	    int *p;
	    GetStringPartition (a, p, Nr);
#ifndef TERMINAL
	    TDialog d (IDD_XPLMenu);
            d.SetTitel (b);
            TList l;
            for (i = 0; i < Nr; i++)
                l.Append (new TStringEl (a (p [i] + 1, p [i + 1] - p [i] - 1)));
            TListField f (IDD_List, &l);
            d << f;
            if (d.Handle (PlotWindowManager.GetFramePtr ()) == IDOK)
                IPush (f.Selection () + 1);
            else
                IPush (0);
            l.DeleteAll (TRUE);
#else
	    puts (b);
	    for (i = 0; i < Nr; i++) {
                printf ("%d: %s\n", i + 1, (const char *) a (p [i] + 1, p [i + 1] - p [i] - 1));
            }
	    int answer;
	    do {
		printf ("Enter choice (0-%d, 0 to abort): ", Nr);
                char buf [5];
                fgets (buf, 3, stdin);
		answer = atoi (buf);
	    } while (answer < 0 || answer > Nr);
	    IPush (answer);
#endif
            delete [] p;
            }
            break;
        case LDU:
            if (!ActSample) return RT_NOSAMPLE;
            k = IPop ();
            t = ActSample->GetSampleType ();
            if (t != UNIVARIATE_SAMPLE) return RT_ARGUMENT;
            if (k < 1 || k > ActSample->GetDataSize ()) return RT_ARGUMENT;
            DPush ((*((TUnivariateSample *) ActSample)) [(int) k - 1]);
            break;
	case LDA:
            if (!ActSample) return RT_NOSAMPLE;
            l = IPop ();
            k = IPop ();
            t = ActSample->GetSampleType ();
            if (k < 1 || k > ActSample->GetDataSize ()) return RT_ARGUMENT;
            if (t == UNIVARIATE_SAMPLE) {
                if (l != 1) return RT_ARGUMENT;
                DPush (((TUnivariateSample *) ActSample)->GetRawData ((int) k - 1));
            } else if (t == MULTIVARIATE_SAMPLE) {
                if (l < 1 || l > ActSample->GetDimension ()) return RT_ARGUMENT;
		DPush ((*((TMultivariateSample *) ActSample)) ((int) k - 1, (int) l - 1));
            } else if (t == TIMESERIES_SAMPLE)
                if (l == 1)
                    DPush ((*((TBivariateBaseSample *) ActSample))[(int) k - 1].x);
                else if (l == 2)
                    DPush ((*((TBivariateBaseSample *) ActSample))[(int) k - 1].y);
                else return RT_ARGUMENT;
#ifndef TERMINAL
            else if (t == CENSORED_SAMPLE)
                switch (l) {
                    case 1:
                        DPush (((TCensoredSample *) ActSample)->operator [] ((int) k - 1));
                        break;
                    case 2:
                        DPush (((TCensoredSample *) ActSample)->GetDelta ((int) k - 1));
                        break;
                    case 3:
                        DPush (((TCensoredSample *) ActSample)->GetWeights ((int) k - 1));
			break;
                    default:
                        return RT_ARGUMENT;
                }
#endif
            else if (t == GROUPED_SAMPLE)
		if (l == 1)
                    DPush (((TGroupedSample *) ActSample)->GetPartition ((int) k - 1));
                else if (l == 2)
                    DPush (((TGroupedSample *) ActSample)->GetFrequency ((int) k - 1));
                else
                    DPush (0.0);        // Fehler !!!! ?
                // Hier Prfung auf andere Werte -> Fehler !!!!
            break;
        case CTS: {
            String des = PopStringFromVector (),
                   fn = PopStringFromVector ();
            PopDoubleVector (y, k);
	    PopDoubleVector (x, m);
	    if (k < 1 || m < 1) return RT_ARGUMENT;
            if (ActSample && ActSample->GetFileName () == String ("")) delete ActSample;
            TTimeSeriesSample *p = new TTimeSeriesSample;
            p->SetName (des);
            p->SetData (min (k, m), x, y, fn);
            ActSample = p; }
            break;
        case CCS:
#ifndef TERMINAL
            k = IPop ();
	    l = IPop ();
            m = IPop ();
            n = IPop ();
            o = IPop ();
            if (ActSample && ActSample->GetFileName () == String ("")) delete ActSample;
            { TCensoredSample *p = new TCensoredSample;
              String des = PopStringFromVector (),
                     fn = PopStringFromVector ();
                  p->SetName (des);
//                TList s;
//                for (int i = 0; i < l; i++)
//                    s.Append (new TStringEl (sstore [k + i]));
//                p->SetDescription (&s);
//                s.DeleteAll (FALSE);
	      int *v = new int [(int) o];
              for (int i = 0; i < o; i++)
                  v[i] = (int) store [(int) m + i].intval;
              p->SetData ((int) o, (double *)   store + (int) n, v, fn);
	      ActSample = p;
              delete [] v; }
#endif
            break;
        case CS: {
	    String des = PopStringFromVector (),
                   fn = PopStringFromVector ();
	    PopDoubleVector (x, k);
	    if (k < 1) return RT_ARGUMENT;
            if (ActSample && ActSample->GetFileName () == String ("")) delete ActSample;
            TUnivariateSample *p = new TUnivariateSample;
            if (fn != String ("")) p->SetName (des);
            ActSample = p;
            p->SetData (k, x, fn);   }
            break;
        case MCMS:
            cms = new TMultivariateSample;
            k = IPop ();        // dimension
            l = IPop ();        // size
	    if (k < 1 || l < 1) return RT_ARGUMENT;
            cms->SetData (k, l);
	    break;
        case MSC: {
            String h = PopStringFromVector ();
	    PopDoubleVector (x, k);
            l = IPop ();        // column
            cms->SetColumn (l - 1, x, h); }
            break;
        case CNAME:
            if (!ActSample) return RT_NOSAMPLE;
            k = IPop ();
            if (k < 1 || k > ActSample->GetDimension ()) return RT_ARGUMENT;
	    PushStringAsVector (ActSample->GetColumnName (k - 1));
            break;
        case CDATA:
            if (!ActSample) return RT_NOSAMPLE;
            k = IPop ();
            if (k < 1 || k > ActSample->GetDimension ()) return RT_ARGUMENT;
            for (l = 0; l < ActSample->GetDataSize (); l++)
                DPush (ActSample->GetData (l, k - 1));
            IPush (ActSample->GetDataSize (), 1);
            IPush (ActSample->GetDataSize () + 1, 1);
	    break;
        case RDATA:
            if (!ActSample) return RT_NOSAMPLE;
            k = IPop ();
            if (k < 1 || k > ActSample->GetDataSize ()) return RT_ARGUMENT;
            for (l = 0; l < ActSample->GetDimension (); l++)
                DPush (ActSample->GetData (k - 1, l));
            IPush (ActSample->GetDimension (), 1);
            IPush (ActSample->GetDimension () + 1, 1);
	    break;
        case MSF: {
            if (ActSample && ActSample->GetFileName () == String ("")) delete ActSample;
	    String des = PopStringFromVector ();
	    String h = PopStringFromVector ();
            if (h != String ("")) {
                cms->SetFilename (h, des);
                cms->SaveData (0);
                SampleList.InsertSample (cms);
            }
            ActSample = cms; }
            break;
        case CMS: {
            String hdr = PopStringFromVector (),
                   des = PopStringFromVector (),
                   fn = PopStringFromVector ();
            k = IPop ();
	    m = IPop ();	// rows
            n = IPop ();        // columns
	    csp -= (k - 2);
            if (ActSample && ActSample->GetFileName () == String ("")) delete ActSample;
            TMultivariateSample *p = new TMultivariateSample;
            p->SetFilename (fn, des);
            p->SetData (n, m);
            double *x = new double [m];
	    int *q, Nr;
	    GetStringPartition (hdr, q, Nr);
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < m; j++)
                    x [j] = store [csp + i + j * n].doubleval;
		if (i < Nr)
		    p->SetColumn (i, x, hdr (q [i] + 1, q [i + 1] - q [i] - 1));
		else
                    p->SetColumn (i, x, form (20, "Column %d", i + 1));
            }
            delete [] x;
	    delete [] q;
            p->SaveData (0);
            ActSample = p;
            SampleList.InsertSample (p);
            }
            break;

// GP estimators

        case GP1HILL:
	    rt = CallPOTEstimator (HillGP1, TRUE);
	    break;
	case GP1HILLA:
	    rt = CallPOTEstimator (HillGP1, FALSE);
	    break;
	case GPMLE:
	    rt = CallPOTEstimator (MLEGP, TRUE);
	    break;
	case GPMLEA:
	    rt = CallPOTEstimator (MLEGP, FALSE);
	    break;
	case GPMOMENT:
	    rt = CallPOTEstimator (MomentGP, TRUE);
	    break;
	case GPMOMENTA:
	    rt = CallPOTEstimator (MomentGP, FALSE);
	    break;
	case GPDREES:
	    rt = CallPOTEstimator (DreesPickandsGP, TRUE);
	    break;
	case GPDREESA:
	    rt = CallPOTEstimator (DreesPickandsGP, FALSE);
	    break;
	case GP0MLE:
	    rt = CallPOTEstimator (MLEExponential, TRUE, 2);
	    break;
	case GP0MLEA:
	    rt = CallPOTEstimator (MLEExponential, FALSE, 2);
	    break;
#ifndef TERMINAL
	case GP1ME:
	    rt = CallPOTEstimator (MEGP1, TRUE, 3);
	    break;
	case GP1MEA:
	    rt = CallPOTEstimator (MEGP1, FALSE, 3);
	    break;
#endif

// EV estimators

	case EV1MLE:
            rt = CallOtherEstimator (MLEEV1, TRUE);
            break;
        case EV1MLEA:
            rt = CallOtherEstimator (MLEEV1, FALSE);
            break;
	case EVMLE:
	    rt = CallOtherEstimator (MLEEV, TRUE);
	    break;
	case EVMLEA:
	    rt = CallOtherEstimator (MLEEV, FALSE);
	    break;
	case EVRDE:
	    rt = CallOtherEstimator (RDEEV, TRUE);
	    break;
	case EVRDEA:
	    rt = CallOtherEstimator (RDEEV, FALSE);
	    break;
	case EV3MLE:
	    rt = CallOtherEstimator (MLEGumbel, TRUE, 2);
	    break;
	case EV3MLEA:
	    rt = CallOtherEstimator (MLEGumbel, FALSE, 2);
	    break;
        case EVMDE:
	    rt = CallOtherEstimator (MDEEV, TRUE);
	    break;
        case EVMDEA:
	    rt = CallOtherEstimator (MDEEV, FALSE);
	    break;

// Student estimators

	case TMLE:
            rt = CallOtherEstimator (MLET, TRUE, 2);
            break;
	case TMLEA:
            rt = CallOtherEstimator (MLET, FALSE, 2);
            break;
	case NCTMLE:
            rt = CallOtherEstimator (MLENCT, TRUE, 3);
            break;
	case NCTMLEA:
            rt = CallOtherEstimator (MLENCT, FALSE, 3);
            break;

// Gaussian estimators

	case GAUSSMLE:
	    rt = CallOtherEstimator (MLEGaussian, TRUE, 2);
	    break;
	case GAUSSMLEA:
	    rt = CallOtherEstimator (MLEGaussian, FALSE, 2);
	    break;

	case GAUSSLSE:
	    rt = CallOtherEstimator (LSEGaussian, TRUE, 2);
	    break;
	case GAUSSLSEA:
	    rt = CallOtherEstimator (LSEGaussian, FALSE, 2);
	    break;
#ifndef TERMINAL
	case GAUSSMDE:
	    rt = CallOtherEstimator (MDEGaussian, TRUE, 2);
            break;
	case GAUSSMDEA:
	    rt = CallOtherEstimator (MDEGaussian, FALSE, 2);
            break;
        case GAUSSMHDE:
            rt = CallOtherEstimator (MHDEGaussian, TRUE, 2);
	    break;
        case GAUSSMHDEA:
            rt = CallOtherEstimator (MHDEGaussian, FALSE, 2);
	    break;
#endif



#ifndef TERMINAL
	case GP1HILLCEN:
// !!!!	    rt = CallPOTEstimator (HillCen);
	    break;
	case GPMLECEN:
// !!!!	    rt = CallPOTEstimator (MLECen);
	    break;
#endif


#ifndef TERMINAL
        case PLOT3D:
        case CONT3D: {
            BOOL Contour = p.code == CONT3D;
	    int dummy;
	    TStorage *ptr = VectorStart ();
	    IPop (k, dummy);
	    IPop (m, n);
	    ReleaseCalcStack (k - 1);
	    double y1 = DPop (), y0 = DPop (), x1 = DPop (), x0 = DPop (), **v;
            v = new double * [(int) n];
            for (int i = 0; i < n; i++) {
                v [i] = new double [(int) m];
                for (int j = 0; j < m; j++)
                    v[i][j] = ptr [i * m + j].doubleval;
            }
            TBivariateFunction *f = new TXPLBivFunction (x0, x1, y0, y1, v, n, m);
            if (Contour)
                PlotWindowManager.GetOpenWindow ("Contour")->InsertFunction (new TContourPlot (f, n, m));
	    else
		new TFunctionPlot3D (PlotWindowManager.GetFramePtr (), f);
	    }
            break;

#endif
        case SEPS: {
	    double b = DPop (), t = DPop (), r = DPop (), l = DPop (), h = DPop (), br = DPop ();
	    String fn = PopStringFromVector (), win = PopStringFromVector ();
#ifndef TERMINAL
	    TPlotWindow *w = PlotWindowManager.GetOpenWindow (win);
	    w->CreateEPS (fn, br, h, l, r, t, b);
#endif
	    }
	    break;

	case SETCS: {
	    double y1 = DPop (), y0 = DPop (), x1 = DPop (), x0 = DPop ();
	    String win = PopStringFromVector ();
#ifndef TERMINAL
	    TPlotWindow *w = PlotWindowManager.GetOpenWindow (win);
	    w->SetNewCoordinates (TDblRectangle (x0, x1, y0, y1), FALSE);
	    w->Repaint ();
#endif
	    } break;
	case SETCT: {
	    double ys = DPop (), yt = DPop (), yf = DPop (),
		   xs = DPop (), xt = DPop (), xf = DPop ();
	    String win = PopStringFromVector ();
#ifndef TERMINAL
	    TPlotWindow *w = PlotWindowManager.GetOpenWindow (win);
	    w->SetTicks (xf, xt, xs, yf, yt, ys);
	    w->Repaint ();
#endif
	    } break;
	case SETCM: {
	    double ys = DPop (), yt = DPop (), yf = DPop (),
		   xs = DPop (), xt = DPop (), xf = DPop ();
	    String win = PopStringFromVector ();
#ifndef TERMINAL
	    TPlotWindow *w = PlotWindowManager.GetOpenWindow (win);
	    w->SetMarkers (xf, xt, xs, yf, yt, ys);
	    w->Repaint ();
#endif
	    } break;
	case SETCF: {
	    int col = IPop (), fontsize = IPop (), fontstyle = IPop (), fonttype = IPop ();
	    String win = PopStringFromVector ();
#ifndef TERMINAL    
	    TPlotWindow *w = PlotWindowManager.GetOpenWindow (win);
	    TFont tf ((TFontType) fonttype, (TFontStyle) fontstyle, fontsize);
	    TColor c (col);
	    w->SetMarkerFont (tf, c);
	    w->Repaint ();
#endif
	    } break;
	case SETLF: {
	    int fontsize = IPop (), fontstyle = IPop (), fonttype = IPop ();
#ifndef TERMINAL
	    PlotFont = TFont ((TFontType) fonttype, (TFontStyle) fontstyle, fontsize);
#endif
	    } break;
	case SETLB: {
	    String lab = PopStringFromVector ();
	    double y = DPop (), x = DPop ();
	    k = IPop ();
	    String win = PopStringFromVector ();
#ifndef TERMINAL
	    TPlotWindow *w = PlotWindowManager.GetOpenWindow (win);
	    TTextField *tf = new TTextField (TDblPoint (x, y), (int) k, lab);
	    tf->SetFont (PlotFont);
	    tf->SetColor (PlotColor);
	    w->InsertFunction (tf);
#endif
	    } break;
	case CLWIN: {
	    String win = PopStringFromVector ();
#ifndef TERMINAL
	    PlotWindowManager.GetOpenWindow (win)->DeleteFunctions (); 
#endif
	    }
	    break;
	case PLOT: {
            String des = PopStringFromVector (),
                   win = PopStringFromVector ();
	    PopDoubleVector (y, k);
	    PopDoubleVector (x, l);
	    m = min (k, l);
#ifndef TERMINAL
            TFunction *f = new TXPLFunction (m, x, y, des, TRUE);
            f->SetColor (PlotColor);
            f->SetLineStyle ((TLineStyle) LineStyle, LineThickness, LineLine, LineGap);
            f->SetPlotMode ((TPlotMode) PlotMode);
            PlotWindowManager.GetOpenWindow (win)->InsertFunction (f, TRUE);
#endif
	    }
            break;
	case BOXPLOT: {
            String des = PopStringFromVector (),
                   win = PopStringFromVector ();
	    PopDoubleVector (x, k);
	    a = DPop ();
#ifndef TERMINAL
            TFunction *f = new TBoxPlot (k, x, a, des);
            f->SetColor (0);
            PlotWindowManager.GetOpenWindow (win)->InsertFunction (f, TRUE);
#endif
	    }
            break;

	case FRUINSIM: {	// parameters collected in vector
	    PopDoubleVector (x, k);
	    if (k != 8) return RT_ARGUMENT;
	    DPush (SimulateSingleGPRuinTime (x)); }
	    break;
	case FRUINSIMB: {	//  single parameters
	    double p [8];
	    for (int i = 7; i >= 0; i--)
		p [i] = DPop ();
	    DPush (SimulateSingleGPRuinTime (p)); }
	    break;

        case FUNIFORM: {
            k = l = IPop ();
	    ptr = CalcStackFirstFree ();
	    while (l--)
		(ptr++)->doubleval = Random ();
	    ReserveCalcStack (k);
            IPush (k, 1);
	    IPush (k + 1, 1);  }
            break;

	case TST:
	    PopDoubleVector (x, n);
	    PopDoubleVector (y, m);
	    DPush (ttest (x, n, y, m));
	    break;
	case TSWELCH:
	    PopDoubleVector (x, n);
	    PopDoubleVector (y, m);
	    DPush (welchtest (x, n, y, m));
	    break;
	case TSWILCOXON:
	    PopDoubleVector (x, n);
	    PopDoubleVector (y, m);
	    DPush (wilcoxontest (x, n, y, m));
	    break;
	    
	case LDT:
            if (ActSample)
                switch (ActSample->GetSampleType ()) {
                    case UNIVARIATE_SAMPLE:
                        PushStringAsVector ("univariate");
                        break;
                    case TIMESERIES_SAMPLE:
                        PushStringAsVector ("time series");
                        break;
                    case GROUPED_SAMPLE:
                        PushStringAsVector ("grouped");
                        break;
                    case CENSORED_SAMPLE:
                        PushStringAsVector ("censored");
                        break;
                    case MULTIVARIATE_SAMPLE:
                        PushStringAsVector ("multivariate");
                        break;
                    case FREQUENCY_SAMPLE:
                        PushStringAsVector ("frequency");
			break;
                    default:
                        PushStringAsVector ("unknown");
                        break;
                }
            else PushStringAsVector ("nothing");
	    break;

        case SGV: {
	    PopDoubleVector (x, n);
            double *p = new double [n];
	    memcpy (p, x, n * sizeof (double));
            String s = PopStringFromVector ();
            for (TObject *q = GlobalVarList.GetFirst (); q; q = GlobalVarList.GetNext ())
                if (q->Name () == s) {
                    GlobalVarList.DeleteActive (TRUE);
                    break;
                }
	    GlobalVarList.Append (new TGlobalVectorVarStorage (s, p, n)); }
            break;
        case GGV: {
            String s = PopStringFromVector ();
            BOOL Found = FALSE;
            for (TObject *p = GlobalVarList.GetFirst (); p; p = GlobalVarList.GetNext ())
                if (p->Name () == s) {
                    double *q = ((TGlobalVectorVarStorage *) p)->GetValue ();
                    int n = ((TGlobalVectorVarStorage *) p)->GetSize ();
                    for (int i = 0; i < n; i++)
                        DPush (q [i]);
                    IPush (n, 1);
		    IPush (n + 1, 1);
                    Found = TRUE;
                    break;
                }
            if (!Found) {
                IPush (0, 1);
		IPush (1, 1); 
	    }
	    }
            break;

        case FMEAN:
	    PopDoubleVector (x, n);
	    DPush (mean (x, n));
            break;
        case FVAR:
	    PopDoubleVector (x, n);
	    DPush (variance (x, n));
            break;
        case FMOM:
            l = IPop ();
	    PopDoubleVector (x, n);
            a = 0.0;
            for (m = 0; m < n; m++)
                a += pow (fabs (x [m]), l);
            if (n) DPush (a / n);
            else DPush (0.0);
            break;

        case FSMOOTH: {
            k = IPop ();	// bandwidth: neighbors
	    ptr = VectorStart ();
            n = VectorLength ();
            double *v = new double [n];
            a = 0.0;
            for (l = 0; l < k; l++)
                a += ptr[l].doubleval;
            for (l = 0; l < n; l++) {
                if (l > k / 2 && l < n - k / 2)
                    a = a - ptr [l - k / 2 - 1].doubleval + ptr [l + k / 2].doubleval;
                v [l] = a / k;
            }
            for (l = 0; l < n; l++)
                ptr [l].doubleval = v [l];
            delete [] v; }
	    break;

        case FMEDIAN: 
	    PopDoubleVector (x, k);
	    if (!k) DPush (0.0);
	    else {
                qsort (x, k, sizeof (double), realcomp);
                DPush (x [k / 2]);
            }
            break;

        case REG: {
            k = IPop ();        // Degree
	    PopDoubleVector (y, l);
	    PopDoubleVector (x, n);
            double *c = PolynomialRegression (k, min (l, n), x, y);
	    for (l = 0; l <= k; l++) 
		DPush (c [l]);
	    IPush (k + 1, 1);
	    IPush (k + 2, 1);
            delete [] c; }
            break;

	case POLY: 
	    a = DPop ();
	    PopDoubleVector (x, k);
	    b = 0.0;
	    while (k--) 
		b = b * a + x [k];
	    DPush (b);
	    break;

#ifndef TERMINAL
        case SETCOL:
            PlotColor = IPop ();
            break;
        case SETOPT:
	    LineGap = IPop ();
	    LineLine = IPop ();
            LineThickness = IPop ();
            break;
        case SETLIN:
            LineStyle = IPop ();
            break;
        case SETSTY:
            PlotMode = IPop ();
            break;
#endif
        case SCA: {
            String des = PopStringFromVector ();
	    PopDoubleVector (y, k);
	    PopDoubleVector (x, l);
	    m = min (k, l);
#ifndef TERMINAL
	    new TScatterPlotWindow (PlotWindowManager.GetFramePtr (), des, m, x, y, TRUE);
#endif
	    }
	    break;
	default:
	    return TPCodeInterpreter::ExtendedCode (p);
    }
    return rt;
}

void TXPLPCodeInterpreter::GetStringPartition (String a, int *&p, int &Nr) {
    int i, j;
    Nr = 1;
    for (i = 0; i < a.Len (); i++)
        if (((const char *) a) [i] == '|') Nr++;
    p = new int [Nr + 1];
    p [0] = -1;
    for (i = 0, j = 1; i < a.Len (); i++)
        if (((const char *) a) [i] == '|')
            p [j++] = i;
    p [Nr] = a.Len () + 1;
}



/*
 * Add Xtremes specific predefined identifiers to the XPL compiler
 */


TXPLCompiler::TXPLCompiler (TPCodeInterpreter *p, int ProgramSize, String PathName)
 : TCompiler (p, ProgramSize, PathName) {
}

void TXPLCompiler::Compile () {
    SymbolTable = new TSymbolTable;
    PredefinedFunctionList = new TPredefinedFunctionList;
    Compiler = this;
#ifdef COMPONENT
    Scanner->GetNextToken ();
    ComponentFlag = Scanner->GetToken () == _COMPONENT;
    if (ComponentFlag) {
	Component ();
	return;
    }
    Scanner->Restart ();
#endif
    Program ();
    delete PredefinedFunctionList;
    delete SymbolTable;
}


#ifdef COMPONENT

String TXPLCompiler::GetComponentName () {
    return ComponentName;
}

int TXPLCompiler::GetNumberOfInports () {
    return InputParameters->GetNumberOfParameters ();
}

int TXPLCompiler::GetNumberOfOutports () {
    TList l;
    SymbolTable->GetVariableTypes (0, &l);
    return l.Entries ();
}

XGPLNodes::TPortType *TXPLCompiler::GetInportTypes () {
    XGPLNodes::TPortType *p = new XGPLNodes::TPortType [GetNumberOfInports ()];
    TType *t = InputParameters->GetFirstParaType ();
    for (int i = 0; i < GetNumberOfInports (); i++) {
	p [i] = ConvertType (t);
	t = InputParameters->GetNextParaType ();
    }
    return p;
}

XGPLNodes::TPortType *TXPLCompiler::GetOutportTypes () {
    TList l;
    SymbolTable->GetVariableTypes (0, &l);
    XGPLNodes::TPortType *p = new XGPLNodes::TPortType [l.Entries ()];
    TType *t = (TType *) l.GetFirst ();
    for (int i = 0; i < l.Entries (); i++) {
	p [i] = ConvertType (t);
	t = (TType *) l.GetNext ();
    }
    return p;         
}

XGPLNodes::TPortType TXPLCompiler::ConvertType (TType *t) {
    if (t == SymbolTable->TINT)
        return XGPLNodes::PORT_INT;
    if (t == SymbolTable->TREAL)
        return XGPLNodes::PORT_REAL;
    if (t == SymbolTable->TREAL2)
   	return XGPLNodes::PORT_REAL2;
    if (t == SymbolTable->TREAL3)
        return XGPLNodes::PORT_REAL3;
    if (t == SymbolTable->TREALVECT)
	return XGPLNodes::PORT_DATA;
    return XGPLNodes::PORT_MATRIX;
}

void TXPLCompiler::Component () {
    Scanner->GetNextToken ();
    if (Scanner->GetToken () != IDENTIFIER) {
    	ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "Component must be followed by an identifier");
	ComponentName = "DUMMY";
    } else
    	ComponentName = Scanner->GetString();
    Scanner->GetNextToken ();
    if (!Scanner->Try (SEMICOLON))
        ErrorMessage (SYNTAX_ERROR, Scanner->GetLine (), "';' expected");
    SymbolTable->EnterLevel (0);
    SymbolTable->BuildTypes ();

    InputParameters = new TRoutine;
    if (Scanner->Try (INPORTS)) {
        Declaration (TRUE, FALSE);
        SymbolTable->AssignParameters (InputParameters);
    }
    if (Scanner->Try (OUTPORTS))
    	Declaration (FALSE, FALSE);

    OutCode (Scanner->GetLine (), LOAD, 0);
    OutCode (Scanner->GetLine (), PUSH, 1);	// simulate call by pushing dummy return address

    MainBlock ();
}

BOOL TXPLCompiler::IsComponent () {
    return ComponentFlag;
}

#endif


void TXPLCompiler::BuildUserSymbols () {
    char *name [] = {
	"verticalpositiontop",
	"verticalpositioncentered",
	"verticalpositionbottom",
	"verticalpositionuser",

	"horizontalpositionleft",
	"horizontalpositioncentered",
	"horizontalpositionright",
	"horizontalpositionuser",

	"verticaldisplay",
	"horizontaldisplay",

	"verticalalignmenttop",
	"verticalalignmentcentered",
	"verticalalignmentbottom",

	"horizontalalignmentleft",
	"horizontalalignmentcentered",
	"horizontalalignmentright",

	"colorblack",
	"colorred",
	"colorgreen",
	"colorblue",
	"colorgrey",

	"solidline",
	"dashedline",
	"dottedline",
	"dashdottedline",
	"userline",

	"linestyle",
	"pointstyle",
	"barstyle",
	"boxstyle",
	"staircasestyle",
	"areastyle",
	
	"times",
	"helvetica",
	"courier",
	"symbol",
	
	"normal",
	"bold",
	"italic",
	"bolditalic" };

    int value [] = {
	0,	1,	2,	3,
	0,	4,	8,	12,
	16,	0,
	0,	32,	64,
	0,	128,	256,
	0,	1,	2,	3,	4,
	0,	1,	2,	3,	4,
	0,	1,	2,	3,	4,	5,
	0,	1,	2,	3,
	0,	1,	2,	3
    };
    for (int i = 0; i < (int) (sizeof (value) / sizeof (int)); i++) {
	TSymbol *p = SymbolTable->NewSymbol (name [i], SymbolTable->TINT, 0);
	p->SetConstant (TRUE);
	p->SetIntVal (value [i]);
    }
}

void TXPLCompiler::AddGPEstimator (String Name, int CodeFull, int CodeActiveSample) {
    TType *R = SymbolTable->TREAL,
          *N = SymbolTable->TINT,
          *RV = SymbolTable->TREALVECT,
//          *NV = SymbolTable->TINTVECT,
//          *B = SymbolTable->TBOOL,
//          *S = SymbolTable->TSTRING,
          *V = SymbolTable->TVOID;
    PredefinedFunctionList->Append (Name, CodeActiveSample, CodeActiveSample, V, N, R, R, R, N);
    PredefinedFunctionList->SetReference (FALSE, TRUE, TRUE, TRUE, TRUE);
    PredefinedFunctionList->Append (Name, CodeFull, CodeFull, V, RV, N, R, R, R, N);
    PredefinedFunctionList->SetReference (FALSE, FALSE, TRUE, TRUE, TRUE, TRUE);
}

void TXPLCompiler::AddOtherEstimator (String Name, int CodeFull, int CodeActiveSample, int NrOfParameters) {
    TType *R = SymbolTable->TREAL,
          *N = SymbolTable->TINT,
          *RV = SymbolTable->TREALVECT,
//          *NV = SymbolTable->TINTVECT,
//          *B = SymbolTable->TBOOL,
//          *S = SymbolTable->TSTRING,
          *V = SymbolTable->TVOID;
    if (NrOfParameters == 3) {
        PredefinedFunctionList->Append (Name, CodeActiveSample, CodeActiveSample, V, R, R, R, N);
        PredefinedFunctionList->SetReference (TRUE, TRUE, TRUE, TRUE);
        PredefinedFunctionList->Append (Name, CodeFull, CodeFull, V, RV, R, R, R, N);
        PredefinedFunctionList->SetReference (FALSE, TRUE, TRUE, TRUE, TRUE);
    } else {
        PredefinedFunctionList->Append (Name, CodeActiveSample, CodeActiveSample, V, R, R, N);
        PredefinedFunctionList->SetReference (TRUE, TRUE, TRUE);
        PredefinedFunctionList->Append (Name, CodeFull, CodeFull, V, RV, R, R, N);
        PredefinedFunctionList->SetReference (FALSE, TRUE, TRUE, TRUE);
    }
}

void TXPLCompiler::BuildPredefinedFunctions () {
    inherited::BuildPredefinedFunctions ();

    TType *R = SymbolTable->TREAL,
          *N = SymbolTable->TINT,
          *RV = SymbolTable->TREALVECT,
	  *RM = SymbolTable->TREALMATR,
//          *NV = SymbolTable->TINTVECT,
//          *B = SymbolTable->TBOOL,
          *S = SymbolTable->TSTRING,
          *V = SymbolTable->TVOID;

// Distributions

    PredefinedFunctionList->Append ("paretodata", FPARETORV, FPARETORV, R, R);
    PredefinedFunctionList->Append ("gaussiandata", FGAUSSRV, FGAUSSRV, R);
    PredefinedFunctionList->Append ("normaldata", FNORMALRV, FNORMALRV, R);
    PredefinedFunctionList->Append ("gcauchydata", FCAUCHYRV, FCAUCHYRV, R, R);
    PredefinedFunctionList->Append ("frechetdata", FFRECHETRV, FFRECHETRV, R, R);
    PredefinedFunctionList->Append ("betadata", FBETARV, FBETARV, R, R);
    PredefinedFunctionList->Append ("weibulldata", FWEIBULLRV, FWEIBULLRV, R, R);
    PredefinedFunctionList->Append ("gumbeldata", FGUMBELRV, FGUMBELRV, R);
    PredefinedFunctionList->Append ("exponentialdata", FEXPONENTIALRV, FEXPONENTIALRV, R);
    PredefinedFunctionList->Append ("gpdata", FGPDRV, FGPDRV, R, R);
    PredefinedFunctionList->Append ("evdata", FEVDRV, FEVDRV, R, R);
    PredefinedFunctionList->Append ("gammadata", FGAMMARV, FGAMMARV, R, R),
    PredefinedFunctionList->Append ("chisquaredata", FCHIRV, FCHIRV, R, N);
    PredefinedFunctionList->Append ("tdata", FTRV, FTRV, R, R);

    PredefinedFunctionList->Append ("paretodata", VFPARETORV, VFPARETORV, RV, R, N);
    PredefinedFunctionList->Append ("gaussiandata", VFGAUSSRV, VFGAUSSRV, RV, N);
    PredefinedFunctionList->Append ("normaldata", VFNORMALRV, VFNORMALRV, RV, N);
    PredefinedFunctionList->Append ("gcauchydata", VFCAUCHYRV, VFCAUCHYRV, RV, R, N);
    PredefinedFunctionList->Append ("frechetdata", VFFRECHETRV, VFFRECHETRV, RV, R, N);
    PredefinedFunctionList->Append ("betadata", VFBETARV, VFBETARV, RV, R, N);
    PredefinedFunctionList->Append ("weibulldata", VFWEIBULLRV, VFWEIBULLRV, RV, R, N);
    PredefinedFunctionList->Append ("gumbeldata", VFGUMBELRV, VFGUMBELRV, RV, N);
    PredefinedFunctionList->Append ("exponentialdata", VFEXPONENTIALRV, VFEXPONENTIALRV, RV, N);
    PredefinedFunctionList->Append ("gpdata", VFGPDRV, VFGPDRV, RV, R, N);
    PredefinedFunctionList->Append ("evdata", VFEVDRV, VFEVDRV, RV, R, N);
    PredefinedFunctionList->Append ("gammadata", VFGAMMARV, VFGAMMARV, RV, R, N),
    PredefinedFunctionList->Append ("chisquaredata", VFCHIRV, VFCHIRV, RV, N, N);
    PredefinedFunctionList->Append ("tdata", VFTRV, VFTRV, RV, R, N);

    PredefinedFunctionList->Append ("paretodensity", FPARETODEN, FPARETODEN, R, R, R);
    PredefinedFunctionList->Append ("paretoqf", FPARETOQF, FPARETOQF, R, R, R);
    PredefinedFunctionList->Append ("paretodf", FPARETODF, FPARETODF, R, R, R);
    PredefinedFunctionList->Append ("gaussiandensity", FGAUSSDEN, FGAUSSDEN, R, R);
    PredefinedFunctionList->Append ("gaussiandensity", FGAUSSDENV, FGAUSSDENV, RV, RV);
    PredefinedFunctionList->Append ("gaussianqf", FGAUSSQF, FGAUSSQF, R, R);
    PredefinedFunctionList->Append ("gaussianqf", FGAUSSQFV, FGAUSSQFV, RV, RV);
    PredefinedFunctionList->Append ("gaussiandf", FGAUSSDF, FGAUSSDF, R, R);
    PredefinedFunctionList->Append ("gaussiandf", FGAUSSDFV, FGAUSSDFV, RV, RV);
    PredefinedFunctionList->Append ("normaldensity", FGAUSSDEN, FGAUSSDEN, R, R);
    PredefinedFunctionList->Append ("normaldensity", FGAUSSDENV, FGAUSSDENV, RV, RV);
    PredefinedFunctionList->Append ("normalqf", FGAUSSQF, FGAUSSQF, R, R);
    PredefinedFunctionList->Append ("normalqf", FGAUSSQFV, FGAUSSQFV, RV, RV);
    PredefinedFunctionList->Append ("normaldf", FGAUSSDF, FGAUSSDF, R, R);
    PredefinedFunctionList->Append ("normaldf", FGAUSSDFV, FGAUSSDFV, RV, RV);
    PredefinedFunctionList->Append ("gcauchydensity", FCAUCHYDEN, FCAUCHYDEN, R, R, R);
    PredefinedFunctionList->Append ("frechetdensity", FFRECHETDEN, FFRECHETDEN, R, R, R);
    PredefinedFunctionList->Append ("frechetqf", FFRECHETQF, FFRECHETQF, R, R, R);
    PredefinedFunctionList->Append ("frechetdf", FFRECHETDF, FFRECHETDF, R, R, R);
    PredefinedFunctionList->Append ("betadensity", FBETADEN, FBETADEN, R, R, R);
    PredefinedFunctionList->Append ("betaqf", FBETAQF, FBETAQF, R, R, R);
    PredefinedFunctionList->Append ("betadf", FBETADF, FBETADF, R, R, R);
    PredefinedFunctionList->Append ("weibulldensity", FWEIBULLDEN, FWEIBULLDEN, R, R, R);
    PredefinedFunctionList->Append ("weibullqf", FWEIBULLQF, FWEIBULLQF, R, R, R);
    PredefinedFunctionList->Append ("weibulldf", FWEIBULLDF, FWEIBULLDF, R, R, R);
    PredefinedFunctionList->Append ("exponentialdensity", FEXPONENTIALDEN, FEXPONENTIALDEN, R, R);
    PredefinedFunctionList->Append ("exponentialqf", FEXPONENTIALQF, FEXPONENTIALQF, R, R);
    PredefinedFunctionList->Append ("exponentialdf", FEXPONENTIALDF, FEXPONENTIALDF, R, R);
    PredefinedFunctionList->Append ("gumbeldensity", FGUMBELDEN, FGUMBELDEN, R, R);
    PredefinedFunctionList->Append ("gumbelqf", FGUMBELQF, FGUMBELQF, R, R);
    PredefinedFunctionList->Append ("gumbeldf", FGUMBELDF, FGUMBELDF, R, R);
    PredefinedFunctionList->Append ("gpdensity", FGPDDEN, FGPDDEN, R, R, R);
    PredefinedFunctionList->Append ("gpqf", FGPDQF, FGPDQF, R, R, R);
    PredefinedFunctionList->Append ("gpdf", FGPDDF, FGPDDF, R, R, R);
    PredefinedFunctionList->Append ("evdensity", FEVDDEN, FEVDDEN, R, R, R);
    PredefinedFunctionList->Append ("evqf", FEVDQF, FEVDQF, R, R, R);
    PredefinedFunctionList->Append ("evdf", FEVDDF, FEVDDF, R, R, R);
    PredefinedFunctionList->Append ("chisquaredensity", FCHIDEN, FCHIDEN, R, R, R);
    PredefinedFunctionList->Append ("chisquareqf", FCHIQF, FCHIQF, R, R, R);
    PredefinedFunctionList->Append ("chisquaredf", FCHIDF, FCHIDF, R, R, R);
    PredefinedFunctionList->Append ("gammadensity", FGAMMADEN, FGAMMADEN, R, R, R);
    PredefinedFunctionList->Append ("gammadf", FGAMMADF, FGAMMADF, R, R, R);
    PredefinedFunctionList->Append ("tdensity", FTDEN, FTDEN, R, R, R);
    PredefinedFunctionList->Append ("tdf", FTDF, FTDF, R, R, R);
    PredefinedFunctionList->Append ("tqf", FTQF, FTQF, R, R, R);
    PredefinedFunctionList->Append ("nctdensity", FNCTDEN, FNCTDEN, R, R, R, R);
    PredefinedFunctionList->Append ("nctdf", FNCTDF, FNCTDF, R, R, R, R);
    PredefinedFunctionList->Append ("nctqf", FNCTQF, FNCTQF, R, R, R, R);
    PredefinedFunctionList->Append ("hrevdensity", FHRDEN, FHRDEN, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("hrevdf", FHRDF, FHRDF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("hrevsf", FHRSF, FHRSF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("hrgpdensity", FHRGPDEN, FHRGPDEN, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("hrgpdf", FHRGPDF, FHRGPDF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("hrgpsf", FHRGPSF, FHRGPSF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gmfevdensity", FGMFDEN, FGMFDEN, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gmfevdf", FGMFDF, FGMFDF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gmfevsf", FGMFSF, FGMFSF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gmfgpdensity", FGMFGPDEN, FGMFGPDEN, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gmfgpdf", FGMFGPDF, FGMFGPDF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gmfgpsf", FGMFGPSF, FGMFGPSF, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("gamma", FGAMMA, FGAMMA, R, R);
    PredefinedFunctionList->Append ("random", FUNIFORM, FUNIFORM, RV, N);

    PredefinedFunctionList->Append ("mean", FMEAN, FMEAN, R, RV);
    PredefinedFunctionList->Append ("median", FMEDIAN, FMEDIAN, R, RV);
    PredefinedFunctionList->Append ("polynomialregression", REG, REG, RV, RV, RV, N);
    PredefinedFunctionList->Append ("poly", POLY, POLY, R, RV, R);
    PredefinedFunctionList->Append ("variance", FVAR, FVAR, R, RV);
    PredefinedFunctionList->Append ("moment", FMOM, FMOM, R, RV, N);
    PredefinedFunctionList->Append ("smooth", FSMOOTH, FSMOOTH, RV, RV, N);
    PredefinedFunctionList->Append ("simulateruintime", FRUINSIMB, FRUINSIMB, R, R, R, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("simulateruintime", FRUINSIM, FRUINSIM, R, RV);

// Access active sample

    PredefinedFunctionList->Append ("samplesize", LDS, LDS, N);
    PredefinedFunctionList->Append ("dimension", LDD, LDD, N);
    PredefinedFunctionList->Append ("columnname", CNAME, CNAME, S, N);
    PredefinedFunctionList->Append ("columndata", CDATA, CDATA, RV, N);
    PredefinedFunctionList->Append ("rowdata", RDATA, RDATA, RV, N);
    PredefinedFunctionList->Append ("data", LDA, LDA, R, N, N);
    PredefinedFunctionList->Append ("data", LDU, LDU, R, N);
    PredefinedFunctionList->Append ("sampleqf", SQF, SQF, R, R);
    PredefinedFunctionList->Append ("sampledf", SDF, SDF, R, R);
    PredefinedFunctionList->Append ("samplemeanclustersize", SMCS, SMCS, R, N, R);
    PredefinedFunctionList->Append ("kerneldensity", SDEN, SDEN, R, R, R, N);
    PredefinedFunctionList->Append ("datatype", LDT, LDT, S);
    PredefinedFunctionList->Append ("readdata", SREAD, SREAD, N, S);

// Generate samples

    PredefinedFunctionList->Append ("createunivariate", CS, CS, V, RV, S, S);
    PredefinedFunctionList->Append ("createtimeseries", CTS, CTS, V, RV, RV, S, S);
    PredefinedFunctionList->Append ("createmultivariate", CMS, CMS, V, RM, S, S, S);
    PredefinedFunctionList->Append ("beginmultivariate", MCMS, MCMS, V, N, N);
    PredefinedFunctionList->Append ("setcolumn", MSC, MSC, V, N, RV, S);
    PredefinedFunctionList->Append ("endmultivariate", MSF, MSF, V, S, S);

// Estimators

    AddGPEstimator ("hillgp1", GP1HILL, GP1HILLA);
    AddGPEstimator ("megp1", GP1ME, GP1MEA);
    AddGPEstimator ("mlegp", GPMLE, GPMLEA);
    AddGPEstimator ("momentgp", GPMOMENT, GPMOMENTA);
    AddGPEstimator ("dreespickandsgp", GPDREES, GPDREESA);

    PredefinedFunctionList->Append ("mlegp0", GP0MLEA, GP0MLEA, V, N, R, R, N);
    PredefinedFunctionList->SetReference (FALSE, TRUE, TRUE, TRUE);
    PredefinedFunctionList->Append ("mlegp0", GP0MLE, GP0MLE, V, RV, N, R, R, N);
    PredefinedFunctionList->SetReference (FALSE, FALSE, TRUE, TRUE, TRUE);

    AddOtherEstimator ("mleev1", EV1MLE, EV1MLEA);
    AddOtherEstimator ("mleev", EVMLE, EVMLEA);
    AddOtherEstimator ("lrsev", EVRDE, EVRDEA);
    AddOtherEstimator ("mdeev", EVMDE, EVMDEA);
    AddOtherEstimator ("mleev0", EV3MLE, EV3MLEA, 2);

    AddOtherEstimator ("mlet", TMLE, TMLEA, 2);
    AddOtherEstimator ("mlenct", NCTMLE, NCTMLEA);

    AddOtherEstimator ("mlegaussian", GAUSSMLE, GAUSSMLEA, 2);
    AddOtherEstimator ("lsegaussian", GAUSSLSE, GAUSSLSEA, 2);
    AddOtherEstimator ("mdegaussian", GAUSSMDE, GAUSSMDEA, 2);
    AddOtherEstimator ("mhdegaussian", GAUSSMHDE, GAUSSMHDEA, 2);

// Tests

    PredefinedFunctionList->Append ("ttest", TST, TST, R, RV, RV);
    PredefinedFunctionList->Append ("welchtest", TSWELCH, TSWELCH, R, RV, RV);
    PredefinedFunctionList->Append ("wilcoxontest", TSWILCOXON, TSWILCOXON, R, RV, RV);

// Plots

    PredefinedFunctionList->Append ("messagebox", MBOX, MBOX, V, S);
    PredefinedFunctionList->Append ("dialogbox", VBOX, VBOX, RV, S, S, RV);
    PredefinedFunctionList->Append ("menubox", CBOX, CBOX, N, S, S);

    PredefinedFunctionList->Append ("scatterplot", SCA, SCA, V, RV, RV, S);
    PredefinedFunctionList->Append ("plot", PLOT, PLOT, V, RV, RV, S, S);
    PredefinedFunctionList->Append ("boxplot", BOXPLOT, BOXPLOT, V, R, RV, S, S);
    PredefinedFunctionList->Append ("plotsurface", PLOT3D, PLOT3D, V, R, R, R, R, RM);
    PredefinedFunctionList->Append ("plotcontour", CONT3D, CONT3D, V, R, R, R, R, RM);
    PredefinedFunctionList->Append ("clearwindow", CLWIN, CLWIN, V, S);

    PredefinedFunctionList->Append ("execute", EXEC, EXEC, V, S, R, R, R);
    PredefinedFunctionList->Append ("playwav", PLAY, PLAY, V, S);
    PredefinedFunctionList->Append ("storeglobalvector", SGV, SGV, V, S, RV);
    PredefinedFunctionList->Append ("getglobalvector", GGV, GGV, RV, S);

    PredefinedFunctionList->Append ("setcoordinates", SETCS, SETCS, V, S, R, R, R, R);
    PredefinedFunctionList->Append ("setticks", SETCT, SETCT, V, S, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("setmarkers", SETCM, SETCM, V, S, R, R, R, R, R, R);
    PredefinedFunctionList->Append ("setmarkerfont", SETCF, SETCF, V, S, N, N, N, N);
    PredefinedFunctionList->Append ("setlabel", SETLB, SETLB, V, S, N, R, R, S);
    PredefinedFunctionList->Append ("setlabelfont", SETLF, SETLF, V, N, N, N);
    PredefinedFunctionList->Append ("setcolor", SETCOL, SETCOL, V, N);
    PredefinedFunctionList->Append ("setplotstyle", SETSTY, SETSTY, V, N);
    PredefinedFunctionList->Append ("setlinestyle", SETLIN, SETLIN, V, N);
    PredefinedFunctionList->Append ("setlineoptions", SETOPT, SETOPT, V, N, N, N);
    PredefinedFunctionList->Append ("saveeps", SEPS, SEPS, V, S, S, R, R, R, R, R, R);

}

#ifdef COMPONENT

// CORBA server for component programs

static PortableServer::POA_ptr RootPOA;

class TNode_impl : public virtual POA_XGPLNodes::TXGPLNode,
                   public virtual PortableServer::RefCountServantBase  {
    public:
	TNode_impl (CORBA::ORB_ptr orb, String name, TXPLPCodeInterpreter *p, int memsize0, int ptsize0, int cssize0, XGPLNodes::TPortType *Inports, int NrInports, XGPLNodes::TPortType *Outports, int NrOutports);

        virtual char *Name () throw (CORBA::SystemException);
        virtual char *Description () throw (CORBA::SystemException);
        virtual char *GetXMLDescription() throw (CORBA::SystemException);
        virtual XGPLNodes::TXGPLNode_ptr Clone () throw (CORBA::SystemException);

        virtual CORBA::Long GetNumberOfInports () throw (CORBA::SystemException);
        virtual XGPLNodes::TPortType GetInportType (CORBA::Long Nr) throw (CORBA::SystemException);
        virtual CORBA::Long GetNumberOfOutports () throw (CORBA::SystemException);
        virtual XGPLNodes::TPortType GetOutportType (CORBA::Long Nr) throw (CORBA::SystemException);

        virtual void SetInport (CORBA::Long Nr, const CORBA::Any &a) throw (CORBA::SystemException);
        virtual void SetInportData (CORBA::Long Nr, const XGPLNodes::doubleseq &x) throw (CORBA::SystemException);

        virtual CORBA::Any *GetOutport (CORBA::Long Nr) throw (CORBA::SystemException);
        virtual XGPLNodes::doubleseq *GetOutportData (CORBA::Long Nr) throw (CORBA::SystemException);

        virtual void SetInportMatrix (CORBA::Long Nr, const XGPLNodes::doublematrix &x) throw (CORBA::SystemException);
        virtual XGPLNodes::doublematrix *GetOutportMatrix (CORBA::Long Nr) throw (CORBA::SystemException);

        virtual void Destroy () throw (CORBA::SystemException);
        virtual void Perform () throw (CORBA::SystemException);

	XGPLNodes::TXGPLNode_ptr RegisterServant ();
    private:
	String name;
	TXPLPCodeInterpreter *p;
	int memsize, ptsize, cssize;
	XGPLNodes::TPortType *Inports, *Outports;
	int NrInports, NrOutports;

        CORBA::Any InValue [20], OutValue [20];
        XGPLNodes::doubleseq InData [20], OutData [20];
	XGPLNodes::doublematrix InMatrix [20], OutMatrix [20];
        CORBA::ORB_ptr orb;
};

TNode_impl::TNode_impl (CORBA::ORB_ptr orb0, String name0, TXPLPCodeInterpreter *p0, int memsize0, int ptsize0, int cssize0, XGPLNodes::TPortType *Inports0, int NrInports0, XGPLNodes::TPortType *Outports0, int NrOutports0) {
    orb = orb0;
    name = name0;
    p = p0;
    Inports = Inports0; Outports = Outports0;
    NrInports = NrInports0; NrOutports = NrOutports0;
    memsize = memsize0; ptsize = ptsize0; cssize = cssize0;
}

void TNode_impl::Destroy () throw (CORBA::SystemException) {
    PortableServer::ObjectId *id = RootPOA->servant_to_id (this);
    RootPOA->deactivate_object (*id);
    delete id;
}

char *TNode_impl::Name () throw (CORBA::SystemException) {
    return CORBA::string_dup ((const char *) name);
}

char *TNode_impl::Description () throw (CORBA::SystemException) {
    return CORBA::string_dup ((const char *) name);
}

char *TNode_impl::GetXMLDescription () throw (CORBA::SystemException) {
    return Description ();
}

XGPLNodes::TXGPLNode_ptr TNode_impl::Clone () throw (CORBA::SystemException) {
    return (new TNode_impl (orb, name, p, memsize, ptsize, cssize, Inports, NrInports, Outports, NrOutports))->RegisterServant ();
}

XGPLNodes::TXGPLNode_ptr TNode_impl::RegisterServant () {
    XGPLNodes::TXGPLNode_ptr p = _this ();
    _remove_ref ();
    return p;
}

CORBA::Long TNode_impl::GetNumberOfInports () throw (CORBA::SystemException) {
    return NrInports;
}

XGPLNodes::TPortType TNode_impl::GetInportType (CORBA::Long Nr) throw (CORBA::SystemException) {
    return Inports [Nr];
}

CORBA::Long TNode_impl::GetNumberOfOutports () throw (CORBA::SystemException) {
    return NrOutports;
}

XGPLNodes::TPortType TNode_impl::GetOutportType (CORBA::Long Nr) throw (CORBA::SystemException) {
    return Outports [Nr];
}

void TNode_impl::SetInport (CORBA::Long Nr, const CORBA::Any &a) throw (CORBA::SystemException) {
    InValue [Nr] = a;
}

void TNode_impl::SetInportData (CORBA::Long Nr, const XGPLNodes::doubleseq &x) throw (CORBA::SystemException) {
    InData [Nr] = x;
}

void TNode_impl::SetInportMatrix (CORBA::Long Nr, const XGPLNodes::doublematrix &x) throw (CORBA::SystemException) {
    InMatrix [Nr] = x;
}

CORBA::Any *TNode_impl::GetOutport (CORBA::Long Nr) throw (CORBA::SystemException) {
    return new CORBA::Any (OutValue [Nr]);
}

XGPLNodes::doubleseq *TNode_impl::GetOutportData (CORBA::Long Nr) throw (CORBA::SystemException) {
    return new XGPLNodes::doubleseq (OutData [Nr]);
}

XGPLNodes::doublematrix *TNode_impl::GetOutportMatrix (CORBA::Long Nr) throw (CORBA::SystemException) {
    return new XGPLNodes::doublematrix (OutMatrix [Nr]);
}
 
void TNode_impl::Perform () throw (CORBA::SystemException) {
    p->CreateDataMemory (memsize, ptsize, cssize);
    p->Init ();
    int i, j;
    for (i = 0; i < NrInports; i++) 
	switch (Inports [i]) {
	    case XGPLNodes::PORT_INT: {
		CORBA::Long l;
		InValue [i] >>= l;
		p->PushValue ((int) l);
		break;
	    }
	    case XGPLNodes::PORT_REAL: {
		CORBA::Double d;
		InValue [i] >>= d;
		p->PushValue ((double) d);
		break;
	    }
	    case XGPLNodes::PORT_REAL2: {
		XGPLNodes::TReal2 *para;
		InValue [i] >>= para;
		p->PushValue (para->y);
		p->PushValue (para->x);
		break;
	    }
	    case XGPLNodes::PORT_REAL3: {
		XGPLNodes::TReal3 *para;
		InValue [i] >>= para;
		p->PushValue (para->z);
		p->PushValue (para->y);
		p->PushValue (para->x);
		break;
	    }
	    case XGPLNodes::PORT_DATA: {
		int n = InData [i].length ();
    		double *x = new double [n];
    		for (int j = 0; j < n; j++)
        	    x [j] = InData [i][j];
		p->PushVector (n, x);
		delete [] x;
		break;
	    }
	    case XGPLNodes::PORT_MATRIX: {
		int cols = InMatrix [i].length (),
                    rows = InMatrix [i][0].length ();
		double *x = new double [rows * cols], *q = x;
		for (int j = 0; j < cols; j++)
		    for (int k = 0; k < rows; k++)
			*q++ = InMatrix [i][j][k];
		p->PushMatrix (rows, cols, x);
		delete [] x;		
	    }
	}
    p->Execute ();
    for (i = 0, j = 0; i < NrOutports; i++, j++)
	switch (Outports [i]) {
	    case XGPLNodes::PORT_INT: {
		CORBA::Long l = p->GetIntValue (j);
		OutValue [i] <<= l;
		break;
	    }
	    case XGPLNodes::PORT_REAL: {
		CORBA::Double d = p->GetDataValue (j);
		OutValue [i] <<= d;
		break;
	    }
	    case XGPLNodes::PORT_REAL2: {
		XGPLNodes::TReal2 para;
		para.x = p->GetDataValue (j++);
		para.y = p->GetDataValue (j);
		OutValue [i] <<= para;
		break;
	    }
	    case XGPLNodes::PORT_REAL3: {
		XGPLNodes::TReal3 para;
		para.x = p->GetDataValue (j++);
		para.y = p->GetDataValue (j++);
		para.z = p->GetDataValue (j);
		OutValue [i] <<= para;
		break;
	    }
	    case XGPLNodes::PORT_DATA: {
		double *x; int n;
		p->GetVector (j, n, x);
		double *r = XGPLNodes::doubleseq::allocbuf (n);
		for (int k = 0; k < n; k++)
		    r [k] = x [k];
		delete [] x;
		XGPLNodes::doubleseq_var q = new XGPLNodes::doubleseq (n, n, r, TRUE);
		OutData [i] = q;
		break;
	    }
	}
    p->DeleteDataMemory ();
}

#endif


/*
 *  Simple text oriented front end 
 */

#ifdef TERMINAL

int myatoi (char *s) {
    char *t;
    long l = strtol (s, &t, 10);
    if (*t) {
	printf ("Fatal: Error in number (%s)\n", s);
	exit (1);
    }
    return (int) l;
}

int codesize = 10000,
    datasize = 1000000,
    ptsize = 10000,
    cssize = 100000,
    rterror;

BOOL Listing = FALSE,
     CodeRangeCheck = TRUE,
     CodeDivisionCheck = TRUE,
     CodeBoolShortCuts = TRUE,
     CodeArgCheck = TRUE,
     SearchRunTimeError = FALSE,
     Components = FALSE;

TXPLCompiler *compiler = NULL;

// CompileFile returns NULL in case of an error.

TXPLPCodeInterpreter *CompileFile (String filename) {
    TXPLPCodeInterpreter *p = new TXPLPCodeInterpreter;

    const int BufSize = 100000;
    char *buf;
    buf = new char [BufSize];
    memset (buf, 0, BufSize);
    FILE *f = fopen (filename,  "rt");
    if (!f) { perror ("XPL"); exit (1); }
    fread (buf, 1, BufSize, f);
    fclose (f);

    if (!p->CreateMemory (datasize, ptsize, cssize, codesize)) {
	delete p;
	puts ("Fatal: Cannot allocate memory for virtual machine");
        exit (1);
    }

    TFileList *fl = new TFileList;
    fl->SetStandardOutput (new TConsoleFileDescriptor);
    fl->SetStandardInput (new TConsoleFileDescriptor);
    p->SetFileList (fl);

    if (compiler) delete compiler;
    compiler = new TXPLCompiler (p, codesize, ".");
    compiler->CodeRangeCheck = CodeRangeCheck;
    compiler->CodeDivisionCheck = CodeDivisionCheck;
    compiler->CodeBoolShortcuts = CodeBoolShortCuts;
    compiler->CodeArgCheck = CodeArgCheck;

    Scanner = new TScanner;
    Scanner->SetProgram (buf, filename);

    if (SearchRunTimeError) compiler->SearchRunTimeError (rterror);

    try {
        compiler->Compile ();
    }
    catch (TErrorClass e) {
	puts ("Fatal: Error handler in control program (this shouldn't happen)");
        puts (e.GetErrorMessage ());
 	abort ();
    }
    if (compiler->GetErrorCount ())  {
	for (TErrorClass *e = ErrorList.GetFirstError (); e; e = ErrorList.GetNextError ())
	    puts (e->GetErrorMessage ());
	delete p;
	p = NULL;
    } else if (Listing) p->Listing ();
    delete [] buf;
    return p;
}

void ShowArgs () {
    puts ("This is the terminal version of StatPascal 0.2-beta7. Usage is:\n\nsp [options] filename [arguments]\n");
    puts ("Options are (default values in parantheses):\n");
    puts ("-codesize n           allocate n code cells (10,000)");
    puts ("-datasize n           allocate n data cells (1,000,000)");
    puts ("-ptsize n             reserve n data cells for pointer table (10,000)");
    puts ("-cssize n             reserve n data cells for calculator stack (100,000)\n");
    puts ("-noboolshortcuts      do not use boolean shortcuts");
    puts ("-nodivisioncheck      do not check for division by zero");
    puts ("-noargcheck           do not check arguments of predefined  functions");
    puts ("-norangecheck         do not perform range checks\n");
    puts ("-find n               locate run time error position n");
    puts ("-listing              list virtual machine codes\n");
    puts ("Additional arguments are passed to the StatPascal program.");
    puts ("The StatPascal server uses the following arguments:\n");
    puts ("-component s1, ...    following arguments are CORBA components");
    puts ("-bootstrap s          set filename with reference to component factory\n");
}

void StartServer (char *bootstrapfile, char **filename, int filenumber, int argc, char **argv) {
#ifdef COMPONENT
#ifdef ACETAO
    CORBA::ORB_ptr orb =  CORBA::ORB_init (argc, argv, "");
#else
    CORBA::ORB_ptr orb =  CORBA::ORB_init (argc, argv, "mico-local-orb");
#endif
    CORBA::Object_var obj = orb->resolve_initial_references ("RootPOA");
    RootPOA = PortableServer::POA::_narrow(obj);
    PortableServer::POAManager_var mgr = RootPOA->the_POAManager();
    mgr->activate();

    ifstream in (bootstrapfile);
    char buf [1000];
    in >> buf;
    in.close ();

    CORBA::Object_var factoryobj = orb->string_to_object (buf);
    XGPLNodes::TNodeFactory_var factory = XGPLNodes::TNodeFactory::_narrow (factoryobj);

    TXPLPCodeInterpreter **p = new TXPLPCodeInterpreter* [filenumber];
    TNode_impl **component = new TNode_impl* [filenumber];
    BOOL Error = FALSE;
    for (int i = 0; i < filenumber; i++) {
	p [i] = CompileFile (filename [i]);
	if (!p [i]) Error = TRUE;
        else {
            p [i]->DeleteDataMemory ();
            component [i] = new TNode_impl (orb, compiler->GetComponentName (), p [i], datasize, ptsize, cssize,
                                            compiler->GetInportTypes (), compiler->GetNumberOfInports (),
		  	 	 	    compiler->GetOutportTypes (), compiler->GetNumberOfOutports ());
        }
    }
    if (!Error) {
	for (int i = 0; i < filenumber; i++) {
            factory->RegisterNode (component [i]->RegisterServant ());
            const char *t = compiler->GetComponentName ();
            printf ("Registering %s\n", t);
        }
        orb->run ();
    } else puts ("Server not started.");
#endif	
}

int main (int argc, char **argv) {
    if (argc <  2) {
	ShowArgs ();
        exit (1);
    }

    char **filename, *bootstrapfile = "objid.tmp";

    int argnumber = 1, filenumber;
    for (;;) {
	if (!strcmp (argv [argnumber], "-codesize")) {
	    argnumber++;
	    codesize = myatoi (argv [argnumber++]);
	} else if (!strcmp (argv [argnumber], "-datasize")) {
	    argnumber++;
	    datasize = myatoi (argv [argnumber++]);
	} else if (!strcmp (argv [argnumber], "-ptsize")) {
	    argnumber++;
	    ptsize = myatoi (argv [argnumber++]);
	} else if (!strcmp (argv [argnumber], "-cssize")) {
	    argnumber++;
	    cssize = myatoi (argv [argnumber++]);
	} else if (!strcmp (argv [argnumber], "-listing")) {
	    argnumber++;
	    Listing = TRUE;
	} else if (!strcmp (argv [argnumber], "-find")) {
	    argnumber++;
	    SearchRunTimeError = TRUE;
	    rterror = myatoi (argv [argnumber++]);
	} else if (!strcmp (argv [argnumber], "-noboolshortcuts")) {
	    argnumber++;
	    CodeBoolShortCuts = FALSE;
	} else if (!strcmp (argv [argnumber], "-norangecheck")) {
	    argnumber++;
	    CodeRangeCheck = FALSE;
	} else if (!strcmp (argv [argnumber], "-norargcheck")) {
	    argnumber++;
	    CodeArgCheck = FALSE;
	} else if (!strcmp (argv [argnumber], "-nodivisioncheck")) {
	    argnumber++;
	    CodeDivisionCheck = FALSE;
	} else if (!strcmp (argv [argnumber], "-bootstrap")) {
	    argnumber++;
	    bootstrapfile = argv [argnumber++];
	} else if (!strcmp (argv [argnumber], "-component")) {
	    argnumber++;
	    filename = argv + argnumber;
	    filenumber = argc - argnumber;
	    Components = TRUE;
	    break;
	} else {
	    filename = argv + argnumber++;
	    filenumber = 1;
	    break;
        }
    }

    BuildEstimators ();
    Randomize ();
    MAXSAMPLESIZE = INT_MAX;
    if (Components)
        StartServer (bootstrapfile, filename, filenumber, argc, argv);
    else {
    	TXPLPCodeInterpreter *p = CompileFile (filename [0]);
	if (p) {
            p->SetProgramArguments (argc - argnumber, argv + argnumber);
	    TRunTimeErrorType t = p->Run ();
            if (t != RT_OK) {
		puts (p->GetRunTimeErrorMessage (t));
		for (int i = 0; i < 5; i++)
		    puts (p->ListingLine (max (0, p->GetPC () - 2) + i));
	    }
	}
    } 
    return 0;
}

#endif
